module output
  
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !!! This module will be used if you configure the code with                !!!
  !!! --enable-BSH_output. At the moment it is exactly the same as the       !!!
  !!! default_output-module. Only an additional comment in this module shows,!!!
  !!! that this module is used.                                              !!!
  !!! Later this module will of course be filled with the output-routines,   !!!
  !!! which are used at BSH!                                                 !!!
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


  implicit none
  private
  
  public :: output_init, output_close, output_write
  
  
  contains
  
    subroutine output_init(aufdat,ctim,tim)
      ! Output-files will be opend in this file
    
      use cmod_miscvars,      only : itime, tidal
      use constants,          only : iptracers, calctrans, ldebug
      use dmi_mpi,            only : mpi_io_serial, ice_serial
      use dmi_mpi_global,     only : iu06
      use io_subs,            only : io_new_unit
      use transport_variables,only : transdat_new
      use wlstat_wrap,        only : wlstat 
      use writenetcdf_BSH,    only : netcdf_new, netcdf_icefile_new,           &
                                     write_c_file, write_m_file, write_t_file, &
                                     write_z_file, write_i_file

      implicit none
    
      character(11), intent(in)           :: aufdat
      character(19), intent(in)           :: ctim  !<==  watch out here
      real(8),       intent(in)           :: tim
      
      integer(4)                          :: lun, ios
      character(8),  parameter            :: fmtl='(a15,l2)'
  
      namelist /outputlist/   write_c_file, write_m_file, write_t_file,        &
                              write_z_file, write_i_file

      if (mpi_io_serial) then
        !- Initialise the output - read possibly output-namelist ---------------
 
        !- set defaults --------------------------------------------------------
        write_c_file = .true.
        write_m_file = .true.
        write_t_file = .true.
        write_z_file = .true.
        write_i_file = .true.

        !-----Arguments from NAMELIST ------------------------------------------
        lun = io_new_unit()
        open (unit=lun, file='output.nml', status='old', iostat=ios)
        if (ios /= 0) then
          write(iu06,'(a47)')  'Warning: Cannot open file output.nml ...       '
          write(iu06,'(a47)')  '       ... continue with default output-options'
        else
          read (unit=lun, nml=outputlist, iostat=ios)
          if (ios /= 0) then
            write(iu06,'(a45)')'Warning: Cannot read output.nml                '
            write(iu06,'(a45)')'         ... continue without                  '
          endif
          close (unit=lun)
        endif
        if (ldebug) then
          write(iu06,'(a44)') 'Dump of namelist outputlist from output.nml:'
          write(iu06,fmtl) 'write_m_file  =',write_m_file
          write(iu06,fmtl) 'write_c_file  =',write_c_file
          write(iu06,fmtl) 'write_t_file  =',write_t_file
          write(iu06,fmtl) 'write_z_file  =',write_z_file
          write(iu06,fmtl) 'write_i_file  =',write_i_file
        endif

        !- create nedtcdf output files (m, c, t and z) -------------------------
        call netcdf_new ()
 
        !- open ascii sea level time series files, read station list file ------
        call wlstat(0,ctim,tim,itime)

        !- open binary file for inline transport output ------------------------
        if (calctrans) call transdat_new ()
      endif

      if (ice_serial .and. write_i_file) then
        !-  create netcdf ice parameter output files ---------------------------
        if (.not.tidal) call netcdf_icefile_new ()
      endif

    end subroutine output_init
  
  !============================================================================!
  
  subroutine output_write(izeits,aufdat,ctim,tim)
    ! Output-files will be closed in this file
    
    use cmod_arrays,        only : mm1k, u, v, h, z, wu, wv, cmp, tke,         &
                                   diss, avv, dispv, wrk,                      &
                                   casus, ice, tsnei, ueis, veis, iota
    use cmod_local_arrays,  only : ind_l, mm1k_l, u_l, v_l, h_l, wi_l, z_l,    &
                                   wu_l, wv_l, cmp_l, tke_l, diss_l, avv_l,    &
                                   dispv_l,                                    &
                                   casus_l, ice_l, tsnei_l, ueis_l, veis_l,    &
                                   iota_l
    use cmod_miscvars,      only : izeite, itime, lmn1, lmn2, lmne, tidal
    use cmod_params,        only : narea, kmx
    use compress,           only : compress_i4, compress_i4_ice
    use constants,          only : iptracers, printturb,dyn,calctrans,iotamodel
    use dmi_mpi,            only : mpi_io_serial, dmpi_gather
    use dmi_mpi_global,     only : mpi_io_rank
    use transports,         only : transport_output
    use transport_variables,only : NDTTRANSOUT, transdat_new
    use turbout_wrap,       only : turbout
    use wlstat_wrap,        only : wlstat 
    use writenetcdf_BSH,    only : netcdf_next, netcdf_icefile_next,           &
                                   write_netcdf_icefile, netcdf_write,         &
                                   write_i_file 
    
 
    implicit none
    
    integer(4), intent(in)       :: izeits
    character(11), intent(in)    :: aufdat
    character(19), intent(in)    :: ctim  !<==  watch out here!
    real(8),       intent(in)    :: tim
    
    integer(4)                   :: ia       


    !- flow field varibles to netcdf-files, etc.:
    if (mod(izeits,lmn2) == 0) then

      !- Gather to I/O task:
      do ia=1,narea
        call dmpi_gather(ia,kmx(ia),  u_l(ia)%p,  u(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia),  v_l(ia)%p,  v(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia), wi_l(ia)%p,wrk(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,(1),      z_l(ia)%p,  z(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,(1),     wu_l(ia)%p, wu(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,(1),     wv_l(ia)%p, wv(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia),cmp_l(ia)%p,cmp(ia)%p,2,mpi_io_rank)
      enddo
      if (dyn) then
        do ia=1,narea
          call dmpi_gather(ia,kmx(ia),h_l(ia)%p,h(ia)%p,mpi_io_rank)
        enddo
      endif          

      if (mpi_io_serial) then
        !- first compress to i4:
        call compress_i4 ()

        !- output to netcdf files:
        call netcdf_write (ctim,tim,aufdat,izeits,lmn2)
      endif

      !- output turbulence profiles:
      if (printturb) then 
        do ia=1,narea
          call dmpi_gather(ia,kmx(ia),  tke_l(ia)%p,  tke(ia)%p,mpi_io_rank)
          call dmpi_gather(ia,kmx(ia), diss_l(ia)%p, diss(ia)%p,mpi_io_rank)
          call dmpi_gather(ia,kmx(ia),  avv_l(ia)%p,  avv(ia)%p,mpi_io_rank)
          call dmpi_gather(ia,kmx(ia),dispv_l(ia)%p,dispv(ia)%p,2,mpi_io_rank)
        enddo
        if (mpi_io_serial)  call turbout(ctim)
      endif
 
    endif

    if (mod(izeits,lmn1) == 0) then    ! hourly
      !- ice data to file:
      if (.not.tidal .and. write_i_file) then 
        do ia=1,narea
          call dmpi_gather(ia,    casus_l(ia)%p,casus(ia)%p,mpi_io_rank)
          call dmpi_gather(ia,(1),  ice_l(ia)%p,  ice(ia)%p,3,mpi_io_rank)
          call dmpi_gather(ia,(1), ueis_l(ia)%p, ueis(ia)%p,mpi_io_rank)
          call dmpi_gather(ia,(1), veis_l(ia)%p, veis(ia)%p,mpi_io_rank)
          call dmpi_gather(ia,(1),tsnei_l(ia)%p,tsnei(ia)%p,mpi_io_rank)
        enddo
        if (iotamodel) then
          do ia=1,narea
            call dmpi_gather(ia,(1), iota_l(ia)%p, iota(ia)%p,3,mpi_io_rank)
          enddo
        end if

        if (mpi_io_serial) then
          !- first compress to i4:
          call compress_i4_ice ()

          call write_netcdf_icefile (ctim,tim,izeits,lmn1)
        endif
      endif
    endif

    !- ascii sea level time series files:
    if (mod(izeits,lmne) == 0) then
      if (mod(izeits,lmn2) /= 0) then
        do ia=1,narea
          call dmpi_gather(ia,(1),z_l(ia)%p,z(ia)%p,mpi_io_rank)
        enddo
      endif
      if (mpi_io_serial) call wlstat(1,ctim,tim,itime)
    endif

    if (calctrans) then
      if (mod(izeits, NDTTRANSOUT) == 0) then
        call transport_output( narea, ctim )
      endif
    endif

    if (izeits == izeite) then
      !- netcdf files in numbered 6 or 12 hour chunks:
      !    With the option archive=.true. the first chunk goes to the 
      !    '...-file_area_archive.nc' files, and subsequent chunks will go to 
      !    '...-file_area.00.nc', '...-file_area.01.nc', ... .   
      !    With archive=.false., the first files are '...-file_area.00.nc', and
      !    subsequent files are '...-file_area.01.nc', '...-file_area.02.nc, 
      !    ... .
      
      if (mpi_io_serial) then
        call netcdf_next ()
        if (.not.tidal .and. write_i_file) then 
          call netcdf_icefile_next ()
        endif
        if (calctrans) call transdat_new () 
      endif
      
    endif
      
  end subroutine output_write

  !============================================================================!
  
  subroutine output_close(ctim,tim)
      ! Output-files will be closed in this file
    
    use constants,          only : iptracers, calctrans
    use dmi_mpi,            only : mpi_io_serial, ice_serial
    use transport_variables,only : transdat_close
    use wlstat_wrap,        only : wlstat 
    use writenetcdf_BSH,    only : netcdf_close, netcdf_icefile_close
      
    implicit none
    
    character(19), intent(in)           :: ctim  !<==  watch out here!
    real(8),       intent(in)           :: tim

    if (mpi_io_serial) then
      call wlstat(2,ctim,tim)
      if (calctrans) call transdat_close()
    endif

  end subroutine output_close

end module output
