!==============================================================================!
!
! parameters for the biogeochemical model
!
!==============================================================================!

module bioparam 

  
  implicit none
  private
  save

!==============================================================================!
!
!   B A L T I C         Parameter Block
!
!  different params to be defined for other water types (ATLANTIC, ..)
!
!       NOTE:
!       All timeunits are at the moment in days
!
!==============================================================================!
  integer(4),public    :: version      = 1
  !- these can be changed in coupler_usage depending on the model vesion
  !-
  integer(4),public    :: ntbio        = 9       ! No. of biological fields
  integer(4),public    :: ntben        = 1       ! No. of benthos fields
  logical, public      :: zpmodel      = .false. ! mirco and mezo zooplankton
  !-
  logical, public      :: ppispmmod    = .false. ! SPM paramterisation in PAR
  logical, public      :: npread       = .false. ! Spatially varying NP ratip
  logical, public      :: impl         = .false. ! Implicit scheme
  logical, public      :: resuspension = .false.  ! Sediment resuspension
  real(8), public      :: nb           = 0.01_8  ! Loss rate P to N  
  real(8), public      :: deltao       = 0.02_8  ! Loss Rate PP to D 
  real(8), public      :: dn           = 0.01_8  ! Loss Rate Detritus to N
  real(8), public      :: ds           = 0.004_8 ! Loss Rate S-Detritus to S
  real(8), public      :: ksdenit      = 0.10_8  ! Half-sat. inhib NO3 denit
  real(8), public      :: ksnit        = 0.01_8  ! Half-sat. O2 nitrification
  real(8), public      :: rnit         = 0.1_8   ! Max nitrification rate at T0
  real(8), public      :: rp0          = 1.0_8   ! Max dia uptake rate at T0
  real(8), public      :: rf0          = 0.7_8   ! Max Flag uptake rate at T0
  real(8), public      :: rb0          = 0.5_8   ! Max cyano uptake rate at T0
  real(8), public      :: cyanotll     = 15.0_8  ! Lower T limit Cyanos
  real(8), public      :: cyanosul     = 12.0_8  ! Upper S limit Cyanos
  real(8), public      :: flagtll      = 0.06_8  ! flag temp scaling
  real(8), public      :: flagtsc      = 0.8_8   ! flag temp dep.
  real(8), public      :: alphap       = 0.3_8   ! Dia half-sat. constant
  real(8), public      :: alphaf       = 0.15_8  ! Flag half-sat. constant
  real(8), public      :: alphab       = 0.5_8   ! Cyano half-sat. constant
  real(8), public      :: p0           = 0.001_8 ! background values dia
  real(8), public      :: f0           = 0.001_8 ! Background values flag
  real(8), public      :: b0           = 0.001_8 ! Background values cyano
  real(8), public      :: z0           = 0.001_8 ! Background values Zoo
  real(8), public      :: Imin1        = 75.0_8  ! Min opt. dia light W/m^2
  real(8), public      :: Imin2        = 75.0_8  ! Min opt. flag light W/m^2
  real(8), public      :: Imin3        = 75.0_8  ! Min opt. cyan light W/m^2
  real(8), public      :: att_par_1    = 0.18_8  ! Backr. light attenuation,[/m]
  real(8), public      :: att_par_2    = 0.03_8  ! Light attenuation, POM []
  real(8), public      :: q10_rec      = 0.15_8  ! Recycling temp dep [/degC]
  real(8), public      :: anit         = 0.11_8  ! Nitrification temp dep[/degC]
  real(8),parameter    :: Nnorm        = 4.5_8   ! Normalizing const. Nutrients
  real(8),parameter    :: Onorm        = 375.0_8 ! Normalizing const. Oxygen
  real(8),parameter    :: NOR          = Nnorm/Onorm
  real(8),parameter    :: ONnitr       = 2.0_8   ! Stoechiometric O/N ratio nitr
  real(8),parameter    :: NNdenit      = 5.3_8   ! Stoechiometric Norg/Nit denit
  real(8),parameter    :: ONniup       = 8.625_8 ! Stoech. O/N ratio nitr uptake
  real(8),parameter    :: ONamup       = 6.625_8 ! Stoech. O/N ratio amm up/rel
  real(8),parameter    :: rfs          = 0.94_8  ! Redfield Ratio S/N
  real(8), public      :: rfr          = 0.072_8 ! Redfield Ratio P/N
  real(8), public      :: wdz          = -1.5_8  ! Sinking velocity detritus m/d
  real(8), public      :: wpz          = -0.2_8  ! Sinking velocity dia [m/d]
  real(8), public      :: wbz          = 0.1_8   ! Rising vel. cyano [m/d]
  real(8), public      :: srfamflx     = 62.1e-3_8 ! Atm. ammonia deposition
  real(8), public      :: srfniflx     = 62.1e-3_8 ! Atm. nitrate deposition
  real(8), public      :: srfpflx      = 4.7097e-4_8 ! Atm. phos deposition

  !----------------------------------------------------------------------------!
  ! Optional paramters
  !----------------------------------------------------------------------------!
  ! ppispmmod
  real(8), public      :: att_par_3    = 0.02_8  ! Light attenuation SPM 
  real(8), public      :: att_vdiff    = 0.01_8  ! Reference vert. diffusivity
  real(8), public      :: att_depth    = 64.0_8  ! Reference depth

  ! Zooplankton rates (zpmodel)
  real(8), public      :: mezgraz      = 0.2_8  ! Mesozoopl. grazing constant
  real(8), public      :: mizgraz      = 0.4_8 ! Microzoopl. grazing constant
  real(8), public      :: mezgrazf     = 0.40_8 ! Mesozoopl. food dep. constant 
  real(8), public      :: mizgrazf     = 0.20_8 ! Microzoopl. food dep. const
  real(8), public      :: miztk        = 150.0_8! Microzoopl. grazing temp dep.
  real(8), public      :: meztk        = 150.0_8! Mezozoopl. grazing temp dep.
  real(8), public      :: mez_tscale   = 2.0_8  ! Mezozoopl. temp scaling
  real(8), public      :: mezcl        = 0.02_8 ! Mesozoopl. closure
  real(8), public      :: ldmez        = 0.02_8 ! Mesozoopl. mortality
  real(8), public      :: ldmiz        = 0.05_8 ! Micrezoopl. mortality
  real(8), public      :: lnmez        = 0.45_8 ! Mesozoopl. excretion
  real(8), public      :: lnmiz        = 0.30_8 ! Microzoopl. excretion
  real(8), public      :: mezprefdia   = 1.0_8  ! Preference of mezoo on dia
  real(8), public      :: mezpreffla   = 0.3_8  ! Preference of mezoo on flag
  real(8), public      :: mezprefcya   = 0.3_8  ! Preference of mezoo on cyano
  real(8), public      :: mezprefmiz   = 1.0_8  ! Preference of mezoo on mizoo
  real(8), public      :: mizprefdia   = 0.3_8  ! Preference of mizoo on dia
  real(8), public      :: mizpreffla   = 1.0_8  ! Preference of mizoo on flag
  real(8), public      :: mizprefcya   = 0.3_8  ! Preference of mizoo on cyano
  ! Zooplankton rates (orig)
  real(8), public      :: zcl1         = 30.0_8 ! Zooplankton closure
  real(8), public      :: toptz        = 20.0_8 ! Opt. grazing temp
  real(8), public      :: sigma_b      = 0.03_8 ! Loss Rate Z to Detritus
  real(8), public      :: nue          = 0.01_8 ! Loss Rate Z to N, resp.
  real(8), public      :: iv           = 0.8_8  ! Ivlev constant, quadratic
  real(8), public      :: graz         = 0.5_8  ! Zooplankton grazing rate

  !----------------------------------------------------------------------------!
  ! Sediment parameters 
  !----------------------------------------------------------------------------!
  real(8), public       :: reslim      = 0.018_8  ! Critical fric.v. for resusp
  real(8), public       :: resusprate  = 10.0_8   ! Resuspension rate
  real(8), public       :: dn_sed      = 0.01_8   ! Loss Rate Sediment to N
  real(8), public       :: dn_sed_anox = 0.005_8  ! Loss Rate Sed. to N, anoxia
  real(8), public       :: q10_recs    = 0.15_8   ! Recycling temp dep sed. [/C]
  real(8), public       :: fnitr       = 0.4_8    ! Fraction of recycled N
                                                  ! nitrified in sediment
  real(8), public       :: pfrac       = 0.0_8    ! fraction of phosphate
                                                  ! buried in oxic conditions
  real(8), public       :: kspburial   = 0.1_8    ! Half-saturation p burial
                                                  ! Oxygen dependance
  !----------------------------------------------------------------------------!
  ! some control parameters:
  !----------------------------------------------------------------------------!
  integer(4), public :: ibiofields = 0, NDTBIOMOD = 2
  real(8),    public :: biooutdt = 3600.0_8

  !----------------------------------------------------------------------------!
  ! index of biological params
  !----------------------------------------------------------------------------!
  integer(4), public :: idx_amm   = 1
  integer(4), public :: idx_nit   = 2
  integer(4), public :: idx_phos  = 3
  integer(4), public :: idx_sil   = 0
  integer(4), public :: idx_dia   = 4
  integer(4), public :: idx_flag  = 5
  integer(4), public :: idx_cyano = 6
  integer(4), public :: idx_mez   = 7
  integer(4), public :: idx_miz   = 0
  integer(4), public :: idx_det   = 8
  integer(4), public :: idx_dets  = 0
  integer(4), public :: idx_oxy   = 9
  integer(4), public :: bidx_nitr = 1
  integer(4), public :: bidx_sil  = 0

  ! public parameters:
  public :: Nnorm, Onorm, NOR, ONnitr, NNdenit, ONniup, ONamup, rfs

  ! public methods:
  public :: ReadBioparam, bioInd
  
  namelist /bioparameters /nb, deltao, nue, sigma_b, dn, ksdenit, ksnit, rp0,  &
                       rf0, rb0, cyanotll, cyanosul, flagtll, alphap, alphaf,  &
                       alphab, iv, graz, toptz, zcl1, p0, f0, b0, z0, Imin1,   &
                       Imin2, Imin3, q10_rec, anit, wdz, wpz, wbz, dn_sed,     &
                       dn_sed_anox, q10_recs, fnitr, pfrac, kspburial,         &
                       ibiofields, NDTBIOMOD, biooutdt, srfamflx, srfniflx,    &
                       srfpflx, att_vdiff, att_depth, rfr, impl, ppispmmod,    &
                       att_par_3, att_vdiff, att_depth, npread, meztk,         &
                       reslim, resusprate, resuspension, version, lnmez, lnmiz,&
                       ldmez, ldmiz, mezcl, mezprefmiz, mezprefdia, mezpreffla,&
                       mezprefcya, mizprefdia, mizpreffla, mizprefcya, miztk,  &
                       mez_tscale, mezgraz, mezgrazf, mizgrazf, mizgraz,       &
                       flagtsc

contains

!===============================================================================

  subroutine ReadBioparam(iu06)

    use exits,          only : exitme
    use io_subs,        only : io_new_unit
    implicit none
  
    integer(4), intent(in) :: iu06

    integer(4) :: ios, lun

    lun = io_new_unit()
    open(lun,file='bioparams.txt',status='old',iostat=ios)
    if (ios /= 0) then
      write(iu06,*) 'Warning: File bioparams.txt not found. Continue with ...'
      write(iu06,*) '        ... default biological parameter values'
    else
      read (unit=lun, nml=bioparameters, iostat=ios)
      if (ios /= 0) then
        write(iu06,*) 'Error: Cannot read values from bioparameters namelist'
        write(iu06,*) 'Error: The iostat failure was: ', ios
        call exitme(1)
      endif
    endif
    close(lun)
  
  end subroutine ReadBioparam

!===============================================================================

  subroutine bioInd()
    implicit none

    if (version == 2) then
      idx_miz = 8
      idx_det = 9
      idx_oxy = 10
    elseif (version == 3) then
      idx_sil   = 4
      idx_dia   = 5
      idx_flag  = 6
      idx_cyano = 7
      idx_mez   = 8
      idx_miz   = 9
      idx_det   = 10
      idx_dets  = 11
      idx_oxy   = 12
      !- silicate benthos pool
      bidx_sil  = 2
    endif
  end subroutine
end module bioparam
