!==============================================================================!
!
! parameters for the biogeochemical model
!
!==============================================================================!

module bioparam 

  
  implicit none
  private
  save

!==============================================================================!
!
!   B A L T I C         Parameter Block
!
!  different params to be defined for other water types (ATLANTIC, ..)
!
!       NOTE:
!       needs to be checked for units and scaling
!       All timeunits are at the moment in days
!       Erosion-sedimentation rates need to verified
!
!==============================================================================!
  integer(4),parameter :: ntbio        = 11      ! No. of biological fields
  logical, public      :: stage_model  = .false. ! zooplankton model 1
  logical, public      :: zpmodel      = .true.  ! zooplankton model 2
  logical, public      :: ecofish      = .true.  ! fish grazing on zooplankton
  real(8), public      :: nb           = 0.01_8  ! loss rate P to N  
  real(8), public      :: deltao       = 0.02_8  ! Loss Rate P to detritus
  real(8), public      :: nue          = 0.02_8  ! Loss Rate Z to N (0.01)
                                                 !   background respiration
  real(8), public      :: sigma_b      = 0.03_8  ! Loss Rate Z to Detritus
  real(8), public      :: dn           = 0.006_8 ! Loss Rate Detritus to N
                                                 !   0.01 0.003
  real(8), public      :: Ksdenit      = 0.03_8  ! Half-saturation/
                                                 !   inhibition  
                                                 !   NO3 denitrification
  real(8), public      :: Ksnit        = 0.01_8  ! Half-saturation O2 
                                                 !   nitrification
  real(8), public      :: rnit         = 0.1_8   ! Max nitrification rate 
                                                     !   at 0 deg C
  real(8), public      :: rp0          = 1.2_8   ! Max diatom uptake (1.1/1.3)
                                                 !   rates at 0 deg C
  real(8), public      :: rf0          = 0.7_8   ! Max flag uptake (0.4)
  real(8), public      :: rb0          = 0.5_8   ! Max cyano uptake (0.75)
  real(8), public      :: cyanotll     = 13.0_8  ! lower T limit Cyanos
  real(8), public      :: cyanosll     = 1.0_8   ! lower S limit Cyanos
  real(8), public      :: cyanosul     = 12.0_8  ! upper S limit Cyanos
  real(8), public      :: flagtll      = 100.0_8 ! Flagellates half 
                                                 !   saturation temperature 
                                                 !   square
  real(8), public      :: alphap       = 0.20_8  ! Halfsaturation constants
                                                 !   (0.30/0.35)
  real(8), public      :: alphaf       = 0.15_8  ! Flag km (0.10/0.15/0.2)
  real(8), public      :: alphab       = 0.50_8  ! Cyano km
  real(8), public      :: iv           = 1.2_8   ! Ivlev constant, quadratic
                                                 !  (km =0.12)   
  real(8), public      :: graz         = 0.16_8  ! zooplankton grazing rate 
                                                 !  (0.5)
  real(8), public      :: ivi          = 0.189_8 ! zoostuff (Marie help me here)
  real(8), public      :: lzd_a        = 0.08_8  ! zoostuff (Marie help me here)
  real(8), public      :: lzd_b        = 0.35_8  ! zoostuff (Marie help me here)
  real(8), public      :: lzn_a        = 0.35_8  ! zoostuff (Marie help me here)
  real(8), public      :: toptz        = 20.0_8  ! opt grazing temp, param 1
  real(8), public      :: zcl1         = 0.10_8  ! Zooplankton closure, param 1
  real(8), public      :: fish         = 1.0_8   ! default fish feeding pressure
                                                 !   index for zooplankton

  real(8), public      :: p0           = 0.01_8  ! background values P,F and B
                                                 !   (0.001)
  real(8), public      :: f0           = 0.01_8  !
  real(8), public      :: b0           = 0.01_8  !
  real(8), public      :: z0           = 0.001_8 ! background values Z
  real(8), public      :: Imin         = 100.0_8 ! minimum of optimal light 
                                                 !   radiation,  W/m^2 (25)
  real(8), public      :: att_par_1    = 0.16_8  ! Light attenuation, i
                                                 !   background, (/m) (0.18)
  real(8), public      :: att_par_1_w  = 0.60_8  ! Light attenuation, i
                                                 !   background, (/m)
  real(8), public      :: att_par_2    = 0.015_8 ! Light attenuation, POM,
                                                 !   (/m) ()
  real(8), public      :: q10_rec      = 0.15_8  ! recycling temp dep 
                                                 !   (/degC)
  real(8), public      :: anit         = 0.11_8  ! nitrification temp dep 
                                                 !   (/degC)
  real(8), parameter   :: Nnorm        = 4.5_8   ! normalizing constant 
                                                 !   nutrients
  real(8), parameter   :: Onorm        = 375.0_8 ! normalizing constant 
                                                 !   oxygen
                                                 ! 1 rel O2 => 375 mmol/m**3
  real(8), parameter   :: NOR          = Nnorm/Onorm
  real(8), parameter   :: ONnitr       = 2.0_8   ! Stoechiometric O/N ratio 
                                                 !   nitrication
  real(8), parameter   :: NNdenit      = 5.3_8   ! Stoechiometric Norg/Nit  
                                                 !   denitrication
  real(8), parameter   :: ONniup       = 8.625_8 ! Stoechiometric O/N ratio 
                                                 !   nitrate uptake
  real(8), parameter   :: ONamup       = 6.625_8 ! Stoechiometric O/N ratio 
                                                 !   ammonium uptake/release
  real(8), public      :: pistonvel    = 5.1_8   ! Piston velocity m/d
  real(8), public      :: rfr          = 1._8/16._8
                                                 ! Redfield Ratio P/N
  real(8), public      :: rfr_s        = 15._8/16._8
                                                 ! Redfield Ratio Si/N
  real(8), public      :: wdz          = -3.50_8 ! sinking velocity 
                                                 !   Detritus m/d
  real(8), public      :: wpz          = -0.5_8  ! sinking velocity 
                                                 !   Pytoplankton  m/d
  real(8), public      :: wbz          = 0.1_8   ! rising BlueGreen Algae 
                                                 !   m/d
  real(8), public      :: zooprefdia   = 1.0_8   ! Preference factor 
                                                 !   dia->zoo
  real(8), public      :: zooprefflag  = 1.0_8   ! Preference factor 
                                                 !   fla->zoo
  real(8), public      :: zooprefcyano = 0.5_8   ! Preference factor 
                                                 !   cyano->zoo
  real(8), public      :: zooprefzoo   = 0.0_8   ! Preference factor 
                                                 !  (1.0 is way to high)
  real(8), public      :: srfamflx     = 62.1e-3_8
                                                 ! atm. ammonia deposition
  real(8), public      :: srfniflx     = 62.1e-3_8
                                                 ! atm. nitrate deposition
  real(8), public      :: srfpflx      = 4.7e-3_8
                                                 ! atm. phos deposition
                                                     
  !----------------------------------------------------------------------------!
  ! sediment parameters 
  !----------------------------------------------------------------------------!
  real(8), public      :: dn_sed       = 0.002_8  ! 0.001 Loss Rate Sediment 
                                                  !   to N
  real(8), public      :: dn_sed_anox  = 0.002_8  ! Loss Rate Sediment 
                                                  !   to N, anoxic bw
  real(8), public      :: q10_recs     = 0.15_8   ! recycling temp dep 
                                                  !   sediment/C
  real(8), public      :: fnitr        = 0.50_8   ! fraction of recycled N
                                                  !   nitrified in sediment
                                                  !   0.50 0.25, 0.67 
  real(8), public      :: pfrac        = 0.15_8   ! fraction of phosphate
                                                  !   buried in oxic conditions
  real(8), public      :: Kspburial    = 0.10_8   ! Half-saturation p burial
                                                  !   Oxygen dependance

  !----------------------------------------------------------------------------!
  ! Protozooplankton (PZ) parameters
  !----------------------------------------------------------------------------!
  real(8), public      :: pz_lpzd    = 0.40_8   ! Loss Rate Z to detritus
  real(8), public      :: pz_lpzn    = 0.35_8   ! Loss Rate Z to NH4
  real(8), public      :: pz_imax    = 0.4_8    ! Max ingestion (per day at
                                                !   0C temp) (0.3)
  real(8), public      :: pz_kpz     = 0.21_8   ! Half saturation (mmol N / m3)
                                                !   (0.13)
  real(8), public      :: zooprefpz  = 1.0_8    ! Mesozooplankton pref fact. for
                                                !   protozoop.
  real(8), public      :: pzprefdia  = 1.0_8    ! PZ pref factor for diatoms
  real(8), public      :: pzprefflag = 1.0_8    ! PZ pref factor for flag
  real(8), public      :: pzprefcya  = 0.2_8    ! PZ pref factor for cyano
  real(8), public      :: lzproto_a = 0.03_8   ! HELP ME MARIE!!
  real(8), public      :: gg_pz_a    = 50.0_8   ! HELP ME MARIE!!

  !----------------------------------------------------------------------------!
  ! aggregation coefficient for diatoms
  !----------------------------------------------------------------------------!

  real(8), public      :: wpz_alpha = 0.0_8     !valid range: 0.0-0.2

  !----------------------------------------------------------------------------!
  !  stage structered zooplankton model, rates are per day except for sinking
  !  cop-biomass unit is ug-C, 
  !  Calanus finmarchicus  Sept. 2009 Marie Maar and Eva Friis Møller
  !----------------------------------------------------------------------------!

  real(8), parameter   :: Fnorm    =  2.795e-3_8    ! from mg-C/m3 to 
                                                    !  (mmol-N/m3/norm)
  real(8), parameter   :: Znorm    =  1.587e-5_8    ! from µg-N to mmol-N/ norm
  real(8), public      :: betan    =  1.41_8        ! Max. grazing rate nauplii 
                                                    !   (14 degrees C)
                                                    !   (1.12)
  real(8), public      :: betac1   =  0.81_8        ! Max. grazing rate cop1 
                                                    !   (14 degrees C)
                                                    !   (0.64)
  real(8), public      :: betac2   =  0.47_8        ! Max. grazing rate cop2 
                                                    !   (14 degrees C)
                                                    !   (0.34)
  real(8), public      :: betaa    =  0.41_8        ! Max. grazing rate aduts 
                                                    !   (14 degrees C)
                                                    !   (0.30)
  real(8), parameter   :: kmn      =  25.0_8*Fnorm  ! Half-saturation constant
                                                    !   nauplii
                                                    !   (118)
  real(8), parameter   :: kmc1     =  25.0_8*Fnorm  ! Half-sat. constant cop1 
  real(8), parameter   :: kmc2     =  25.0_8*Fnorm  ! Half-sat. constant cop2
  real(8), parameter   :: kma      =  25.0_8*Fnorm  ! Half-sat. constant adults
  
  real(8), parameter   :: xn       = 0.33_8*Znorm   ! max mass nauplii (Table 2)
  real(8), parameter   :: xc1      = 4.17_8*Znorm   ! max mass cop1 
                                                    !   (now as norm.-N)
  real(8), parameter   :: xc2      = 30.00_8*Znorm  ! max mass cop2
  real(8), parameter   :: xa       = 36.5_8*Znorm   ! max mass adults
  real(8), parameter   :: moe      = 0.035_8*Znorm  ! mean egg mass
  real(8), parameter   :: mon      = 0.287_8*Znorm  ! mean mass nauplii stage, 
                                                    !   set manuelly since 
                                                    !   to small diff to max
                                                    !   (0.327)
  real(8), parameter   :: moc1     = 3.53_8*Znorm   ! mean mass cop1
  real(8), parameter   :: moc2     = 24.78_8*Znorm  ! mean mass cop2
  real(8), parameter   :: moa      = 31.5_8*Znorm   ! mean mass adults
  real(8), parameter   :: templ    = 287.9_8        ! Lower temp. tolerance 
                                                    !   boundary (K)
  real(8), parameter   :: temph    = 286.2_8        ! upper temp. tolerance 
                                                    !   boundary (K)
  real(8), parameter   :: tempal   = 14352._8       ! rate of decrease at 
                                                    !   lower boundary
  real(8), parameter   :: tempah   = 31310._8       ! rate of decrease at 
                                                    !   upper boundary
  real(8), parameter   :: hatch    = 1747._8        ! maximum hatching 
                                                    !   (Cook et al 2007)
  real(8), parameter   :: hatexp   = -2.05_8        ! hatching exponent
  real(8), parameter   :: hatalpha = -9.8_8         ! hatching alpha
  real(8), public      :: swa      = moc2*0.5_8     ! starvation weight adults
                                                    !   (0.9)
  real(8), public      :: swc2     = moc1*0.5_8     ! starvation weight cop2
  real(8), public      :: swc1     = mon*0.5_8      ! starvation weight cop1
  real(8), public      :: swn      = moe*0.5_8      ! starvation weight nauplii
  real(8), public      :: w_adult  =  -5.0_8        ! migration velcity (m/d)

  !----------------------------------------------------------------------------!
  !  stage structered zooplankton model, rates are per day except for sinking
  !  cop-biomass unit is ug-C, 
  !
  !  Calanus helgolandicus Sept. 2009 Marie Maar and Eva Friis Møller
  !  ingestion is the same for the two Calanus spp.  
  !----------------------------------------------------------------------------!
  real(8), public      :: betan_ch    =  1.35_8     ! Max. grazing rate nauplii 
                                                    !   (14 degrees C)
                                                    !   (1.12)
  real(8), public      :: betac1_ch   =  0.89_8     ! Max. grazing rate cop1 
                                                    !   (14 degrees C)
                                                    !   (0.64)
  real(8), public      :: betac2_ch   =  0.54_8     ! Max. grazing rate cop2 
                                                    !   (14 degrees C)
                                                    !   (0.34)
  real(8), public      :: betaa_ch    =  0.45_8     ! Max. grazing rate aduts 
                                                    !   (14 degrees C)
                                                    !   (0.3)
  
  real(8), parameter   :: xn_ch       = 0.49_8 *Znorm  !max mass nauplii Table 2
  real(8), parameter   :: xc1_ch      = 2.49_8 *Znorm 
  real(8), parameter   :: xc2_ch      = 15.25_8 *Znorm  
  real(8), parameter   :: xa_ch       = 21.7_8 *Znorm  
  real(8), parameter   :: moe_ch      = 0.071_8 *Znorm !mean egg mass
  real(8), parameter   :: mon_ch      = 0.40_8 *Znorm  !mean mass nauplii stage
  real(8), parameter   :: moc1_ch     = 2.16_8 *Znorm   
  real(8), parameter   :: moc2_ch     = 11.68_8 *Znorm  
  real(8), parameter   :: moa_ch      = 17.82_8 *Znorm !mean mass adults
  real(8), parameter   :: templ_ch    = 287.2_8        ! Lower temp. 
                                                       !   tolerance boundary(K)
  real(8), parameter   :: temph_ch    = 280.5_8        ! upper temp. 
                                                       !   tolerance boundary(K)
  real(8), parameter   :: tempal_ch   = 40296._8       ! rate of decrease at 
                                                       !   lower boundary
  real(8), parameter   :: tempah_ch   = 14970._8       ! rate of decrease at 
                                                       !   upper boundary
  real(8), parameter   :: hatch_ch    = 1378._8        ! maximum hatching 
                                                       !   (Bonnet et al 2009)
  real(8), parameter   :: hatexp_ch   = -2.05_8        ! hatching exponent
  real(8), parameter   :: hatalpha_ch = -6.01_8        ! hatching alpha
  real(8), public      :: w_adult_ch  = 3000.0_8       ! migration velocity 
                                                       !   pseudoca.
  real(8), public      :: swa_ch      = moa_ch*0.5_8   ! starvation weight 
                                                       !   adults
  real(8), public      :: swc2_ch     = moc2_ch*0.5_8  ! starvation weight cop2
  real(8), public      :: swc1_ch     = moc1_ch*0.5_8  ! starvation weight cop1
  real(8), public      :: swn_ch      = mon_ch*0.5_8   ! starvation weight 
                                                       !   nauplii
           
  !----------------------------------------------------------------------------!
  ! some control parameters:
  !----------------------------------------------------------------------------!
  integer(4), public :: ibiofields = 0, NDTBIOMOD = 2
  real(8),    public :: biooutdt = 3600.0_8
  
  !----------------------------------------------------------------------------!
  ! index of biological params
  !----------------------------------------------------------------------------!
  integer(4), parameter, public :: idx_amm       = 1
  integer(4), parameter, public :: idx_nit       = 2
  integer(4), parameter, public :: idx_phos      = 3
  integer(4), parameter, public :: idx_dia       = 4
  integer(4), parameter, public :: idx_flag      = 5
  integer(4), parameter, public :: idx_cyano     = 6
  integer(4), parameter, public :: idx_zoo       = 7
  integer(4), parameter, public :: idx_det       = 8
  integer(4), parameter, public :: idx_oxy       = 9
  integer(4), parameter, public :: idx_sil       = 10
  integer(4), parameter, public :: idx_proto     = 11
  integer(4), parameter, public :: idx_eggs      = 12
  integer(4), parameter, public :: idx_naup_m    = 13
  integer(4), parameter, public :: idx_naup_a    = 14
  integer(4), parameter, public :: idx_cop1_m    = 15
  integer(4), parameter, public :: idx_cop1_a    = 16
  integer(4), parameter, public :: idx_cop2_m    = 17
  integer(4), parameter, public :: idx_cop2_a    = 18
  integer(4), parameter, public :: idx_adul_m    = 19
  integer(4), parameter, public :: idx_adul_a    = 20
  integer(4), parameter, public :: idx_eggs_ch   = 21
  integer(4), parameter, public :: idx_naup_m_ch = 22
  integer(4), parameter, public :: idx_naup_a_ch = 23
  integer(4), parameter, public :: idx_cop1_m_ch = 24
  integer(4), parameter, public :: idx_cop1_a_ch = 25
  integer(4), parameter, public :: idx_cop2_m_ch = 26
  integer(4), parameter, public :: idx_cop2_a_ch = 27
  integer(4), parameter, public :: idx_adul_m_ch = 28
  integer(4), parameter, public :: idx_adul_a_ch = 29

  ! public parameters:
  public :: ntbio, Nnorm, Onorm,NOR,ONnitr,NNdenit, ONniup,ONamup

  ! public stage model parameters:
  public :: hatalpha_ch,hatexp_ch,hatch_ch,tempah_ch,tempal_ch,temph_ch,       &
            moa_ch,moc2_ch,moc1_ch,mon_ch,moe_ch,xa_ch,xc2_ch,xc1_ch,xn_ch,    &
            hatalpha,hatexp,hatch,tempah,tempal,temph,templ,moa,moc2,          &
            moc1,mon,moe,xa,xc2,xc1,xn,kma,kmc2,kmc1,kmn,Znorm,Fnorm, templ_ch 

  ! public methods
  public :: ReadBioparam

  namelist/bioparameters/nb, deltao, nue, sigma_b, dn, Ksdenit, Ksnit, rp0,    &
                       rf0, rb0, cyanotll, cyanosll, cyanosul, flagtll, alphap,&
                       alphaf, alphab, iv, graz, toptz, zcl1, p0, f0, b0, z0,  &
                       Imin, att_par_1, att_par_2, q10_rec, anit, pistonvel,   &
                       wdz, wpz, wbz, zooprefdia, zooprefflag, zooprefcyano,   &
                       dn_sed, dn_sed_anox, q10_recs, fnitr, pfrac, Kspburial, &
                       wpz_alpha, pz_lpzd, pz_lpzn, pz_imax, pz_kpz, zooprefpz,&
                       pzprefdia, pzprefflag, pzprefcya, att_par_1_w,          &
                       betan, betac1, betac2, betaa, zooprefzoo,               &
                       w_adult, betan_ch, betac1_ch, betac2_ch, betaa_ch,      &
                       w_adult_ch, stage_model, zpmodel,ecofish
contains

  subroutine ReadBioparam(iu06)

    use exits,          only : exitme
    use io_subs,        only : io_new_unit

    implicit none

    ! args
    integer(4), intent(in) :: iu06

    ! locals
    integer:: ios, lun

    lun = io_new_unit()
    open(lun,file='bioparams.txt',status='old',iostat=ios)
    if (ios /= 0) then
      write(iu06,*) 'Warning: File bioparams.txt not found. Continue with ...'
      write(iu06,*) '        ... default biological parameter values'
    else
      read (unit=lun, nml=bioparameters, iostat=ios)
      if (ios /= 0) then
        write(iu06,*) 'Error: Cannot read values from bioparameters namelist'
        write(iu06,*) 'Error: The iostat failure was: ', ios
        call exitme(1)
      endif
    endif
    close(lun)

  end subroutine ReadBioparam

end module bioparam
