module biozoo
  implicit none

  private

  ! parameters for grazing rates
  real(8), parameter :: arrhen_par    = 3.28_8
  real(8), parameter :: arrhen_ch_par = 5.46_8
  real(8), parameter :: tab_par       = 0.1_8
  real(8), parameter :: tab_ch_par    = 0.1_8
 
  ! parameters for Loss rates respiration C. finmarchicus
  real(8), parameter, private :: lna_par1 = 0.7_8, lna_par2 = 0.05_8
  real(8), parameter, private :: lc1_par1 = 0.64_8, lc1_par2 = 0.02_8
  real(8), parameter, private :: lc2_par1 = 0.64_8, lc2_par2 = 0.02_8
  real(8), parameter, private :: lad_par1 = 0.60_8, lad_par2 = 0.02_8
    
  real(8), parameter, private  :: lna_ch_par1 = 0.70_8, lna_ch_par2 = 0.05_8
  real(8), parameter, private  :: lc1_ch_par1 = 0.64_8, lc1_ch_par2 = 0.02_8
  real(8), parameter, private  :: lc2_ch_par1 = 0.64_8, lc2_ch_par2 = 0.02_8
  real(8), parameter, private  :: lad_ch_par1 = 0.60_8, lad_ch_par2 = 0.02_8

  real(8), parameter, private  :: mea_par1  = 0.072_8, mea_par2  = 0.64_8
  real(8), parameter, private  :: mnb_par1  = 0.102_8, mnb_par2  = 3.291_8
  real(8), parameter, private  :: mc1b_par1 = 1.102_8, mc1b_par2 = 3.291_8
  real(8), parameter, private  :: mc2b_par1 = 1.102_8, mc2b_par2 = 3.291_8
  real(8), parameter, private  :: mab_par1  = 1.4_8, mab_par2 = 0.102_8,       &
                                  mab_par3  = 3.291_8, mab_par4 = 0.003_8

  ! parameters for  C. helgolandicus mortality rates 
  real(8), parameter, private :: mea_ch_par1  = 1.0_8, mea_ch_par2 = 0.072_8,  &
                                 mea_ch_par3  = 0.64_8
  real(8), parameter, private :: mnb_ch_par1  = 1.0_8, mnb_ch_par2 = 1.02_8,   &
                                 mnb_ch_par3  = 3.291_8
  real(8), parameter, private :: mc1b_ch_par1 = 1.0_8, mc1b_ch_par2 = 1.02_8,  &
                                 mc1b_ch_par3 = 3.291_8
  real(8), parameter, private :: mc2b_ch_par1 = 1.0_8, mc2b_ch_par2 = 1.02_8,  &
                                 mc2b_ch_par3 = 3.291_8
  real(8), parameter, private :: mab_ch_par1 = 1.4_8, mab_ch_par2 = 1.02_8,    &
                                 mab_ch_par3 = 3.291_8, mab_ch_par4 = 0.012_8
  
  ! parameters formigration of C. finmar from Norwegian trench in April

  real(8),    parameter, private :: doy_low = 45.0_8, doy_up = 135.0_8
  integer(8), parameter, private :: dep_lim = 44
  real(8),    parameter, private :: migrate_par      = 3.0_8
  real(8),    parameter, private :: migr_cop2m_par   = 27.4_8
  real(8),    parameter, private :: migr_cop2mch_par = 13.5_8
  
  public :: stage_dynamics

contains
  !----------------------------------------------------------------------------!
  !  Stage structured zooplankton model
  !----------------------------------------------------------------------------!
  subroutine stage_dynamics(i,k,kb,dt,doy,ecoa,foodtau,temp,salt,oxy,          &
                            gpress, lzd_hz, lzn_hz)
    
    use bioparam, only : idx_eggs, idx_naup_m, idx_naup_a, idx_cop1_m,         &
                         idx_cop1_a, idx_cop2_m, idx_cop2_a, idx_adul_m,       &
                         idx_adul_a, idx_eggs_ch, idx_naup_m_ch, idx_naup_a_ch,&
                         idx_cop1_m_ch, idx_cop1_a_ch, idx_cop2_m_ch,          &
                         idx_cop2_a_ch, idx_adul_m_ch, idx_adul_a_ch,          &
                         tempal, templ, tempah, temph, betan, kmn, betac1,     &
                         kmc1, betac2, kmc2, betaa, kma, tempal_ch, templ_ch,  &
                         tempah_ch, temph_ch, betan_ch, betac1_ch, betac2_ch,  &
                         betaa_ch, xc2, xc2_ch, hatch, hatalpha, hatexp,       &
                         xc2_ch, moa_ch, hatch_ch, hatalpha_ch, hatexp_ch,     &
                         moe, moa, mon, xn, moc1, xc1, moc2, moe_ch, mon_ch,   &
                         xn_ch, moc1_ch, xc1_ch, moc2_ch, xa, xa_ch, Nnorm,    &
                         swa, swc2, swc1, swn, swa_ch, swc1_ch, swc2_ch,       &
                         swn_ch, Znorm

    implicit none

    ! --------------------------------------------------------------------------
    ! Arguments 
    ! --------------------------------------------------------------------------
    real(8),    intent(inout) :: ecoa(1:,1:)
    !lossrate to amm, det & grazing
    real(8),    intent(out)   :: gpress, lzd_hz, lzn_hz

    integer(4), intent(in)    :: i,k,kb
    real(8),    intent(in)    :: foodtau(:),temp,salt,oxy,dt, doy

    ! --------------------------------------------------------------------------
    ! Local variables 
    ! --------------------------------------------------------------------------

    ! local copies of ecoa array
    real(8)  :: ega,npb,npa,c1b,c1a,c2b,c2a,adb,ada,ega_ch,npb_ch,npa_ch,      &
                c1b_ch,c1a_ch,c2b_ch,c2a_ch,adb_ch,ada_ch

    ! - rates
    real(8)  :: arrhen, arrhen_ch
    real(8)  :: im_na,im_c1,im_c2,im_ad
    real(8)  :: im_na_ch,im_c1_ch,im_c2_ch,im_ad_ch
    real(8)  :: mea,mnb,mna,mc1b,mc1a,mc2b,mc2a,mab,maa
    real(8)  :: mea_ch,mnb_ch,mna_ch,mc1b_ch,mc1a_ch,mc2b_ch
    real(8)  :: mc2a_ch,mab_ch,maa_ch
    real(8)  :: gn,gc1,gc2,ga
    real(8)  :: gn_ch,gc1_ch,gc2_ch,ga_ch
    real(8)  :: lna,lc1,lc2,lad
    real(8)  :: lna_ch,lc1_ch,lc2_ch,lad_ch
    real(8)  :: tea,tnb,tna,tc1b,tc1a,tc2b,tc2a,tab,taa
    real(8)  :: tea_ch,tnb_ch,tna_ch,tc1b_ch,tc1a_ch,tc2b_ch
    real(8)  :: tc2a_ch,tab_ch,taa_ch
    real(8)  :: fx1

    ! other variables
    real(8)  :: tempk
    real(8)  :: food, food_eps

    ! Parameters ---------------------------------------------------------------
    real(8), parameter :: eps     = 0.00000000001_8 ! no division by 0
    real(8), parameter :: cel2kel = 273.15_8 ! Celcius to Kelvin
    real(8), parameter :: zero    = 0.0_8, one  = 1.0_8, two = 2.0_8,          &
                          twenty  = 20.0_8, onetenth = 0.1_8

    ! copy array values
    ega = ecoa(k,idx_eggs)
    npb = ecoa(k,idx_naup_m)
    npa = ecoa(k,idx_naup_a)
    c1b = ecoa(k,idx_cop1_m)
    c1a = ecoa(k,idx_cop1_a)
    c2b = ecoa(k,idx_cop2_m)
    c2a = ecoa(k,idx_cop2_a)
    adb = ecoa(k,idx_adul_m)
    ada = ecoa(k,idx_adul_a)
    ega_ch = ecoa(k,idx_eggs_ch)
    npb_ch = ecoa(k,idx_naup_m_ch)
    npa_ch = ecoa(k,idx_naup_a_ch)
    c1b_ch = ecoa(k,idx_cop1_m_ch)
    c1a_ch = ecoa(k,idx_cop1_a_ch)
    c2b_ch = ecoa(k,idx_cop2_m_ch)
    c2a_ch = ecoa(k,idx_cop2_a_ch)
    adb_ch = ecoa(k,idx_adul_m_ch)
    ada_ch = ecoa(k,idx_adul_a_ch)

    ! init stuff
    food     = foodtau(k)
    food_eps = max(food,eps)
    fx1      = food/food_eps
    
    tempk = temp + cel2kel

    ! Individual mass (norm-mmol-N/m3)/(ind/m3)
    im_na = npb/(npa+eps)
    im_c1 = c1b/(c1a+eps)
    im_c2 = c2b/(c2a+eps)
    im_ad = adb/(ada+eps)

    im_na_ch = npb_ch/(npa_ch+eps)
    im_c1_ch = c1b_ch/(c1a_ch+eps)
    im_c2_ch = c2b_ch/(c2a_ch+eps)
    im_ad_ch = adb_ch/(ada_ch+eps)

    ! --------------------------------------------------------------------------
    ! Grazing rates C. fin. and C. hel.
    ! --------------------------------------------------------------------------
    if(oxy > zero .and. salt > twenty .and. i <= 148) then 
      !FIXME: grid hardcoded

      ! Grazing rates Arrhenius temperature function C. fin.
      arrhen = arrhen_par/(one+exp(tempal/tempk-tempal/templ)+                 &
      exp(tempah/temph-tempah/tempk))

      gn     = betan*arrhen*food**2/(food**2+kmn**2)

      gc1    = betac1*arrhen*food**2/(food**2+kmc1**2)

      gc2    = betac2*arrhen*food**2/(food**2+kmc2**2)

      ga     = betaa*arrhen*food**2/(food**2+kma**2)

      ! Arrhenius temperature function C. hel.
      arrhen_ch = arrhen_ch_par/(1+exp(tempal_ch/tempk-tempal_ch/templ_ch)     &
      +exp(tempah_ch/temph_ch-tempah_ch/tempk))
      gn_ch     = betan_ch*arrhen_ch*food**2/(food**2+(one*kmn)**2)

      gc1_ch    = betac1_ch*arrhen_ch*food**2/(food**2+(one*kmc1)**2)

      gc2_ch    = betac2_ch*arrhen_ch*food**2/(food**2+(one*kmc2)**2)

      ga_ch     = betaa_ch*arrhen_ch*food**2/(food**2+(one*kma)**2)
    else
      ! no grazing under anoxic conditions, salt<20 or below 51N
      gn     = zero
      gc1    = zero
      gc2    = zero
      ga     = zero
      gn_ch  = zero
      gc1_ch = zero
      gc2_ch = zero
      ga_ch  = zero
    endif

    ! --------------------------------------------------------------------------
    ! SEP of C. finmarchicus (transfer)
    ! 50% are females and max-SEP = 20% of BW               
    ! --------------------------------------------------------------------------
    if(im_ad.gt.xc2) then
      tab     = tab_par*(im_ad-xc2)**4/((im_ad-xc2)**4+(moa-xc2)**4)
    else
      tab     = zero
    endif
    
    ! --------------------------------------------------------------------------
    ! Hatching of eggs of C. finmarchicus (Cook et al. 2007)
    ! --------------------------------------------------------------------------
    tea    = one/(hatch*(temp-hatalpha)**hatexp)*ega

    ! --------------------------------------------------------------------------
    ! SEP of eggs C. helgolandicus (transfer)
    !50% are females and max-SEP = 20% of BW
    ! --------------------------------------------------------------------------
     if(im_ad_ch.gt.xc2_ch) then
       tab_ch  = tab_ch_par*(im_ad_ch-xc2_ch)**4/((im_ad_ch-xc2_ch)**4 +       &
                 (moa_ch-xc2_ch)**4)
     else
       tab_ch  = zero
     endif

    ! --------------------------------------------------------------------------
    ! Hatching of eggs C. helgolandicus (Cook et al. 2007)
    ! --------------------------------------------------------------------------
    tea_ch = one/(hatch_ch*(temp-hatalpha_ch)**hatexp_ch)*ega_ch
    
    ! --------------------------------------------------------------------------
    ! Transfer rates C. finmarchicus: high pass filter (Stegert et al. 2007)
    ! --------------------------------------------------------------------------
    taa     = tab*adb/moe
    if(im_na.gt.mon) then
      tnb     = (im_na-mon)**4/((im_na-mon)**4+(((mon+xn)/2)-mon)**4)
      tna     =  tnb*min(npb/xn,npa)
    else
      tnb     = zero
      tna     = zero
    endif

    if(im_c1.gt.moc1) then
      tc1b     = (im_c1-moc1)**4/((im_c1-moc1)**4+(((moc1+xc1)/2)-moc1)**4)
      tc1a     =  tc1b*min(c1b/xc1,c1a)
    else
      tc1b     = zero
      tc1a     = zero
    endif

    if(im_c2.gt.moc2) then
      tc2b     = (im_c2-moc2)**4/((im_c2-moc2)**4+(((moc2+xc2)/2)-moc2)**4)
      tc2a     =  tc2b*min(c2b/xc2,c2a)
    else
      tc2b     = zero
      tc2a     = zero
    endif

    ! --------------------------------------------------------------------------
    ! Transfer rates C. helgolandicus: high pass filter (Stegert et al. 2007)
    ! --------------------------------------------------------------------------
    taa_ch  = tab_ch*adb_ch/moe_ch
    if(im_na_ch.gt.mon_ch) then
      tnb_ch     = (im_na_ch-mon_ch)**4/((im_na_ch-mon_ch)**4+                 &
                   (((mon_ch+xn_ch)/two)-mon_ch)**4)
      tna_ch     =  tnb_ch*min(npb_ch/xn_ch,npa_ch)
    else
      tnb_ch     = zero
      tna_ch     = zero
    endif

    if(im_c1_ch.gt.moc1_ch) then
      tc1b_ch     = (im_c1_ch-moc1_ch)**4/((im_c1_ch-moc1_ch)**4+              &
                    (((moc1_ch+xc1_ch)/two)-moc1_ch)**4)
      tc1a_ch     =  tc1b_ch*min(c1b_ch/xc1_ch,c1a_ch)
    else
      tc1b_ch     = zero
      tc1a_ch     = zero
    endif

    if(im_c2_ch.gt.moc2_ch) then
      tc2b_ch     = (im_c2_ch-moc2_ch)**4/((im_c2_ch-moc2_ch)**4+              &
                    (((moc2_ch+xc2_ch)/two)-moc2_ch)**4)
      tc2a_ch     =  tc2b_ch*min(c2b_ch/xc2_ch,c2a_ch)
    else
      tc2b_ch     = zero
      tc2a_ch     = zero
    endif


    ! --------------------------------------------------------------------------
    ! C. finmarchicus
    !
    ! stage-dependent low-pass filter function (fig.4 in Stegert et al. 2007)-
    ! decrease growth if im approaches molting biomass
    ! --------------------------------------------------------------------------
    if(im_na.gt.(mon+xn)/two) then
      gn = gn*max(zero,one-(im_na-((mon+xn)/two))**2/                          &
           (xn-((mon+xn)/two))**2)
    endif

    if(im_c1.gt.(moc1+xc1)/two) then
      gc1 = gc1*max(zero,one-(im_c1-((moc1+xc1)/two))**2/                      &
            (xc1-((moc1+xc1)/two))**2)
    endif

    if(im_c2.gt.(moc2+xc2)/two) then
      gc2 = gc2*max(zero,one-(im_c2-((moc2+xc2)/two))**2/                      &
            (xc2-((moc2+xc2)/two))**2)
    endif

    if(im_ad.gt.(moa+xa)/two) then
      ga = ga*max(zero,one-(im_ad-((moa+xa)/two))**2/(xa-((moa+xa)/two))**2)
    endif

    ! C. helgolandicus
    if(im_na_ch.gt.(mon_ch+xn_ch)/two) then
      gn_ch = gn_ch*max(zero,one-(im_na_ch-((mon_ch+xn_ch)/two))**2            &
              /(xn_ch-((mon_ch+xn_ch)/two))**2)
    endif

    if(im_c1_ch.gt.(moc1_ch+xc1_ch)/two) then
      gc1_ch = gc1_ch*max(zero,one-(im_c1_ch-((moc1_ch+xc1_ch)/2))**2          &
               /(xc1_ch-((moc1_ch+xc1_ch)/two))**2)
    endif

    if(im_c2_ch.gt.(moc2_ch+xc2_ch)/2) then
      gc2_ch = gc2_ch*max(zero,one-(im_c2_ch-((moc2_ch+xc2_ch)/2))**2          &
               /(xc2_ch-((moc2_ch+xc2_ch)/two))**2)
    endif

    if(im_ad_ch.gt.(moa_ch+xa_ch)/2) then
      ga_ch = ga_ch*max(zero,one-(im_ad_ch-((moa_ch+xa_ch)/2))**2              &
              /(xa_ch-((moa_ch+xa_ch)/two))**2)
    endif

    ! --------------------------------------------------------------------------
    ! Loss rates respiration C. finmarchicus (fraction per day)
    ! egestion= 0.35, respiration: SDA 10% of assimilated (kiørboe et al 1985),
    ! exceretion  10%, moulting 15% & basal resp stegert
    ! no loss for eggs
    ! --------------------------------------------------------------------------
       
    lna = gn* lna_par1 + lna_par2
    lc1 = gc1*lc1_par1 + lc1_par2 
    lc2 = gc2*lc2_par1 + lc2_par2 
    lad = ga* lad_par1 + lad_par2

    ! Loss rates respiration C. helgolandicus
    lna_ch = gn_ch* lna_ch_par1 + lna_ch_par2
    lc1_ch = gc1_ch*lc1_ch_par1 + lc1_ch_par2 
    lc2_ch = gc2_ch*lc2_ch_par1 + lc2_ch_par2
    lad_ch = ga_ch* lad_ch_par1 + lad_ch_par2 

    ! C. finmarchicus Mortality rates (Hirst et al. 2007)
    mea  = exp(mea_par1*temp-mea_par2)
    mnb  = exp(mnb_par1*temp-mnb_par2) !*npb/(npb+0.012/Nnorm)
    mc1b = exp(mc1b_par1*temp-mc1b_par2) !*c1b/(c1b+0.012/Nnorm)
    mc2b = exp(mc2b_par2*temp-mc2b_par2) !*c2b/(c2b+0.012/Nnorm)
    mab  = mab_par1*exp(mab_par2*temp-mab_par3)*adb/(adb+mab_par4/Nnorm)

    ! Mortality rates on abundance
    mna  = mnb
    mc1a = mc1b
    mc2a = mc2b
    maa  = mab

    ! --------------------------------------------------------------------------
    ! C. helgolandicus
    ! Mortality rates on mass
    ! Hirst et al. 2007
    ! --------------------------------------------------------------------------
    mea_ch  = mea_ch_par1*exp(mea_ch_par2*temp-mea_ch_par3)
    !*npb_ch/(npb_ch+0.012/Nnorm)
    mnb_ch  = mnb_ch_par1*exp(mnb_ch_par2*temp-mnb_ch_par3)
    !*c1b_ch/(c1b_ch+0.012/Nnorm) 
    mc1b_ch = mc1b_ch_par1*exp(mc1b_ch_par2*temp-mc1b_ch_par3) 
    !*c2b_ch/(c2b_ch+0.012/Nnorm)
    mc2b_ch = mc2b_ch_par1*exp(mc2b_ch_par2*temp-mc2b_ch_par3)
    mab_ch  = mab_ch_par1*exp(mab_ch_par2*temp-mab_ch_par3)*adb_ch/            &
             (adb_ch+mab_ch_par4/Nnorm)

    ! Mortality rates on abundance
    mna_ch  = mnb_ch
    mc1a_ch = mc1b_ch
    mc2a_ch = mc2b_ch
    maa_ch  = mab_ch

    ! --------------------------------------------------------------------------
    ! grazpress(i,j,k)=grazpress(i,k,j) - mab - lad
    ! lets die out guys suffering from starvation
    ! --------------------------------------------------------------------------
    maa = max(maa,(tc2a - (adb/swa - ada)/dt)/ada)
    if(im_ad < swa) then
      maa = maa + onetenth
    endif
    if(im_c2 < swc2) then
      mc2a = mc2a + onetenth
    endif
    if(im_c1 < swc1) then
      mc1a = mc1a + onetenth
    endif
    if(im_na < swn) then
      mna = mna + onetenth
    endif

    if(im_ad_ch < swa_ch) then
      maa_ch = maa_ch + onetenth
    endif
    if(im_c2_ch < swc2_ch) then
      mc2a_ch = mc2a_ch + onetenth
    endif
    if(im_c1_ch < swc1_ch) then
      mc1a_ch = mc1a_ch + onetenth
    endif
    if(im_na_ch < swn_ch) then
      mna_ch = mna_ch + onetenth
    endif
    
    ! --------------------------------------------------------------------------
    ! Interaction rates with other ERGOM variables
    ! --------------------------------------------------------------------------
    
    ! Total loss rates of Calanus to NH4
    lzn_hz = lna*npb+ lc1*c1b+ lc2*c2b+ lad*adb                                &
           + lna_ch*npb_ch + lc1_ch*c1b_ch + lc2_ch*c2b_ch                     &
           + lad_ch*adb_ch

    ! Total loss rates of Calanus to detritus
    lzd_hz = mea*ega*moe + mnb*npb + mc1b*c1b + mc2b*c2b + mab*adb             &
           + mea_ch*ega_ch*moe_ch + mnb_ch*npb_ch + mc1b_ch*c1b_ch             &
           + mc2b_ch*c2b_ch + mab_ch*adb_ch

    ! Total grazing pressure by Calanus on phytoplankton
    gpress = (gn*npb + gc1*c1b + gc2*c2b + ga*adb + gn_ch*npb +                &
             gc1_ch*c1b + gc2_ch*c2b_ch + ga_ch*adb_ch)/food_eps

    !--------------------------------------------------------------------------!
    !
    !            Stage model tracer dynamics
    !
    !--------------------------------------------------------------------------!

    !-----------------
    !  Biomass C. finmarchicus
    !-----------------
    ecoa(k,idx_naup_m) = ecoa(k,idx_naup_m)                                    &
                       + dt*(                                                  &
                               tea*moe + (gn*fx1 - mnb - lna - tnb)*npb        &
                            )
    ecoa(k,idx_cop1_m) = ecoa(k,idx_cop1_m)                                    &
                       + dt*(                                                  &
                              tnb*npb + (gc1*fx1 - mc1b - lc1 - tc1b)*c1b      &
                            )
    ecoa(k,idx_cop2_m) = ecoa(k,idx_cop2_m)                                    &
                       + dt*(                                                  &
                              tc1b*c1b + (gc2*fx1 - mc2b - lc2 - tc2b)*c2b     &
                            )
    ecoa(k,idx_adul_m) = ecoa(k,idx_adul_m)                                    &
                       + dt*(                                                  &
                              tc2b*c2b + (ga*fx1 - mab - lad - tab) *adb       &
                            )

    !-----------------
    !  Abundance C. finmarchicus
    !-----------------
    ecoa(k,idx_eggs) = ecoa(k,idx_eggs)                                        &
                     + dt*(                                                    &
                            taa -  mea*ega - tea                               &
                          )
    ecoa(k,idx_naup_a) = ecoa(k,idx_naup_a)                                    &
                     + dt*(                                                    &
                            tea - mna*npa - tna                                &
                          )
    ecoa(k,idx_cop1_a) = ecoa(k,idx_cop1_a)                                    &
                     + dt*(                                                    &
                            tna - mc1a*c1a - tc1a                              &
                          )
    ecoa(k,idx_cop2_a) = ecoa(k,idx_cop2_a)                                    &
                     + dt*(                                                    &
                            tc1a - mc2a*c2a - tc2a                             &
                          )
    ecoa(k,idx_adul_a) = ecoa(k,idx_adul_a)                                    &
                     + dt*(                                                    &
                            tc2a - maa*ada                                     &
                          )
    ! migration of C. finmar from Norwegian trench in April !59-145
    ! FIXME: hardcoded to specific grid
    if((doy > doy_low).and.(doy < doy_up).and.(kb > dep_lim)) then
      ecoa(k,idx_cop2_a)    = ecoa(k,idx_cop2_a) + dt*migrate_par 
      ecoa(k,idx_cop2_m)    = ecoa(k,idx_cop2_m) + dt                          &
                              *migrate_par*migr_cop2m_par*Znorm

      ecoa(k,idx_cop2_a_ch) = ecoa(k,idx_cop2_a_ch) +dt*migrate_par
      ecoa(k,idx_cop2_m_ch) = ecoa(k,idx_cop2_m_ch)+ dt                        &
                              *migrate_par*migr_cop2mch_par*Znorm
    endif
    
    !-----------------
    !  Biomass C. helgolandicus
    !-----------------
    ecoa(k,idx_naup_m_ch) = ecoa(k,idx_naup_m_ch)                              &
                     + dt*(                                                    &
                            tea_ch*moe_ch*npb_ch +                             &
                            (                                                  &
                              gn_ch*fx1 - mnb_ch - lna_ch - tnb_ch             &
                            )                                                  &
                          )
    ecoa(k,idx_cop1_m_ch) = ecoa(k,idx_cop1_m_ch)                              &
                     + dt*(                                                    &
                            tnb_ch*npb_ch*c1b_ch +                             &
                            (                                                  &
                              gc1_ch*fx1 - mc1b_ch - lc1_ch - tc1b_ch          &
                            )                                                  &
                          )
    ecoa(k,idx_cop2_m_ch) = ecoa(k,idx_cop2_m_ch)                              &
                     + dt*(                                                    &
                            tc1b_ch*c1b_ch*c2b_ch +                            &
                            (                                                  &
                              gc2_ch*fx1 - mc2b_ch - lc2_ch - tc2b_ch          &
                            )                                                  &
                          )
    ecoa(k,idx_adul_m_ch) = ecoa(k,idx_adul_m_ch)                              &
                     + dt*(                                                    &
                            tc2b_ch*c2b_ch*adb_ch +                            &
                            (                                                  &
                              ga_ch*fx1 - mab_ch - lad_ch - tab_ch             &
                            )                                                  &
                          )
    !-----------------
    !  Abundance C. helgolandicus
    !-----------------
    ecoa(k,idx_eggs_ch) = ecoa(k,idx_eggs_ch)                                  &
                     + dt*(                                                    &
                            taa_ch - mea_ch*ega_ch - tea_ch                    &
                          )
    ecoa(k,idx_naup_a_ch) = ecoa(k,idx_naup_a_ch)                              &
                     +dt*(                                                     &
                           tea_ch - mna_ch*npa_ch - tna_ch                     &
                         )
    ecoa(k,idx_cop1_a_ch) = ecoa(k,idx_cop1_a_ch)                              &
                     +dt*(                                                     &
                           tna_ch - mc1a_ch*c1a_ch - tc1a_ch                   &
                         )
    ecoa(k,idx_cop2_a_ch) = ecoa(k,idx_cop2_a_ch)                              &
                     +dt*(                                                     &
                           tc1a_ch - mc2a_ch*c2a_ch - tc2a_ch                  &
                         )
    ecoa(k,idx_adul_a_ch) = ecoa(k,idx_adul_a_ch)                              &
                     +dt*(                                                     &
                           tc2a_ch - maa_ch*ada_ch                             &
                         )

  end subroutine stage_dynamics

end module biozoo
