module cfl_wrap 
  implicit none
  private
  public :: cfl

contains

  subroutine cfl(ia,dx,dy,d,dt,iw,ind)

    ! calculate the local CFL number, then evaluate max(CFL) < 1.

    use exits,          only : exitme
    use constants,      only : g, zero, one, two
    use dmi_mpi_global, only : iu06

    implicit none

    integer(4), intent(in) :: ia, iw, ind(:,:) 
    real(8), intent(in)    :: dt, dx(:), dy, d(0:)

    integer(4) :: nsrf
    real(8)    :: hm , xm, cflvar, ddy, ddx
    real(8), parameter :: CFLmax = one, dd = 0.001_8

    ddy    = two/(dy*dy)
    cflvar = zero
    do nsrf=1,iw
      ddx    = two/(dx(ind(1,nsrf))*dx(ind(1,nsrf)))
      cflvar = max(cflvar,sqrt(g*max(zero,d(nsrf))*(ddx + ddy)))
    enddo
    cflvar = dt*cflvar
 
    hm = maxval( d(1:iw) )
    xm = min( dy, minval(dx(:)) )

    write(iu06,'(a8,i5,a7,f6.3)') 'Area No.',ia,', CFL =', cflvar
    if (abs(dt - real(int(dt),8)) < dd*dt) then
      write(iu06,'(a6,f8.2,a8,f8.2,a4,i4)')                                    &
        'mindx=', xm, ' maxdpt=', hm, ' dt=', int(dt)
    else
      write(iu06,'(a6,f8.2,a8,f8.2,a4,f8.3)')                                  &
        'mindx=', xm, ' maxdpt=', hm, ' dt=', dt
    endif

    if (cflvar >= CFLmax) call exitme (1,'Error: Time step size too large ')

  end subroutine cfl
end module cfl_wrap 
