module checkuvborder_wrap 
  
  implicit none
  private
  public :: checkuvborder

contains

subroutine checkuvborder ( kg, mmg, krug,krvg, mmf, igaf,jgaf, kfg,            &
                           nu1, nu2, nv1, nv2, iac, iaf )
  !-----------------------------------------------------------------------------
  ! check for errors along u/v borders
  !-----------------------------------------------------------------------------

  !- modules -------------------------------------------------------------------
  use dmi_mpi_global, only : iu06
  use exits,          only : exitme

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in) :: kg 
  integer(4), intent(in) :: mmg(0:,0:,:), mmf(0:,0:,:)
  integer(4), intent(in) :: igaf(:), jgaf(:)
  integer(4), intent(in) :: krug(:,0:)
  integer(4), intent(in) :: krvg(:,0:)
  integer(4), intent(in) :: kfg(:)
  integer(4), intent(in) :: nu1, nu2, nv1, nv2, iac, iaf


  !- local vars ----------------------------------------------------------------
  integer(4) :: ira, i, j, ii, jj, iii, jjj, iff, jf
  integer(4) :: k, kff, kff0, idx, ir
  logical    :: err 


  !- On every entry, assume no errors ------------------------------------------
  err = .false.

  !- Run through U-boundary points ---------------------------------------------
  do ira=max(1,nu1),nu2
    i = krug(1,ira)
    j = krug(2,ira)

    if (krug(3,ira) == 3) j = j+1
    ii = igaf(2)*(i-igaf(1))+1
    jj = jgaf(2)*(j-jgaf(1))+1

    kff0 = 1
    do k=1,kg
      if (mmg(i,j,k) > 0) then
        ir     = 0
        do kff = kff0,kff0+kfg(k)-1
          do iii=0,igaf(2)-1
            do jjj=0,jgaf(2)-1
              iff = ii+iii
              jf  = jj+jjj

              idx = mmf(iff,jf,kff)
              if (idx > 0) ir = ir + 1
            enddo
          enddo
        enddo

        kff0 = kff0+kfg(k)

        if (ir == 0) then
          if (.not.err) then
            write(iu06,*) 'Error: inconsistent number of layers'
            err = .true.
          endif
          write(iu06,*) '  U border point no: ',ira
          write(iu06,*) '  coarse grid no:    ',iac
          write(iu06,*) '  fine   grid no:    ',iaf
          write(iu06,*) '  coarse grid (ijk): ',i,j,k
          write(iu06,*) '  fine   grid (ijk): ',ii,jj,kff0-kfg(k)
          write(iu06,*) ' '
        endif

      endif
    enddo
  enddo


  !- Run through V-boundary points ---------------------------------------------
  do ira=max(1,nv1),nv2
    i = krvg(1,ira)
    j = krvg(2,ira)

    if (krvg(3,ira) == 4) i = i+1
    ii = igaf(2)*(i-igaf(1))+1
    jj = jgaf(2)*(j-jgaf(1))+1

    kff0 = 1
    do k=1,kg
      if (mmg(i,j,k) > 0) then
        ir     = 0
        do kff = kff0,kff0+kfg(k)-1
          do iii=0,igaf(2)-1
            do jjj=0,jgaf(2)-1
              iff = ii+iii
              jf = jj+jjj

              idx = mmf(iff,jf,kff)
              if (idx > 0) ir = ir + 1
            enddo
          enddo
        enddo

        kff0 = kff0+kfg(k)

        if (ir == 0) then
          if (.not.err) then
            write(iu06,*) 'Error: inconsistent number of layers'
            err = .true.
          endif
          write(iu06,*) '  V border point no: ',ira
          write(iu06,*) '  coarse grid no:    ',iac
          write(iu06,*) '  fine   grid no:    ',iaf
          write(iu06,*) '  coarse grid (ijk): ',i,j,k
          write(iu06,*) '  fine   grid (ijk): ',ii,jj,kff0-kfg(k)
          write(iu06,*) ' '
        endif

      endif
    enddo
  enddo

  if (err) call exitme(1)

end subroutine checkuvborder

end module checkuvborder_wrap 
