program cmod_main

!===============================================================================
!
!  CMOD - the Circulation MODel
!
!  - a generalisation of BSHcmod originally developed by Bundesamt fuer 
!    Seeschiffahrt und Hydrographie, Hamburg (thanks guys!)
!
!  - now completely re-written, optimised, parallelised and further developed 
!    at DMI
!
!===============================================================================

  !- Modules -------------------------------------------------------------------
  use dmi_mpi_global,     only : iu06, mpi_io_rank
  use io_control,         only : ios_init, in_hbm, ios_loop, ioc_finalize,     &
                                 io_server, ioc_writerestart, ioc_validate,    &
                                 ioc_readrestart,  io_readrestart
  use dmi_mpi,            only : dmpi_init, dmpi_decompose, dmpi_debug,        &
                                 mpi_io_serial, dmpi_gather, dd, mpi_size,     &
                                 dmpi_halo, ice_serial, ice_rank,              &
                                 dmpi_broadcast, dmpi_scatter,                 &
                                 dmpi_scatter_table, mpi_rank,                 &
                                 nctable, nftable, uctable, vctable,           &
                                 uftable, vftable, cctable, cftable,           &
                                 uhtable, vhtable
  use mpi_task_permute,   only : task_permute, task_global2local_met,          &
                                 task_local_arrays, task_local_khuv,           &
                                 task_permute_index, task_local_numa_ft,       &
                                 task_permute_all
  use cmod_date,          only : caltime
  use cmod_usage,         only : usage
  use adcheck,            only : ADcheck2, ADcheckIce, STlimit
  use cmod_sponge,        only : sponge, sponge_e, sponge_ice_io, sponge_init, &
                                 sponge_io
  use boundary,           only : checkopenbnd, rand_st, bndstcl, bndbclz,      &
                                 bndiceno, daybnd, bndstz, bndzst, bnduvst,    &
                                 bndicez, bndzice, bnduvice, ice_vel
  use exits,              only : exitme
  use dmi_omp,            only : domp_init, domp_estimate_weighted_load
  use cmod_timer,         only : timer, timer_init, timer_destroy, timer_print
  use cmod_init,          only : prep, entr, init_z, init, init_uvdams,        &
                                 initrwqk, inituvwzi, ice_vel_init, kh_numa
  use cmod_params,        only : narea, mainarea, mmx, nmx, kmx, iw2, iw3, mm1,&
                                 krz, iga, jga, kfg, mmxp, nmxp, me1, hz, nz,  &
                                 nthreads, nestingfrom, nestingto, enclosing,  &
                                 znesting, unesting, vnesting, nestinglevels,  &
                                 znest_serial, unest_serial, vnest_serial,     &
                                 cfgfile, nu, nv, nq, timelevel, onthislevel,  &
                                 nud, nvd, nweir, getbasicparams, getparams,   &
                                 PrintComputationalIndex,                      &
                                 ReadElevationReference, elev_ref,             &
                                 un_c_mpi, vn_c_mpi, nweir, nsoilpar
  use cmod_arrays,        only : krq, kru, krv, mm1k, kh, ind, idx, nqlist,    &
                                 khu, khv, lud, lvd, weir_idx, zweir, wflowr,  &
                                 kh_fake, xkoor, ykoor, ti, ft, rt, cosphi,    &
                                 fcorio, ty, tx, gty, gtx, covo, covu, ndtp,   &
                                 dt, dx, dxx, dy, dyy, dz, khbnd,              &
                                 rwz, rwq, rwqk, zetac2,                       &
                                 bndz, bndzk, bndq, bndice, zetac, iceobc,     &
                                 atemp, humid, cloud, pl, wu, wv, precip,      &
                                 atempn, humidn, cloudn, pln, wun, wvn,        &
                                 atempa, humida, clouda, pla, wua, wva,        &
                                 precipn, precipo, h,                          &
                                 z, u, v, cmp, t_soil, avv, tke, diss,         &
                                 a_lastasstime, a_errvar, a_asstemp, a_obserr, &
                                 dxk, dyk, ahc, avc, kdry,ueis, veis, tsnei,   &
                                 ice, casus, nqfile, allocarrays, wrk, wrk2,   &
                                 weir_idx, iota
  use cmod_local_arrays,  only : iw2_l, iw3_l, mm1_l, mm1k_l, me_l, ind_l,     &
                                 kh_l, khu_l, khv_l,                           &
                                 h_old_l, h_new_l, h_i_l, ti_l, hflux_l,       &
                                 eddyh_l, sm1_l, sm2_l, sm3_l, smu_l, smv_l,   &
                                 avv_l, tpe_l, dispv_l,                        &
                                 hz_l, h_l, hn_l, z_l, elev_ref_l, zn_l,       &
                                 hu_l, hv_l, rho_l, ui_l, vi_l, wi_l,          &
                                 u_l, v_l, w_l, cmp_l, press_l, geopot_l,      &
                                 t_soil_l, pla_l, wua_l, wva_l, pln_l,         &
                                 wun_l, wvn_l, pl_l, wu_l, wv_l, atemp_l,      &
                                 cloud_l, humid_l, precip_l, z0srf_l,          &
                                 atempn_l, cloudn_l, atempa_l, clouda_l,       &
                                 humida_l, humidn_l, precipo_l, precipn_l,     &
                                 taux_l, tauy_l, un_l, vn_l, tke_l, diss_l,    &
                                 ice_l, casus_l, tsnei_l, ueis_l, veis_l,      &
                                 evapo_l, alloc_local_arrays, divrad_l,        &
                                 divrad_s_l, uvdam_l, sbfmap_l, tidemap_l,     &
                                 iota_l,iotac
  use constants,          only : aufdat, coldst, dyn, elevation_ref,           &
                                 icedynactive, init_only, isponge, izeits,     &
                                 kalman, konst_set, ldebug, precevap,          &
                                 printturb, rzeite, spinup, starta, nclss,     &
                                 tim, usezbcl, updateSTI, zzeits, zeitstu,     &
                                 nproci, nprocj, coupler_active, pen_type,     &
                                 nicepar, calctrans, weirs, one, half,         &
                                 twentyfour, qflx, coupler_sponge, zbnd_type,  &
                                 shortcircuit, iotamodel, ct10m, niceact,      &
                                 omp_loadbalance_experiment, snow, niotapar,   &
                                 setTSbnd
  use turbmodels,         only : InitTurb
  use io_subs,            only : io_new_unit, flush_unit
  use tflow,              only : tflow_int, tflow_ft
  use eisdyn,             only : eisdyn_int
  use eflow,              only : eflow_int
  use updatearrays,       only : zerofy, h_srfplusconst, hplusconst, set_h_srf,&
                                 set_h
  use thermodynamic,      only : thermodyn, thermodyn_x, copy_iota2iotac
  use cmod_miscvars,      only : cmet, c2, ctim, DoAdvec, DoDiff, dtmain,      &
                                 dticedyn, dtadvec, dteddy, dtday,             &
                                 dtspong, dtthermo, dtvdiff, NDTEDDY, iao, i2, &
                                 ia, iyyy, iday, imon, ihour, ii, iif, iia,    &
                                 indt, nv2, minut, itime, ios, lunsurge,       &
                                 izeita, izeite, jstuna, lun, lmn2, lmn1,lmnts,&
                                 lvalidate, mdatei, NDTTHERMO, nb1, NDTADVEC,  &
                                 nb2, ndt, ndticedyn, NDTVDIFF, nsecs, nv1,    &
                                 ntracers, nu2, nu1, nz1, nz2, sstmin, sstmax, &
                                 sstsat, tim1, tidal, tidal_surge, time, tima, &
                                 tmet, trest, upd, wforc, advecstab,sch,       &
                                 schfactor, ischfac, irc, ioff, rclsurge,      &
                                 DoGather1, DoGather2, DoGatherC, irmaxa
  use cmod_miscsubs,      only : readtask, simtime, checkrivers,               &
                                 checktimestepsizes, freshwatersources,        &
                                 RiverTSsources, timestepsettings, validate
  use assimilation,       only : assimilate
  use restart,            only : ReadRestart, WriteRestart
  use metsubs,            only : metinit, metread, metsave, metnull, metclose, &
                                 meteo_interp_t
  use cmod_hydrodynamics, only : SolveHydrodynamics
  use nesting,            only : copy_f_g, checknesting, copy_f_g_ice, brdcopy
  use cmod_precip,        only : save_old_sfc, tracer_precevap
  use cmod_dens,          only : dens
  use intfeis_wrap,       only : intfeis
  use checkborder_wrap,   only : checkborder
  use checknesting_wrap,  only : checknesting_dyn
  use checkuvborder_wrap, only : checkuvborder
  use cfl_wrap,           only : cfl
  use intface_wrap,       only : intface
  use wlstat_wrap,        only : forward
  use cmod_advecsubs,     only : traceroutflow, resuvwzi
  use mom_heights_wrap,   only : mom_heights
  use coupler,            only : coupler_init, coupler_dynamics, coupler_finish
  use output,             only : output_init, output_close, output_write
  use geopotential,       only : hydrostat
  use transports,         only : transport_init, transport_calc
  use transport_variables,only : NDTTRANS
  use weireqs,            only : weir_diff, weir_init
  use botfricmaps,        only : get_simple_bf_map
  use tideparam,          only : get_tide_map
  use icedyn_drift,       only : icedyn, icedyn_ft, ice_vel_drift
  use icedyn_nest,        only : ice_mom_f_c
  use iflow,              only : iflow_int, iflow_ft
  use t10m_mod,           only : t10m_init
  use old_ice_dyn,        only : iceite, iuvp, muvse1, induvse, iwuvse,        &
                                 taux, tauy, OldIceModelParams

  !- Directives ----------------------------------------------------------------
  implicit none

  !- Local vars ----------------------------------------------------------------
  !  are found in module cmod_miscvars

  !- Initialize MPI ------------------------------------------------------------
  call dmpi_init(.true.)

  !- Scan argument list --------------------------------------------------------
  call usage(ntracers)

  !- Initialize io_control -----------------------------------------------------
  call ios_init()

  !- Initialize openmp ---------------------------------------------------------
  call domp_init(nthreads)

  !- Initialize timers ---------------------------------------------------------
  call timer_init(.true.,nthreads,nclss)

  call timer(1,'init')

  !- Read in the task specification --------------------------------------------
  if (in_hbm) call ReadTask()

  !- Simulation from cta to cte ------------------------------------------------
  if (in_hbm) call SimTime()

  !- Obtain the basic parameters from bathymetry file and specs files ----------
  call GetBasicParams ()

  do ia=1,narea
    call GetParams (ia)
  enddo

  if (in_hbm) then
    if (elevation_ref) then
      call ReadElevationReference(  )
    endif
  endif

  !- Set some MPI module vars based on model specs -----------------------------
  call dmpi_init(ntracers, nz(mainarea), mainarea, nicepar+1, nthreads)

  !- Prepare the MPI decomposition ---------------------------------------------
  !... reading the task decomposition from a file:
  call dmpi_decompose(narea, mmx, nmx, kmx, iw2, iw3, mm1, krz(mainarea)%p,    &
                      nproci, nprocj, cfgfile, nz, nu, nv)
  
  !- Print some MPI debug info -------------------------------------------------
  call dmpi_debug(narea)

  call ios_loop(coupler_active)

  !- Allocate memory for main arrays -------------------------------------------
  do ia=1,narea
    call AllocArrays(ia)
  enddo
  call Alloc_Local_Arrays(narea,nthreads)

  !- Setup the old ice model stuff (TO BE OUTPHASED) ---------------------------
  if (icedynactive .and. (.not.shortcircuit)) then
    call OldIceModelParams(narea)
  endif

  if (zbnd_type == 1) then
    !- Set tidal harmonics, main area only -------------------------------------
    call konst_set (nz(mainarea))
  endif

  !- Read remaining specs, initialise some fields, set up grid quantities ------
  do ia=1,narea
    lun  = io_new_unit()
    open(lun,file=trim(cfgfile(ia)),status='old',iostat=ios)
    if (ios /= 0) then
      write(c2,'(i2.2)') ia
      call exitme (1,'Error: Cannot open configuration file for area '//c2)
    else
      !  ... initialise grid:
      call init( lun,mmx(ia),nmx(ia),mmxp(ia),nmxp(ia),kmx(ia),                &
                 mm1(ia)%p,nestingfrom(ia)%p(0),iga(ia)%p,jga(ia)%p,           &
                 dxk(ia),xkoor(ia)%p,dyk(ia),ykoor(ia)%p,dz(ia)%p,             &
                 dt(ia),dx(ia)%p,dy(ia),ft(ia)%p,rt(ia),                       &
                 ty(ia),tx(ia)%p,gty(ia),gtx(ia)%p,covo(ia)%p,covu(ia)%p,      &
                 z(ia)%p,hz(ia)%p, h(ia)%p, elev_ref(ia)%p,ti(ia)%p,           &
                 nz(ia),nu(ia),kru(ia)%p,nv(ia),krv(ia)%p,nq(ia),krq(ia)%p,    &
                 rwq(ia)%p,nud(ia),lud(ia)%p,nvd(ia),lvd(ia)%p, nweir(ia),     &
                 weir_idx(ia)%p, zweir(ia)%p, wflowr(ia)%p, ahc(ia),avc(ia),   &
                 idx(ia)%p,nestinglevels(ia),znesting(ia)%p, unesting(ia)%p,   &
                 vnesting(ia)%p, narea, znest_serial(ia)%p, unest_serial(ia)%p,&
                 vnest_serial(ia)%p, kdry(ia) )
      close(lun)

      !  ... check rivers:
      call checkrivers(mmxp(ia),nmxp(ia),mm1(ia)%p,nq(ia),krq(ia)%p,hz(ia)%p,  &
                       ia,nqfile(ia),nqlist(ia)%p)
    endif
  enddo
  dtmain = dt(mainarea)

  !  ... get and print the computational intensity of this setup:
  call PrintComputationalIndex(narea, iw3, dt, irmaxa)

  !  ... check time step sizes:
  call CheckTimeStepSizes ()

  !  ... update river inflow:
  if (.not.tidal) call FreshWaterSources ( narea, maxval(nqfile) )
  !-----------------------------------------------------------------------------

  !- do some more initialisations ----------------------------------------------
  do ia=1,narea
    call intface (dxk(ia),dyk(ia),dxx(ia)%p,dyy(ia)%p)

    call prep (mmxp(ia),nmxp(ia),kmx(ia),mm1(ia)%p,mm1k(ia)%p,ind(ia)%p,       &
               iw2(ia),ykoor(ia)%p,dyy(ia)%p,kh_fake(ia)%p,khu(ia)%p,khv(ia)%p,&
               cosphi(ia)%p,fcorio(ia)%p)

    !  ... calc CFL.
    if (ldebug) call cfl(ia,dx(ia)%p,dy(ia),ti(ia)%p,dt(ia),iw2(ia),ind(ia)%p)

    !- Do the weir initialisations:
    if (weirs .and. .not.tidal) then
      call weir_init ( ia, nweir(ia), weir_idx(ia)%p, me1(ia)%p, khu(ia)%p,    &
                       khv(ia)%p, zweir(ia)%p, wflowr(ia)%p )
    endif
  enddo

!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
  ! NUMA first touch layout of kh
  do ia=1,narea
    call kh_numa(iw2(ia), iw3(ia), kh(ia)%p, kh_fake(ia)%p)
  enddo
!$OMP END PARALLEL
  do ia=1,narea
    deallocate( kh_fake(ia)%p )
  enddo
  deallocate( kh_fake )

  !- Check border between fine and coarse grid, and define vertical nesting ----
  do ia=1,narea
    !- run through all nesting levels ii of actual area ia:
    do ii=1,nestinglevels(ia)
      !- check border to inclosed area iia:
      iia = nestingto(ia)%p(ii)
      iao = enclosing(iia,ia) 
      iif = 1
      do while(ia /= nestingfrom(iia)%p(iif))
        iif = iif + 1
      enddo
      nz1 = znesting(iia)%p(iif,1)
      nz2 = znesting(iia)%p(iif,2)
      nu1 = unesting(ia)%p(ii,1)
      nu2 = unesting(ia)%p(ii,2)
      nv1 = vnesting(ia)%p(ii,1)
      nv2 = vnesting(ia)%p(ii,2)

      call checkborder( kmx(ia), mm1(ia)%p, dz(ia)%p, iga(iia)%p(:,iao),       &
                       jga(iia)%p(:,iao), kmx(iia), mm1(iia)%p, dz(iia)%p,     &
                       krz(iia)%p, kfg(ia)%p(:,ii), nz1, nz2, ia, iia)

      call checkuvborder (kmx(ia), mm1(ia)%p, kru(ia)%p,krv(ia)%p, mm1(iia)%p, &
                          iga(iia)%p(:,iao), jga(iia)%p(:,iao),                &
                          kfg(ia)%p(:,ii), nu1, nu2, nv1, nv2, ia, iia )

      !- if dyn=.true.: check additionally the nesting consistency of the set up
      if ( dyn ) then
        call checknesting_dyn ( mmx(ia), nmx(ia), kh(ia)%p, mm1(ia)%p,         &
                                mmx(iia), nmx(iia), kmx(iia), mm1(iia)%p,      &
                                iga(iia)%p(:,iao), jga(iia)%p(:,iao),          &
                                kfg(ia)%p(:,ii), dx(iia)%p, dy(iia), hz(ia)%p, &
                                hz(iia)%p )   
      endif
    enddo
  enddo

  !- Only treat nesting border for areas that are on the same MPI task ---------
  call checknesting()
  !- Set up the mpi-tasks specific scatter table:
  call dmpi_scatter_table(narea, mmx, nmx, iga, jga,                           &
                          nestinglevels, enclosing, nestingto)


  !- Only treat open boundaries that are on this MPI task ----------------------
  call checkopenbnd( kh(mainarea)%p, me1(mainarea)%p, khbnd )


  !- Read the restart file (unless its a coldstart) ----------------------------
  if (io_server .and. io_readrestart) then
    call timer(1,'io_readr')
    call ioc_readrestart(aufdat)
    call timer(2,'io_readr')
    call timer_print('TS - IO read restart took:','io_readr',100)

    !- Initialise some ice model variables and velocities ----------------------
    if (.not.icedynactive) then
      do ia=1,narea
        call ice_vel_init(ueis(ia)%p, veis(ia)%p)
      enddo
    endif

    do ia=1,narea
      ueis_l(ia)%p = 0.0_8
      veis_l(ia)%p = 0.0_8
      call task_permute_all( ia, mmx(ia),nmx(ia),kmx(ia), iw3(ia),iw2(ia),     &
                             mm1(ia)%p, mm1_l(ia)%p, mm1k(ia)%p, mm1k_l(ia)%p, &
                             me1(ia)%p, me_l(ia)%p, ind(ia)%p, ind_l(ia)%p,    &
                             kh(ia)%p, kh_l(ia)%p, hz(ia)%p, hz_l(ia)%p,       &
                             h(ia)%p, h_l(ia)%p, h_old_l(ia)%p, h_new_l(ia)%p, &
                             ti(ia)%p, ti_l(ia)%p, elev_ref(ia)%p,             &
                             elev_ref_l(ia)%p, khu(ia)%p, khv(ia)%p )
    enddo
  else
    !- No IO-server
    call timer(1,'io_readr')
    call ReadRestart( narea, iw2, iw3, ind, kh, mmxp, nmxp, kmx, mm1, hz,      &
                      starta, h, u, v, z, cmp, t_soil,casus, ice, tsnei, ueis, &
                      veis,iota, tke, diss, avv, a_lastasstime, a_errvar,      &
                      aufdat, maxval(iw2(:)) )
    call flush_unit(iu06)
    call timer(2,'io_readr')
    call timer_print('TS - IO read restart took: ','io_readr',100)


    !- Initialise some ice model variables and velocities ----------------------
    if (coldst .or. (.not.icedynactive)) then
      do ia=1,narea
        call ice_vel_init(ueis(ia)%p, veis(ia)%p)
      enddo
    endif


    !- Do the permuting --------------------------------------------------------
    ! do the mpi task permuation here:
    do ia=1,narea
      ! task-permutation of data to present task and its neighbours:
      call task_permute(ia,mm1(ia)%p,iw3(ia),iw2(ia),kh(ia)%p,                 &
                          h(ia)%p,u(ia)%p,v(ia)%p,                             &
                          cmp(ia)%p,ntracers,avv(ia)%p,tke(ia)%p,diss(ia)%p,   &
                          z(ia)%p,t_soil(ia)%p,casus(ia)%p,ice(ia)%p,          &
                          tsnei(ia)%p,ueis(ia)%p,veis(ia)%p,ti(ia)%p,coldst,   &
                          a_lastasstime(ia)%p,a_errvar(ia)%p,kalman,           &
                          coupler_active, iotamodel, iota(ia)%p )
      ! task-permutation of indices and some data to full domain:
      call task_permute_index(ia,mmx(ia),nmx(ia),kmx(ia),mm1(ia)%p,me1(ia)%p,  &
                              mm1k(ia)%p,iw3(ia),iw2(ia),                      &
                              kh(ia)%p,khu(ia)%p,khv(ia)%p,hz(ia)%p,ind(ia)%p)
    enddo
  
!$OMP PARALLEL DEFAULT(shared) PRIVATE(ia)
    do ia=1,narea
      call task_local_numa_ft(ia, kh(ia)%p, mm1(ia)%p, ntracers, nicepar,      &
                              nsoilpar, iw2_l(ia), mm1k_l(ia)%p,mm1_l(ia)%p,   &
                              me_l(ia)%p, ind_l(ia)%p, kh_l(ia)%p, khu_l(ia)%p,&
                              khv_l(ia)%p, u_l(ia)%p, v_l(ia)%p, hz_l(ia)%p,   &
                              h_l(ia)%p, h_old_l(ia)%p, h_new_l(ia)%p,         &
                              cmp_l(ia)%p, avv_l(ia)%p, tke_l(ia)%p,           &
                              diss_l(ia)%p, un_l(ia)%p, vn_l(ia)%p, ui_l(ia)%p,&
                              vi_l(ia)%p, wi_l(ia)%p, hu_l(ia)%p, hv_l(ia)%p,  &
                              rho_l(ia)%p, press_l(ia)%p, geopot_l(ia)%p,      &
                              dispv_l(ia)%p, eddyh_l(ia)%p,                    &
                              smu_l(ia)%p, smv_l(ia)%p,                        &
                              sm1_l(ia)%p, sm2_l(ia)%p, sm3_l(ia)%p,           &
                              z_l(ia)%p, elev_ref_l(ia)%p, zn_l(ia)%p,         &
                              w_l(ia)%p, taux_l(ia)%p, tauy_l(ia)%p,           &
                              casus_l(ia)%p, ueis_l(ia)%p, veis_l(ia)%p,       &
                              ice_l(ia)%p, z0srf_l(ia)%p, uvdam_l(ia)%p,       &
                              tsnei_l(ia)%p, ti_l(ia)%p, t_soil_l(ia)%p,       &
                              idx(ia)%p, niotapar, iotamodel, iota_l(ia)%p,    &
                              iotac(ia)%p )
    enddo
!$OMP END PARALLEL

    !- Make some local arrays, indices, ...:
    do ia=1,narea
      call task_local_arrays (ia, coldst, coupler_active, ntracers,            &
                              mm1(ia)%p, mm1_l(ia)%p, mm1k_l(ia)%p,            &
                              me_l(ia)%p, ind_l(ia)%p, kh(ia)%p, kh_l(ia)%p,   &
                              hz(ia)%p, hz_l(ia)%p, h(ia)%p, h_l(ia)%p,        &
                              h_old_l(ia)%p, h_new_l(ia)%p,                    &
                              z(ia)%p, elev_ref(ia)%p,                         &
                              z_l(ia)%p, elev_ref_l(ia)%p, zn_l(ia)%p,         &
                              ti(ia)%p,ti_l(ia)%p,t_soil(ia)%p, t_soil_l(ia)%p,&
                              u(ia)%p, v(ia)%p, u_l(ia)%p, v_l(ia)%p,w_l(ia)%p,&
                              cmp(ia)%p, cmp_l(ia)%p, avv(ia)%p, avv_l(ia)%p,  &
                              tke(ia)%p, tke_l(ia)%p, diss(ia)%p, diss_l(ia)%p,&
                              casus(ia)%p, casus_l(ia)%p,                      &
                              tsnei(ia)%p,tsnei_l(ia)%p,ice(ia)%p,ice_l(ia)%p, &
                              ueis(ia)%p,ueis_l(ia)%p,veis(ia)%p,veis_l(ia)%p, &
                              iotamodel, iota(ia)%p, iota_l(ia)%p)
    enddo
  endif

  !- Estimate and print weighted load thread decomposition, and exit -----------
  if (omp_loadbalance_experiment) then
    do ia=1,narea
      call domp_estimate_weighted_load (iu06, ia, iw2_l(ia), mm1k_l(ia)%p,     &
                                        ind_l(ia)%p, kh_l(ia)%p, idx(ia)%p)
    enddo 
    call exitme(0,'Finished load balancing experiment ... bailing out')
  endif

!$OMP PARALLEL DEFAULT(shared) PRIVATE(ia)
  !- Set up local khu, khv:
  do ia=1,narea
    call task_local_khuv(iw2_l(ia), mm1k_l(ia)%p, ind_l(ia)%p, kh_l(ia)%p,     &
                         khu_l(ia)%p, khv_l(ia)%p, idx(ia)%p, dd(ia)%halo2)
  enddo
  ia = irmaxa
  call tflow_ft(mm1k_l(ia)%p,iw2_l(ia),ind_l(ia)%p,kh_l(ia)%p,idx(ia)%p,       &
                iw3_l(ia))
  if (icedynactive .and. shortcircuit) then
    call iflow_ft (iw2_l(ia),dd(ia)%halo2,kh_l(ia)%p,2,idx(ia)%p)
    call icedyn_ft(iw2_l(ia),dd(ia)%halo2,kh_l(ia)%p,  idx(ia)%p)
  endif
!$OMP END PARALLEL

  !- if transports should be calculated inline- do the initialisation ----------
  if (calctrans) then
    call transport_init( narea )
  endif

  !- Initialise stretch and shear here, not in each call of deform() -----------
!$OMP PARALLEL DEFAULT(shared) PRIVATE(ia)
  do ia=1,narea
    ! Note: Need not be permuted since the initialisation is zeroes all over, 
    !       regardless of task decompo
    call zerofy(sm1_l(ia)%p, sm2_l(ia)%p, sm3_l(ia)%p, eddyh_l(ia)%p,          &
                smu_l(ia)%p, smv_l(ia)%p,                                      &
                kh_l(ia)%p, iw2_l(ia), mm1k_l(ia)%p, ind_l(ia)%p, idx(ia)%p)
  enddo

  !- Perform 1st time setup of momentum heights --------------------------------
  do ia=1,narea
    call mom_heights(iw2_l(ia),mm1k_l(ia)%p, kh_l(ia)%p, ind_l(ia)%p,          &
                     idx(ia)%p, h_l(ia)%p, hu_l(ia)%p, hv_l(ia)%p )
  enddo
!$OMP BARRIER
  do ia=1,narea
    call dmpi_halo(kmx(ia),mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,                 &
                   hu_l(ia)%p,hv_l(ia)%p)
  enddo
!$OMP END PARALLEL
  do ia=1,narea
    call dmpi_gather(ia,kmx(ia),h_l(ia)%p,h(ia)%p,mpi_io_rank)
    call dmpi_gather(ia,1,ti_l(ia)%p,ti(ia)%p,mpi_io_rank)
  enddo

  !- Initialisation based on the starting fields -------------------------------
!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
  do ia=1,narea
    call init_z (iw2_l(ia), nz(ia), me_l(ia)%p, krz(ia)%p,                     &
                 hz_l(ia)%p, h_l(ia)%p, elev_ref_l(ia)%p, z_l(ia)%p,zn_l(ia)%p,&
                 rwz(ia)%p, zetac)

    !- Initialise also the new h-filed (only if dyn = .true.)
    if (dyn) then
      call set_h (mm1k_l(ia)%p, ind_l(ia)%p, kh_l(ia)%p, h_l(ia)%p, iw2_l(ia), &
                  hn_l(ia)%p, idx(ia)%p)
    endif
  enddo
!$OMP END PARALLEL

  do ia=1,narea
    call initrwqk (rwqk(ia)%p)
    call entr (mm1k_l(ia)%p,ind_l(ia)%p,iw2_l(ia),dxx(ia)%p,dyy(ia)%p,         &
               kh_l(ia)%p,khu_l(ia)%p,khv_l(ia)%p,hu_l(ia)%p,hv_l(ia)%p,       &
               u_l(ia)%p,v_l(ia)%p,w_l(ia)%p,nz(ia),krz(ia)%p,                 &
               nq(ia),krq(ia)%p,rwq(ia)%p,rwqk(ia)%p,cosphi(ia)%p,             &
               un_l(ia)%p, vn_l(ia)%p)
    call init_uvdams( uvdam_l(ia)%p, nud(ia), nvd(ia), lud(ia)%p, lvd(ia)%p,   &
                      mm1k_l(ia)%p, iw2_l(ia), dd(ia)%halo2 )
    if (icedynactive .and. (.not.shortcircuit))                                &
      call intfeis(mmxp(ia), nmxp(ia), me1(ia)%p, muvse1(ia)%p, induvse(ia)%p, &
                 iceite(ia), iuvp(ia)%p, nud(ia), nvd(ia), lud(ia)%p, lvd(ia)%p)
  enddo

     
  !- initialise the water density and (in the default case) the water pressure -
  !    OBS: call of dens requires that starting S,T field is sane  ie. that it
  !         is correctly limited:   0 <= S <= Smax ,   Tfreeze <= T <= Tmax
!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
  do ia=1,narea
    call dens (mm1k_l(ia)%p,ind_l(ia)%p,iw2_l(ia),kh_l(ia)%p,dz(ia)%p,         &
               cmp_l(ia)%p,rho_l(ia)%p,press_l(ia)%p,idx(ia)%p)
  enddo
!$OMP END PARALLEL


  !- Simple bottom friction map ------------------------------------------------
  do ia=1,narea
    call get_simple_bf_map(ia,iw2_l(ia),mmx(ia),nmx(ia),mm1k_l(ia)%p,          &
                           sbfmap_l(ia)%p)
  enddo


  !- Parameterization of effects of tides --------------------------------------
  do ia=1,narea
    call get_tide_map(ia,iw2_l(ia),mmx(ia),nmx(ia),mm1k_l(ia)%p,tidemap_l(ia)%p)
  enddo


  !- T + S input from rivers, or outtake at sinks ------------------------------
  !  ... T,S for river input (S=0)
  do ia=1,narea
    call rand_st (mm1k_l(ia)%p,kh_l(ia)%p,cmp_l(ia)%p,                         &
                  nq(ia),krq(ia)%p,rwqk(ia)%p,bndq(ia)%p)
  enddo

  !  ... update river inflow:
  !FIXME: this only works for sources, it does not support sinks!
  if (.not.tidal) call RiverTSsources ( )

  !-----------------------------------------------------------------------------


  !- Do some boundary data reading ---------------------------------------------
  ia = mainarea
  if (mpi_io_serial) then
    !- Climatological S,T boundaries, read from rwzksbnd/rwzktbnd:
    call bndstcl (kmx(ia),khbnd,nz(ia),bndzk,jstuna)

    !- Baroclinic correction from file (store in zetac2; zetac not used here):
    if (usezbcl) then
      call bndbclz(nz(ia),zetac2,jstuna,zetac,krz(mainarea)%p,nb1,nb2,0)
    endif
  endif

  !- Initialize the sub-model coupler ------------------------------------------
  if (coupler_active) then
    do ia=1,narea
      call dmpi_gather(ia,kmx(ia),rho_l(ia)%p,wrk(ia)%p,mpi_io_rank)
    enddo
    do ia=1,narea
    !FIXME: dmi_ergom and other couplers do not support sinks!
      call coupler_init(ia, narea, iu06, (ia==mainarea), (2),                  &
                        mmx(ia), nmx(ia), kmx(ia), nq(ia), nz(ia), mm1k(ia)%p, &
                        kh(ia)%p, hz(ia)%p, cmp(ia)%p, wrk(ia)%p,              &
                        bndq(ia)%p, bndzk, mpi_io_serial, irc )
      if (irc /= 0) call exitme(irc, 'COUPLER ERROR: Failed to initialise')
    enddo
    ! FIXME: if quantities other than tracers are touched in coupler_init above,
    !        they must be scattered/swapped too.
    do ia=1,narea
      ! scatter only the tracers to task-local arrays (don't touch S and T):
      call dmpi_scatter(ia,iw3(ia),mm1k_l(ia)%p,cmp_l(ia)%p,kh_l(ia)%p,        &
                        mm1k(ia)%p,cmp(ia)%p,(ntracers-2),(2),mpi_io_rank)
    enddo
!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
    do ia=1,narea
      ! a haloswap is generally needed here:
      call dmpi_halo(kmx(ia),mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,cmp_l(ia)%p)
    enddo
!$OMP END PARALLEL
  endif

  !- Broadcast the open boundary data ------------------------------------------
  ia = mainarea
  call dmpi_broadcast( nz(ia), zetac2, nb1, nb2, usezbcl )
  call dmpi_broadcast( ntracers, kmx(ia), nz(ia), bndzk )
  !-----------------------------------------------------------------------------

  call timer(1,'inituvwzi')
!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
! MPI: NO need for halo swap 
  do ia=1,narea
    ! NUMA initialisation, ensure that ui/vi/wi are initialized on the cores 
    !  that afterwards uses them
    call inituvwzi (ui_l(ia)%p,vi_l(ia)%p,wi_l(ia)%p,iw2_l(ia),kh_l(ia)%p,     &
                    ind_l(ia)%p,mm1k_l(ia)%p,idx(ia)%p)
  enddo
!$OMP END PARALLEL
  call timer(2,'inituvwzi')
  call timer_print('TS - inituvwzi took: ','inituvwzi',100)
  

  !- Optionally initialise sponge T&S values from climatology ------------------
  !  Optionally initialise tracer sponges.
  ia = mainarea
  ! FIXME: hmmm, why bother calling with ui, vi which are ALWAYS zero here?
  call sponge_init( kmx(ia), mm1k_l(ia)%p, khu_l(ia)%p, khv_l(ia)%p, nz(ia),   &
                    krz(ia)%p, bndzk, bndz(ia)%p, cmp_l(ia)%p, ui_l(ia)%p,     &
                    vi_l(ia)%p, coupler_active, ntracers )
  if (.not.isponge) then
    if (mpi_io_serial) then
      call sponge_io( kmx(ia), nz(ia), khbnd, bndz(ia)%p, coupler_sponge,      &
                      (.true.))
    endif
    call dmpi_broadcast( ntracers, kmx(ia), nz(ia), bndz(ia)%p )
  endif
  !-----------------------------------------------------------------------------

  !-  Ice climatology and sponges ----------------------------------------------
  ia = mainarea
  if (ice_serial) then
    !-  Ice climatology, read from files:
    call bndiceno( nz(ia), iceobc, jstuna)
    !-  Initialise ice sponges w/zero as defaults:
    call sponge_ice_io( nz(ia), bndice(ia)%p, (.true.))
  endif
  call dmpi_broadcast( (3), nz(ia), iceobc )
  call dmpi_broadcast( (3), nz(ia), bndice(ia)%p )

  !- initialise the output
  call output_init(aufdat,ctim,tim)

  !-----------------------------------------------------------------------------
  !-- Open a number of i/o files
  !-----------------------------------------------------------------------------

  if (zbnd_type == 1) then
    !- open ascii open boundary surge input file -------------------------------
    if ((.not.tidal .or. tidal_surge) .and. nz(mainarea) > 0) then 
      if (mpi_io_serial) then
        lunsurge = io_new_unit()
        rclsurge = 2048
        ii = 35 + nz(mainarea)*7 + 1 
        do while (ii > rclsurge)
          ! adjust record length for surge data file: 
          rclsurge = rclsurge + 1024
        enddo
        open(lunsurge,file='surge',status='old',iostat=ios,pad='NO',           &
             recl=rclsurge)
        if (ios /= 0) call exitme(1, 'Cannot open surge data file' )
        call forward(lunsurge,tima)
      endif
    endif
  endif

  !- Treat meteorology ---------------------------------------------------------
  !    Always, make sure to do a nice initialization for later use:
  do ia=1,narea
    call metnull(pl(ia)%p,wu(ia)%p,wv(ia)%p,                                   &
                 atemp(ia)%p,humid(ia)%p,cloud(ia)%p,precip(ia)%p)
  enddo

  itime=-1
  if (wforc) then 

    call timer(1,'met1')
    !- only access the meteorological input from one task:
    if (mpi_io_serial) then
      !- open and read the meteorological input at starting time (tima):
      call metinit( mdatei )
      tmet = tima
      do while (tmet <= tima)
        itime = itime + 1
        call metread(cmet,tmet,narea,pln,wun,wvn,atempn,humidn,cloudn,precipn)
        if (tmet == tima) then
!$OMP PARALLEL DEFAULT (shared)
          !- save old met. fields:
          call metsave(narea, iw2, .false., .false., dd,                       &
                             pln, wun, wvn, atempn, humidn, cloudn,            &
                             pla, wua, wva, atempa, humida, clouda,            &
                             precipn, precipo )
!$OMP END PARALLEL
        endif
      enddo
    endif
    call dmpi_broadcast(cmet, tmet, itime, mdatei)
    do ia=1,narea
      call dmpi_broadcast(iw2(ia), pln(ia)%p, wun(ia)%p, wvn(ia)%p,            &
                          atempn(ia)%p, humidn(ia)%p, cloudn(ia)%p,            &
                          precipn(ia)%p, precevap)
      call task_global2local_met (ia, mmx(ia), nmx(ia), precevap,              &
                                      me1(ia)%p, me_l(ia)%p,                   &
                                      wun(ia)%p, wvn(ia)%p, pln(ia)%p,         &
                                      atempn(ia)%p, cloudn(ia)%p,              &
                                      humidn(ia)%p, precipn(ia)%p,             &
                                      wun_l(ia)%p, wvn_l(ia)%p, pln_l(ia)%p,   &
                                      atempn_l(ia)%p, cloudn_l(ia)%p,          &
                                      humidn_l(ia)%p, precipn_l(ia)%p )
      call dmpi_broadcast(iw2(ia), pla(ia)%p, wua(ia)%p, wva(ia)%p,            &
                          atempa(ia)%p, humida(ia)%p, clouda(ia)%p,            &
                          precipo(ia)%p, precevap)
      call task_global2local_met (ia, mmx(ia), nmx(ia), precevap,              &
                                      me1(ia)%p, me_l(ia)%p,                   &
                                      wua(ia)%p, wva(ia)%p, pla(ia)%p,         &
                                      atempa(ia)%p, clouda(ia)%p,              &
                                      humida(ia)%p, precipo(ia)%p,             &
                                      wua_l(ia)%p, wva_l(ia)%p, pla_l(ia)%p,   &
                                      atempa_l(ia)%p, clouda_l(ia)%p,          &
                                      humida_l(ia)%p, precipo_l(ia)%p )
    enddo
    call timer(2,'met1')
    call timer_print('TS - IO met took: ','met1',100)
  else
    !- Without meteorology, set up neutral conditions:
    write(iu06,'(a35)') 'running without atmospheric forcing'
    do ia=1,narea
      call metnull(pla_l(ia)%p,wua_l(ia)%p,wva_l(ia)%p,                        &
                   atempa_l(ia)%p,humida_l(ia)%p,clouda_l(ia)%p,precipo_l(ia)%p)
      call metnull(pln_l(ia)%p,wun_l(ia)%p,wvn_l(ia)%p,                        &
                   atempn_l(ia)%p,humidn_l(ia)%p,cloudn_l(ia)%p,precipn_l(ia)%p)
      call metnull(pl_l(ia)%p,wu_l(ia)%p,wv_l(ia)%p,                           &
                   atemp_l(ia)%p,humid_l(ia)%p,cloud_l(ia)%p,precip_l(ia)%p)
    enddo
  endif
  !- end of meteorology --------------------------------------------------------


  !- time step settings --------------------------------------------------------
  call TimeStepSettings ()

  ! ... set time step for ice dynamics:
  dticedyn = ndticedyn*dtmain
  if (ndticedyn /= NDTADVEC*(ndticedyn/NDTADVEC)) then
    write(iu06,'(a57)')                                                        &
                   'ERROR: Time step size for ice dynamics must be an integer'
    write(iu06,'(a57)')                                                        &
                   '       multiple of the time step size for advection      '
    call exitme(1,' ... bailing out')
  endif

  ! ... check time step sizes for precipitation/evaporation:
  if (precevap .and. (NDTADVEC /= NDTTHERMO)) then
    call exitme(1,'Option precevap requires NDTADVEC equal to NDTTHERMO')
  endif


  !- if dyn=true initialise water pressure and geopotential --------------------
  if (dyn) then
!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
    do ia=1,narea
      call hydrostat ( mm1k_l(ia)%p, ind_l(ia)%p, iw2_l(ia), iw3_l(ia),        &
                       kh_l(ia)%p, h_l(ia)%p, rho_l(ia)%p, geopot_l(ia)%p,     &
                       press_l(ia)%p )
    enddo
!$OMP END PARALLEL
  endif

  !- intialise quantities for two-equation turbulence models -------------------
  !  (requires previous initialisation of hz, rho, met- and ice-vars)
  upd      = .false.
  dtthermo = NDTTHERMO*dtmain
  dteddy   = NDTEDDY*dtmain
!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
  !- tke, diss, avv and avt/avs:
  do ia=1,narea
    call InitTurb (coldst, kmx(ia), iw2_l(ia), mm1k_l(ia)%p, ind_l(ia)%p,      &
                   kh_l(ia)%p, h_l(ia)%p, rho_l(ia)%p, u_l(ia)%p, v_l(ia)%p,   &
                   avv_l(ia)%p, dispv_l(ia)%p, tke_l(ia)%p, diss_l(ia)%p,      &
                   z0srf_l(ia)%p, cmp_l(ia)%p, coupler_active, idx(ia)%p,      &
                   dteddy, ia, tidemap_l(ia)%p)
  enddo

  ! experiment: cast to 10m Tair and Qair:
  if (ct10m) call t10m_init()

! barrier here due to different decomp used in InitTurb and thermodyn.
!$OMP BARRIER

  !- heat flux, use dummy time step with "old"="new":
  !  don't update temperature
  if (qflx) then
    do ia=1,narea
      call thermodyn_x(pen_type, mm1k_l(ia)%p, iw2_l(ia), ind_l(ia)%p,         &
                      xkoor(ia)%p, ykoor(ia)%p, h_l(ia)%p,                     &
                      atempa_l(ia)%p, humida_l(ia)%p, clouda_l(ia)%p,          &
                      wua_l(ia)%p, wva_l(ia)%p, casus_l(ia)%p, cmp_l(ia)%p,    &
                      tsnei_l(ia)%p, t_soil_l(ia)%p, dtthermo, ice_l(ia)%p,    &
                      hflux_l(ia)%p, upd, kh_l(ia)%p, evapo_l(ia)%p,           &
                      divrad_l(ia)%p, divrad_s_l(ia)%p, rho_l(ia)%p,           &
                      pla_l(ia)%p, idx(ia)%p, iotamodel, cosphi(ia)%p,         &
                      ti_l(ia)%p, uvdam_l(ia)%p, iota_l(ia)%p, iotac(ia)%p)
    enddo
  else
    do ia=1,narea
      call thermodyn (pen_type, mm1k_l(ia)%p, iw2_l(ia), ind_l(ia)%p,          &
                      xkoor(ia)%p, ykoor(ia)%p, h_l(ia)%p,                     &
                      atempa_l(ia)%p, humida_l(ia)%p, clouda_l(ia)%p,          &
                      wua_l(ia)%p, wva_l(ia)%p,                                &
                      casus_l(ia)%p, cmp_l(ia)%p, tsnei_l(ia)%p,               &
                      t_soil_l(ia)%p, dtthermo, ice_l(ia)%p,                   &
                      hflux_l(ia)%p, upd, kh_l(ia)%p, evapo_l(ia)%p,           &
                      divrad_l(ia)%p, divrad_s_l(ia)%p, snow,                  &
                      cosphi(ia)%p, ti_l(ia)%p, uvdam_l(ia)%p, iotamodel,      &
                      iota_l(ia)%p, iotac(ia)%p ) 
    enddo
  endif    

!$OMP END PARALLEL
  upd = .true.

  !- communicate the initialised quantities:
  if (coldst) then
!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
    do ia=1,narea
      !  need to swap halo of avv (can possibly be avoided later):
      call dmpi_halo(kmx(ia),mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,avv_l(ia)%p)
    enddo
!$OMP END PARALLEL
  else
    iif = nicepar + 1
    if (shortcircuit) then
      !  need to halo swap ice model vars:
!fixme: check if this is really necessary
!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
      do ia=1,narea
        call dmpi_halo(mm1k_l(ia)%p,ia,casus_l(ia)%p)
        call dmpi_halo(1,      mm1k_l(ia)%p,kh_l(ia)%p,ia,  1,  1,1,6,         &
                       ice_l(ia)%p,nicepar)
        call dmpi_halo(kmx(ia),mm1k_l(ia)%p,kh_l(ia)%p,ia,iif,iif,2,6,         &
                       cmp_l(ia)%p,1)
      enddo
      ! iotamodel only when shortcircuit=.true.
      if (iotamodel) then
        do ia=1,narea
          call dmpi_halo(1,mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,3,6,iota_l(ia)%p,3)
        enddo
      endif
!$OMP END PARALLEL
    else
      !  need to gather ice model vars as long as ice model has no MPI support:
      do ia=1,narea
        call dmpi_gather(ia,(1),    tsnei_l(ia)%p,tsnei(ia)%p,ice_rank)
        call dmpi_gather(ia,(1),      ice_l(ia)%p,  ice(ia)%p,3,ice_rank)
        call dmpi_gather(ia,kmx(ia),  cmp_l(ia)%p,  cmp(ia)%p,1,ice_rank)
        call dmpi_gather(ia,        casus_l(ia)%p,casus(ia)%p,ice_rank)
      enddo
!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
      do ia=1,narea
        call dmpi_halo(        mm1k_l(ia)%p,           ia,casus_l(ia)%p)
        call dmpi_halo(1,      mm1k_l(ia)%p,kh_l(ia)%p,ia,  1,  1,1,6,         &
                       ice_l(ia)%p,nicepar)
        call dmpi_halo(kmx(ia),mm1k_l(ia)%p,kh_l(ia)%p,ia,iif,iif,2,6,         &
                       cmp_l(ia)%p,1)
      enddo
!$OMP END PARALLEL
    endif
  endif
  !-----------------------------------------------------------------------------

  !- possibly print stuff for validation:
  if (lvalidate) then
    if (io_server) then
      call ioc_validate()
    else
      do ia=1,narea
        call dmpi_gather(ia,kmx(ia),  u_l(ia)%p,  u(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia),  v_l(ia)%p,  v(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,(1),      z_l(ia)%p,  z(ia)%p,mpi_io_rank)
        call dmpi_gather(ia,kmx(ia),cmp_l(ia)%p,cmp(ia)%p,ntracers,mpi_io_rank)
        if (shortcircuit) then
          call dmpi_gather(ia,(1), ueis_l(ia)%p, ueis(ia)%p,  mpi_io_rank)
          call dmpi_gather(ia,(1), veis_l(ia)%p, veis(ia)%p,  mpi_io_rank)
          call dmpi_gather(ia,    casus_l(ia)%p,casus(ia)%p,  mpi_io_rank)
          call dmpi_gather(ia,(1),  ice_l(ia)%p,  ice(ia)%p,3,mpi_io_rank)
          call dmpi_gather(ia,(1),tsnei_l(ia)%p,tsnei(ia)%p,  mpi_io_rank)
          if (iotamodel) then
             !Note: Currently, the use of the iota model under conditions of
             !      shortcircuit=.false. is not foreseen.
            call dmpi_gather(ia,(1), iota_l(ia)%p, iota(ia)%p,3,mpi_io_rank)
          endif
        endif
      enddo
      if (mpi_io_serial) call Validate ()
    endif
  endif

  !- Prepare for non-blocking mpi-comm around nesting --------------------------
  call dmpi_init( narea, nestingfrom, znesting, krz, kmx, iga, jga,            &
                  mm1k, kh, mm1k_l, kh_l )
  call dmpi_init( narea, nestinglevels, nestingto, enclosing,                  &
                  unesting, vnesting, kru, krv, iga, jga,                      &
                  mm1k, kh, mm1k_l, kh_l )
  call dmpi_init( narea, nestinglevels, nestingto, nestingfrom, enclosing,     &
                  kmx, znesting, znest_serial, krz, iga, jga )

  call timer(2,'init')
  call timer_print('TS - Initialisation took: ','init',1)

  !- possibly terminate when initialization have been performed ----------------
  if (init_only) then
    call flush_unit(iu06)
    call exitme(66,'Initializations completed ... bailing out')
  endif
  !=============================================================================
  !     end of initialisations
  !     start of the time loop
  !=============================================================================
  i2 = 0
  call timer(1,'tloop')
  TIMELOOP: do while (i2 >= 0) 
    ! terminated by EXIT TIMELOOP stmnt inside next loop
    call timer(1,sch//'hstep')
    i2 = i2 + 1
    do izeits=izeita,izeite !- one run through this loop is 6 or 12hr simulation

      call timer(1,'nstep')
 
      !- check if end time is exceeded:
      tim = tim + dtday
      zzeits = zzeits + one
      if ((tim-dtday*schfactor) > time) exit TIMELOOP
      call caltime(tim, iday, imon, iyyy, ihour, minut, nsecs, trest, ctim)


      !- find actual date+time and time step No:
      zeitstu = (tim-tim1)*twentyfour
      lun = io_new_unit()
      if (mpi_io_serial) then
        open (lun,file='zeitschritt')
        write(lun,'(i5,2x,a,f15.8,f10.4)')izeits,ctim,zeitstu
        close(lun)
      endif


      !- Get the new meteorological inputs:
      if (wforc) then
        if (zzeits > rzeite) then
          zzeits = one
          itime  = itime + 1

!$OMP PARALLEL DEFAULT (shared)
          !- save old met. fields:
          call metsave(narea, iw2_l, .false., .true., dd,                      &
                           pln_l,wun_l,wvn_l,atempn_l,humidn_l,cloudn_l,       &
                           pla_l,wua_l,wva_l,atempa_l,humida_l,clouda_l,       &
                           precipn_l,precipo_l )
!$OMP END PARALLEL

          !- read the global met fields:
          if (mpi_io_serial) then
            call metread(cmet,tmet,narea,pln,wun,wvn,atempn,humidn,cloudn,     &
                         precipn)
          endif

          ! FIXME: temporary work around a NAG bug:
          !- need to sync the tasks before extraction to local arrays:
          call timer(1,'MPI_met')
          call dmpi_broadcast(cmet, tmet)

          !- get local met fields from the global ones:
          do ia=1,narea
            call dmpi_broadcast(iw2(ia), pln(ia)%p, wun(ia)%p, wvn(ia)%p,      &
                                atempn(ia)%p, humidn(ia)%p, cloudn(ia)%p,      &
                                precipn(ia)%p, precevap)
          enddo
          call timer(2,'MPI_met')
          do ia=1,narea
            call task_global2local_met(ia, mmx(ia), nmx(ia), precevap,         &
                                       me1(ia)%p, me_l(ia)%p,                  &
                                       wun(ia)%p, wvn(ia)%p, pln(ia)%p,        &
                                       atempn(ia)%p, cloudn(ia)%p,             &
                                       humidn(ia)%p, precipn(ia)%p,            &
                                       wun_l(ia)%p, wvn_l(ia)%p, pln_l(ia)%p,  &
                                       atempn_l(ia)%p, cloudn_l(ia)%p,         &
                                       humidn_l(ia)%p, precipn_l(ia)%p )
          enddo
        endif

!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
! MPI: NO need for halo swap
        do ia = 1,narea
          call meteo_interp_t (iw2_l(ia), dd(ia)%halo2,                        &
                               atempa_l(ia)%p, atempn_l(ia)%p, atemp_l(ia)%p,  &
                               humida_l(ia)%p, humidn_l(ia)%p, humid_l(ia)%p,  &
                               clouda_l(ia)%p, cloudn_l(ia)%p, cloud_l(ia)%p,  &
                               wua_l(ia)%p, wun_l(ia)%p, wu_l(ia)%p,           &
                               wva_l(ia)%p, wvn_l(ia)%p, wv_l(ia)%p,           &
                               pla_l(ia)%p, pln_l(ia)%p, pl_l(ia)%p,           &
                               precipo_l(ia)%p, precipn_l(ia)%p, precip_l(ia)%p)
        enddo
!$OMP END PARALLEL        
! FIXME:  It is assumed that the meteo vars are gathered to global arrays at the
!         places where they are needed. This has not been designed nor tested 
!         carefully yet, so please be careful and please help verifying this
!         approach. 

      endif
        
      !- Get daily values for S/T at lateral open boundaries (lmnts == 0):
      if ( lmnts == 0 .and. izeits == izeita .and.                             &
             (i2 == 1 .or. i2-ioff == ischfac*((i2-ioff)/ischfac)) ) then
        !  the relevant task was selected through checkopenbnd() 
        ia = mainarea
        if (mpi_io_serial) then
          call daybnd (ctim, lmnts, kmx(ia), khbnd, nz(ia), bndzk, setTSbnd,   &
                       bndz(ia)%p)
        endif
        call timer(1,'MPI_bnd')
        call dmpi_broadcast( 2, kmx(ia), nz(ia), bndzk )
        if (setTSbnd) call dmpi_broadcast( 2, kmx(ia), nz(ia), bndz(ia)%p )
        call timer(2,'MPI_bnd')

      !- Or get temporal high resolution S/T values at lateral open boundaries 
      !- (lmnts > 0)
      elseif ( lmnts > 0 ) then 
        if ((mod(izeits,lmnts) == 0) .or. (izeits==izeita .and. i2==1) ) then
          !  the relevant task was selected through checkopenbnd() 
          ia = mainarea
          if (mpi_io_serial) then
            call daybnd (ctim, lmnts, kmx(ia), khbnd, nz(ia), bndzk, setTSbnd, &
                       bndz(ia)%p)
          endif
          call timer(1,'MPI_bnd')
          call dmpi_broadcast( 2, kmx(ia), nz(ia), bndzk )
          if (setTSbnd) call dmpi_broadcast( 2, kmx(ia), nz(ia), bndz(ia)%p )
          call timer(2,'MPI_bnd')
        endif
      endif


      !- every NDTADVEC time step, initialise *i quantities,
      !  and obtain h_old from h:
      if (mod(izeits-1,NDTADVEC) == 0) then
!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
! MPI: NO need for halo swap 
        do ia = 1,narea
          call inituvwzi(ui_l(ia)%p,vi_l(ia)%p,wi_l(ia)%p,iw2_l(ia),kh_l(ia)%p,&
                         ind_l(ia)%p,mm1k_l(ia)%p,idx(ia)%p)
        enddo

        if (dyn) then
! MPI: NO need for halo swap ( set_h )
          do ia = 1,narea
            call set_h (mm1k_l(ia)%p, ind_l(ia)%p, kh_l(ia)%p, h_l(ia)%p,      &
                        iw2_l(ia), h_old_l(ia)%p, idx(ia)%p)
          enddo
        else
! MPI: NO need for halo swap ( set_h_srf )
          do ia = 1,narea
            call set_h_srf (h_l(ia)%p, iw2_l(ia), h_old_l(ia)%p, kh_l(ia)%p,   &
                            idx(ia)%p)

          enddo
        endif
!$OMP END PARALLEL
      endif


      call timer(1,'HD',.true.)
      ! - Solve the hydrodynamic equations -------------------------------------
      call SolveHydrodynamics ( timelevel(mainarea) )
      call timer(2,'HD')
      call timer_print('TS - SolveHydrodynamics took: ','HD',40)

      !- update salt, temperature, ice: ----------------------------------------
      if (.not.tidal .and. updateSTI) then 
        if (mod(izeits, NDTADVEC)  == 0 .or.                                   &
            mod(izeits, NDTVDIFF)  == 0 .or.                                   &
            mod(izeits, NDTTHERMO) == 0      ) then


          call timer(1,'adv_diff')

          !- Do tracer advection:
          if (mod(izeits, NDTADVEC) == 0) then
            dtadvec = NDTADVEC*dtmain
            DoAdvec = .true.
            DoDiff  = .false.


!$OMP PARALLEL DEFAULT (shared)                                                &
!$OMP          PRIVATE (ia,ii,iia,iao,iif,nz1,nz2,nu1,nu2,nv1,nv2,dtspong)
! MPI: NO need for halo swap 
            !- convert cell face fluxes to velocities,
            !  and obtain h_new from h:
            do ia = 1,narea
              call resuvwzi (iw2_l(ia), mm1k_l(ia)%p, ind_l(ia)%p,             &
                             kh_l(ia)%p, khu_l(ia)%p, khv_l(ia)%p,             &
                             hu_l(ia)%p, hv_l(ia)%p, ui_l(ia)%p, vi_l(ia)%p,   &
                             idx(ia)%p)

              if (dyn) then
! MPI: NO need for halo swap ( set_h )
                call set_h (mm1k_l(ia)%p, ind_l(ia)%p, kh_l(ia)%p, h_l(ia)%p,  &
                            iw2_l(ia), h_new_l(ia)%p, idx(ia)%p)
              else
! MPI: NO need for halo swap ( set_h_srf )
                call set_h_srf (h_l(ia)%p, iw2_l(ia), h_new_l(ia)%p,           &
                                kh_l(ia)%p, idx(ia)%p)
              endif

! MPI: NO need for halo swap 
              !- make sure to save the old sst, 
              !  divide by h_new here 'cause we got the correct h_new now:
              if (precevap) call save_old_sfc( iw2_l(ia), cmp_l(ia)%p,         &
                                               h_new_l(ia)%p, tpe_l(ia)%p )
            enddo

!$OMP BARRIER

            if (dyn) then
! MPI: NO need for halo swap
              !- add 1m to layer thickness as a stabilizer:
              do ia=1,narea
                call hplusconst (mm1k_l(ia)%p,ind_l(ia)%p,kh_l(ia)%p,iw2_l(ia),&
                                advecstab,h_old_l(ia)%p,h_new_l(ia)%p,idx(ia)%p)
              enddo
            else
! MPI: NO need for halo swap
              !- add 1m to layer thickness as a stabilizer:
              do ia=1,narea
                call h_srfplusconst (iw2_l(ia), kh_l(ia)%p, advecstab,         &
                                     h_old_l(ia)%p, h_new_l(ia)%p, idx(ia)%p)
              enddo
            endif
!$OMP BARRIER

!FIXME: this only work as intended if NDTVDIFF==NDTADVEC, but hey this has
!       always been the case, so later we MUST assure that NDTVDIFF==NDTADVEC
!       and thereby simplify the code and get rid of most of these 
!       "if (mod(izeits,NDT*) == 0)" statements.
            ! do vertical diffusion:
            if (mod(izeits, NDTVDIFF) == 0 .and. qflx) then
              call timer(1,'tflow_vdif')

              do ia=1,narea
                ! do only vertical diffusion:
                call tflow_int(kmx(ia), mm1k_l(ia)%p, ind_l(ia)%p, iw2_l(ia),  &
                               dxx(ia)%p, dyy(ia)%p, advecstab,NDTVDIFF*dtmain,&
                               kh_l(ia)%p, khu_l(ia)%p, khv_l(ia)%p,           &
                               h_l(ia)%p, hu_l(ia)%p, hv_l(ia)%p,              &
                               h_old_l(ia)%p, h_new_l(ia)%p,                   &
                               ui_l(ia)%p, vi_l(ia)%p, wi_l(ia)%p, cmp_l(ia)%p,&
                               dispv_l(ia)%p, eddyh_l(ia)%p,                   &
                               nz(ia), krz(ia)%p,                              &
                               nq(ia), krq(ia)%p, rwqk(ia)%p,                  &
                               nu(ia), kru(ia)%p, nv(ia), krv(ia)%p,           &
                               bndz(ia)%p, bndq(ia)%p, cosphi(ia)%p,           &
                               .false., .true.,                                &
                               idx(ia)%p, dd(ia)%halo2, ia, dyn, uvdam_l(ia)%p,&
                               .true., hflux_l(ia)%p)
              enddo
              call timer(2,'tflow_vdif')
              call timer_print('TS - tflow_vdiff took: ','tflow_vdif',50)
            endif ! NDTVDIFF

!FIXME: Check if MPI+weir acts like intendet
            if (weirs) then
!$OMP BARRIER
!$OMP MASTER
              do ia=1,narea
                call weir_diff ( nweir(ia), weir_idx(ia)%p, mm1k_l(ia)%p,      &
                                 ui_l(ia)%p, vi_l(ia)%p,uvdam_l(ia)%p )
              enddo
!$OMP END MASTER
            endif

            call timer(1,'MPI_halo')
!$OMP BARRIER
            do ia=1,narea
              call dmpi_halo(kmx(ia),mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,       &
                             h_new_l(ia)%p,h_old_l(ia)%p)
              call dmpi_halo(kmx(ia),mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,       &
                             ui_l(ia)%p,vi_l(ia)%p,wi_l(ia)%p,h_l(ia)%p)
              call dmpi_halo(kmx(ia),mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,       &
                             cmp_l(ia)%p)
            enddo
            call timer(2,'MPI_halo')

            !- sponge layers for S and T:
!$OMP BARRIER
!$OMP MASTER
            dtspong = half*dtadvec
            ia = mainarea
            call sponge(kmx(ia), mm1k_l(ia)%p, khu_l(ia)%p, khv_l(ia)%p,       &
                        nz(ia), krz(ia)%p, bndzk, bndz(ia)%p,                  &
                        cmp_l(ia)%p, ui_l(ia)%p, vi_l(ia)%p,                   &
                        dtspong, dxx(ia)%p, dyy(ia)%p, cosphi(ia)%p)
!$OMP END MASTER
!$OMP BARRIER

            !- do tracer nesting: 
            do ia=1,narea
              do ii=1,nestinglevels(ia)
                !  iia: fine grid
                !  ia:  coarse grid
                iia = nestingto(ia)%p(ii)
                iao = enclosing(iia,ia)
                iif = 1
                do while(ia /= nestingfrom(iia)%p(iif))
                  iif = iif + 1
                enddo

                !  pre-post RECV:
                if (mpi_io_serial) then
                  nz1 = znest_serial(iia)%p(iif,1)
                  nz2 = znest_serial(iia)%p(iif,2)
                  call timer(1,'MPI_nest_7')
                  call dmpi_gather(ia, ntracers, kmx(iia), nz1, nz2,krz(iia)%p,&
                            bndz(iia)%p, iga(iia)%p(:,iao), jga(iia)%p(:,iao), &
                            ii, (1) ) 
                  call timer(2,'MPI_nest_7')
                endif

                !- Tracer boundary values coarse grid-->fine grid:
                !- step 1: find coarse grid values just outside border:
                if (nctable(mpi_rank+1,ia,iia)) then
                  nz1 = znesting(iia)%p(iif,1)
                  nz2 = znesting(iia)%p(iif,2)
                  call bndstz(mm1k_l(ia)%p, iga(iia)%p(:,iao),                 &
                              jga(iia)%p(:,iao), cmp_l(ia)%p,                  &
                              kmx(iia), mm1k(iia)%p, krz(iia)%p, bndz(iia)%p,  &
                              kfg(ia)%p(:,ii), nz1, nz2 )
!$OMP BARRIER
                endif

                call timer(1,'MPI_nest_7')
                if (mpi_io_serial) then
                  !  Wait for the RECV requests and decode to bndz:
                  nz1 = znest_serial(iia)%p(iif,1)
                  nz2 = znest_serial(iia)%p(iif,2)
                  call dmpi_gather(ia, ntracers, kmx(iia), nz1, nz2,krz(iia)%p,&
                            bndz(iia)%p, iga(iia)%p(:,iao), jga(iia)%p(:,iao), &
                            ii, (3) ) 
!$OMP BARRIER
                elseif (nctable(mpi_rank+1,ia,iia)) then
                  !  Encode bndz and SEND:
                  nz1 = znesting(iia)%p(iif,1)
                  nz2 = znesting(iia)%p(iif,2)
                  call dmpi_gather(ia, ntracers, kmx(iia), nz1, nz2,krz(iia)%p,&
                            bndz(iia)%p, iga(iia)%p(:,iao), jga(iia)%p(:,iao), &
                            ii, (2) ) 
!$OMP BARRIER
                endif
                if (mpi_io_serial .or. nftable(mpi_rank+1,ia,iia)) then
                  call timer(1,'MPI_nest7s')
                  nz1 = znest_serial(iia)%p(iif,1)
                  nz2 = znest_serial(iia)%p(iif,2)
                  call dmpi_scatter(ia, iia, ii, ntracers, kmx(iia), nz1, nz2, &
                                    krz(iia)%p, bndz(iia)%p)
                  call timer(2,'MPI_nest7s')
                endif
                call timer(2,'MPI_nest_7')

                !- Tracer boundary values coarse grid-->fine grid:
                !- step 2: advect boundary values at inflow points:
                if (nftable(mpi_rank+1,ia,iia)) then
                  nz1 = znesting(iia)%p(iif,1)
                  nz2 = znesting(iia)%p(iif,2)
                  call bndzst(mm1k_l(iia)%p, kh_l(iia)%p, krz(iia)%p,          &
                              bndz(iia)%p,cmp_l(iia)%p,ui_l(iia)%p,vi_l(iia)%p,&
                              dtadvec,dxx(iia)%p,dyy(iia)%p,cosphi(iia)%p,nz1, &
                              nz2)
!$OMP BARRIER
                endif

                !- Tracer boundary values fine grid-->coarse grid:
                !  (previous steps 1+2 merged)
                !- step 1: coarse grid S/T by cell-averaging from fine grid:
                !- step 2: advect boundary values at inflow points (f-->c):
                if (uctable(mpi_rank+1,ia,iia) .or.                            &
                    vctable(mpi_rank+1,ia,iia)) then
                  nu1 = un_c_mpi(ia)%p(ii,1)
                  nu2 = un_c_mpi(ia)%p(ii,2)
                  nv1 = vn_c_mpi(ia)%p(ii,1)
                  nv2 = vn_c_mpi(ia)%p(ii,2)
                  call bnduvst(mm1k_l(ia)%p, kh_l(ia)%p, kru(ia)%p, krv(ia)%p, &
                               cmp_l(ia)%p, ui_l(ia)%p, vi_l(ia)%p, dtadvec,   &
                               dxx(ia)%p,dyy(ia)%p,cosphi(ia)%p,nu1,nu2,nv1,nv2)
!$OMP BARRIER
                endif

                !  Wait for the SEND requests.
                !  (stickly this must be there; in reality we might not need it)
                if (nctable(mpi_rank+1,ia,iia)) then
                  nz1 = znesting(iia)%p(iif,1)
                  nz2 = znesting(iia)%p(iif,2)
                  call timer(1,'MPI_nest_7')
                  call dmpi_gather(ia, ntracers, kmx(iia), nz1, nz2,krz(iia)%p,&
                            bndz(iia)%p, iga(iia)%p(:,iao), jga(iia)%p(:,iao), &
                            ii, (4) ) 
                  call timer(2,'MPI_nest_7')
                endif

              enddo
            enddo

!$OMP BARRIER
            call timer(1,'MPI_halo')
            do ia=1,narea
! FIXME: consider swapping only tasks needed to be swapped:
              call dmpi_halo(kmx(ia),mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,       &
                             cmp_l(ia)%p)
            enddo
            call timer(2,'MPI_halo')
!$OMP END PARALLEL
            
            call timer(1,'tflow_adv')

            !  Run through all areas and do advection
            do ia=1,narea

!$OMP PARALLEL DEFAULT (shared)
! MPI: Halo swap already avail: ui, vi
              !- check stability limits:
              call ADcheck2(mm1k_l(ia)%p, iw2_l(ia), kh_l(ia)%p, dtadvec,      &
                            ui_l(ia)%p, vi_l(ia)%p, wi_l(ia)%p,dx(ia)%p,dy(ia),&
                            h_old_l(ia)%p, hu_l(ia)%p, hv_l(ia)%p, ndtp,       &
                            (NDTADVEC*timelevel(ia)), ind_l(ia)%p, ia,idx(ia)%p)
!$OMP END PARALLEL

              !- update the temperature and salinity:
              ndt = maxval( ndtp(:) )
! FIXME: With mpi_size > 1, this only works fully correct if all tasks return
!        the same ndt value. If tasks return different ndt, the code will 
!        deadlock in the halo-swap below. Therefore, 'till a solution has been
!        implemented, we will not allow ndt /= 1 when playing with any other 
!        number of tasks than one; we will force an exit here, leaving the rest
!        of the code as is for now:
              if (mpi_size > 1 .and. ndt /= 1) then
                write(iu06,*)'Cannot sub-divide advection time step.'
                write(iu06,*)'Decrease the advection time step size and re-run.'
                call exitme(1,'Advection time step size too large. Bailing out')
              endif

              if (ndt > 1) then
                dtadvec  = dtadvec/real(ndt,8)
                h_i_l(ia)%p(:) = (h_new_l(ia)%p(:)-h_old_l(ia)%p(:))/real(ndt,8)
              endif
              do indt=1,ndt
                if (ndt > 1) then
                  h_new_l(ia)%p(:) = h_old_l(ia)%p(:) + h_i_l(ia)%p(:)
 
! MPI: Halo swap needed for: cmp for subsequent sub-time steps
                  if (indt >= 2) then
                    call timer(1,'halo_2')
!$OMP PARALLEL DEFAULT (shared)
                    call dmpi_halo(kmx(ia),mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6, &
                                   cmp_l(ia)%p)
!$OMP END PARALLEL
                    call timer(2,'halo_2')
                    call timer_print('TS - halo_2 took: ','halo_2',100)
                  endif
                endif

!$OMP PARALLEL DEFAULT (shared)
                ! do the advection:
                call tflow_int(kmx(ia), mm1k_l(ia)%p, ind_l(ia)%p, iw2_l(ia),  &
                               dxx(ia)%p, dyy(ia)%p, advecstab, dtadvec,       &
                               kh_l(ia)%p, khu_l(ia)%p, khv_l(ia)%p,           &
                               h_l(ia)%p, hu_l(ia)%p, hv_l(ia)%p,              &
                               h_old_l(ia)%p, h_new_l(ia)%p,                   &
                               ui_l(ia)%p, vi_l(ia)%p, wi_l(ia)%p, cmp_l(ia)%p,&
                               dispv_l(ia)%p, eddyh_l(ia)%p,                   &
                               nz(ia), krz(ia)%p,                              &
                               nq(ia), krq(ia)%p, rwqk(ia)%p,                  &
                               nu(ia), kru(ia)%p, nv(ia), krv(ia)%p,           &
                               bndz(ia)%p, bndq(ia)%p, cosphi(ia)%p,           &
                               DoAdvec, DoDiff,                                &
                               idx(ia)%p, dd(ia)%halo2, ia, dyn, uvdam_l(ia)%p)
!$OMP END PARALLEL

! fixme: need a halo swap of cmp_l if we generalize OBC and nesting across tasks
!        (at least those tasks involved in OBC treament)
                call traceroutflow(ntracers, nz(ia), krz(ia)%p,                &
                                   kh_l(ia)%p, mm1k_l(ia)%p, dtadvec,          &
                                   dxx(ia)%p, dyy(ia)%p, cosphi(ia)%p,         &
                                   ui_l(ia)%p, vi_l(ia)%p, cmp_l(ia)%p)
                if (ndt > 1) h_old_l(ia)%p(:) = h_new_l(ia)%p(:)
              enddo
              dtadvec = dtadvec*real(ndt,8)

              !- add the precipitation/evaporation;
              !  only temp is treated here, all other tracers have zero
              !  contration in precip/evapo.
              if (precevap) then
!$OMP PARALLEL DEFAULT (shared)
! MPI: NO need for halo swap
                call tracer_precevap( iw2_l(ia), dtadvec,                      &
                                      tpe_l(ia)%p, tpe_l(ia)%p,                &
                                      precip_l(ia)%p, evapo_l(ia)%p,           &
                                      cmp_l(ia)%p, 1 )
!$OMP END PARALLEL
              endif
            enddo
            call timer(2,'tflow_adv')
            call timer_print('TS - tflow_adv took: ','tflow_adv',50)

          endif ! NDTADVEC

          !- Do tracer diffusion:
          if (mod(izeits, NDTVDIFF) == 0) then
            dtvdiff = NDTVDIFF*dtmain
            DoAdvec = .false.
            DoDiff  = .true.
            call timer(1,'tflow_diff')

            !- do tracer nesting 'coz it was likely modified during advection:
!$OMP PARALLEL DEFAULT (shared) PRIVATE (ia, ii, iia, iao, nu1, nu2, nv1, nv2)

! MPI: Halo swap needed for: cmp 1:ntracers
            call timer(1,'MPI_halo')
            do ia=1,narea
              call dmpi_halo(kmx(ia),mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,       &
                             cmp_l(ia)%p)
            enddo
!$OMP BARRIER
            call timer(2,'MPI_halo')

            do ia=1,narea
              do ii=1,nestinglevels(ia)
                !  iia: fine grid
                !  ia:  coarse grid
                iia = nestingto(ia)%p(ii)
                iao = enclosing(iia,ia)
                nu1 = unesting(ia)%p(ii,1)
                nu2 = unesting(ia)%p(ii,2)
                nv1 = vnesting(ia)%p(ii,1)
                nv2 = vnesting(ia)%p(ii,2)

                call timer(1,'MPI_nest_8')
                if ((nu1 <= nu2 .or. uftable(mpi_rank+1,ia,iia)) .or.          &
                    (nv1 <= nv2 .or. vftable(mpi_rank+1,ia,iia))     ) then

                  call timer(1,'MPI_nest8a')
                  if (nu1 <= nu2 .or. uftable(mpi_rank+1,ia,iia))              &
                    ! gather fine h_l along ubrd, RECV
                    call dmpi_gather(ia,iia,nu1,nu2,kru(ia)%p,(0),             &
                                     uctable(mpi_rank+1,ia,iia),               &
                                     uhtable(ia,iia),kmx(iia),                 &
                                     mm1k_l(iia)%p,mm1k(iia)%p,mm1k(ia)%p,     &
                                     kh_l(iia)%p,kh(iia)%p,                    &
                                     iga(iia)%p(:,iao),jga(iia)%p(:,iao),      &
                                     (1),ii,(1),h_l(iia)%p,h(iia)%p)
                  if (nv1 <= nv2 .or. vftable(mpi_rank+1,ia,iia))              &
                    ! gather fine h_l along vbrd, RECV
                    call dmpi_gather(ia,iia,nv1,nv2,krv(ia)%p,(0),             &
                                     vctable(mpi_rank+1,ia,iia),               &
                                     vhtable(ia,iia),kmx(iia),                 &
                                     mm1k_l(iia)%p,mm1k(iia)%p,mm1k(ia)%p,     &
                                     kh_l(iia)%p,kh(iia)%p,                    &
                                     iga(iia)%p(:,iao),jga(iia)%p(:,iao),      &
                                     (2),ii,(1),h_l(iia)%p,h(iia)%p)
                  call timer(2,'MPI_nest8a')

                  call timer(1,'MPI_nest8b')
                  if (nu1 <= nu2 .or. uftable(mpi_rank+1,ia,iia)) then
                    ! gather fine cmp_l along ubrd, RECV
                    call dmpi_gather(ia,iia,nu1,nu2,kru(ia)%p,                 &
                                   uctable(mpi_rank+1,ia,iia), uhtable(ia,iia),&
                                   ntracers,kmx(iia),mm1k_l(iia)%p,mm1k(iia)%p,&
                                   mm1k(ia)%p, kh_l(iia)%p, kh(iia)%p,         &
                                   iga(iia)%p(:,iao),jga(iia)%p(:,iao),(1),    &
                                   ii,(1),cmp_l(iia)%p,cmp(iia)%p)
                  endif
                  if (nv1 <= nv2 .or. vftable(mpi_rank+1,ia,iia)) then
                    ! gather fine cmp_l along vbrd, RECV
                    call dmpi_gather(ia,iia,nv1,nv2,krv(ia)%p,                 &
                                   vctable(mpi_rank+1,ia,iia), vhtable(ia,iia),&
                                   ntracers,kmx(iia),mm1k_l(iia)%p,mm1k(iia)%p,&
                                   mm1k(ia)%p, kh_l(iia)%p, kh(iia)%p,         &
                                   iga(iia)%p(:,iao),jga(iia)%p(:,iao),(2),    &
                                   ii,(1),cmp_l(iia)%p,cmp(iia)%p)
                  endif
                  call timer(2,'MPI_nest8b')

                  call timer(1,'MPI_nest8a')
                  if (nu1 <= nu2 .or. uftable(mpi_rank+1,ia,iia))              &
                    ! gather fine h_l along ubrd, encode+SEND
                    call dmpi_gather(ia,iia,nu1,nu2,kru(ia)%p,(0),             &
                                     uctable(mpi_rank+1,ia,iia),               &
                                     uhtable(ia,iia),kmx(iia),                 &
                                     mm1k_l(iia)%p,mm1k(iia)%p,mm1k(ia)%p,     &
                                     kh_l(iia)%p,kh(iia)%p,                    &
                                     iga(iia)%p(:,iao),jga(iia)%p(:,iao),      &
                                     (1),ii,(2),h_l(iia)%p,h(iia)%p)
                  if (nv1 <= nv2 .or. vftable(mpi_rank+1,ia,iia))              &
                    ! gather fine h_l along vbrd, encode+SEND
                    call dmpi_gather(ia,iia,nv1,nv2,krv(ia)%p,(0),             &
                                     vctable(mpi_rank+1,ia,iia),               &
                                     vhtable(ia,iia),kmx(iia),                 &
                                     mm1k_l(iia)%p,mm1k(iia)%p,mm1k(ia)%p,     &
                                     kh_l(iia)%p,kh(iia)%p,                    &
                                     iga(iia)%p(:,iao),jga(iia)%p(:,iao),      &
                                     (2),ii,(2),h_l(iia)%p,h(iia)%p)
                  call timer(2,'MPI_nest8a')

                  call timer(1,'MPI_nest8b')
                  if (nu1 <= nu2 .or. uftable(mpi_rank+1,ia,iia)) then
                    ! gather fine cmp_l along ubrd, encode+SEND
                    call dmpi_gather(ia,iia,nu1,nu2,kru(ia)%p,                 &
                                   uctable(mpi_rank+1,ia,iia), uhtable(ia,iia),&
                                   ntracers,kmx(iia),mm1k_l(iia)%p,mm1k(iia)%p,&
                                   mm1k(ia)%p, kh_l(iia)%p, kh(iia)%p,         &
                                   iga(iia)%p(:,iao),jga(iia)%p(:,iao),(1),    &
                                   ii,(2),cmp_l(iia)%p,cmp(iia)%p)
                  endif
                  if (nv1 <= nv2 .or. vftable(mpi_rank+1,ia,iia)) then
                    ! gather fine cmp_l along vbrd, encode+SEND
                    call dmpi_gather(ia,iia,nv1,nv2,krv(ia)%p,                 &
                                   vctable(mpi_rank+1,ia,iia), vhtable(ia,iia),&
                                   ntracers,kmx(iia),mm1k_l(iia)%p,mm1k(iia)%p,&
                                   mm1k(ia)%p, kh_l(iia)%p, kh(iia)%p,         &
                                   iga(iia)%p(:,iao),jga(iia)%p(:,iao),(2),    &
                                   ii,(2),cmp_l(iia)%p,cmp(iia)%p)
                  endif
                  call timer(2,'MPI_nest8b')

                  call timer(1,'MPI_nest8a')
                  if (nu1 <= nu2 .or. uftable(mpi_rank+1,ia,iia))              &
                    ! gather fine h_l along ubrd, WAIT+decode
                    call dmpi_gather(ia,iia,nu1,nu2,kru(ia)%p,(0),             &
                                     uctable(mpi_rank+1,ia,iia),               &
                                     uhtable(ia,iia),kmx(iia),                 &
                                     mm1k_l(iia)%p,mm1k(iia)%p,mm1k(ia)%p,     &
                                     kh_l(iia)%p,kh(iia)%p,                    &
                                     iga(iia)%p(:,iao),jga(iia)%p(:,iao),      &
                                     (1),ii,(3),h_l(iia)%p,h(iia)%p)
                  if (nv1 <= nv2 .or. vftable(mpi_rank+1,ia,iia))              &
                    ! gather fine h_l along vbrd, WAIT+decode
                    call dmpi_gather(ia,iia,nv1,nv2,krv(ia)%p,(0),             &
                                     vctable(mpi_rank+1,ia,iia),               &
                                     vhtable(ia,iia),kmx(iia),                 &
                                     mm1k_l(iia)%p,mm1k(iia)%p,mm1k(ia)%p,     &
                                     kh_l(iia)%p,kh(iia)%p,                    &
                                     iga(iia)%p(:,iao),jga(iia)%p(:,iao),      &
                                     (2),ii,(3),h_l(iia)%p,h(iia)%p)
                  call timer(2,'MPI_nest8a')

                  call timer(1,'MPI_nest8b')
                  if (nu1 <= nu2 .or. uftable(mpi_rank+1,ia,iia)) then
                    ! gather fine cmp_l along ubrd, WAIT+decode
                    call dmpi_gather(ia,iia,nu1,nu2,kru(ia)%p,                 &
                                   uctable(mpi_rank+1,ia,iia), uhtable(ia,iia),&
                                   ntracers,kmx(iia),mm1k_l(iia)%p,mm1k(iia)%p,&
                                   mm1k(ia)%p, kh_l(iia)%p, kh(iia)%p,         &
                                   iga(iia)%p(:,iao),jga(iia)%p(:,iao),(1),    &
                                   ii,(3),cmp_l(iia)%p,cmp(iia)%p)
                  endif
                  if (nv1 <= nv2 .or. vftable(mpi_rank+1,ia,iia)) then
                    ! gather fine cmp_l along vbrd, WAIT+decode
                    call dmpi_gather(ia,iia,nv1,nv2,krv(ia)%p,                 &
                                   vctable(mpi_rank+1,ia,iia), vhtable(ia,iia),&
                                   ntracers,kmx(iia),mm1k_l(iia)%p,mm1k(iia)%p,&
                                   mm1k(ia)%p, kh_l(iia)%p, kh(iia)%p,         &
                                   iga(iia)%p(:,iao),jga(iia)%p(:,iao),(2),    &
                                   ii,(3),cmp_l(iia)%p,cmp(iia)%p)
                  endif
                  call timer(2,'MPI_nest8b')
!$OMP BARRIER
                endif
                call timer(2,'MPI_nest_8')

                !- Tracer boundary values fine grid-->coarse grid:
                call brdcopy(ia, kmx(ia), mm1k_l(ia)%p, kru(ia)%p, krv(ia)%p,  &
                             cmp_l(ia)%p, mm1k(iia)%p,                         &
                             iga(iia)%p(:,iao), jga(iia)%p(:,iao),             &
                             cmp(iia)%p, kfg(ia)%p(:,ii), h(iia)%p,            &
                             dx(iia)%p, dy(iia), nu1, nu2, nv1, nv2, ntracers )
              enddo
              if (nestinglevels(ia) > 0) then
!$OMP BARRIER
                call timer(1,'MPI_nest_9')
                call dmpi_halo(kmx(ia),mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,     &
                               cmp_l(ia)%p)
                call timer(2,'MPI_nest_9')
              endif
            enddo

!$OMP BARRIER
            ! do the diffusion:
            do ia=1,narea
              if (qflx) then
                ! do only horizontal diffusion:
                call tflow_int(kmx(ia), mm1k_l(ia)%p, ind_l(ia)%p, iw2_l(ia),  &
                               dxx(ia)%p, dyy(ia)%p, advecstab, dtvdiff,       &
                               kh_l(ia)%p, khu_l(ia)%p, khv_l(ia)%p,           &
                               h_l(ia)%p, hu_l(ia)%p, hv_l(ia)%p,              &
                               h_old_l(ia)%p, h_new_l(ia)%p,                   &
                               ui_l(ia)%p, vi_l(ia)%p, wi_l(ia)%p, cmp_l(ia)%p,&
                               dispv_l(ia)%p, eddyh_l(ia)%p,                   &
                               nz(ia), krz(ia)%p,                              &
                               nq(ia), krq(ia)%p, rwqk(ia)%p,                  &
                               nu(ia), kru(ia)%p, nv(ia), krv(ia)%p,           &
                               bndz(ia)%p, bndq(ia)%p, cosphi(ia)%p,           &
                               DoAdvec, DoDiff,                                &
                               idx(ia)%p, dd(ia)%halo2, ia, dyn, uvdam_l(ia)%p,&
                               .false.)
              else
                ! do both horizontal and vertical diffusion:
                call tflow_int(kmx(ia), mm1k_l(ia)%p, ind_l(ia)%p, iw2_l(ia),  &
                             dxx(ia)%p, dyy(ia)%p, advecstab, dtvdiff,         &
                             kh_l(ia)%p, khu_l(ia)%p, khv_l(ia)%p,             &
                             h_l(ia)%p, hu_l(ia)%p, hv_l(ia)%p,                &
                             h_old_l(ia)%p, h_new_l(ia)%p,                     &
                             ui_l(ia)%p, vi_l(ia)%p, wi_l(ia)%p, cmp_l(ia)%p,  &
                             dispv_l(ia)%p, eddyh_l(ia)%p,                     &
                             nz(ia), krz(ia)%p,                                &
                             nq(ia), krq(ia)%p, rwqk(ia)%p,                    &
                             nu(ia), kru(ia)%p, nv(ia), krv(ia)%p,             &
                             bndz(ia)%p, bndq(ia)%p, cosphi(ia)%p,             &
                             DoAdvec, DoDiff,                                  &
                             idx(ia)%p, dd(ia)%halo2, ia, dyn, uvdam_l(ia)%p )
              endif
!$OMP BARRIER
            enddo
!$OMP END PARALLEL
            call timer(2,'tflow_diff')
          endif ! NDTVDIFF


          !- Copy fine grid t,s and tracers to coarse grid:
          if (mod(izeits,NDTADVEC) == 0 .or. mod(izeits,NDTVDIFF) == 0) then
            call timer(1,'gs_1')

!$OMP PARALLEL DEFAULT (shared) PRIVATE (ia,iao,iia,ii)
            do ia=narea,1,-1
              do iao=1,nestingfrom(ia)%p(0)
                !  ia:  fine grid
                !  iia: coarse grid
                iia = nestingfrom(ia)%p(iao)
                ii  = 1
                do while (ia /= nestingto(iia)%p(ii))
                  ii = ii + 1
                enddo

                call timer(1,'MPI_nest_a')
                !  gather S&T, tracers and h in fine grid to global arrays on 
                !  this task
                call dmpi_gather (iia,ia,cctable(iia,ia),cftable(iia,ia),      &
                                  (1),ntracers,mmx(ia),nmx(ia),kmx(ia),        &
                                  mm1k_l(ia)%p,mm1k(ia)%p,mm1k(iia)%p,         &
                                  iga(ia)%p(:,iao),jga(ia)%p(:,iao),           &
                                  cmp_l(ia)%p,cmp(ia)%p)
! fixme: consider gathering ALL h during initializations and then only k=1 here.
                call dmpi_gather (iia,ia,cctable(iia,ia),cftable(iia,ia),      &
                                  mmx(ia),nmx(ia),kmx(ia),                     &
                                  mm1k_l(ia)%p,mm1k(ia)%p,mm1k(iia)%p,         &
                                  iga(ia)%p(:,iao),jga(ia)%p(:,iao),           &
                                  h_l(ia)%p,h(ia)%p)
                call timer(2,'MPI_nest_a')
!$OMP BARRIER

                !  do fine-->coarse grid copy in parallel mode
                !  use local coarse grid vars with the global fine grid vars 
                !  gathered above
                call copy_f_g ( mmx(ia), nmx(ia), mm1k(ia)%p,                  &
                                dd(iia)%low_i, dd(iia)%up_i,                   &
                                dd(iia)%low_j, dd(iia)%up_j,                   &
                                kmx(iia), mm1k_l(iia)%p,                       &
                                iga(ia)%p(:,iao), jga(ia)%p(:,iao),            &
                                dx(ia)%p, dy(ia), h(ia)%p,                     &
                                cmp(ia)%p, cmp_l(iia)%p, kfg(iia)%p(:,ii),     &
                                ntracers ) 
              enddo
            enddo
!$OMP END PARALLEL
            call timer(2,'gs_1')
          endif
          call timer(2,'adv_diff')
          call timer_print('TS - adv_diff took: ','adv_diff',50)


          !- Do ice dynamics:
          if (mod(izeits, ndticedyn) == 0 .and. icedynactive) then
            call timer(1,'icemodel')
            dticedyn = ndticedyn*dtmain

            if (shortcircuit) then
              ! the shortcircuit ice model -------------------------------------
! MPI: Haloswap avail for u/vi, h above before tflow.
!      Make sure that icevars (ice, u/veis, casus, fastmask, iota) are swapped
!      properly elsewhere.
!REQUIRES: ndticedyn = NDTADVEC for u/vi to be sensible
              call timer(1,'MPI_halo')
!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
              do ia=1,narea
                !FIXME: check if this one can be avoided:
                call dmpi_halo(1,mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,z_l(ia)%p)
              enddo
!$OMP END PARALLEL
              call timer(2,'MPI_halo')
              ! need to loop inside-out, i.e. from fine to coarser grids
              do ii=maxval(timelevel(1:)),1,-1
                do ia=1,narea
                  if (.not.onthislevel(ia,ii)) cycle
                  call icedyn(me_l(ia)%p,iw2_l(ia),ind_l(ia)%p,kh_l(ia)%p,     &
                              idx(ia)%p,mmxp(ia),dxx(ia)%p(1),dyy(ia)%p(1),    &
                              cosphi(ia)%p,ft(ia)%p,dticedyn,                  &
                              ui_l(ia)%p,vi_l(ia)%p,h_l(ia)%p,z_l(ia)%p,       &
                              wu_l(ia)%p,wv_l(ia)%p,casus_l(ia)%p,             &
                              ice_l(ia)%p,ueis_l(ia)%p,veis_l(ia)%p,ia,        &
                              mm1k_l(ia)%p,nud(ia),nvd(ia),uvdam_l(ia)%p,      &
                              z0srf_l(ia)%p,iotamodel,iota_l(ia)%p)
                  call timer(1,'MPI_halo')
!$OMP PARALLEL DEFAULT (shared)
                  call dmpi_halo(1,mm1k_l(ia)%p,kh_l(ia)%p,ia,4,4,2,6,         &
                                 ueis_l(ia)%p,veis_l(ia)%p)
                  if (iotamodel) then
                    call dmpi_halo(1,mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,       &
                                   iota_l(ia)%p,3)
                  endif
!$OMP END PARALLEL
                  call timer(2,'MPI_halo')
                enddo
              enddo
!$OMP PARALLEL DEFAULT (shared) PRIVATE (ia, ii, iia, iao, nu1, nu2, nv1, nv2)
              ! nesting of ice momentum, fine-->coarse.
              do ia=1,narea
                if (nestinglevels(ia) > 0) then
                  !- fix velocity along border:
                  do ii=1,nestinglevels(ia)
                    !  iia: fine grid
                    !  ia:  coarse grid
                    iia = nestingto(ia)%p(ii)
                    iao = enclosing(iia,ia)
                    nu1 = unesting(ia)%p(ii,1)
                    nu2 = unesting(ia)%p(ii,2)
                    nv1 = vnesting(ia)%p(ii,1)
                    nv2 = vnesting(ia)%p(ii,2)

                    call timer(1,'MPI_nest_b')
                    !  gather fine grid u/veis to global arrays along u/v brds:
                    call dmpi_gather(ia,iia,nu1,nu2,kru(ia)%p,(1),             &
                         uctable(mpi_rank+1,ia,iia), uhtable(ia,iia),          &
                         (1),mm1k_l(iia)%p,mm1k(iia)%p,mm1k(ia)%p,             &
                         kh_l(iia)%p,kh(iia)%p,                                &
                         iga(iia)%p(:,iao),jga(iia)%p(:,iao),(1),ii,(4),       &
                         ueis_l(iia)%p,wrk(iia)%p,veis_l(iia)%p,wrk2(iia)%p)
                    call dmpi_gather(ia,iia,nv1,nv2,krv(ia)%p,(1),             &
                         vctable(mpi_rank+1,ia,iia), vhtable(ia,iia),          &
                         (1),mm1k_l(iia)%p,mm1k(iia)%p,mm1k(ia)%p,             &
                         kh_l(iia)%p,kh(iia)%p,                                &
                         iga(iia)%p(:,iao),jga(iia)%p(:,iao),(2),ii,(4),       &
                         ueis_l(iia)%p,wrk(iia)%p,veis_l(iia)%p,wrk2(iia)%p)
                    call timer(2,'MPI_nest_b')

!$OMP BARRIER
                    call ice_mom_f_c(me_l(ia)%p, ueis_l(ia)%p, veis_l(ia)%p,   &
                                     kru(ia)%p, krv(ia)%p, me1(iia)%p,         &
                                     iga(iia)%p(:,iao), jga(iia)%p(:,iao),     &
                                     wrk(iia)%p, wrk2(iia)%p, nu1,nu2,nv1,nv2)

!$OMP BARRIER
                    call timer(1,'MPI_nest_c')
                    call dmpi_halo(1,mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,       &
                                   ueis_l(ia)%p,veis_l(ia)%p)
                    call timer(2,'MPI_nest_c')
!$OMP BARRIER
                  enddo
                endif
              enddo
!$OMP END PARALLEL

              !- sponge layer, ice:
              ia = mainarea
              call sponge_e (me_l(ia)%p,nz(ia),krz(ia)%p,iceobc,               &
                           bndice(ia)%p,ice_l(ia)%p,ueis_l(ia)%p,veis_l(ia)%p, &
                           half*dticedyn,dxx(ia)%p,dyy(ia)%p,cosphi(ia)%p,     &
                           khu_l(ia)%p,khv_l(ia)%p)

              !- Nesting of ice varaiables:
              do ia=1,narea
                do ii=1,nestinglevels(ia)
                  !  iia: fine grid
                  !  ia:  coarse grid
                  iia = nestingto(ia)%p(ii)
                  iao = enclosing(iia,ia)
                  iif = 1
                  do while(ia /= nestingfrom(iia)%p(iif))
                    iif = iif + 1
                  enddo

                  !- ice boundary values coarse-->fine grid:
                  ! Step #1: find coarse grid values just outside border:
                  if (nctable(mpi_rank+1,ia,iia)) then
                    nz1 = znesting(iia)%p(iif,1)
                    nz2 = znesting(iia)%p(iif,2)
                    call bndicez(mm1k_l(ia)%p, iga(iia)%p(:,iao),              &
                                 jga(iia)%p(:,iao), ice_l(ia)%p, mm1k(iia)%p,  &
                                 krz(iia)%p, bndice(iia)%p, nz1, nz2 )
                  endif

                  if (mpi_io_serial) then
                    nz1 = znest_serial(iia)%p(iif,1)
                    nz2 = znest_serial(iia)%p(iif,2)
                  elseif (nctable(mpi_rank+1,ia,iia)) then
                    nz1 = znesting(iia)%p(iif,1)
                    nz2 = znesting(iia)%p(iif,2)
                  else
                    nz1 =  1
                    nz2 = -1
                  endif
                  call timer(1,'MPI_bnd')
                  call dmpi_gather(ia, (3), nz1, nz2, krz(iia)%p,              &
                                   bndice(iia)%p,                              &
                                   iga(iia)%p(:,iao), jga(iia)%p(:,iao))
  
                  if (mpi_io_serial .or. nftable(mpi_rank+1,ia,iia)) then
                    nz1 = znest_serial(iia)%p(iif,1)
                    nz2 = znest_serial(iia)%p(iif,2)
                    call dmpi_scatter(iia, (3), nz1, nz2, krz(iia)%p,          &
                                      bndice(iia)%p  )
                  endif
                  call timer(2,'MPI_bnd')

                  !- ice boundary values coarse-->fine grid:
                  ! Step #2: advect boundary values at inflow points:
                  if (nftable(mpi_rank+1,ia,iia)) then
                    nz1 = znesting(iia)%p(iif,1)
                    nz2 = znesting(iia)%p(iif,2)
                    call bndzice(me_l(iia)%p, krz(iia)%p, bndice(iia)%p,       &
                                 ice_l(iia)%p, ueis_l(iia)%p, veis_l(iia)%p,   &
                                 casus_l(iia)%p, dticedyn,                     &
                                 dxx(iia)%p, dyy(iia)%p, cosphi(iia)%p, nz1,nz2)
                  endif

                  !- ice boundary values fine-->coarse grid:
                  !  (previous steps 1+2 merged)
                  !  #1: coarse grid ice vars by cell-averaging from fine grid:
                  !  #2: advect boundary values at inflow points (f-->c):
                  if (uctable(mpi_rank+1,ia,iia) .or.                          &
                      vctable(mpi_rank+1,ia,iia)) then
                    nu1 = un_c_mpi(ia)%p(ii,1)
                    nu2 = un_c_mpi(ia)%p(ii,2)
                    nv1 = vn_c_mpi(ia)%p(ii,1)
                    nv2 = vn_c_mpi(ia)%p(ii,2)
                    call bnduvice (me1(ia)%p, kru(ia)%p, krv(ia)%p, ice(ia)%p, &
                                 ueis(ia)%p, veis(ia)%p, casus(ia)%p,          &
                                 dticedyn, dxx(ia)%p, dyy(ia)%p, cosphi(ia)%p, &
                                 nu1, nu2, nv1, nv2 )
                  endif
                enddo
              enddo

              call timer(1,'MPI_halo')
! FIXME: consider swapping only tasks needed to be swapped:
!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
              do ia=1,narea
                call dmpi_halo(1,mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,           &
                               ice_l(ia)%p,niceact)
                call dmpi_halo(mm1k_l(ia)%p,ia,casus_l(ia)%p)
              enddo
!$OMP END PARALLEL
              call timer(2,'MPI_halo')

              do ia=1,narea
!$OMP PARALLEL DEFAULT (shared)
              !- check stability limits:
                call ADcheckIce(me_l(ia)%p, iw2_l(ia), dticedyn,               &
                                ueis_l(ia)%p, veis_l(ia)%p, dx(ia)%p, dy(ia),  &
                                ndtp, ind_l(ia)%p)
!$OMP END PARALLEL
                ndt = maxval( ndtp(:) )
                if (ndt > 1) then
                  !- print stability violations:
                  call ADcheckIce( ndtp, (ndticedyn*timelevel(ia)), ia )
                  dticedyn = dticedyn/real(ndt,8)
                  if (mpi_size > 1) then
                    write(iu06,*)'Cannot sub-divide ice advection time step.'
                    write(iu06,*)'Decrease the time step size and re-run.'
                    call exitme(1,                                             &
                                'Error: Ice advection time step size too large')
                  endif
                endif
                do indt=1,ndt
                  ! advection, conservative form (ice)
                  call iflow_int(me_l(ia)%p,iw2_l(ia),ind_l(ia)%p,             &
                                 kh_l(ia)%p,idx(ia)%p,                         &
                                 casus_l(ia)%p,ueis_l(ia)%p,veis_l(ia)%p,      &
                                 ice_l(ia)%p,dxx(ia)%p,dyy(ia)%p,dticedyn,     &
                                 cosphi(ia)%p,nz(ia),krz(ia)%p,bndice(ia)%p,   &
                                 nu(ia),kru(ia)%p,nv(ia),krv(ia)%p,niceact,    &
                                 ia,mm1k_l(ia)%p )
                  call traceroutflow(niceact,nz(ia),krz(ia)%p,me_l(ia)%p,      &
                                     dticedyn,dxx(ia)%p,dyy(ia)%p,cosphi(ia)%p,&
                                     ueis_l(ia)%p,veis_l(ia)%p,ice_l(ia)%p )
                enddo
                dticedyn = dticedyn*real(ndt,8)
              enddo
              call timer(1,'MPI_halo')
!$OMP PARALLEL DEFAULT (shared) PRIVATE(ia)
              do ia=1,narea
                call dmpi_halo(1,mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,           &
                               ice_l(ia)%p,niceact)
              enddo
!$OMP END PARALLEL
              call timer(2,'MPI_halo')

            else
              ! original ice model ---------------------------------------------
              !  FIXME: consider getting rid of this old model completely
              do ia=1,narea
                call dmpi_gather(ia,    casus_l(ia)%p,casus(ia)%p,ice_rank)
                call dmpi_gather(ia,(1),  ice_l(ia)%p,  ice(ia)%p,3,ice_rank)
                call dmpi_gather(ia,(1), ueis_l(ia)%p, ueis(ia)%p,ice_rank)
                call dmpi_gather(ia,(1), veis_l(ia)%p, veis(ia)%p,ice_rank)
                call dmpi_gather(ia,(1),tsnei_l(ia)%p,tsnei(ia)%p,ice_rank)
                call dmpi_gather(ia,(1),    z_l(ia)%p,    z(ia)%p,ice_rank)
                call dmpi_gather(ia,(1), taux_l(ia)%p, taux(ia)%p,ice_rank)
                call dmpi_gather(ia,(1), tauy_l(ia)%p, tauy(ia)%p,ice_rank)
                call dmpi_gather(ia,(1),   ui_l(ia)%p,  wrk(ia)%p,ice_rank)
                call dmpi_gather(ia,(1),   vi_l(ia)%p, wrk2(ia)%p,ice_rank)
              enddo

              ! do old ice model only in on one task:
              if (ice_serial) then

                !- do the ice dynamics:
                do ia=1,narea
                  call eisdyn_int (me1(ia)%p, muvse1(ia)%p, iw2(ia), ind(ia)%p,&
                                 iwuvse(ia),induvse(ia)%p, casus(ia)%p,        &
                                 ueis(ia)%p, veis(ia)%p, ice(ia)%p,            &
                                 dxx(ia)%p, dyy(ia)%p, dticedyn, nz(ia),       &
                                 krz(ia)%p, ti(ia)%p, wrk(ia)%p, wrk2(ia)%p,   &
                                 z(ia)%p, taux(ia)%p, tauy(ia)%p,              &
                                 cosphi(ia)%p,fcorio(ia)%p, ctim, izeits,      &
                                 iceite(ia), iuvp(ia)%p, nthreads)
                enddo

                !- sponge layer, ice:
                ia = mainarea
                call sponge_e (me1(ia)%p, nz(ia), krz(ia)%p, iceobc,           &
                             bndice(ia)%p, ice(ia)%p, ueis(ia)%p, veis(ia)%p,  &
                             half*dticedyn, dxx(ia)%p, dyy(ia)%p, cosphi(ia)%p,&
                             khu(ia)%p, khv(ia)%p )

                !- Nesting of ice varaiables:
!FIXME: this doesn't work with MPI and will be remove together with the rest of
!       the old ice model. For a short while, we will leave the code in here.
!       Some of the calls will no longer compile afetr having fixed nesting
!       across tasks.
                do ia=1,narea
                  do ii=1,nestinglevels(ia)
                    !  iia: fine grid
                    !  ia:  coarse grid
                    iia = nestingto(ia)%p(ii)
                    iao = enclosing(iia,ia)
                    iif = 1
                    do while(ia /= nestingfrom(iia)%p(iif))
                      iif = iif + 1
                    enddo
                    nz1 = znest_serial(iia)%p(iif,1)
                    nz2 = znest_serial(iia)%p(iif,2)
                    nu1 = unest_serial(ia)%p(ii,1)
                    nu2 = unest_serial(ia)%p(ii,2)
                    nv1 = vnest_serial(ia)%p(ii,1)
                    nv2 = vnest_serial(ia)%p(ii,2)
  
                    !- ice boundary values coarse-->fine grid:
                    !  #1: find coarse grid values just outside border:
                    call bndicez(mm1k(ia)%p, iga(iia)%p(:,iao),                &
                                 jga(iia)%p(:,iao), ice(ia)%p, mm1k(iia)%p,    &
                                 krz(iia)%p, bndice(iia)%p, nz1, nz2 )
                    !  #2: advect boundary values at inflow points:
                    call bndzice(me1(iia)%p, krz(iia)%p, bndice(iia)%p,        &
                               ice(iia)%p,                                     &
                               ueis(iia)%p, veis(iia)%p, casus(iia)%p,dticedyn,&
                               dxx(iia)%p, dyy(iia)%p, cosphi(iia)%p, nz1, nz2 )

                    !- ice boundary values fine-->coarse grid:
                    !  (previous steps 1+2 merged)
                    ! #1: coarse grid ice vars by cell-averaging from fine grid:
                    ! #2: advect boundary values at inflow points (f-->c):
                    call bnduvice (me1(ia)%p, kru(ia)%p, krv(ia)%p, ice(ia)%p, &
                                 ueis(ia)%p, veis(ia)%p, casus(ia)%p,          &
                                 dticedyn, dxx(ia)%p, dyy(ia)%p, cosphi(ia)%p, &
                                 nu1, nu2, nv1, nv2 )
                  enddo
                enddo


                !- ice thickness and compactness and snow depth:
                do ia=1,narea
!$OMP PARALLEL DEFAULT (shared)
                  !- check stability limits:
                  call ADcheckIce(me1(ia)%p, iw2(ia), dticedyn,               &
                                ueis(ia)%p, veis(ia)%p, dx(ia)%p, dy(ia),     &
                                ndtp, ind(ia)%p)
!$OMP END PARALLEL
                  ndt = maxval( ndtp(:) )
                  if (ndt > 1) then
                    !- print stability violations:
                    call ADcheckIce( ndtp, (ndticedyn*timelevel(ia)), ia )
                    dticedyn = dticedyn/real(ndt,8)
                  endif
                  do indt=1,ndt
                    ! do the advection:
                    call eflow_int(me1(ia)%p, muvse1(ia)%p, iw2(ia), ind(ia)%p,&
                               casus(ia)%p, ueis(ia)%p, veis(ia)%p, ice(ia)%p, &
                               dxx(ia)%p, dyy(ia)%p, dticedyn, cosphi(ia)%p,   &
                               nz(ia), krz(ia)%p, bndice(ia)%p,                &
                               nu(ia), kru(ia)%p, nv(ia), krv(ia)%p, (3) ) 

                    call traceroutflow( (3), nz(ia), krz(ia)%p, me1(ia)%p,     &
                                      dticedyn, dxx(ia)%p, dyy(ia)%p,          &
                                      cosphi(ia)%p, ueis(ia)%p, veis(ia)%p,    &
                                      ice(ia)%p )
                  enddo 
                  dticedyn = dticedyn*real(ndt,8)
                enddo 
              endif ! ice_serial

              do ia=1,narea
                call dmpi_scatter(ia,iw2(ia),mm1k_l(ia)%p,                     &
                                ueis_l(ia)%p,veis_l(ia)%p,ice_l(ia)%p,         &
                                mm1k(ia)%p,ueis(ia)%p,veis(ia)%p,ice(ia)%p,    &
                                (3),ice_rank,casus_l(ia)%p,casus(ia)%p)
              enddo

            endif
            call timer(2,'icemodel')
            call timer_print('TS - icemodel took: ','icemodel',60)

          endif ! ndticedyn


!$OMP PARALLEL DEFAULT (shared) PRIVATE (ia,iao,iia,ii,iif)
          !- Do thermodynamics:
          if (mod(izeits, NDTTHERMO) == 0) then
            call timer(1,'thermo')
            dtthermo = NDTTHERMO*dtmain

!FIXME: there is already one IF(qflx) branch and that must be fully sufficient!
            if (qflx.and.iotamodel) then
              do ia=1,narea
                call copy_iota2iotac (iw2_l(ia), dd(ia)%halo2, kh_l(ia)%p,     &
                     idx(ia)%p, iota_l(ia)%p, iotac(ia)%p)
              enddo
!$OMP BARRIER
            endif

            !- thermodynamics:
            if (qflx) then
              do ia=1,narea
                call thermodyn_x (pen_type, mm1k_l(ia)%p, iw2_l(ia),           &
                      ind_l(ia)%p, xkoor(ia)%p, ykoor(ia)%p, h_l(ia)%p,        &
                      atemp_l(ia)%p, humid_l(ia)%p, cloud_l(ia)%p,             &
                      wu_l(ia)%p, wv_l(ia)%p, casus_l(ia)%p, cmp_l(ia)%p,      &
                      tsnei_l(ia)%p, t_soil_l(ia)%p, dtthermo, ice_l(ia)%p,    &
                      hflux_l(ia)%p, upd, kh_l(ia)%p, evapo_l(ia)%p,           &
                      divrad_l(ia)%p, divrad_s_l(ia)%p, rho_l(ia)%p,           &
                      pl_l(ia)%p, idx(ia)%p, iotamodel, cosphi(ia)%p,          &
                      ti_l(ia)%p, uvdam_l(ia)%p, iota_l(ia)%p, iotac(ia)%p)
              enddo
!$OMP BARRIER
            else
              do ia=1,narea
                call thermodyn (pen_type, mm1k_l(ia)%p, iw2_l(ia),             &
                      ind_l(ia)%p, xkoor(ia)%p, ykoor(ia)%p, h_l(ia)%p,        &
                      atemp_l(ia)%p, humid_l(ia)%p, cloud_l(ia)%p,             &
                      wu_l(ia)%p, wv_l(ia)%p,                                  &
                      casus_l(ia)%p, cmp_l(ia)%p, tsnei_l(ia)%p,               &
                      t_soil_l(ia)%p, dtthermo, ice_l(ia)%p,                   &
                      hflux_l(ia)%p, upd, kh_l(ia)%p, evapo_l(ia)%p,           &
                      divrad_l(ia)%p, divrad_s_l(ia)%p, snow,                  &
                      cosphi(ia)%p, ti_l(ia)%p, uvdam_l(ia)%p, iotamodel,      &
                      iota_l(ia)%p, iotac(ia)%p )
              enddo
!$OMP BARRIER
            endif  

            do ia=narea,1,-1
              do iao=1,nestingfrom(ia)%p(0)
                !  ia:  fine grid
                !  iia: coarse grid
                iia = nestingfrom(ia)%p(iao)
                ii  = 1
                do while (ia /= nestingto(iia)%p(ii))
                  ii = ii + 1
                enddo

                call timer(1,'MPI_nest_d')
                !  gather T, surface S, and ice vars in fine grid to global 
                !  arrays on this task
                call dmpi_gather (iia,ia,cctable(iia,ia),cftable(iia,ia),      &
                                  (1),(1),mmx(ia),nmx(ia),kmx(ia),             &
                                  mm1k_l(ia)%p,mm1k(ia)%p,mm1k(iia)%p,         &
                                  iga(ia)%p(:,iao),jga(ia)%p(:,iao),           &
                                  cmp_l(ia)%p,cmp(ia)%p)
                call dmpi_gather (iia,ia,cctable(iia,ia),cftable(iia,ia),      &
                                  (2),(2),mmx(ia),nmx(ia),(1),                 &
                                  mm1k_l(ia)%p,mm1k(ia)%p,mm1k(iia)%p,         &
                                  iga(ia)%p(:,iao),jga(ia)%p(:,iao),           &
                                  cmp_l(ia)%p,cmp(ia)%p)
                call dmpi_gather (iia,ia,cctable(iia,ia),cftable(iia,ia),      &
                                  (1),(3),mmx(ia),nmx(ia),(1),                 &
                                  mm1k_l(ia)%p,mm1k(ia)%p,mm1k(iia)%p,         &
                                  iga(ia)%p(:,iao),jga(ia)%p(:,iao),           &
                                  ice_l(ia)%p,ice(ia)%p)
                call timer(2,'MPI_nest_d')
!$OMP BARRIER
                call timer(1,'MPI_nest_e')
                call dmpi_halo(1,mm1k_l(iia)%p,kh_l(iia)%p,iia,                &
                               1,1,1,1,ice_l(iia)%p,niceact)
                iif = niceact + 1
                call dmpi_halo(1,mm1k_l(iia)%p,kh_l(iia)%p,iia,                &
                               iif,iif,2,1,cmp_l(iia)%p,1)
                if (iao == 1) call dmpi_halo(1,mm1k_l(ia)%p,kh_l(ia)%p,ia,     &
                                            iif+1,iif+1,3,1,ice_l(ia)%p,niceact)
                call timer(2,'MPI_nest_e')
!$OMP BARRIER

                call copy_f_g ( mmx(ia), nmx(ia), mm1k(ia)%p,                  &
                                dd(iia)%low_i, dd(iia)%up_i,                   &
                                dd(iia)%low_j, dd(iia)%up_j,                   &
                                kmx(iia), mm1k_l(iia)%p,                       &
                                iga(ia)%p(:,iao), jga(ia)%p(:,iao),            &
                                dx(ia)%p, dy(ia), h(ia)%p, cmp(ia)%p,          &
                                cmp_l(iia)%p, kfg(iia)%p(:,ii), (1) ) 
 
                ! FIXME: For longer term development, we should consider
                !        merging the handling of T and ice vars
                call copy_f_g_ice (mmx(ia), nmx(ia), me1(ia)%p,                &
                                   dd(iia)%low_i, dd(iia)%up_i,                &
                                   dd(iia)%low_j, dd(iia)%up_j,                &
                                   mm1k_l(iia)%p,                              &
                                   iga(ia)%p(:,iao), jga(ia)%p(:,iao),         &
                                   dx(ia)%p, dy(ia), h_l(iia)%p, ice(ia)%p,    &
                                   ice_l(iia)%p, casus_l(iia)%p,               &
                                   cmp_l(iia)%p, tsnei_l(iia)%p )

                call timer(1,'MPI_nest_e')
!$OMP BARRIER
                ! complete the halo swaps:
                call dmpi_halo(1,mm1k_l(iia)%p,kh_l(iia)%p,iia,                &
                               1,1,1,3,ice_l(iia)%p,niceact)
                call dmpi_halo(1,mm1k_l(iia)%p,kh_l(iia)%p,iia,                &
                               1,1,1,4,ice_l(iia)%p,niceact)
                iif = niceact + 1
                call dmpi_halo(1,mm1k_l(iia)%p,kh_l(iia)%p,iia,                &
                               iif,iif,2,3,cmp_l(iia)%p,1)
                call dmpi_halo(1,mm1k_l(iia)%p,kh_l(iia)%p,iia,                &
                               iif,iif,2,4,cmp_l(iia)%p,1)
                if (iao == 1) then
                  call dmpi_halo(1,mm1k_l(ia)%p,kh_l(ia)%p,ia,                 &
                                 iif+1,iif+1,3,3,ice_l(ia)%p,niceact)
                  call dmpi_halo(1,mm1k_l(ia)%p,kh_l(ia)%p,ia,                 &
                                 iif+1,iif+1,3,4,ice_l(ia)%p,niceact)
                endif
                call dmpi_halo(mm1k_l(iia)%p,iia,casus_l(iia)%p)
                if (iao == 1) call dmpi_halo(mm1k_l(ia)%p,ia,casus_l(ia)%p)
                call dmpi_halo(1,mm1k_l(iia)%p,kh_l(iia)%p,iia,                &
                               1,1,1,2,ice_l(iia)%p,niceact)
                call dmpi_halo(1,mm1k_l(iia)%p,kh_l(iia)%p,iia,                &
                               iif,iif,2,2,cmp_l(iia)%p,1)
                if (iao == 1) call dmpi_halo(1,mm1k_l(ia)%p,kh_l(ia)%p,ia,     &
                                 iif+1,iif+1,3,2,ice_l(ia)%p,niceact)
                call dmpi_halo(1,mm1k_l(iia)%p,kh_l(iia)%p,iia,                &
                               1,1,1,5,ice_l(iia)%p,niceact)
                call dmpi_halo(1,mm1k_l(iia)%p,kh_l(iia)%p,iia,                &
                               iif,iif,2,5,cmp_l(iia)%p,1)
                if (iao == 1) call dmpi_halo(1,mm1k_l(ia)%p,kh_l(ia)%p,ia,     &
                                 iif+1,iif+1,3,5,ice_l(ia)%p,niceact)
!$OMP BARRIER
                call timer(2,'MPI_nest_e')
              enddo
            enddo

            if (icedynactive) then
              !- modify ice velocities:
              if (shortcircuit) then
                do ia=1,narea
                  call ice_vel_drift(iw2_l(ia), mm1k_l(ia)%p, ind_l(ia)%p,     &
                                     kh_l(ia)%p, idx(ia)%p,                    &
                                     casus_l(ia)%p, ueis_l(ia)%p, veis_l(ia)%p)
                enddo
                call timer(1,'MPI_halo')
!$OMP BARRIER
                do ia=1,narea
                  call dmpi_halo(1,mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,         &
                                 ueis_l(ia)%p,veis_l(ia)%p)
                enddo
                call timer(2,'MPI_halo')
              else
!$OMP BARRIER
!$OMP MASTER
                do ia=1,narea
                  call ice_vel(iwuvse(ia), induvse(ia)%p, mm1k_l(ia)%p,        &
                               dd(ia)%low_i, dd(ia)%up_i,                      &
                               dd(ia)%low_j, dd(ia)%up_j,                      &
                               casus_l(ia)%p, ueis_l(ia)%p, veis_l(ia)%p)
                enddo
!$OMP END MASTER
                do ia=1,narea
                  call dmpi_halo(1,mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,         &
                                 ueis_l(ia)%p,veis_l(ia)%p)
                enddo
              endif
            endif

            call timer(2,'thermo')
!$OMP BARRIER
          endif ! NDTTHERMO


! MPI: NO need for halo swap 
          do ia=1,narea
            !- check S&T limits:
            call STlimit( iw2_l(ia), kh_l(ia)%p, ind_l(ia)%p,                  &
                          mm1k_l(ia)%p, cmp_l(ia)%p, idx(ia)%p )

            !- update density:
            call dens (mm1k_l(ia)%p, ind_l(ia)%p, iw2_l(ia), kh_l(ia)%p,       &
                       dz(ia)%p, cmp_l(ia)%p, rho_l(ia)%p, press_l(ia)%p,      &
                       idx(ia)%p )
          enddo
!$OMP END PARALLEL


          !- Kalman Filter
          if (kalman) then
            call timer(1,'kalman')
            !- do SST assimilation
            do ia=1,narea
              call assimilate(minut,nsecs,tim,mmxp(ia),nmxp(ia),kmx(ia),       &
                              iw2_l(ia),mm1k_l(ia)%p,mm1k(ia)%p,               &
                              z_l(ia)%p,h_l(ia)%p,                             &
                              dispv_l(ia)%p,cmp_l(ia)%p,a_asstemp(ia)%p,       &
                              a_obserr(ia)%p,a_lastasstime(ia)%p,              &
                              a_errvar(ia)%p,xkoor(ia)%p,ykoor(ia)%p,          &
                              sstmin,sstmax,sstsat,iu06, 1)
            enddo
            call timer(2,'kalman')
            call timer_print('TS - kalman took: ','kalman',70)
          endif
        endif
      endif

      !- Do the sub-model dynamics ---------------------------------------------
      if (coupler_active) then
        call timer(1,'coupler')
        do ia=1,narea
          call coupler_dynamics(ia, iu06, mmx(ia), nmx(ia), kmx(ia),           &
                                mm1k_l(ia)%p, kh_l(ia)%p, h_l(ia)%p,           &
                                cmp_l(ia)%p, dispv_l(ia)%p, rho_l(ia)%p,       &
                                ui_l(ia)%p, vi_l(ia)%p, z_l(ia)%p,             &
                                casus_l(ia)%p, atemp_l(ia)%p, humid_l(ia)%p,   &
                                cloud_l(ia)%p, wu_l(ia)%p, wv_l(ia)%p,         &
                                pl_l(ia)%p, precip_l(ia)%p, (ia==narea),       &
                                idx(ia)%p, irc)
          if (irc /= 0) call exitme(irc, 'coupler_dynamics failed')
        enddo
        call timer(1,'MPI_halo')
!$OMP PARALLEL DEFAULT (shared) PRIVATE (ia)
        do ia=1,narea
          call dmpi_halo(kmx(ia),mm1k_l(ia)%p,kh_l(ia)%p,ia,1,1,1,6,cmp_l(ia)%p)
        enddo
!$OMP END PARALLEL
        call timer(2,'MPI_halo')
        call timer(2,'coupler')
        call timer_print('TS - coupler took: ','coupler',50)
      endif

      !- Calculate poss. water-, temperature and salinity-transport through 
      !- defined transects
      if (calctrans) then
        if (mod(izeits, NDTTRANS) == 0) then
          call transport_calc( )
        endif
      endif


      !=========================================================================
      !- output:
      !=========================================================================

      call timer(1,'io_tmpdat')
      call output_write(izeits,aufdat,ctim,tim)
      call timer(2,'io_tmpdat')
      call timer_print('TS - IO tmpdat took: ','io_tmpdat',100)

      call timer(2,'nstep')
      call timer_print('TS - nstep took: ','nstep',10)

      DoGather1 = (mod(izeits,lmn2) /= 0)
      DoGather2 = (mod(izeits,lmn1) /= 0)
    enddo !- end of 6 or 12 hrs simulation

    !- possibly print stuff for validation:
    if (lvalidate) then
      if (io_server) then
        call ioc_validate()
      else
        DoGatherC = (coupler_active)
        if (DoGatherC) then
          do ia=1,narea
            call dmpi_gather(ia,kmx(ia),cmp_l(ia)%p,cmp(ia)%p,ntracers,        &
                             mpi_io_rank)
          enddo
        endif
        !-  check if vars have just been gathered above:
        if (DoGather1) then
          do ia=1,narea
            call dmpi_gather(ia,kmx(ia),  u_l(ia)%p,  u(ia)%p,mpi_io_rank)
            call dmpi_gather(ia,kmx(ia),  v_l(ia)%p,  v(ia)%p,mpi_io_rank)
            call dmpi_gather(ia,(1),      z_l(ia)%p,  z(ia)%p,mpi_io_rank)
            if (.not.DoGatherC) then
              call dmpi_gather(ia,kmx(ia),cmp_l(ia)%p,cmp(ia)%p,2,mpi_io_rank)
            endif
          enddo
          if ((.not.tidal) .and. DoGather2) then
            do ia=1,narea
              call dmpi_gather(ia,casus_l(ia)%p,casus(ia)%p,mpi_io_rank)
              if (shortcircuit) then
                call dmpi_gather(ia,(1), ueis_l(ia)%p, ueis(ia)%p,mpi_io_rank)
                call dmpi_gather(ia,(1), veis_l(ia)%p, veis(ia)%p,mpi_io_rank)
                call dmpi_gather(ia,(1),  ice_l(ia)%p,ice(ia)%p,3,mpi_io_rank)
                call dmpi_gather(ia,(1),tsnei_l(ia)%p,tsnei(ia)%p,mpi_io_rank)
                if (iotamodel) then
                  !Note: Currently, the use of the iota model under conditions
                  !      of shortcircuit=.false. is not foreseen.
                  call dmpi_gather(ia,(1),iota_l(ia)%p,iota(ia)%p,3,mpi_io_rank)
                endif
              else
                call dmpi_gather(ia,(1),  ice_l(ia)%p,ice(ia)%p,2,mpi_io_rank)
              endif
            enddo
          endif
        endif
        if (mpi_io_serial) call Validate ()
      endif
    endif

    ! write restart file only once - after 6 or 12hrs
    ! if spinup then write every 6 or 12 hours
    if (i2 == 1 .or. spinup) then
      call timer(1,'io_restart')
      if (io_server) then
        !- gather tracers at boundaries:
        ia = mainarea
        call dmpi_gather(ia,ntracers,kmx(ia),1,nz(ia),krz(ia)%p,bndz(ia)%p)

        !FIXME: IO server still not writes sponge files
        if (mpi_io_serial) then
          !- write sponge files:
          call sponge_io( kmx(ia), nz(ia), khbnd, bndz(ia)%p, coupler_active,  &
                        (.false.))
          call sponge_ice_io(nz(ia),bndice(ia)%p,(.false.))
        endif
        !- write restart file
        call ioc_writerestart(tim)
      else
        if (DoGather1) then
          do ia=1,narea
            call dmpi_gather(ia,kmx(ia),  u_l(ia)%p,  u(ia)%p,mpi_io_rank)
            call dmpi_gather(ia,kmx(ia),  v_l(ia)%p,  v(ia)%p,mpi_io_rank)
            call dmpi_gather(ia,(1),      z_l(ia)%p,  z(ia)%p,mpi_io_rank)
            call dmpi_gather(ia,kmx(ia),cmp_l(ia)%p,cmp(ia)%p,2,mpi_io_rank)
          enddo
        endif
        if ((.not.tidal) .and. DoGather2) then
          do ia=1,narea
            call dmpi_gather(ia,    casus_l(ia)%p,casus(ia)%p,mpi_io_rank)
            call dmpi_gather(ia,(1),  ice_l(ia)%p,  ice(ia)%p,3,mpi_io_rank)
            call dmpi_gather(ia,(1), ueis_l(ia)%p, ueis(ia)%p,mpi_io_rank)
            call dmpi_gather(ia,(1), veis_l(ia)%p, veis(ia)%p,mpi_io_rank)
            call dmpi_gather(ia,(1),tsnei_l(ia)%p,tsnei(ia)%p,mpi_io_rank)
          enddo
          if (iotamodel) then
             call dmpi_gather(ia,(1), ice_l(ia)%p, iota(ia)%p,3,mpi_io_rank)
          endif
        endif
        do ia=1,narea
          call dmpi_gather(ia,3,1,ind_l(ia)%p,mm1k_l(ia)%p,                    &
                           t_soil_l(ia)%p,mm1k(ia)%p,t_soil(ia)%p,mpi_io_rank)
        enddo
        if (.not.printturb .or. DoGather1) then
          do ia=1,narea
            call dmpi_gather(ia,kmx(ia), tke_l(ia)%p, tke(ia)%p,mpi_io_rank)
            call dmpi_gather(ia,kmx(ia),diss_l(ia)%p,diss(ia)%p,mpi_io_rank)
            call dmpi_gather(ia,kmx(ia), avv_l(ia)%p, avv(ia)%p,mpi_io_rank)
          enddo
        endif
        if (kalman) then
          do ia=1,narea
              call dmpi_gather(ia,ind_l(ia)%p,mm1k_l(ia)%p,                    &
                               mm1k(ia)%p,a_lastasstime(ia)%p,mpi_io_rank)
              call dmpi_gather(ia,kmx(ia),ind_l(ia)%p,mm1k_l(ia)%p,            &
                               mm1k(ia)%p,a_errvar(ia)%p,mpi_io_rank)
          enddo
        endif
  
        !- gather tracers at boundaries:
        ia = mainarea
        call dmpi_gather(ia,ntracers,kmx(ia),1,nz(ia),krz(ia)%p, bndz(ia)%p)
  
        if (mpi_io_serial) then
          !- write sponge files:
          call sponge_io( kmx(ia), nz(ia), khbnd, bndz(ia)%p, coupler_active,  &
                          (.false.))
          !- write restart file:
          call WriteRestart ()
        endif

        !- write ice-vars sponges:
        if (ice_serial) then 
          ia = mainarea
          call sponge_ice_io(nz(ia),bndice(ia)%p,(.false.))
        endif
      endif ! io-server
      call timer(2,'io_restart')
      call timer_print('TS - IO restart took: ','io_restart',100)
    endif

    call timer(2,sch//'hstep')
    call timer_print('TS - '//sch//'hstep took: ',sch//'hstep',5)

    !- print timings of MPI related segments:
    call timer_print('MPI - meteo   :','MPI_met'   ,10)
    call timer_print('MPI - boundary:','MPI_bnd'   ,10)
    call timer_print('MPI - nest  1 :','MPI_nest_1',10)
    call timer_print('MPI - nest  2 :','MPI_nest_2',10)
    call timer_print('MPI - nest  3 :','MPI_nest_3',10)
    call timer_print('MPI - nest  4 :','MPI_nest_4',10)
    call timer_print('MPI - nest  5 :','MPI_nest_5',10)
    call timer_print('MPI - nest  6 :','MPI_nest_6',10)
    call timer_print('MPI - nest  7 :','MPI_nest_7',10)
    call timer_print('MPI - nest  7s:','MPI_nest7s',10)
    call timer_print('MPI - nest  8 :','MPI_nest_8',10)
    call timer_print('MPI - nest  8a:','MPI_nest8a',10)
    call timer_print('MPI - nest  8b:','MPI_nest8b',10)
    call timer_print('MPI - nest  9 :','MPI_nest_9',10)
    call timer_print('MPI - nest 10 :','MPI_nest_a',10)
    call timer_print('MPI - nest 11 :','MPI_nest_b',10)
    call timer_print('MPI - nest 12 :','MPI_nest_c',10)
    call timer_print('MPI - nest 13 :','MPI_nest_d',10)
    call timer_print('MPI - nest 14 :','MPI_nest_e',10)
    call timer_print('MPI - nest 15 :','MPI_nest_f',10)
    call timer_print('MPI - haloswap:','MPI_halo'  ,10)
    call timer_print('MPI - TF halo :','MPI_haloTF',10)

    ! other print-outs from timer:
    if (dyn) call timer_print('TS - hydrostat took: ','hydrostat',50)
    call timer_print('TS - turbulence took: ','turbulence',50)
    call timer_print('TS - momeqs took: ','momeqs',50)
    call timer_print('TS - masseqs took: ','masseqs',50)
    call timer_print('TS - thermo took: ','thermo',50)
    call timer_print('TS - gs_1 took: ','gs_1',100)
    call timer_print('TS - tflow_diff took: ','tflow_diff',50)
  enddo TIMELOOP
  call timer(2,'tloop')
  call timer_print('TS - timeloop took: ','tloop',50)

  call timer(1,'finish')


  !- close some output files ---------------------------------------------------
  call output_close(ctim,tim)

  if (mpi_io_serial) then
    if (wforc) call metclose()
  endif

  ! kill the sub-model coupler. 
  if (coupler_active) then
    if (mpi_io_serial) then
      call coupler_finish(iu06, irc)
      if (irc /= 0) call exitme(irc,'COUPLER ERROR: Unable to finish sub-model')
    endif
  endif

  call timer(2,'finish')
  call timer_print('TS - Finalisation took: ','finish',50)
  call timer_destroy()

  !- That's all folks! ---------------------------------------------------------
  call ioc_finalize()
  call exitme(0, 'END TIME IN CMOD' )

end program cmod_main
