module cmod_miscsubs

  use dmi_mpi_global,only : iu06
  use exits,         only : exitme
  use io_subs,       only : io_new_unit, flush_unit
  use cmod_params,   only : narea, nfjord, mainarea, maxtimelevels, iw2, iw3,  &
                            timelevelsize, timelevel, nq
  use cmod_arrays,   only : rwq, krq, bndq, nqlist, nqfile, dt, u, v, z, casus,&
                            cmp, ice, ueis, veis, tsnei, iota
  use constants,     only : iotamodel
  ! lahey bug: remove the 'only' below to avoid an internal compiler bug
  use cmod_miscvars, only : cmodcfg, ntracers, tidal, tidal_surge,qforc, wforc,&
                            tim1, ctim, ctim1, cta, cte, jstuna, mdatei,       &
                            dtice, dt3600, dtday, dtsurge, dttsbnd, dtmain,    &
                            tima, time, izeite, NDTFC, NDTADVEC,               &
                            ia, ndt, ios, lmn, lmne, lmn1, lmn2, lmns,         &
                            ija_akt, imon_akt, itg_akt,                        &
                            schedule, sch, schfactor, ischfac, ioff,           &
                            NDTVDIFF, NDTEDDY, NDTTHERMO, NDTICEDYN, OPRFMT,   &
                            lvalidate, sstmin, sstmax, sstsat, dtdscale,       &
                            printmetstats, lmnts, NDTSMAG

  implicit none
  private 

  public :: ReadTask, SimTime, FreshWaterSources, RiverTSsources,              &
            CheckTimeStepSizes, TimeStepSettings, CheckWL, Validate,           &
            checkrivers

  private :: CheckTimeStep, ReadSources, TracerStats, ReadRiverTS, PrintRiverTS


  ! some baisc stats tootls:
  interface sum_of_values
    module procedure priest_sum
    module procedure priest_sum_nc
  end interface
  interface sum_of_squares
    module procedure sum_of_squares_1
    module procedure sum_of_squares_nc
  end interface
  
  public  :: sum_of_values
  private :: priest_sum, priest_sum_nc, fast2sum
  private :: sum_of_squares, sum_of_squares_1, sum_of_squares_nc

contains

!===============================================================================

  subroutine ReadTask()
    use constants, only : aufdat, ldebug

    implicit none

    integer(4) :: lun
    character(8),  parameter :: fmti = '(a15,i3)', fmtl = '(a15,l2)'
    character(10), parameter :: fmtr = '(a15,f8.2)'

    !- Read in the setup specification -----------------------------------------
    lun = io_new_unit()
    open(unit=lun,file='cmod.nml',STATUS='old',ACTION='read',IOSTAT=ios)
    if (ios/=0) then
      write(iu06,'(a54)')                                                      &
        'Missing namelist file: cmod.nml. Using default values.'
    else
      read(lun,nml=cmodcfg)
      close(unit=lun)
      if (ldebug) then
        write(iu06,'(a39)') 'Dump of namelist cmodcfg from cmod.nml:'
        write(iu06,fmti) 'NDTADVEC      =',NDTADVEC
        write(iu06,fmti) 'NDTVDIFF      =',NDTVDIFF
        write(iu06,fmti) 'NDTEDDY       =',NDTEDDY
        !FIXME: when we find time to modify the references, include this always:
        if (NDTSMAG /= 0) write(iu06,fmti) 'NDTSMAG       =',NDTSMAG
        write(iu06,fmti) 'NDTTHERMO     =',NDTTHERMO
        write(iu06,fmti) 'ndticedyn     =',ndticedyn
        write(iu06,fmtl) 'OPRFMT        =',OPRFMT
        write(iu06,fmtl) 'lvalidate     =',lvalidate
        write(iu06,fmtr) 'sstmin        =',sstmin
        write(iu06,fmtr) 'sstmax        =',sstmax
        write(iu06,fmtr) 'sstsat        =',sstsat
        write(iu06,fmtr) 'dtdscale      =',dtdscale
        write(iu06,fmtr) 'dtsurge       =',dtsurge
        write(iu06,fmtr) 'dttsbnd       =',dttsbnd
        write(iu06,fmti) 'schedule      =',schedule
        write(iu06,fmtl) 'printmetstats =',printmetstats
      endif
    endif

    lun = io_new_unit()
    open(lun,file='metast_neu',form='formatted',status='old',iostat = ios)
    if (ios /= 0) then
      call exitme ( 'input error - could open read metast_neu' )
    else
      read(lun,'(a11,9x,a19,8x,a19)',iostat = ios) aufdat,cta,cte
      if (ios /= 0) then
        call exitme ( 'input error - could not read metast_neu' )
      endif
      read(lun,'(a)',iostat = ios) mdatei
      if (ios /= 0) then
        call exitme ( 'input error - could open read metast_neu' )
      endif
    endif
    close(lun)

    tidal = ( aufdat(1:1) == 'G' .or. aufdat(1:1) == 'T' )
    tidal_surge = aufdat(1:1) == 'T'
    wforc = aufdat(1:1) == 'V'
    qforc = aufdat(1:1) == 'Q'

    if (tidal) then
      write(iu06,'(a21)') 'tidal simulation only'
      if (tidal_surge) then
        write(iu06,'(a40)') ' ... but reading from a tidal surge file'
      endif
    else if (wforc) then
      write(iu06,'(a25)') 'weather forced simulation'
    else if (qforc) then
      write(iu06,'(a27)') 'heat flux forced simulation'
    else
      call exitme ( 'input error - could open read metast_neu' )
    endif
  end subroutine ReadTask

!===============================================================================

  subroutine SimTime()
    use constants, only : jahra, tim, zeitstu, twentyfour
    use cmod_date, only : jultimea

    implicit none

    call jultimea(cta,tima)

    read(cta,'(i4)')jahra
    tim=tima
    call jultimea(cte,time)

    if (tima >= time) then
      call exitme ( 'input error - endtime before starttime, metast_neu' )
    endif

    ctim='yyyy.01.01 00:00:00'
    ctim(1:4)=cta(1:4)
    call jultimea(ctim,tim1)
    jstuna=(tim-tim1)*twentyfour

    zeitstu = jstuna

    write(iu06,'(a21,a19)') 'Start simulation at: ',cta
  end subroutine SimTime

!===============================================================================

  subroutine FreshWaterSources ( na, nqfmax )

    use constants, only : zero
    use dmi_mpi,   only : mpi_rank

    implicit none

    integer(4), intent(in) :: na, nqfmax

    integer(4)    :: lun, iq, iqa
    real(8)       :: rwqa(na,0:nqfmax)
    character(12) :: qlogfile
    
    lun = io_new_unit()
    write (qlogfile,'(a,i3.3)') 'qnew.out.',mpi_rank
    open(lun,file=qlogfile,status='unknown',iostat=ios)
    if (ios /= 0) then
      call exitme ( 'input error - could open '//qlogfile )
    else
      read(cta,'(i4,1x,i2,1x,i2)')ija_akt,imon_akt,itg_akt

      if (na /= narea .or. nqfmax /= maxval(nqfile)) then
        call exitme ( 'error in fresh water source specifications' )
      endif

      do ia=1,narea
        ! better safe than sorry:
        rwq(ia)%p(0) = zero

        if (nq(ia) > 0) then

          ! save the climatological river discharges:
          rwqa(ia,0) = zero
          do iq=1,nqfile(ia)
            iqa = nqlist(ia)%p(iq)
            if (iqa > 0) rwqa(ia,iqa) = rwq(ia)%p(iq)
          enddo

          ! read in the source data:
          write(iu06,'(a30,i3)') 'Reading river inflow for area ',ia
          call ReadSources( ia, nqfile(ia), rwq(ia)%p )

          ! shuffle the read source data:
          do iq=1,nqfile(ia)
            iqa = nqlist(ia)%p(iq)
            if (iqa > 0) rwq(ia)%p(iqa) = rwq(ia)%p(iq)
          enddo

          ! print the source data:
          write(lun,'(/,'' Freshwater sources in area '',i3)') ia
          do iq=1,nqfile(ia)
            iqa = nqlist(ia)%p(iq)
            if (iqa > 0) write(lun,'(i3,2f7.1)') iqa,rwqa(ia,iqa),rwq(ia)%p(iqa)
          enddo

        else
          write(lun,'(/,'' No freshwater sources in area '',i3)') ia
        endif
      enddo

    endif
    close (lun)

  end subroutine FreshWaterSources

!===============================================================================

  subroutine RiverTSsources (  )

    use cmod_local_arrays, only : mm1k_l, kh_l
    use dmi_mpi,   only : mpi_rank

    implicit none
    
    integer(4)    :: lun
    character(12) :: qlogfile
    
    logical, save :: TSsource = .false.


    !- if existing read also T and S source!
    do ia = 1,narea     
      call ReadRiverTS( ia,mm1k_l(ia)%p,kh_l(ia)%p, nqfile(ia),nqlist(ia)%p,   &
                        krq(ia)%p, bndq(ia)%p, TSsource )
    enddo

    if (TSsource) then
      lun = io_new_unit()
      write (qlogfile,'(a,i3.3)') 'qTS.out.',mpi_rank
      open(lun,file=qlogfile,status='unknown',iostat=ios)
      if (ios /= 0) then
        call exitme ( 'input error - could open '//qlogfile )
      else
        read(cta,'(i4,1x,i2,1x,i2)')ija_akt,imon_akt,itg_akt
        do ia = 1,narea
          ! print the TS-source data:
          call PrintRiverTS( lun, ia,mm1k_l(ia)%p,kh_l(ia)%p,nqfile(ia),       &
                             nqlist(ia)%p, krq(ia)%p, bndq(ia)%p )
        enddo
      endif
      close (lun)
    endif


  end subroutine RiverTSsources

!===============================================================================

  subroutine ReadSources( ia, nq, rwq )
    ! straight forward read of fresh water sources from an ascii file

    use cmod_date, only : jultimea
    use constants, only : TIM, twenty

    implicit none

    integer(4), intent(in)    :: ia, nq
    real(8),    intent(inout) :: rwq(0:)

    character(1)  :: c1
    character(2)  :: c2
    character(3)  :: c3
    character(4)  :: iformat
    character(13) :: cformat
    character(14) :: fname
    character(19) :: cdat
    integer(4)    :: lun, i
    real(8)       :: timfl

    !- set up file name, format, etc ... ---------------------------------------
    if (nq < 1) then
      return
    elseif (nq < 10) then
      iformat = '(i1)'
      write(c1,iformat) nq
      cformat = '(a19, '//c1//'f8.1)'
    elseif (nq < 100) then
      iformat = '(i2)'
      write(c2,iformat) nq
      cformat = '(a19,'//c2//'f8.1)'
    elseif (nq < 1000) then
      iformat = '(i3)'
      write(c3,iformat) nq
      cformat = '(a19,'//c3//'f8.1)'
    else
      ! trap
      call exitme(1,'ERROR: Number of fresh water sources too large.')
    endif
    if (ia < 10) then
      iformat = '(i1)'
      write(c1,iformat) ia
      fname   = 'riverinflow_0'//c1
    elseif (ia < 100) then
      iformat = '(i2)'
      write(c2,iformat) ia
      fname   = 'riverinflow_'//c2
    else
      ! trap
      call exitme(1,'ERROR: Number of areas too large.')
    endif

    !- open the file and read it -----------------------------------------------
    lun = io_new_unit()
    open (lun,file=fname,action='read',status='old',iostat=ios)
    if (ios /= 0) then
      write(iu06,'(a49)') 'Unable to open '//fname//', continuing without'
    else
      read (lun,*)
      read (lun,'(a19)') cdat
      call jultimea(cdat,timfl)
      if ( abs(TIM-timfl) <= twenty ) then
        backspace (lun)
        read (lun,cformat) cdat, (rwq(i),i=1,nq)
        close (lun)
      endif
    endif


  end subroutine ReadSources

!===============================================================================

  subroutine ReadRiverTS( ia, mmk, kh, nqf, nql, krq, bndq, TSsource )
    ! straight forward read T/S of rivers from an ascii file

    use cmod_date,      only : jultimea
    use constants,      only : TIM, twenty

    implicit none

    integer(4), intent(in)    :: ia, nqf
    integer(4), intent(in)    :: mmk(:,0:,0:)
    integer(4), intent(in)    :: kh(0:)
    integer(4), intent(in)    :: nql(0:)
    integer(4), intent(in)    :: krq(:,0:)
    real(8),    intent(inout) :: bndq(1:,1:,0:)
    logical,    intent(inout) :: TSsource

    character(1)  :: c1
    character(2)  :: c2
    character(3)  :: c3
    character(4)  :: iformat
    character(13) :: cformat
    character(10) :: fname
    character(19) :: cdat
    integer(4)    :: lun, n, nta, i, j, k, kb, ios
    real(8)       :: timfl
    real(8)       :: rivert(0:nqf), rivers(0:nqf)

    !- set up file name, format, etc ... -------------------------------------
    if (nqf < 1) then
      return
    elseif (nqf < 10) then
      iformat = '(i1)'
      write(c1,iformat) nqf
      cformat = '(a19, '//c1//'f8.1)'
    elseif (nqf < 100) then
      iformat = '(i2)'
      write(c2,iformat) nqf
      cformat = '(a19,'//c2//'f8.1)'
    elseif (nqf < 1000) then
      iformat = '(i3)'
      write(c3,iformat) nqf
      cformat = '(a19,'//c3//'f8.1)'
    else
      ! trap
      call exitme(1,'ERROR: Number of fresh water sources too large.')
    endif
    if (ia < 10) then
      iformat = '(i1)'
      write(c1,iformat) ia
      fname   = 'riverTS_0'//c1
    elseif (ia < 100) then
      iformat = '(i2)'
      write(c2,iformat) ia
      fname   = 'riverTS_'//c2
    else
      ! trap
      call exitme(1,'ERROR: Number of areas too large.')
    endif

    !- open the file and read it ---------------------------------------------
    lun = io_new_unit()
    open (lun,file=fname,action='read',status='old',iostat=ios)
    if (ios == 0) then
      write(iu06,'(a31,i3)') 'Reading river T and S for area ', ia
      read (lun,*)
      read (lun,'(a19)') cdat
      call jultimea(cdat,timfl)
      if ( abs(TIM-timfl) <= twenty ) then
        backspace (lun)
        read (lun,cformat) cdat, (rivert(n),n=1,nqf)
        read (lun,cformat) cdat, (rivers(n),n=1,nqf)
        close (lun)

        ! shuffle the read TS-source data:
        do n=1,nqf
          nta = nql(n)
          if (nta > 0) then
            i  = krq(1,nta)
            j  = krq(2,nta)
            kb = kh(mmk(1,i,j))
            do k=1,kb
            !- fresh water inflow, s=0
              bndq(1,k,nta) = rivert(n)
              bndq(2,k,nta) = rivers(n)
            enddo
          endif
        enddo

      endif

      TSsource = .true.
    endif

  end subroutine ReadRiverTS

!===============================================================================

  subroutine PrintRiverTS( lun, ia, mmk, kh, nqf, nql, krq, bndq )

    implicit none

    integer(4), intent(in)    :: lun, ia, nqf
    integer(4), intent(in)    :: mmk(:,0:,0:)
    integer(4), intent(in)    :: kh(0:)
    integer(4), intent(in)    :: nql(0:)
    integer(4), intent(in)    :: krq(:,0:)
    real(8),    intent(inout) :: bndq(1:,1:,0:)

    integer(4)    :: iq, iqa, i, j, k, kb

    write(lun,'(/,'' River TS sources in area '',i3)') ia
    do iq=1,nqf
      iqa = nql(iq)
      if (iqa > 0) then
        i  = krq(1,iqa)
        j  = krq(2,iqa)
        kb = kh(mmk(1,i,j))

        do k=1,kb
          write(lun,'(2i4,2f7.1)') iqa, k, bndq(1,k,iqa), bndq(2,k,iqa)
        enddo
      endif
    enddo

  end subroutine PrintRiverTS

!===============================================================================

  subroutine CheckTimeStepSizes ( )

    implicit none

    timelevelsize(mainarea) = 1

    if (maxtimelevels > 2) then
      ! time step sizes is divided by 2 on each successive time level:
      do ia=2,narea
        ndt = 2**(timelevel(ia)-1)
        call CheckTimeStep( mainarea, dt(mainarea), ia, dt(ia), ndt )
        timelevelsize(ia) = ndt
      enddo
    else
      ! dt(mainarea) being any integer(4) multiple of dt on level 2
      if (maxtimelevels == 2) then
        do ia=2,narea
          if (timelevel(ia) /= timelevel(mainarea)) then
            ndt   = nint( dt(mainarea)/dt(ia) )
            NDTFC = ndt
            exit
          endif
        enddo
      endif
      do ia=2,narea
        if (timelevel(ia) == timelevel(mainarea)) then
          ndt = 1
        else
          ndt = NDTFC
        endif
        call CheckTimeStep( mainarea, dt(mainarea), ia, dt(ia), ndt )
        timelevelsize(ia) = ndt
      enddo
    endif

  end subroutine CheckTimeStepSizes

!===============================================================================

  subroutine CheckTimeStep ( a1, dt1, a2, dt2, ndt21 )

    implicit none

    integer(4), intent(in) :: a1, a2, ndt21
    real(8), intent(in)    :: dt1, dt2

    real(8), parameter :: onep = 1.01_8, onem = 0.99_8

    if (dt1 > onep*ndt21*dt2 .or. dt1 < onem*ndt21*dt2) then
      write(iu06,'(a25)')           'error in time step sizes:'
      write(iu06,'(a7,i4,a7,f7.2)') '  area ',        a1,', dt = ',dt1
      write(iu06,'(a7,i4,a7,f7.2)') '  area ',        a2,', dt = ',dt2
      write(iu06,'(a7,f7.2,a8,i4)') 'must be', ndt21*dt2,' in area',a1
      call exitme(1)
    endif
  end subroutine CheckTimeStep

!===============================================================================

  subroutine TimeStepSettings ()

    use constants, only : rzeite, zzeits, mzwei, toutput, dtwlfile, zero,      &
                          twentyfour
    use cmod_date, only : jultimea
!FIXME: this code seems obsolete and will be removed:
!   use dmi_mpi,   only : mpi_io_serial

    implicit none

    real(8) :: rschedule, dtpegel
    real(8), parameter :: onehour = 3600.0_8, m2time = 12.4166666666_8

    if (schedule == 12 .or. schedule == 6) then
      !  only 12 hrs or 6 hrs schedule allowed
      rschedule = real(schedule,8)
      schfactor = rschedule/twentyfour
      ischfac   = 24/schedule
      if (schedule == 6) sch = '06'
    else
      !  trap!!
      call exitme(1,'ERROR: schedule must be 6 or 12 hrs')
    endif

    ! determine that offset for reading dynamic, daily S and T bnd data:
    if (schedule == 12) then
      if (cta(12:13) == '00') then
        ioff = 1
      elseif (cta(12:13) == '12') then
        ioff = 2
      else
        call exitme(1,'ERROR: starting hrs must be 00 or 12')
      endif
    elseif (schedule == 6) then
      if (cta(12:13) == '00') then
        ioff = 1
      elseif (cta(12:13) == '06') then
        ioff = 4
      elseif (cta(12:13) == '12') then
        ioff = 3
      elseif (cta(12:13) == '18') then
        ioff = 2
      else
        call exitme(1,'ERROR: starting hrs must be 00, 06, 12 or 18')
      endif
    else
      !  don't know what to do ??!!!
      call exitme(1,'ERROR: schedule must be 6 or 12 hrs')
    endif

    dt3600 = dtmain/onehour         ! time step size fraction of one hour
    dtday  = dt3600/twentyfour      ! time step size fraction of one day
    izeite = nint(rschedule/dt3600) ! No. of time steps to complete 6 or 12 hrs
    rzeite = real(izeite/schedule,8)! No. of time steps to complete 1 hr as a
                                    ! real. 
    zzeits = zero                   ! actual time step No. as a real within each
                                    ! hour, starting value is 0.0 

    if (mzwei) then
      if (narea /= 2 .or. nfjord /= 0) then 
        call exitme(1,'MZWEI option only valid in the original 2-domain case')
      endif
      dtmain = m2time/izeite
      dt(2)  = dtmain/real(NDTFC,8)
      dt3600 = dtmain/onehour  
      dtday  = dt3600/twentyfour
    endif

    dtice   = onehour                 ! configurable (ice)

    lmns = nint(dtsurge/dtmain)       ! *must* match noamod (surge)
    lmnts= nint(dttsbnd/dtmain)
    lmn  = NDTADVEC

    lmn1 = nint(dtice/dtmain)         ! configurable (ice)
    lmn2 = nint(toutput/dtmain)       ! configurable (tempdat)

    if (dtwlfile <= zero) call exitme(1,'Invalid dtwlfile value')
    dtpegel = max(dtmain,dtwlfile)    ! configurable (pegel)
    lmne    = nint(dtpegel/dtmain)    ! configurable (water level)

    !- write output time steps -------------------------------------------------
!FIXME: this code seems obsolete and will be removed:
!   if (mpi_io_serial) then
!     lun = io_new_unit()
!     open(lun,file='hydro_step')
!     write(lun,*) nint(toutput)
!     close(lun)
!     if (.not.tidal) then
!       lun = io_new_unit()
!       open(lun,file='ice_step')
!       write(lun,*) nint(dtice)
!       close(lun)
!     endif
!   endif

    ctim1='yyyy.01.01 00:00:00'
    ctim1(1:4)=ctim(1:4)
    call jultimea(ctim1,tim1)

    !- try to capture obvious time-setting bugs --------------------------------
    if (lmn2*(izeite/lmn2) /= izeite) then
      write(iu06,*) 'Error: Inconsistent combination of toutput and schedule.' 
      call exitme(1,'Modify toutput and/or schedule, and re-run.')
    endif

  end subroutine TimeStepSettings

!===============================================================================

  subroutine CheckWL ( iw2, zn, hz, z )

    !- modules -----------------------------------------------------------------
    use dmi_omp,   only : domp_get_domain
    use constants, only : hgr

    !- directives --------------------------------------------------------------
    implicit none

    !- args --------------------------------------------------------------------
    integer(4), intent(in)    :: iw2
    real(8),    intent(in)    :: hz(0:), z(0:) ! hz(0:iw3), z(0:iw2)
    real(8),    intent(inout) :: zn(0:)        ! zn(0:iw2)

    !- locals ------------------------------------------------------------------
    integer(4) :: ii, il, iu

    call domp_get_domain(1, iw2, il, iu)

    !- treat possibly dried points:
    do ii=il,iu
      zn(ii) = max( -(hz(ii)-z(ii))+hgr, zn(ii) )
    enddo

  end subroutine CheckWL

!===============================================================================

  subroutine Validate ()

    use constants,   only : zero, one, shortcircuit, icedynactive
    use old_ice_dyn, only : iceite

    implicit none

    integer(4) :: ia, i, ie
    real(8)    :: smean, smin, smax, tmean, tmin, tmax, emean, emin, emax
    real(8)    :: srms, sstd, trms, tstd, erms, estd
    character(8),  parameter :: fmti1 = '(a33,i5)'
    character(12), parameter :: fmtr1 = '(a33,f16.13)'
    character(22), parameter :: fmtr2 = '(a33,f16.13,5x,f16.13)'
    character(12), parameter :: fmtr3 = '(a33,f16.10)'
    character(22), parameter :: fmtr4 = '(a33,f16.10,5x,f16.10)'

! TVSCRIPT: statistics
    write(iu06,'(a18)') 'Validation prints:'
    do ia=1,narea 
      write(iu06,'(a72)')                                                      &
      '------------------------------------------------------------------------'
      write(iu06,fmti1) 'Statistics for domain:           ', ia

      call TracerStats( iw3(ia), 1, cmp(ia)%p, tmean, tmin, tmax, trms, tstd )
      call TracerStats( iw3(ia), 2, cmp(ia)%p, smean, smin, smax, srms, sstd )

      write(iu06,fmtr1) 'Average value for salinity:      ', smean
      write(iu06,fmtr2) 'RMS and STD for salinity:        ', srms, sstd
      write(iu06,fmtr1) 'Average value for temperature:   ', tmean
      write(iu06,fmtr2) 'RMS and STD for temperature:     ', trms, tstd

      write(iu06,fmtr2) 'Min and max for salinity:        ', smin, smax
      write(iu06,fmtr2) 'Min and max for temperature:     ', tmin, tmax

      write(iu06,fmtr2) 'Min and max for u:               ',                   &
        minval(u(ia)%p(1:)), maxval(u(ia)%p(1:))
      write(iu06,fmtr2) 'Min and max for v:               ',                   &
        minval(v(ia)%p(1:)), maxval(v(ia)%p(1:))

      call sum_of_values(iw2(ia),z(ia)%p,emean)
      emean = emean/real(iw2(ia),8)
      call sum_of_squares(iw2(ia),z(ia)%p,zero,erms)
      erms = sqrt( erms/real(iw2(ia),8) )
      if (iw2(ia) > 1) then
        call sum_of_squares(iw2(ia),z(ia)%p,emean,estd)
        estd = sqrt( estd/(real(iw2(ia)-1,8)) )
      else
        estd = -one
      endif

      write(iu06,fmtr1) 'Average value for z:             ', emean
      write(iu06,fmtr2) 'RMS and STD for z:               ', erms, estd
      write(iu06,fmtr2) 'Min and max for z:               ',                   &
        minval(z(ia)%p(1:)), maxval(z(ia)%p(1:))

      ie = 0
      do i=1,iw2(ia)
        if (.not.casus(ia)%p(i)) ie = ie + 1
      enddo
      write(iu06,'(a41,i8)') 'No. of partially ice-covered grid points:', ie
      if (icedynactive .and. (.not.shortcircuit)) then
        if (iceite(ia) > 0)                                                    &
          write(iu06,fmti1)    'No. of ice-solver iterations:    ', iceite(ia)
      endif
      if (shortcircuit) then
        call TracerStats( iw2(ia), 1, ice(ia)%p, emean, emin, emax )
        write(iu06,fmtr3) 'Ice thickness, average:          ', emean
        write(iu06,fmtr4) 'Ice thickness, min and max:      ', emin, emax
        call TracerStats( iw2(ia), 2, ice(ia)%p, emean, emin, emax )
        write(iu06,fmtr3) 'Ice compactness, average:        ', emean
        write(iu06,fmtr4) 'Ice compactness, min and max:    ', emin, emax
        call TracerStats( iw2(ia), 3, ice(ia)%p, emean, emin, emax )
        write(iu06,fmtr3) 'Snow thickness, average:         ', emean
        write(iu06,fmtr4) 'Snow thickness, min and max:     ', emin, emax
        call sum_of_values(iw2(ia),tsnei(ia)%p,emean)
        emean = emean/real(iw2(ia),8)
        write(iu06,fmtr3) 'Surface temperature, average:    ', emean
        write(iu06,fmtr4) 'Surface temperature, min and max:',                 &
          minval(tsnei(ia)%p(1:)), maxval(tsnei(ia)%p(1:))

        write(iu06,fmtr4) 'Ice u, min and max:              ',                 &
          minval(ueis(ia)%p(1:)), maxval(ueis(ia)%p(1:))
        write(iu06,fmtr4) 'Ice v, min and max:              ',                 &
          minval(veis(ia)%p(1:)), maxval(veis(ia)%p(1:))
      endif

      if (iotamodel) then
        call TracerStats( iw2(ia), 1, iota(ia)%p, emean, emin, emax )
        write(iu06,fmtr3) 'Iota, average:                   ', emean
        write(iu06,fmtr4) 'Iota, min and max:               ', emin, emax
        call TracerStats( iw2(ia), 2, iota(ia)%p, emean, emin, emax )
        write(iu06,fmtr3) 'L, average:                      ', emean
        write(iu06,fmtr4) 'L, min and max:                  ', emin, emax
        call TracerStats( iw2(ia), 3, iota(ia)%p, emean, emin, emax )
        write(iu06,fmtr3) 'Kappa, average:                  ', emean
        write(iu06,fmtr4) 'Kappa, min and max:              ', emin, emax
      endif

      if (ntracers > 2) then
        do i=1,ntracers-2
          call TracerStats( iw3(ia), i+2, cmp(ia)%p, emean, emin, emax )

          write(iu06,fmti1) 'Tracer parameter No:             ', i
          write(iu06,fmtr3) '  Average value:                 ', emean
          write(iu06,fmtr4) '  Min and max:                   ', emin, emax
        enddo
      endif
    enddo
    write(iu06,'(a72)')                                                        &
      '------------------------------------------------------------------------'
    call flush_unit(iu06)

  end subroutine Validate

!-------------------------------------------------------------------------------

  subroutine TracerStats( iw3, nc, cmp, emean, emin, emax, erms, estd ) 

    use constants, only : zero, one

    implicit none

    integer(4), intent(in)            :: iw3, nc
    real(8),    intent(in)            :: cmp(1:,0:)
    real(8),    intent(out)           :: emean, emin, emax
    real(8),    intent(out), optional :: erms, estd

    call sum_of_values(iw3,nc,cmp,emean)
    emean = emean/real(iw3,8)
    emin = ( minval(cmp(nc,1:)) )
    emax = ( maxval(cmp(nc,1:)) )

    if (present(erms)) then
      call sum_of_squares(iw3,nc,cmp,zero,erms)
      erms = sqrt( erms/real(iw3,8) )
    endif
    if (present(estd)) then
      if (iw3 > 1) then
        call sum_of_squares(iw3,nc,cmp,emean,estd)
        estd = sqrt( estd/(real(iw3-1,8)) )
      else
        estd = -one
      endif
    endif

  end subroutine TracerStats

!-------------------------------------------------------------------------------

 subroutine fast2sum(a,b,s,t)
    implicit none
    real(8), intent(in)  :: a,b
    real(8), intent(out) :: s,t
    real(8) :: z
    if (b>a) then
      s=a+b
      z=s-b
      t=a-z
    else
      s=a+b
      z=s-a
      t=b-z
    endif
  end subroutine fast2sum

  subroutine priest_sum(nsize,x,sumout)
    ! double compensated sum
    use constants, only : zero
    implicit none
    integer(4), intent(in)  :: nsize
    real(8),    intent(in)  :: x(0:)
    real(8),    intent(out) :: sumout
    real(8)    :: c, u, y, t, s, v, z
    integer(4) :: i
    s = x(1)
    c = zero
    do i=2,nsize
      call fast2sum(c,x(i),y,u)
      call fast2sum(s,y,t,v)
      z = u + v 
      call fast2sum(t,z,s,c)
    enddo
    sumout = s
  end subroutine priest_sum

  subroutine priest_sum_nc(nsize,nc,x,sumout)
    ! double compensated sum
    use constants, only : zero
    implicit none
    integer(4), intent(in)  :: nsize, nc
    real(8),    intent(in)  :: x(:,0:)
    real(8),    intent(out) :: sumout
    real(8)    :: c, u, y, t, s, v, z
    integer(4) :: i
    s = x(nc,1)
    c = zero
    do i=2,nsize
      call fast2sum(c,x(nc,i),y,u)
      call fast2sum(s,y,t,v)
      z = u + v 
      call fast2sum(t,z,s,c)
    enddo
    sumout = s
  end subroutine priest_sum_nc

!==============================================================================+

  subroutine sum_of_squares_1(nsize,x,mean,sumout)
    ! double compensated sum-of-squares
    use constants, only : zero
    implicit none
    integer(4), intent(in)  :: nsize
    real(8),    intent(in)  :: x(0:), mean
    real(8),    intent(out) :: sumout
    real(8)    :: c, u, y, t, s, v, z
    integer(4) :: i
    s = (x(1) - mean)**2
    c = zero
    do i=2,nsize
      call fast2sum(c,(x(i)-mean)**2,y,u)
      call fast2sum(s,y,t,v)
      z = u + v 
      call fast2sum(t,z,s,c)
    enddo
    sumout = s
  end subroutine sum_of_squares_1

  subroutine sum_of_squares_nc(nsize,nc,x,mean,sumout)
    ! double compensated sum-of-squares, components
    use constants, only : zero
    implicit none
    integer(4), intent(in)  :: nsize, nc
    real(8),    intent(in)  :: x(:,0:), mean
    real(8),    intent(out) :: sumout
    real(8)    :: c, u, y, t, s, v, z
    integer(4) :: i
    s = (x(nc,1) - mean)**2
    c = zero
    do i=2,nsize
      call fast2sum(c,(x(nc,i)-mean)**2,y,u)
      call fast2sum(s,y,t,v)
      z = u + v 
      call fast2sum(t,z,s,c)
    enddo
    sumout = s
  end subroutine sum_of_squares_nc

!==============================================================================+
!  Check for rivers on land and print a warning.
!
!  Rivers on land are permitted, though not actively contributing to
!  mass inflow. 
!==============================================================================+
subroutine checkrivers( mmx, nmx, m, nq, krq, h, ia, nqf, nql )

  !- modules -------------------------------------------------------------------
  use constants, only : ldebug
  use dmi_mpi,   only : dd, mpi_rank


  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in)    :: mmx, nmx, ia
  integer(4), intent(in)    :: m(0:,0:,1:)
  integer(4), intent(inout) :: nq
  integer(4), intent(out)   :: nqf, nql(0:)
  integer(4), intent(inout) :: krq(1:,0:)
  real(8),    intent(in)    :: h(0:)
  

  !- local vars ----------------------------------------------------------------
  integer(4)            :: iq, i, j, k, iam, il, iu, jl, ju, nqn, nqa

  !- Always set No. of rivers to read from file and dummy values ---------------
  nqf    = nq
  nql(0) = 0

  !- some initial settings -----------------------------------------------------
  if (nq < 1) return
  k   = 1
  iam = mpi_rank + 1
  il  = dd(ia)%low_i
  iu  = dd(ia)%up_i
  jl  = dd(ia)%low_j
  ju  = dd(ia)%up_j
  nqn = nq
  nqa = 0

  !- run through all rivers and perform check ----------------------------------
  !
  !   check for out of domain/task:
  do iq=1,nq
    i = krq(1,iq)
    j = krq(2,iq)
    if (i < 1 .or. i >= mmx .or. j < 1 .or. j >= nmx) then
      write(iu06,'(a37, i4)') '**** Error: River No:                ', iq
      write(iu06,'(a37,2i4)') '     is placed outside domain at i,j=', i,j
      call exitme(1) 
    elseif (i < il .or. i > iu .or. j < jl .or. j > ju) then
      nqn = nqn - 1 
      write(iu06,'(a14,i5,a15,2i5)') '**** River No:',iq,' placed at i,j=',i,j
      write(iu06,'(a20,i5,a35)')                                               &
               '     is outside task',iam,' and will be deleted from this task'
      nql(iq) = 0
    else
      nqa = nqa + 1
      if (iq > nqa) then
        krq(:,nqa) = krq(:,iq)
      endif
      nql(iq) = nqa
    endif
  enddo
  !
  !    update No. of rivers in this task:
  if (nqa /= nqn) then
    write(iu06,'(a45,i4)') '**** Error: Unable to set up rivers for task:', iam
    call exitme(1)
  elseif (nq > nqn) then
    krq(:,nqn+1:nq) = -99
    nq = nqn
  endif
  !
  !    possibly, print some info:
  if (ldebug) then
    do iq=1,nq
      i = krq(1,iq)
      j = krq(2,iq)
      if ( m(i,j,k) <= 0 ) then
        write(iu06,'(a30, i4)') '**** Warning: River No:       ', iq
        write(iu06,'(a30,2i4)') '     is placed on land at i,j=', i,j
      else
        write(iu06,'(a22, i4  )') '**** River No:        ', iq
        write(iu06,'(a22,2i4  )') '     is placed at i,j=', i,j
        write(iu06,'(a22, f6.3)') '     toplayer depth=  ', h(m(i,j,k))
      endif
    enddo
  endif

  !- That's all folks! ---------------------------------------------------------
  

end subroutine checkrivers
!===============================================================================

end module cmod_miscsubs
