
module cmod_usage 

  implicit none
  private
  public  :: usage
  private :: printhelp

  character(5), private, parameter :: f = '(a62)', v = '(a53)', w = '(a57)'
  character(4), private, parameter :: x = '(a1)'

contains

!===============================================================================

  subroutine usage(ntracers)
  
    use dmi_mpi_global, only : iu06,mpi_io_rank,mpi_decomp_file
    use exits,          only : exitme
    use io_subs,        only : io_new_unit
    use constants,      only : dyn,toutput,coldst,spinup,archive,ldebug,       &
                               setTSbnd,isponge,mzwei,itsfields,zbnd_type,     &
                               pen_type, updateSTI,usezbcl,kalman,ReadNOA,     &
                               sstrs,precevap,printturb,tom,icedynactive, CDA, &
                               CDB,CDC,init_only, filesizemax, nclss,          &
                               coriolis,Q0, coupler_active, dtwlfile, zref,    &
                               nproci, nprocj, calctrans,botfit,               &
                               elevation_ref, weirs, only_islices, ionml,      &
                               coupler_sponge, simplebfmap, CLA, CLB, CLC,     &
                               CSCL, CD1A, CD1B, CD1C, CL1A, CL1B, CL1C, qflx, &
                               wghtz, wghthar, wghteta, wghtturb, gammadt,     &
                               shortcircuit, iotamodel, iotars, ct10m, itmax,  &
                               coupler_psponge, omp_loadbalance_experiment,    &
                               snow, fastice_test, niceact, nicepar
    use turbmodels,     only : prodtype, pdiffass, srfprod, tidetype
    use coupler,        only : coupler_usage
    use io_control,     only : ios_usage
  
    implicit none

    integer(4), intent(inout) :: ntracers
  
    integer(4)     :: lun, ios, coupler_tracers, coupler_2d_fields, irc
    logical        :: help

    real(8), parameter :: quarter = 900.0_8, tenmins = 600.0_8
    real(8), parameter :: Q0default = -100.0_8

    character(8),  parameter :: fmti='(a15,i3)',fmtl='(a15,l2)',fmii='(a15,i6)'
    character(10), parameter :: fmtr='(a15,f8.2)'
    character(11), parameter :: fmrr='(a15,f10.7)', fmrr2='(a15,f15.9)'

    namelist /optionlist/   dyn,toutput,coldst,spinup,archive,ldebug,nclss,    &
                            setTSbnd,isponge,mzwei,itsfields,pen_type,         &
                            elevation_ref,updateSTI,usezbcl,kalman,ReadNOA,    &
                            sstrs,precevap,printturb,tom,                      &
                            omp_loadbalance_experiment,                        &
                            prodtype, srfprod, help,mpi_io_rank, icedynactive, &
                            init_only,mpi_decomp_file,calctrans, botfit,       &
                            filesizemax,coupler_active, dtwlfile, zref,        &
                            pdiffass, CDA, CDB, CDC, Q0, nproci, tidetype,     &
                            nprocj, weirs,only_islices, ionml, simplebfmap,    &
                            zbnd_type, CSCL, CLA, CLB, CLC, CD1A, CD1B, CD1C,  &
                            CL1A, CL1B, CL1C, wghtz, wghthar, wghteta,         &
                            wghtturb, gammadt, qflx, coriolis, shortcircuit,   &
                            iotamodel, ct10m, itmax, coupler_psponge, snow,    &
                            iotars
  
    !-----Defaults -------------------------------------------------------------
  
    ionml        = .false.
    dyn          = .false.
    mzwei        = .false.
    coldst       = .false.
    spinup       = .false.
    archive      = .false.
    ldebug       = .false.
    toutput      = quarter
    setTSbnd     = .false.
    isponge      = .false.
    updateSTI    = .true.
    itsfields    = 0
    zbnd_type    = 1
    pen_type     = 1
    elevation_ref= .false.
    weirs        = .false.
    usezbcl      = .false.
    kalman       = .false.
    sstrs        = .true.
    ReadNOA      = .true.
    precevap     = .false.
    help         = .false.
    printturb    = .false.
    tom          = .true.
    icedynactive = .true.
    init_only    = .false.
    calctrans    = .false.
    botfit       = .false.
    dtwlfile     = tenmins
    coriolis     = .true.
    Q0           = Q0default
    nproci       = 0
    nprocj       = 0
    nclss        = 7
    only_islices = .false.
    simplebfmap  = .false.
    qflx         = .false.
    ct10m        = .false.
    shortcircuit = .false.
    iotamodel    = .false.
    iotars       = .false.
    itmax        = 1
    snow         = .false.

    ! sub-model stuff:
    coupler_active = .false.
    coupler_sponge = .false.
    coupler_psponge= .false.
    coupler_2d_fields = 0
    coupler_tracers   = 0
  
    !-----Arguments from NAMELIST ----------------------------------------------
    lun = io_new_unit()
    open (unit=lun, file='options.nml', status='old', iostat=ios)
    if (ios /= 0) then
      write(iu06,w)  'Warning: Cannot open file options.nml ...                '
      write(iu06,w)  '         ... continue with default options.              '
      write(iu06,'(a25,i5)') 'I/O problem - iostat is: ',ios
    else
      read (unit=lun, nml=optionlist, iostat=ios)
      if (ios /= 0) then
        write(iu06,w)'Warning: Cannot read optionlist in options.nml file ...  '
        write(iu06,w)'         ... continue without                            '
        write(iu06,'(a25,i5)') 'I/O problem - iostat is: ',ios
      endif
      close (unit=lun)
      if (ldebug) then
        write(iu06,'(a44)') 'Dump of namelist optionlist from option.nml:'
        write(iu06,fmtl) 'icedynactive  =',icedynactive
        if (icedynactive) then
          if (.not.shortcircuit) then
             iotamodel = .false.
             iotars    = .false.
          endif
          if (iotamodel) then
            fastice_test = .false.
          else
            fastice_test = .true.
          endif
        else
          shortcircuit = .false.
          iotamodel    = .false.
          iotars       = .false.
          fastice_test = .false.
        endif
        write(iu06,fmtl) 'dyn           =',dyn
        write(iu06,fmtr) 'toutput       =',toutput
        write(iu06,fmtl) 'coldst        =',coldst
        write(iu06,fmtl) 'spinup        =',spinup
        write(iu06,fmtl) 'archive       =',archive
        write(iu06,fmtl) 'ldebug        =',ldebug
!FIXME: Just outcommented for nightly builds - please rm the "!" in future
!        write(iu06,fmtl) 'setTSbnd      =',setTSbnd
        write(iu06,fmtl) 'isponge       =',isponge
        write(iu06,fmtl) 'mzwei         =',mzwei
        write(iu06,fmti) 'itsfields     =',itsfields
        write(iu06,fmti) 'pen_type      =',pen_type
        write(iu06,fmtl) 'elevation_ref =',elevation_ref
        write(iu06,fmtl) 'updateSTI     =',updateSTI
        write(iu06,fmtl) 'usezbcl       =',usezbcl
        write(iu06,fmtl) 'kalman        =',kalman
        write(iu06,fmtl) 'ReadNOA       =',ReadNOA
        write(iu06,fmtl) 'sstrs         =',sstrs
        write(iu06,fmtl) 'precevap      =',precevap
        write(iu06,fmtl) 'printturb     =',printturb
        write(iu06,fmtl) 'tom           =',tom
        write(iu06,fmti) 'prodtype      =',prodtype
        if (srfprod)      write(iu06,fmtl) 'srfprod       =',srfprod
        if (tidetype > 0) write(iu06,fmti) 'tidetype      =',tidetype
        write(iu06,fmtl) 'help          =',help
        write(iu06,fmii) 'mpi_io_rank   =',mpi_io_rank
        write(iu06,fmtl) 'init_only     =',init_only
        write(iu06,'(a)')'mpi_decomp_file=',trim(mpi_decomp_file)
        write(iu06,fmtl) 'calctrans     =',calctrans
        write(iu06,fmtl) 'coriolis      =',coriolis
        write(iu06,fmtl) 'botfit        =',botfit
        write(iu06,fmtr) 'filesizemax   =',filesizemax
        write(iu06,fmtl) 'coupler_active=',coupler_active
        write(iu06,fmtr) 'dtwlfile      =',dtwlfile
        write(iu06,fmtr) 'zref          =',zref
        write(iu06,fmtl) 'pdiffass      =',pdiffass
        write(iu06,fmrr) 'CDA           =',CDA
        write(iu06,fmrr) 'CDB           =',CDB
        write(iu06,fmrr) 'CDC           =',CDC
        write(iu06,fmtr) 'Q0            =',Q0
        write(iu06,fmii) 'nproci        =',nproci
        write(iu06,fmii) 'nprocj        =',nprocj
        write(iu06,fmtl) 'weirs         =',weirs
        write(iu06,fmtl) 'only_islices  =',only_islices
        write(iu06,fmtl) 'ionml         =',ionml
        write(iu06,fmtl) 'simplebfmap   =',simplebfmap
        write(iu06,fmti) 'zbnd_type     =',zbnd_type
        write(iu06,fmrr) 'CSCL          =',CSCL
        write(iu06,fmrr) 'CLA           =',CLA
        write(iu06,fmrr) 'CLB           =',CLB
        write(iu06,fmrr) 'CLC           =',CLC
        write(iu06,fmrr) 'CD1A          =',CD1A
        write(iu06,fmrr) 'CD1B          =',CD1B
        write(iu06,fmrr) 'CD1C          =',CD1C
        write(iu06,fmrr) 'CL1A          =',CL1A
        write(iu06,fmrr) 'CL1B          =',CL1B
        write(iu06,fmrr) 'CL1C          =',CL1C
        write(iu06,fmtl) 'qflx          =',qflx
        !FIXME: when we have time to change the reference results, include this
        !write(iu06,fmtl) 'snow          =',snow
        if (dyn) then
          write(iu06,fmrr2)  'wghtz         =',wghtz
          write(iu06,fmrr2)  'wghthar       =',wghthar
          write(iu06,fmrr2)  'wghteta       =',wghteta
          write(iu06,fmrr2)  'wghtturb      =',wghtturb
          write(iu06,fmrr2)  'gammadt       =',gammadt
        endif
        if (iotamodel) then
          write(iu06,fmtl)  'iotamodel     =',iotamodel
          write(iu06,fmtl)  'iotars        =',iotars
        endif
      endif
      if (help) call printhelp (iu06)
    endif

    !-----Messages -------------------------------------------------------------
  
    if (dyn) then
      write(iu06,v)     'using dynamical vertical coordinates                 '
    else
      write(iu06,v)     'using static vertical coordinates (z-level)          '
    endif
    if (elevation_ref) then
      write(iu06,v)     'An elevation reference field will be read in ...     '
      write(iu06,v)     ' ... probably pts which lie generally over NN exists '
    endif
    if (.not.coriolis) then
      write(iu06,v)     '    !!! Calculating without Coriolis force!!!        '
    endif
    if (coldst) then
      write(iu06,v)     'coldstarted run                                      '
      if (itsfields == 1) then
        write(iu06,v)   'using initial T/S fields from a restart file         '
      elseif (itsfields == 2) then
        write(iu06,v)   'using initial T/S fields from a separate file        '
      endif
    else
      write(iu06,v)     'restart from previous run                            '
      if (itsfields > 0) then
        write(iu06,v)   'using initial T/S fields is not possible.            '
        write(iu06,v)   'Please clear the -f(ields) flag and re-run.          '
        call exitme(1)
      endif
    endif
    
    if (.not.updateSTI) then
      write(iu06,v)     'keeping S, T and I fixed during simulation           '
    endif
  
    if (spinup) then
      write(iu06,v)     'saving restart and sponges every 6 or 12h            '
    else
      write(iu06,v)     'saving restart and sponges after 6 or 12h only       '
    endif
  
    if (archive) then
      write(iu06,v)     'archiving initial 6 or 12h hindcast                  '
    endif
  
    if (setTSbnd) then
      write(iu06,v)     'T / S sponge layer will be replaced by data from     '
      write(iu06,v)     '     external file if such data is available         '
    endif

    if (isponge) then
      write(iu06,v)     'initializing sponge t/s values from climatology      '
    endif
 
    if (zbnd_type == 1) then
      if (mzwei) then
        write(iu06,v)   'only M2 tides                                        '
      else
        write(iu06,v)   'all tidal constituents                               '
      endif
    elseif (zbnd_type == 2) then
      write(iu06,v)     'radiation boundary condition                         '
    else
      call exitme(1,'Sorry, only zbnd_type 1 and 2 implemented... Exiting.')
    endif
  
    write(iu06,v) 'applying k-omega model with stability functions type4'
    if (tom) then
      write(iu06,v) 'including third-order-moments in buoyancy production.'
    endif

    if (kalman) then
      write(iu06,w) 'applying Kalman SST data assimilation                    '
    endif  

    if (iotamodel) then
       write(iu06,v) 'fast ice model (iotamodel) included                  '
       if (iotars) then
         write(iu06,v) 'read iotamodel restart file                          '
       else
         write(iu06,v) 'reading of iotamodel restart file is turned off      '
       endif
    elseif (icedynactive .and. shortcircuit .and. fastice_test) then
         write(iu06,v) 'run with diagnostical fast ice model                 '
    endif
  

    ! FIXME not correct if ionml is set
    write(iu06,'(a21,i6)') 'output frequency (s):',int(toutput)
    ! FIXME review !!!
    if (ionml) then
      write(iu06,'(a37,i5)') 'requisting an MPI IO server at rank: ',mpi_io_rank
    endif

    if (coupler_active) then
      call coupler_usage(iu06,coupler_tracers, coupler_2d_fields, irc)
      if (irc == 0 .or. irc == 1) then 
        if (irc == 0) coupler_sponge = .true.
        ntracers = ntracers + coupler_tracers
        write(iu06,'(a50)') 'COUPLER warning:                                  '
        write(iu06,'(a50)') '  work as intended if time step size used in the  '
        write(iu06,'(a50)') '  coupler_dynamics is an integer multiple of the  '
        write(iu06,'(a50)') '  advection time step size.                       '
        if (coupler_tracers <= 0) then
          call exitme(1,'COUPLER ERROR: Cannot run coupler without tracers')
        endif
      else
        call exitme(irc,'COUPLER ERROR: Failed to determine usage')
      endif
    endif  
    
    if (ionml) then
      write(iu06,v)     'reading io.nml for more advanced IO control          '
      call ios_usage(iu06, coupler_active, coupler_tracers, coupler_2d_fields)
    else
      call ios_usage(iu06, coupler_active, coupler_tracers, coupler_2d_fields,&
                     .true.)
    endif
 
    if (pen_type == 1) then
      write(iu06,v)     'Using short wave penetration of type 1.              '
    elseif (pen_type == 2) then
      write(iu06,v)     'Using short wave penetration of type 2.              '
    else
      call exitme(1,'Only two SW-penetration types are implemented...Exiting.')
    endif

    if (qflx .and. shortcircuit) then
      niceact = nicepar
    else
      niceact = nicepar - 1
    endif

  end subroutine usage

!===============================================================================
  
  subroutine printhelp(iu)

    use exits, only : exitme

    implicit none

    integer(4), intent(in) :: iu

    write(iu,f) 'Usage: $cmod                                                  '
    write(iu,x) ' '
    write(iu,f) 'circulation model with nested grids                           '
    write(iu,x) ' '
    write(iu,x) ' '
    write(iu,f) 'basic configuration files:                                    '
    write(iu,f) '  cmod.nml: containing namelist cmodcfg for modification of   '
    write(iu,f) '            some model parameters (optional)                  '
    write(iu,f) '  cfg.nml:  configuration of the model run with the           '
    write(iu,f) '            optionlist for turning on/off different options,  '
    write(iu,f) '            the cfglist with No. of nested areas and one      '
    write(iu,f) '  options.nml:  namelist containing an                        '
    write(iu,f) '            optionlist for turning on/off different options,  '
    write(iu,f) '            the cfglist with No. of nested areas and one      '
    write(iu,f) '            namelist called cfgfn for each area describing the'
    write(iu,f) '            file names and nesting structure; must be used    '
    write(iu,f) '            with anything else than two areas.                '
    write(iu,x) ' '
    write(iu,f) 'input files:                                                  '
    write(iu,f) '  coarse_grid_depth  default file name for 3d depth data      '
    write(iu,f) '  fine_grid_depth    default file name for 3d depth data      '
    write(iu,f) '  data_coarse        default file name for grid configuration '
    write(iu,f) '  data_fine          default file name for grid configuration '
    write(iu,f) '  data_surge         grid configuration for surge model       '
    write(iu,f) '  surge_model_depth  depth data for for surge model           '
    write(iu,f) '  tide_const         tidal constituents                       '
    write(iu,f) '  aufpkt             restart fields (binary)                  '
    write(iu,f) '  rwzk[st]bnd        S/T climatology at bnds (2 files)        '
    write(iu,f) '  rwze[cd]no,rwzsdno bnd values for ice vars (3 files)        '
    write(iu,f) '  sponge*            bdy sponges (5 files), also output       '
    write(iu,f) '  metast_neu         job spec: start/end dates                '
    write(iu,f) '  Mccyymm.ddh        atm. forcing file                        '
    write(iu,f) '  Pccyymm.ddh        precipation input data                   '
    write(iu,f) '  surge              open boundary waterlevel data            '
    write(iu,f) '  wlstat.dat         station list, print water level          '
    write(iu,f) '  dynbnds            daily boundary S values (optional)       '
    write(iu,f) '  dynbndt            daily boundary T values (optional)       '
    write(iu,f) '  rwzkbcl            baroclinic correction of srf @ open bnd  '
    write(iu,f) '  sinitN, tinitN     initial S/T for coldstart for domain N   '
    write(iu,f) '  oi_yyyymmddhh.asc  SST or tracer surface field              '
    write(iu,f) '                     (only used with -kalman)                 '
    write(iu,f) '  err_yyyymmddhh.asc SST or tracer surface error field        '
    write(iu,f) '                     (only used with -kalman)                 '
    write(iu,f) '  riverinflow_nn     river inflow (m^3/s), area #nn           '
    write(iu,f) '  turbstations.txt   station list, print turbulence params    '
    write(iu,x) ' '
    write(iu,f) 'output files:                                                 '
    write(iu,f) '  restart            new restart data fields                  '
    write(iu,f) '  sponge*            bdy sponges (5-6 files), also input      '
    write(iu,f) '  tempdat.??         6 or 12h model state (binary)            '
    write(iu,f) '  archive            ..do.. hindcast part                     '
    write(iu,f) '  iccyymm.ddh        ice maps                                 '
    write(iu,f) '  hydro_step         time step in tempdat                     '
    write(iu,f) '  ice_step           time step in ice maps                    '
    write(iu,f) '  zeitschritt        current model time step                  '
    write(iu,f) '  turbout.txt        turbulence output as defined in the      '
    write(iu,f) '                     file turbstations.txt                    '
    write(iu,f) '  consistency_errors:        output of consistency check      '
    write(iu,f) '  BSHcmodp_eisdyn_iteration: iteration problems in ice model  '
    write(iu,f) '  Files possibly defined in wlstat.dat                        '
    write(iu,x) ' '
    write(iu,f) 'MPI task specific output files, task No. indicated by xxx:    '
    write(iu,f) '  qnew.out.xxx       log of river input used on this task     '
    write(iu,f) '  logfile.xxx        log of output msg produced for this task '
    write(iu,x) ' '
    call exitme(1)
  
  end subroutine printhelp

end module cmod_usage 
