module coupler
  implicit none
  private
  integer(4), private, parameter :: coupler_tracers=0
  integer(4), private, save      :: coupler_offset =2
  public :: coupler_usage, coupler_init, coupler_dynamics, coupler_finish
contains

  !PLEASE NOTE:  This is a stub just ment to define the coupling interface.
  !              Compilers may generate warnings like "Unused dummy variable"
  !              for the unsed dummy arguments in the subroutines below,
  !              We will accept these unused variables her.

  subroutine coupler_usage(iu06,nt,nt2d, irc)
    implicit none
    integer(4), intent(in)  :: iu06
    integer(4), intent(out) :: nt    ! number of traces to be inc. in cmp array
    integer(4), intent(out) :: nt2d  ! number of 2d variables
    integer(4), intent(out) :: irc   ! return code; 0 means 'no errors'
    ! Container for reading in run parameters from a seperate namelist file.
    ! CMOD should not have any knowledge about these.
    ! This subroutine will be called from cmod_usage if coupler_active.
    nt   = coupler_tracers
    nt2d = 0
    irc  = 0
  end subroutine coupler_usage

  subroutine coupler_init(ia, narea, iu06, mainarea, ncoff,                    &
                          mmx, nmx, kmx, nq, nz, mmk, kh, hz,                  &
                          components, rho, bndq, bndzk, on_IO_task, irc )
    ! Container for reading in initial values into the arrays and reading
    ! in boundaries (river, open). Can also be used to allocate arrays etc.
    ! for couplers that aren't supposed to mess with the cmp array.
    implicit none
    integer(4), intent(in)    :: ia, narea, iu06, mmx, nmx, kmx, nq, nz, ncoff
    logical,    intent(in)    :: mainarea
    integer(4), intent(in)    :: mmk(:,0:,0:), kh(0:)
    real(8),    intent(in)    :: hz(0:), rho(0:)
    real(8),    intent(inout) :: components(:,0:)
    real(8),    intent(inout) :: bndq(:,:,0:), bndzk(:,:,0:)
    logical,    intent(in)    :: on_IO_task
    integer(4), intent(out)   :: irc

    coupler_offset = ncoff 

    irc = 0
  end subroutine coupler_init

  subroutine coupler_dynamics(ia, iu06, mmx, nmx, kmx, mmk, kh, hz,            &
                              components, dispv, rho, u, v, z, casus,          &
                              atemp, humid, cloud, wu, wv, pl, precip,         &
                              update_counters, idx, irc)
    ! Container for the only call in the timeloop. If the coupler needs to do
    ! different things at different timesteps, e.g. dynamics and output, then
    ! it's up to the coupler_dynamics implementation to do this differentiation.
    implicit none
    integer(4), intent(in)    :: ia, iu06, mmx, nmx, kmx, mmk(:,0:,0:), kh(0:)
    real(8),    intent(in)    :: hz(0:)
    real(8),    intent(inout) :: components(:,0:), dispv(:,0:)
    real(8),    intent(in)    :: u(0:), v(0:), z(0:), rho(0:)
    logical,    intent(in)    :: casus(0:)
    real(8),    intent(in)    :: atemp(0:), humid(0:), cloud(0:)
    real(8),    intent(in)    :: wu(0:), wv(0:), pl(0:), precip(0:)
    logical,    intent(in)    :: update_counters
    integer(4), intent(inout) :: idx(1:,1:)
    integer(4), intent(out)   :: irc

    irc = 0
  end subroutine coupler_dynamics

  subroutine coupler_finish(iu06, irc)
    ! Container for the call after the timeloop. Can be used to close files and
    ! deallocate memory and to do final outputs.
    implicit none
    integer(4), intent(in)   :: iu06
    integer(4), intent(out)  :: irc

    irc = 0
  end subroutine coupler_finish

end module coupler

