module dmi_mpi_global

#if defined (MPI)
  use mpi
#endif

  !- implicit directives -------------------------------------------------------
  implicit none
  private

  !- public vars ---------------------------------------------------------------
  logical,            save, public :: debug_print
  integer(4),         save, public :: iu06, mpi_comm_model
  integer(4),         save, public :: mpi_io_rank = 0
  ! A "Initialisation expression for MPI_DECOMP_FILE padded with blanks"-warning
  ! is accepted here:
  character(LEN=256), save, public :: mpi_decomp_file = 'mpi_decompo.txt'

  !- public methods ------------------------------------------------------------
  public :: dmpi_close, dmpi_abort

contains

!===============================================================================

  subroutine dmpi_close ( flag )
    implicit none

    logical, intent(in) :: flag

#if defined (MPI)
    integer(4) :: ierr

    if (flag) then
      write(*,'(a19)') 'Time to die for MPI'
    else
      write(iu06,'(a19)') 'Time to die for MPI'
    endif
#endif

    if (.not.flag) close (iu06)

#if defined (MPI)
    call mpi_finalize(ierr)
#endif

  end subroutine dmpi_close

  subroutine dmpi_abort ( flag )
    implicit none

    logical, intent(in) :: flag

#if defined (MPI)
    integer(4) :: ierr

    if (flag) then
      write(*,'(a19)') 'Time to abort MPI'
    else
      write(iu06,'(a19)') 'Time to abort MPI'
    endif
#endif

    if (.not.flag) close (iu06)

#if defined (MPI)
    call mpi_abort(MPI_COMM_WORLD, 0, ierr)
#endif

  end subroutine dmpi_abort


!===============================================================================

end module dmi_mpi_global

