module drying_cases

  !- directives ----------------------------------------------------------------
  implicit none
  private

  !- interfaces ----------------------------------------------------------------
  interface drying_case
    module procedure drying_case_1
    module procedure drying_case_2
  end interface

  !- private vars & methods ----------------------------------------------------
  private :: drying_case_1, drying_case_2

  !- public vars & methods -----------------------------------------------------
  public  :: drying_case 

contains 

  ! hack to prevent vel2 from being optional in drying_case_static. When it's
  ! optional compilers have problems inlining this subroutine 
  subroutine drying_case_2 (first,secd,vel,hz1,hz2,z,vel2)
    !- modules -----------------------------------------------------------------
    use constants,   only : hgr, heps, zero
    !- directives --------------------------------------------------------------
    implicit none
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)        :: first, secd
    real(8), intent(in)           :: hz1,hz2
    real(8), intent(in)           :: z(0:)
    real(8), intent(inout)        :: vel
    real(8), intent(out) :: vel2
    !- local vars --------------------------------------------------------------
    real(8)                 :: velnew
    real(8), parameter      :: twohgr = 2.0_8 * hgr
    real(8), parameter      :: khgr   = -0.1_8 / hgr
    real(8), parameter      :: hpeps  = hgr + heps

    velnew = vel
    if     (velnew > zero) then
      if     (hz1 <= hpeps) then
        velnew = zero
      elseif (hz1 <= twohgr) then
        velnew = velnew*exp(khgr*abs(z(first)-z(secd)))
      endif
    elseif (velnew < zero) then
      if     (hz2 <= hpeps) then
        velnew = zero
      elseif (hz2 <= twohgr) then
        velnew = velnew*exp(khgr*abs(z(first)-z(secd)))
      endif
    endif
    
    vel = velnew
    vel2 = velnew

  end subroutine drying_case_2

  subroutine drying_case_1 (first,secd,vel,hz1,hz2,z)
    !- modules -----------------------------------------------------------------
    use constants,   only : hgr, heps, zero
    !- directives --------------------------------------------------------------
    implicit none
    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)        :: first, secd
    real(8), intent(in)           :: hz1,hz2
    real(8), intent(in)           :: z(0:)
    real(8), intent(inout)        :: vel
    !- local vars --------------------------------------------------------------
    real(8)                 :: velnew
    real(8), parameter      :: twohgr = 2.0_8 * hgr
    real(8), parameter      :: khgr   = -0.1_8 / hgr
    real(8), parameter      :: hpeps  = hgr + heps

    velnew = vel
    if     (velnew > zero) then
      if     (hz1 <= hpeps) then
        velnew = zero
      elseif (hz1 <= twohgr) then
        velnew = velnew*exp(khgr*abs(z(first)-z(secd)))
      endif
    elseif (velnew < zero) then
      if     (hz2 <= hpeps) then
        velnew = zero
      elseif (hz2 <= twohgr) then
        velnew = velnew*exp(khgr*abs(z(first)-z(secd)))
      endif
    endif
    
    vel = velnew

  end subroutine drying_case_1

end module drying_cases
