module eisdyn
  use constants, only : zero, one, qrt
  implicit none
  private

  public  :: eisdyn_int
  private :: eisdynp1, eisdyn_sss, init_uvse, press, limit_ice, ice_stats,     &
             stress_stats

  real(8), parameter :: quart = qrt
  real(8), parameter :: elast = 1.e-4_8
  real(8), parameter :: corr  = one/(one+(elast/(one+elast))**2)
  real(8), parameter :: eee   = corr/(one+elast)
  
  private :: elast, corr, eee

contains

  subroutine eisdynp1( me,muvs,nuvs2d,induvs,casus,tiefe,zeta,                 &
                     ia,ja,ieq,meq,ina,neq )
    !- modules -----------------------------------------------------------------
    use constants , only: hgr 

    !- directives --------------------------------------------------------------
    implicit none

    !- parameters --------------------------------------------------------------
    real(8) , parameter :: hmin = hgr

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: nuvs2d
    integer(4), intent(in)    :: muvs(:,0:,0:)          ! (8,0:imax,0:jmax)
    integer(4), intent(in)    :: induvs(:,:)            ! (3,nuvs2d)
    integer(4), intent(in)    :: me(0:,0:)              ! (0:imax,0:jmax)
    integer(4), intent(inout) :: ina, neq
    integer(4), intent(out)   :: ia(:), ja(:)           ! (nuvs2d+1), (9*nuvs2d)
    integer(4), intent(inout) :: ieq(:), meq(:)         ! (nuvs2d), (nuvs2d)
    logical,    intent(in)    :: casus(0:)
    real(8),    intent(in)    :: tiefe(0:), zeta(0:)

    !- local vars --------------------------------------------------------------
    integer(4) :: i, j, nl, nlu, nn, nt, nu, nr, no, nol, nul, nur, nlo
    integer(4) :: muvsnt, mi, nro, nru, nor
    logical    :: icel, icer, iceo, iceu, iceol, iceor, iceul, iceur

    ! --------------------------------------------------------------------------
    do nn = 1,nuvs2d
 
      i = induvs(1,nn)
      j = induvs(2,nn)
      nt = induvs(3,nn)
 
      if ( nt == 1 ) then
        nl = me(i,j)
        nr = me(i,j+1)
        icel = .not.casus(nl)
        icer = .not.casus(nr)
        if ( icel .or. icer ) then
          if ( min(tiefe(nl)+zeta(nl),tiefe(nr)+zeta(nr)) > hmin ) then

            muvsnt      = muvs(nt,i,j) 
            neq         = neq + 1
            ieq(muvsnt) = neq
            meq(neq)    = muvsnt
            ia(neq)     = ina
            if ( icel .and. muvs(2,i-1,j) > 0 ) then
              if ( (tiefe(me(i-1,j))+zeta(me(i-1,j))) > hmin ) then
                ja(ina) = muvs(2,i-1,j)
                ina     = ina + 1
              endif
            endif
            if ( icer .and. muvs(2,i-1,j+1) > 0 ) then
              if ( (tiefe(me(i-1,j+1))+zeta(me(i-1,j+1))) > hmin ) then
                ja(ina) = muvs(2,i-1,j+1)
                ina     = ina + 1
              endif
            endif
            ja(ina) = muvsnt
            ina     = ina + 1
            if ( icel .and. muvs(2,i,j) > 0 ) then
              if ( (tiefe(me(i+1,j))+zeta(me(i+1,j))) > hmin ) then
                ja(ina) = muvs(2,i,j)
                ina     = ina + 1
              endif
            endif
            if ( icer .and. muvs(2,i,j+1) > 0 ) then
              if ( (tiefe(me(i+1,j+1))+zeta(me(i+1,j+1))) > hmin ) then
                ja(ina) = muvs(2,i,j+1)
                ina     = ina + 1
              endif
            endif
            if ( muvs(8,i-1,j) > 0 ) then
              nol = me(i-1,j)
              nor = me(i-1,j+1)
              if (.not.(casus(nol).and.casus(nor)) .or. icel .and. icer) then
                if (min(tiefe(nol)+zeta(nol),tiefe(nor)+zeta(nor)) > hmin) then
                  ja(ina) = muvs(8,i-1,j)
                  ina     = ina + 1
                endif
              endif
            endif
            if ( icel ) then
              ja(ina)   = muvs(6,i,j)
              ja(ina+1) = muvs(7,i,j)
              ina       = ina + 2
            endif
            if ( muvs(8,i,j) > 0 ) then
              nul = me(i+1,j)
              nur = me(i+1,j+1)
              if (.not.(casus(nul).and.casus(nur)) .or. icel .and. icer) then
                if (min(tiefe(nul)+zeta(nul),tiefe(nur)+zeta(nur)) > hmin) then
                  ja(ina) = muvs(8,i,j)
                  ina     = ina + 1
                endif
              endif
            endif
            if ( icer ) then
              ja(ina)   = muvs(6,i,j+1)
              ja(ina+1) = muvs(7,i,j+1)
              ina       = ina + 2
            endif
 
          endif
        endif
 
      elseif ( nt == 2 ) then
        no = me(i,j)
        nu = me(i+1,j)
        iceo = .not.casus(no)
        iceu = .not.casus(nu)
        if ( iceo .or. iceu ) then
          if ( min(tiefe(no)+zeta(no),tiefe(nu)+zeta(nu)) > hmin ) then
 
            muvsnt      = muvs(nt,i,j)
            neq         = neq + 1
            ieq(muvsnt) = neq
            meq(neq)    = muvsnt
            ia(neq)     = ina
            if ( iceo .and. muvs(1,i,j-1) > 0 ) then
              if ( (tiefe(me(i,j-1))+zeta(me(i,j-1))) > hmin ) then
                ja(ina) = muvs(1,i,j-1)
                ina     = ina + 1
              endif
            endif
            if ( iceo .and. muvs(1,i,j) > 0 ) then
              if ( (tiefe(me(i,j+1))+zeta(me(i,j+1))) > hmin ) then
                ja(ina) = muvs(1,i,j)
                ina     = ina + 1
              endif
            endif
            ja(ina) = muvsnt
            ina     = ina + 1
            if ( iceu .and. muvs(1,i+1,j-1) > 0 ) then
              if ( (tiefe(me(i+1,j-1))+zeta(me(i+1,j-1))) > hmin ) then
                ja(ina) = muvs(1,i+1,j-1)
                ina     = ina + 1
              endif
            endif
            if ( iceu .and. muvs(1,i+1,j) > 0 ) then
              if ( (tiefe(me(i+1,j+1))+zeta(me(i+1,j+1))) > hmin ) then
                ja(ina) = muvs(1,i+1,j)
                ina     = ina + 1
              endif
            endif
            if ( muvs(8,i,j-1) > 0 ) then
              nlo = me(i,j-1)
              nlu = me(i+1,j-1)
              if (.not.casus(nlo) .or. .not.casus(nlu) .or.                    &
                  iceo .and. iceu                          ) then
                if (min(tiefe(nlo)+zeta(nlo),tiefe(nlu)+zeta(nlu)) > hmin) then
                  ja(ina) = muvs(8,i,j-1)
                  ina     = ina + 1
                endif
              endif
            endif
            if ( iceo ) then
              ja(ina)   = muvs(6,i,j)
              ja(ina+1) = muvs(7,i,j)
              ina       = ina + 2
            endif
            if ( muvs(8,i,j) > 0 ) then
              nro = me(i,j+1)
              nru = me(i+1,j+1)
              if (.not.casus(nro) .or. .not.casus(nru) .or.                    &
                  iceo .and. iceu                          ) then
                if (min(tiefe(nro)+zeta(nro),tiefe(nru)+zeta(nru)) > hmin) then
                  ja(ina) = muvs(8,i,j)
                  ina     = ina + 1
                endif
              endif
            endif
            if ( iceu ) then
               ja(ina)   = muvs(6,i+1,j)
               ja(ina+1) = muvs(7,i+1,j)
               ina       = ina + 2
            endif
 
          endif
        endif
 
      elseif ( nt == 3 .or. nt == 4 ) then
        mi = me(i,j)
        if ( .not.casus(mi) ) then
          if ( (tiefe(mi)+zeta(mi)) > hmin ) then
 
            muvsnt      = muvs(nt,i,j)
            neq         = neq + 1
            ieq(muvsnt) = neq
            meq(neq)    = muvsnt
            ia(neq)     = ina
            if ( muvs(2,i-1,j) > 0 ) then
              if ( (tiefe(me(i-1,j))+zeta(me(i-1,j))) > hmin ) then
                ja(ina) = muvs(2,i-1,j)
                ina     = ina + 1
              endif
            endif
            if ( muvs(1,i,j-1) > 0 ) then
              if ( (tiefe(me(i,j-1))+zeta(me(i,j-1))) > hmin ) then
                ja(ina) = muvs(1,i,j-1)
                ina     = ina + 1
              endif
            endif
            if ( muvs(1,i,j) > 0 ) then
              if ( (tiefe(me(i,j+1))+zeta(me(i,j+1))) > hmin ) then
                ja(ina) = muvs(1,i,j)
                ina     = ina + 1
              endif
            endif
            if ( muvs(2,i,j) > 0 ) then
              if ( (tiefe(me(i+1,j))+zeta(me(i+1,j))) > hmin ) then
                ja(ina) = muvs(2,i,j)
                ina     = ina + 1
              endif
            endif
            ja(ina) = muvsnt
            ina     = ina + 1
 
          endif
        endif
 
      elseif ( nt == 5 ) then
        nol = me(i,j)
        nor = me(i,j+1)
        nul = me(i+1,j)
        nur = me(i+1,j+1)
        if (min(tiefe(nol)+zeta(nol),tiefe(nor)+zeta(nor),                     &
                tiefe(nul)+zeta(nul),tiefe(nur)+zeta(nur)) > hmin) then
          iceol = .not.casus(nol)
          iceor = .not.casus(nor)
          iceul = .not.casus(nul)
          iceur = .not.casus(nur)
          if ( iceol .and. iceor .or.                                          &
               iceol .and. iceul .or.                                          &
               iceol .and. iceur .or.                                          &
               iceor .and. iceul .or.                                          &
               iceor .and. iceur .or.                                          &
               iceul .and. iceur      ) then
 
            muvsnt      = muvs(nt,i,j)
            neq         = neq + 1
            ieq(muvsnt) = neq
            meq(neq)    = muvsnt
            ia(neq)     = ina
            if ( iceol .or. iceor) then
              ja(ina) = muvs(1,i,j)
              ina     = ina + 1
            endif
            if ( iceol .or. iceul ) then
              ja(ina) = muvs(2,i,j)
              ina     = ina + 1
            endif
            if ( iceor .or. iceur ) then
              ja(ina) = muvs(2,i,j+1)
              ina     = ina + 1
            endif
            if ( iceul .or. iceur ) then
              ja(ina) = muvs(1,i+1,j)
              ina     = ina + 1
            endif
            ja(ina) = muvsnt
            ina     = ina + 1
 
          endif
        endif
 
      elseif ( nt == 6 .or. nt == 7 ) then
        mi = me(i,j)
        if ( .not.casus(mi) ) then
          if ( (tiefe(mi)+zeta(mi)) > hmin ) then
 
            muvsnt      = muvs(nt,i,j)
            neq         = neq + 1
            ieq(muvsnt) = neq
            meq(neq)    = muvsnt
            ia(neq)     = ina
            ja(ina)     = muvs(nt-3,i,j)
            ja(ina+1)   = muvsnt
            ina         = ina + 2
 
          endif
        endif
 
      elseif ( nt == 8 ) then
        nol = me(i,j)
        nor = me(i,j+1)
        nul = me(i+1,j)
        nur = me(i+1,j+1)
        if (min(tiefe(nol)+zeta(nol),tiefe(nor)+zeta(nor),                     &
                tiefe(nul)+zeta(nul),tiefe(nur)+zeta(nur)) > hmin) then
          iceol = .not.casus(nol)
          iceor = .not.casus(nor)
          iceul = .not.casus(nul)
          iceur = .not.casus(nur)
          if ( iceol .and. iceor .or.                                          &
               iceol .and. iceul .or.                                          &
               iceol .and. iceur .or.                                          &
               iceor .and. iceul .or.                                          &
               iceor .and. iceur .or.                                          &
               iceul .and. iceur      ) then
 
            muvsnt      = muvs(nt,i,j)
            neq         = neq + 1
            ieq(muvsnt) = neq
            meq(neq)    = muvsnt
            ia(neq)     = ina
            ja(ina)     = muvs(5,i,j)
            ja(ina+1)   = muvsnt
            ina         = ina + 2
 
          endif
        endif
      endif

    enddo

  end subroutine eisdynp1

  subroutine eisdyn_int(me,muvs,n2d,ind,nuvs2d,induvs,casus,ueis,veis,ice,     &
                        dxj,dyi,dt,nbpz,krz,tiefe,uwas,vwas,zeta,tauxs,tauys,  &
                        cosphi,fcorio,ctim,izeits,iceite,iuvp,nthreads)

    !- modules -----------------------------------------------------------------
    use dmi_mpi_global, only : iu06 
    use dmi_omp,        only : domp_get_domain, domp_get_thread_no
    use constants ,     only : g, hgr ,rhow, rhoi, cdrag, ldebug
    use io_subs ,       only : io_new_unit
    use ice_solver,     only : solver

    !- directives --------------------------------------------------------------
    implicit none

    !- parameters --------------------------------------------------------------
    real(8) , parameter :: hmin = hgr

    !- arguments ---------------------------------------------------------------
    character(19), intent(in)    :: ctim
    integer(4),    intent(in)    :: izeits,n2d,nbpz,nuvs2d
    integer(4),    intent(in)    :: ind(:,:)
    integer(4),    intent(in)    :: me(0:,0:)                ! (0:imax,0:jmax)
    logical,       intent(in)    :: casus(0:)
    real(8),       intent(in)    :: dt
    real(8),       intent(in)    :: cosphi(:,0:),fcorio(:,0:)! (2,0:imax)
    real(8),       intent(in)    :: dxj(0:), dyi(0:)         ! (0:jmax),(0:imax)
    real(8),       intent(in)    :: tiefe(0:), zeta(0:)
    real(8),       intent(in)    :: tauxs(0:), tauys(0:)
    real(8),       intent(in)    :: uwas(0:), vwas(0:)
    integer(4),    intent(in)    :: induvs(:,:)              ! (3,nuvs2d)
    integer(4),    intent(in)    :: krz(:,0:)                ! (3,0:nbpz)
    integer(4),    intent(in)    :: muvs(:,0:,0:)            ! (8,0:imax,0:jmax)
    integer(4),    intent(in)    :: iuvp(:), nthreads

    integer(4),    intent(out)   :: iceite
    real(8),       intent(inout) :: ice(1:,0:)
    real(8),       intent(inout) :: ueis(0:), veis(0:)

    !- local vars --------------------------------------------------------------
    real(8)    :: am, diffm, diffu, diffuv, diffv, eiscox, eiscoy,             &
                  eisdix, eisdiy, eta, ppp, strc, zet, ddt,dx,dy,aax1,aax2,    &
                  aay1,aay2,aay3,aay4,ddx,ddy,aaf,ded, tmp, rms, rmsu, rmsv,   &
                  uvs3, uvs4, uvs5, uvs6, uvs7, uvs8, np1, rn
    integer(4) :: i, ii, ina, j, jj, lll, lun, n, nb, neq, nn, nnn, nt,        &
                  induvs3, ind3, nd, mi, mee, n2dl, n2du, mja, tnum, it,       &
                  mnn, mne, mww, msw, mss, mse, mnw, im1, ip1, jm1, jp1, ioff
    integer(4), parameter :: MAXITERATIONS = 10
    real(8),    parameter :: e1 = 0.01_8, d1 = 0.025_8, epsl = 100.0_8
    real(8),    parameter :: w1 = 0.2_8, w2 = one - w1
    real(8),    parameter :: epss = 0.01_8, epsi = 0.001_8, half = 0.5_8
    real(8),    parameter :: compr_lim = 2.e7_8, visko = one/4.0_8
    real(8),    parameter :: exc = 2.0_8, dexc = one/exc
    real(8),    parameter :: drhoe = one/rhoi, ge = g, drr = cdrag*rhow*drhoe
    real(8),    parameter :: tiny_val = 1.0e-10_8
    real(8)    :: eispr(0:n2d), diffq(4,nthreads), nuv(2,nthreads)
    real(8)    :: as(nuvs2d), aa(9*nuvs2d), ar(nuvs2d)
    real(8)    :: ss3(0:iuvp(2)), ss4(0:iuvp(2)), ss5(0:iuvp(3))
    real(8)    :: ss6(0:iuvp(2)), ss7(0:iuvp(2)), ss8(0:iuvp(3))
    integer(4) :: ieq(nuvs2d), meq(nuvs2d), ia(nuvs2d+1), ja(9*nuvs2d)
    integer(4) :: idx(3:8,0:n2d)

    ! --------------------------------------------------------------------------

    ! First, examine the ice field, set the structure of the system of eqs -----
    ieq(:) = 0
    meq(:) = 0
    neq = 0
    ina = 1
    call eisdynp1(me,muvs,nuvs2d,induvs,casus,tiefe,zeta,ia,ja,ieq,meq,ina,neq)

    ! no iterations so far:
    lll = 0

    ! save actual number of iterations:
    iceite = lll

    ! let's get out of here if there's nothing to do ---------------------------
    if (neq <= 0) return

    ! Further initializations --------------------------------------------------

    ! finalize the row/column of the compactly stored nonzero elements:
    ia(neq+1) = ina
    do n=1,ina-1
      ja(n) = ieq(ja(n))
    enddo
 
    ! model discretization:
    ddt = one/dt
    dx  = dxj(1)
    dy  = dyi(1)

    ! some initializations for iteration
    ! fake a large diff initially:
    diffm = epsl*epsi
    rms   = epss*epsi
 
    ! Initialize coefficient arrays; aa is the matrix, ar is the RHS:
    aa(1:9*neq) = zero
    ar(1:neq)   = zero

!$OMP PARALLEL DEFAULT (shared)                                                &
!$OMP PRIVATE (n, i, j, n2dl, n2du, mi, induvs3, nn, nd, mee, eiscox, eisdix,  &
!$OMP          ddx, ded, diffu, diffv, diffuv, strc, aax1, aax2, aay1, aay2,   &
!$OMP          aay3, aay4, aaf, nnn, ind3, jj, ii, ppp, zet, eta, nb, nt, am,  &
!$OMP          tmp, mja, tnum, it, uvs3, uvs4, uvs5, uvs6, uvs7, uvs8, rn, np1,&
!$OMP          mnn, mne, mww, msw, mss, mse, mnw, im1, ip1, jm1, jp1, ioff,    &
!$OMP          rmsu, rmsv )

    ! first, get common domain decompo:
    call domp_get_domain (1, n2d, n2dl, n2du)

!$OMP MASTER
    idx(3:8,0) = 0
!$OMP END MASTER
    do mi=n2dl,n2du
      i = ind(1,mi)
      j = ind(2,mi)
      idx(3,mi) = max(0, muvs(3,i,j)-iuvp(1))
      idx(4,mi) = max(0, muvs(4,i,j)-iuvp(1)-  iuvp(2))
      idx(5,mi) = max(0, muvs(5,i,j)-iuvp(1)-2*iuvp(2))
      idx(6,mi) = max(0, muvs(6,i,j)-iuvp(1)-2*iuvp(2)-iuvp(3))
      idx(7,mi) = max(0, muvs(7,i,j)-iuvp(1)-3*iuvp(2)-iuvp(3))
      idx(8,mi) = max(0, muvs(8,i,j)-iuvp(1)-4*iuvp(2)-iuvp(3))
    enddo

    ! shallow water treatement:
    do mi=n2dl,n2du
      ! no transport to/from "dried" points:
      if (tiefe(mi) + zeta(mi) <= hmin) then
        i = ind(1,mi)
        j = ind(2,mi)
        ueis(mi)        = zero
        ueis(me(i,j-1)) = zero
        veis(mi)        = zero
        veis(me(i-1,j)) = zero
      endif
    enddo
!$OMP BARRIER

    ! make sure to have nice limits on thickness and concentration:
    call limit_ice(n2dl, n2du, casus, ice)

    !  Compute the ice strength eispr:
    call press (n2dl, n2du, casus, ice, eispr, ueis, veis)

    ! initialize uvse
    call init_uvse(n2dl, n2du, idx, ss3, ss4, ss5, ss6, ss7, ss8)

    ! Assign values to the entries of equations for momentum (induvs3=1,2)
    ! and unlimited viscous stress (induvs3=3,4,5)
!$OMP BARRIER
    do mi=n2dl,n2du
      i = ind(1,mi)
      j = ind(2,mi)
      do induvs3=1,5 
        n  = muvs(induvs3,i,j)
        if (n <= 0) cycle
        nn = ieq(n)
        if (nn <= 0) cycle
 
        do nd = ia(nn) , ia(nn+1) - 1
          if (ja(nd) == nn) exit 
        enddo
        ! nd is now the diagonal element of equation #nn
 
        ! U-mom equation
        if (induvs3 == 1) then
          im1 = i-1
          jp1 = j+1
          mee = me(i,jp1)
          mnn = me(im1,j)
          mne = me(im1,jp1)

          eiscox = half*(ice(2,mi)+ice(2,mee))
          eisdix = half*(ice(1,mi)+ice(1,mee))/sqrt(eiscox) + e1

          ddx = one/(dx*cosphi(1,i))
          ded = one/eisdix

          diffu = ueis(mi) - uwas(mi)
          diffv = quart*(  veis(mnn) - vwas(mnn) + veis(mne) - vwas(mne)       &
                         + veis(mi)  - vwas(mi)  + veis(mee) - vwas(mee)  )
          diffuv = sqrt(diffu**2+diffv**2) + d1
          strc   = drr*diffuv

          aax1 = ded*drhoe*ddx
          aax2 = ded*drhoe*ddx*dexc
          aay1 = ded*drhoe*dexc*cosphi(2,im1)/(dy*cosphi(1,i))
          aay2 = ded*drhoe*dexc*cosphi(2,i  )/(dy*cosphi(1,i))
          aaf  = -fcorio(1,i)*quart

          ! Now we are ready to determine ar(:) and aa(:).
          ! The next two comments are some left overs from original code:
          !     aa(nd)= ... +0.5*strc*ded
          !     ar(nn)= ... +strc*(uwas(mi)-0.5*ueis(mi))*ded
          aa(nd) = ddt + strc*ded
          ar(nn) = ueis(mi)*ddt - (zeta(mee)-zeta(mi))*ge*ddx                  &
                 + ( (tauxs(mi) - half*(eispr(mee)-eispr(mi))*ddx)*drhoe       &
                    + strc*uwas(mi)                                     )*ded
          do nnn = ia(nn),ia(nn+1) - 1
            mja  = meq(ja(nnn))
            ii   = induvs(1,mja)
            jj   = induvs(2,mja)
            ind3 = induvs(3,mja)
            if     (ind3 == 6) then
              if     (jj == jp1) then
                aa(nnn) = -aax1
              elseif (jj == j  ) then
                aa(nnn) =  aax1
              endif
            elseif (ind3 == 7) then
              if     (jj == jp1) then
                aa(nnn) = -aax2
              elseif (jj == j  ) then
                aa(nnn) =  aax2
              endif
            elseif (ind3 == 8) then
              if     (ii == im1) then
                aa(nnn) = -aay1
              elseif (ii == i  ) then
                aa(nnn) = +aay2
              endif
            elseif (ind3 == 2 .and. ii <= i .and. jj >= j) then
              aa(nnn) = aaf
            endif
          enddo
 
        ! V-mom equation
        elseif (induvs3 == 2) then
          ip1 = i+1
          jm1 = j-1
          mss = me(ip1,j)
          mww = me(i,jm1)
          msw = me(ip1,jm1)
  
          eiscoy = half*(ice(2,mi)+ice(2,mss))
          eisdiy = half*(ice(1,mi)+ice(1,mss))/sqrt(eiscoy) + e1

          ddy = one/dy
          ded = one/eisdiy

          diffv = veis(mi) - vwas(mi)
          diffu = quart*(  ueis(mww) - uwas(mww) + ueis(mi)  - uwas(mi)        &
                         + ueis(msw) - uwas(msw) + ueis(mss) - uwas(mss) )
          diffuv = sqrt(diffu**2+diffv**2) + d1
          strc   = drr*diffuv
  
          aax1 = ded*drhoe*dexc/(dx*cosphi(2,i))
          aay1 = ded*drhoe*cosphi(1,i  )/(dy*cosphi(2,i))
          aay2 = ded*drhoe*cosphi(1,ip1)/(dy*cosphi(2,i))
          aay3 = aay1*dexc
          aay4 = aay2*dexc
          aaf  = fcorio(2,i)*quart
  
          ! Now we are ready to determine ar(:) and aa(:).
          ! The next two comments are some left overs from original code:
          !     aa(nd)= ... +0.5*strc*ded
          !     ar(nn)= ... +strc*(vwas(mi)-0.5*veis(mi))*ded
          aa(nd) = ddt + strc*ded
          ar(nn) = veis(mi)*ddt - (zeta(mi)-zeta(mss))*ge*ddy                  &
                 + ( (tauys(mi) - half*(eispr(mi)-eispr(mss))*ddy)*drhoe       &
                    + strc*vwas(me(i,j))                                 )*ded
          do nnn = ia(nn),ia(nn+1) - 1
            mja  = meq(ja(nnn))
            ii   = induvs(1,mja)
            jj   = induvs(2,mja)
            ind3 = induvs(3,mja)
            if     (ind3 == 6) then
              if     (ii == i  ) then
                aa(nnn) = -aay1
              elseif (ii == ip1)  then
                aa(nnn) =  aay2
              endif
            elseif (ind3 == 7) then
              if     (ii == i  ) then
                aa(nnn) =  aay3
              elseif (ii == ip1) then
                aa(nnn) = -aay4
              endif
            elseif (ind3 == 8) then
              if     (jj == j  ) then
                aa(nnn) = -aax1
              elseif (jj == jm1) then
                aa(nnn) =  aax1
              endif
            elseif (ind3 == 1 .and. jj <= j .and. ii >= i) then
              aa(nnn) = aaf
            endif
          enddo
 
        elseif (induvs3 == 3) then
          im1    = i-1
          jm1    = j-1
          ppp    = eispr(mi)
          zet    = compr_lim*ppp
          aax1   = zet/(dx*cosphi(1,i))
          aay1   = zet/dy
          aa(nd) = one
          ar(nn) = half*ppp
          do nnn = ia(nn),ia(nn+1) - 1
            mja  = meq(ja(nnn))
            ind3 = induvs(3,mja)
            if (ind3 == 1) then
              jj = induvs(2,mja)
              if     (jj == jm1) then
                aa(nnn) =  aax1
              elseif (jj == j  ) then
                aa(nnn) = -aax1
              endif
            elseif (ind3 == 2) then
              ii = induvs(1,mja)
              if     (ii == im1) then
                aa(nnn) = -aay1
              elseif (ii == i  ) then
                aa(nnn) =  aay1
              endif
            endif
          enddo
 
        elseif (induvs3 == 4) then
          im1    = i-1
          jm1    = j-1
          ppp    = eispr(mi)
          eta    = compr_lim*ppp*visko*exc
          aax1   = eta/(dx*cosphi(1,i))
          aay1   = eta/dy
          aa(nd) = one
          ar(nn) = zero  
          do nnn = ia(nn),ia(nn+1) - 1
            mja  = meq(ja(nnn))
            ind3 = induvs(3,mja)
            if     (ind3 == 1) then
              jj = induvs(2,mja)
              if     (jj == jm1) then
                aa(nnn) =  aax1
              elseif (jj == j  ) then
                aa(nnn) = -aax1
              endif
            elseif (ind3 == 2) then
              ii = induvs(1,mja)
              if     (ii == im1) then
                aa(nnn) =  aay1
              elseif (ii == i  ) then
                aa(nnn) = -aay1
              endif
            endif
          enddo
   
        elseif (induvs3 == 5) then
          ip1 = i+1
          jp1 = j+1
          ppp = quart*( eispr(mi)        + eispr(me(i,jp1))                    &
                       +eispr(me(ip1,j)) + eispr(me(ip1,jp1)) )
          eta = compr_lim*ppp*visko*exc
          aax1 = eta/(dx*cosphi(2,i))
          aay1 = eta/dy
          aa(nd) = one
          ar(nn) = zero  
          do nnn = ia(nn),ia(nn+1) - 1
            mja  = meq(ja(nnn))
            ind3 = induvs(3,mja)
            if     (ind3 == 1) then
              ii = induvs(1,mja)
              if     (ii == i   .and. ieq(muvs(1,ip1,j)) > 0) then
                aa(nnn) = -aay1
              elseif (ii == ip1 .and. ieq(muvs(1,i,  j)) > 0) then
                aa(nnn) =  aay1
              endif
            elseif (ind3 == 2) then
              jj = induvs(2,mja)
              if     (jj == j   .and. ieq(muvs(2,i,jp1)) > 0) then
                aa(nnn) =  aax1
              elseif (jj == jp1 .and. ieq(muvs(2,i,j  )) > 0) then
                aa(nnn) = -aax1
              endif
            endif
          enddo
        endif
 
      enddo ! induvs3
    enddo   ! mi

    if ( nbpz > 0 ) then
!$OMP BARRIER
! usually too small to do in parallel, but let's have a look at that later
!$OMP MASTER
      do nb = 1 , nbpz
        i = krz(1,nb)
        j = krz(2,nb)
        nt = abs(krz(3,nb))
        if (.not.(nt == 1 .or. nt == 2)) cycle
 
        n = muvs(3,i,j)  !  assuming that i,j is ALWAYS a wetpoint
        nn = ieq(n)
        if ( nn > 0 ) then
          do nnn = ia(nn) , ia(nn+1) - 1
            induvs3 = induvs(3,meq(ja(nnn)))
            if ((nt == 1 .and. induvs3 == 1) .or. (nt == 2 .and. induvs3 == 2))&
              aa(nnn) = zero  
          enddo
        endif
 
        n = muvs(4,i,j)  !  assuming that i,j is ALWAYS a wetpoint
        nn = ieq(n)
        if ( nn > 0 ) then
          do nnn = ia(nn) , ia(nn+1) - 1
            induvs3 = induvs(3,meq(ja(nnn)))
            if ((nt == 1 .and. induvs3 == 1) .or. (nt == 2 .and. induvs3 == 2))&
              aa(nnn) = zero  
          enddo
        endif
 
      enddo
!$OMP END MASTER
    endif  ! nbpz > 0
 
    ! Initial values of stresses for the iteration:
!$OMP BARRIER
    do mi=n2dl,n2du
      n  = idx(3,mi)
      nn = ieq(n+iuvp(1))
      if (nn > 0) then
        do nd = ia(nn),ia(nn+1) - 1
          if (ja(nd) == nn) exit 
        enddo
        ! nd is now the diagonal element of equation #nn
        am  = one/aa(nd)
        tmp = ar(nn)*am
        do nnn = ia(nn),ia(nn+1) - 1
          mja  = meq(ja(nnn))
          ind3 = induvs(3,mja)
          if (ind3 == 1) then
            tmp = tmp - aa(nnn)*am*ueis(me(induvs(1,mja),induvs(2,mja)))
          elseif (ind3 == 2) then
            tmp = tmp - aa(nnn)*am*veis(me(induvs(1,mja),induvs(2,mja)))
          endif
        enddo
        ! Initial values of unlimited viscous stress for the iteration:
        ss3(n) = tmp
        ! Initial values of plastically limited stress for the iteration:
        ss6(idx(6,mi)) = tmp
      endif

      n  = idx(4,mi)
      nn = ieq(n+iuvp(1)+iuvp(2))
      if (nn > 0) then
        do nd = ia(nn),ia(nn+1) - 1
          if (ja(nd) == nn) exit 
        enddo
        ! nd is now the diagonal element of equation #nn
        am  = one/aa(nd)
        tmp = ar(nn)*am
        do nnn = ia(nn),ia(nn+1) - 1
          mja  = meq(ja(nnn))
          ind3 = induvs(3,mja)
          if (ind3 == 1) then
            tmp = tmp - aa(nnn)*am*ueis(me(induvs(1,mja),induvs(2,mja)))
          elseif (ind3 == 2) then
            tmp = tmp - aa(nnn)*am*veis(me(induvs(1,mja),induvs(2,mja)))
          endif
        enddo
        ! Initial values of unlimited viscous stress for the iteration:
        ss4(n) = tmp
        ! Initial values of plastically limited stress for the iteration:
        ss7(idx(7,mi)) = tmp
      endif

      n  = idx(5,mi)
      if (n <= 0) cycle
      nn = ieq(n+iuvp(1)+2*iuvp(2))
      if (nn > 0) then
        do nd = ia(nn),ia(nn+1) - 1
          if (ja(nd) == nn) exit 
        enddo
        ! nd is now the diagonal element of equation #nn
        am  = one/aa(nd)
        tmp = ar(nn)*am
        do nnn = ia(nn),ia(nn+1) - 1
          mja  = meq(ja(nnn))
          ind3 = induvs(3,mja)
          if (ind3 == 1) then
            tmp = tmp - aa(nnn)*am*ueis(me(induvs(1,mja),induvs(2,mja)))
          elseif (ind3 == 2) then
            tmp = tmp - aa(nnn)*am*veis(me(induvs(1,mja),induvs(2,mja)))
          endif
        enddo
        ! Initial values of unlimited viscous stress for the iteration:
        ss5(n) = tmp
        ! Initial values of plastically limited stress for the iteration:
        ss8(idx(8,mi)) = tmp
      endif
    enddo
 
!$OMP BARRIER
    do mi=n2dl,n2du
      i = ind(1,mi)
      j = ind(2,mi)
      do induvs3=1,2
        n  = muvs(induvs3,i,j)
        if (n <= 0) cycle
        nn = ieq(n)
        if (nn <= 0) cycle

        ! scale the equation:
        do nd = ia(nn),ia(nn+1) - 1
          if (ja(nd) == nn) exit
        enddo
        ! nd is now the diagonal element of equation #nn
        if (abs(aa(nd)) > tiny_val) then
          !  usual scaling of this equation:
          am     = one/aa(nd)
          tmp    = ar(nn)*am
          ar(nn) = tmp
          as(nn) = tmp
          do nnn = ia(nn),ia(nn+1) - 1
            aa(nnn) = aa(nnn)*am
          enddo
          aa(nd) = one    ! assurance, for numerical robustness
        else
          !  exclude this equation:
          ar(nn) = zero
          as(nn) = zero
          do nnn = ia(nn),ia(nn+1) - 1
            aa(nnn) = zero
          enddo
          aa(nd) = one 
        endif
      enddo
      ioff = iuvp(1) - iuvp(2)
      do induvs3=3,5
        n    = idx(induvs3,mi)
        ioff = ioff + iuvp(2)
        if (induvs3 == 5 .and. n <= 0) cycle
        nn   = ieq(n+ioff)
        if (nn <= 0) cycle

        ! scale the equation:
        do nd = ia(nn),ia(nn+1) - 1
          if (ja(nd) == nn) exit
        enddo
        ! nd is now the diagonal element of equation #nn
        if (abs(aa(nd)) > tiny_val) then
          !  usual scaling of this equation:
          am     = one/aa(nd)
          tmp    = ar(nn)*am
          ar(nn) = tmp
          as(nn) = tmp
          do nnn = ia(nn),ia(nn+1) - 1
            aa(nnn) = aa(nnn)*am
          enddo
          aa(nd) = one    ! assurance, for numerical robustness
        else
          !  exclude this equation:
          ar(nn) = zero
          as(nn) = zero
          do nnn = ia(nn),ia(nn+1) - 1
            aa(nnn) = zero
          enddo
          aa(nd) = one    
        endif
      enddo
    enddo

!$OMP BARRIER
    ! Iteration loop till diff is small:
    do while (diffm > epsi*rms)

      ! during iterations, set up rhs for u and v:
      if (lll > 0) then
        do mi=n2dl,n2du
          i = ind(1,mi)
          j = ind(2,mi)
          do induvs3=1,2
            n  = muvs(induvs3,i,j)
            if (n <= 0) cycle
            nn = ieq(n)
            if (nn > 0) as(nn) = ar(nn)
          enddo
          ioff = iuvp(1) - iuvp(2)
          do induvs3=3,4
            n    = idx(induvs3,mi)
            ioff = ioff + iuvp(2)
            nn   = ieq(n+ioff)
            if (nn > 0) as(nn) = ar(nn)
          enddo
          induvs3 = 5
          n    = idx(induvs3,mi)
          if (n <= 0) cycle
          ioff = ioff + iuvp(2)
          nn   = ieq(n+ioff)
          if (nn <= 0) cycle
          as(nn) = ar(nn)
        enddo
      endif

      ! Update coefficients ar(:) and aa(:) in the relation between
      ! unlimited viscous stress and plastically limited stress:
      do mi=n2dl,n2du
        i    = ind(1,mi)
        j    = ind(2,mi)
        ioff = iuvp(1) + iuvp(2) + iuvp(3)
        do induvs3=6,7
          n    = idx(induvs3,mi)
          ioff = ioff + iuvp(2)
          nn   = ieq(n+ioff)
          if (nn <= 0) cycle

          do nd = ia(nn) , ia(nn+1) - 1
            if ( ja(nd) == nn ) exit 
          enddo
          ! nd is now the diagonal element of equation #nn
 
          aa(nd) = one
          do nnn=ia(nn),ia(nn+1)-1
            if (induvs(3,meq(ja(nnn))) == induvs3-3) aa(nnn) = -one
          enddo
          ar(nn) = zero  
          ppp    = eispr(mi)*half

          uvs3 = ss3(idx(3,mi))
          uvs4 = ss4(idx(4,mi))
          uvs6 = ss6(idx(6,mi))
          uvs7 = ss7(idx(7,mi))

          im1 = i-1
          jm1 = j-1

          mnw = me(im1,jm1)
          call eisdyn_sss (aa, ia, ja, meq, induvs, ppp, nn, induvs3,          &
                           uvs3, uvs4, ss5(idx(5,mnw)),                        &
                           uvs6, uvs7, ss8(idx(8,mnw)))

          mnn = me(im1,j)
          call eisdyn_sss (aa, ia, ja, meq, induvs, ppp, nn, induvs3,          &
                           uvs3, uvs4, ss5(idx(5,mnn)),                        &
                           uvs6, uvs7, ss8(idx(8,mnn)))

          mww = me(i,jm1)
          call eisdyn_sss (aa, ia, ja, meq, induvs, ppp, nn, induvs3,          &
                           uvs3, uvs4, ss5(idx(5,mww)),                        &
                           uvs6, uvs7, ss8(idx(8,mww)))

          call eisdyn_sss (aa, ia, ja, meq, induvs, ppp, nn, induvs3,          &
                           uvs3, uvs4, ss5(idx(5,mi)),                         &
                           uvs6, uvs7, ss8(idx(8,mi)))

          ! scale this equation at every iterative step:
          am     = one/aa(nd)
          tmp    = ar(nn)*am
          ar(nn) = tmp
          as(nn) = tmp
          do nnn = ia(nn),ia(nn+1) - 1
            aa(nnn) = aa(nnn)*am
          enddo
          aa(nd) = one    ! assurance, for numerical robustness
        enddo ! induvs3 == 6 or 7

        induvs3 = 8 
        n  = idx(induvs3,mi)
        if (n <= 0) cycle
        nn = ieq(n+iuvp(1)+4*iuvp(2)+iuvp(3))
        if (nn > 0) then

          do nd = ia(nn) , ia(nn+1) - 1
            if ( ja(nd) == nn ) exit
          enddo
          ! nd is now the diagonal element of equation #nn

          aa(nd) = one
          do nnn=ia(nn),ia(nn+1)-1
            if (induvs(3,meq(ja(nnn))) == induvs3-3) aa(nnn) = -one
          enddo
          ar(nn) = zero
          ppp    = eispr(mi)*half

          uvs5 = ss5(idx(5,mi))
          uvs8 = ss8(n)
 
          ip1 = i+1
          jp1 = j+1

          if (ppp > zero) call eisdyn_sss (aa,ia,ja,meq,induvs,ppp,            &
                                           nn, induvs3,                        &
                                           ss3(idx(3,mi)),                     &
                                           ss4(idx(4,mi)), uvs5,               &
                                           ss6(idx(6,mi)),                     &
                                           ss7(idx(7,mi)), uvs8 )

          mee = me(i,jp1)
          ppp = eispr(mee)*half
          if (ppp > zero) call eisdyn_sss (aa,ia,ja,meq,induvs,ppp,            &
                                           nn, induvs3,                        &
                                           ss3(idx(3,mee)),                    &
                                           ss4(idx(4,mee)), uvs5,              &
                                           ss6(idx(6,mee)),                    &
                                           ss7(idx(7,mee)), uvs8 )

          mss = me(ip1,j)
          ppp = eispr(mss)*half
          if (ppp > zero) call eisdyn_sss (aa,ia,ja,meq,induvs,ppp,            &
                                           nn, induvs3,                        &
                                           ss3(idx(3,mss)),                    &
                                           ss4(idx(4,mss)), uvs5,              &
                                           ss6(idx(6,mss)),                    &
                                           ss7(idx(7,mss)), uvs8 )

          mse = me(ip1,jp1)
          ppp = eispr(mse)*half
          if (ppp > zero) call eisdyn_sss (aa,ia,ja,meq,induvs,ppp,            &
                                           nn, induvs3,                        &
                                           ss3(idx(3,mse)),                    &
                                           ss4(idx(4,mse)), uvs5,              &
                                           ss6(idx(6,mse)),                    &
                                           ss7(idx(7,mse)), uvs8 )

          ! scale this equation at every iterative step:
          am     = one/aa(nd)
          tmp    = ar(nn)*am
          ar(nn) = tmp
          as(nn) = tmp
          do nnn = ia(nn),ia(nn+1) - 1
            aa(nnn) = aa(nnn)*am
          enddo
          aa(nd) = one    ! assurance, for numerical robustness
        endif

      enddo   ! mi
   
!$OMP BARRIER
!$OMP MASTER
      ! update iteration counter:
      lll = lll + 1
 
      ! solve for as(:)
      call solver(neq,as,aa,ia,ja,lll)
!$OMP END MASTER
!$OMP BARRIER
 
      call domp_get_thread_no( tnum )
      diffq(1:4,tnum) = zero
      nuv(1:2,tnum)   = zero
      do mi=n2dl,n2du
        i = ind(1,mi)
        j = ind(2,mi)

        n = muvs(1,i,j)
        if (n > 0) then
          nn = ieq(n)
          if (nn > 0) then
            rn            = nuv(1,tnum)
            np1           = rn + one
            diffq(1,tnum) = (diffq(1,tnum)*rn + (as(nn)-ueis(mi))**2)/np1
            nuv(1,tnum)   = np1
            ueis(mi)      = w1*ueis(mi) + w2*as(nn)
            diffq(3,tnum) = (diffq(3,tnum)*rn + ueis(mi)**2)/np1
          endif
        endif

        n = muvs(2,i,j)
        if (n > 0) then
          nn = ieq(n)
          if (nn > 0) then
            rn            = nuv(2,tnum)
            np1           = rn + one
            diffq(2,tnum) = (diffq(2,tnum)*rn + (as(nn)-veis(mi))**2)/np1
            nuv(2,tnum)   = np1
            veis(mi)      = w1*veis(mi) + w2*as(nn)
            diffq(4,tnum) = (diffq(4,tnum)*rn + veis(mi)**2)/np1
          endif
        endif

        ioff = iuvp(1)
        n    = idx(3,mi)
        nn   = ieq(n+ioff)
        if (nn > 0) ss3(n) = w1*ss3(n) + w2*as(nn)

        ioff = ioff + iuvp(2)
        n    = idx(4,mi)
        nn   = ieq(n+ioff)
        if (nn > 0) ss4(n) = w1*ss4(n) + w2*as(nn)

        ioff = iuvp(1) + 2*iuvp(2) + iuvp(3)
        n    = idx(6,mi)
        nn   = ieq(n+ioff)
        if (nn > 0) ss6(n) = w1*ss6(n) + w2*as(nn)

        ioff = ioff + iuvp(2)
        n    = idx(7,mi)
        nn   = ieq(n+ioff)
        if (nn > 0) ss7(n) = w1*ss7(n) + w2*as(nn)

        n    = idx(5,mi)
        if (n <= 0) cycle
        ioff = iuvp(1) + 2*iuvp(2)
        nn   = ieq(n+ioff)
        if (nn > 0) ss5(n) = w1*ss5(n) + w2*as(nn)

        n    = idx(8,mi)
        ioff = iuvp(1) + 4*iuvp(2) + iuvp(3)
        nn   = ieq(n+ioff)
        if (nn > 0) ss8(n) = w1*ss8(n) + w2*as(nn)
      enddo
!$OMP BARRIER
!$OMP MASTER
      diffu = zero
      rmsu  = zero
      rn    = zero
      do it=1,nthreads
        rn    = rn + nuv(1,it)
        diffu = diffu + diffq(1,it)*nuv(1,it)
        rmsu  = rmsu  + diffq(3,it)*nuv(1,it)
      enddo
      if (rn > zero) then
        diffu = diffu/rn
        rmsu  = rmsu/rn
      endif
      diffv = zero
      rmsv  = zero
      rn    = zero
      do it=1,nthreads
        rn    = rn + nuv(2,it)
        diffv = diffv + diffq(2,it)*nuv(2,it)
        rmsv  = rmsv  + diffq(4,it)*nuv(2,it)
      enddo
      if (rn > zero) then
        diffv = diffv/rn
        rmsv  = rmsv/rn
      endif
      diffm = sqrt(diffu + diffv)
      rms   = sqrt(rmsu  + rmsv)
   
      ! print some statistics:
      if (ldebug) call ice_stats( ss3, ss4, ss5, ss6, ss7, ss8, muvs, idx, ind,&
                                  ieq, n2d, iuvp, ueis, veis, diffm, rms, lll )

      if ( lll > MAXITERATIONS .and. diffm > epsi*rms ) then
        write(iu06,*) ' Iteration did not converge !'
        lun = io_new_unit()
        open (lun,FILE='BSHcmodp_eisdyn_iteration',POSITION='append')
        write(lun,'("Iteration did not converge  ! Time step   ",i5.5,2x,a19)')&
          izeits, ctim
        close (lun)
        ! fake a small diff to exit do-while loop
        diffm = epss*epsi
        rms   = epsl*epsi
      endif
!$OMP END MASTER
!$OMP BARRIER
 
    enddo ! diff > epsi
 
!$OMP MASTER
    ! save actual number of iterations:
    iceite = lll
!$OMP END MASTER

!$OMP END PARALLEL
 
  end subroutine eisdyn_int

  !-----------------------------------------------------------------------------

  !  Compare stresses,
  !  compute scaling coefficients for the reduction of unlimited viscous stress
  !  wrt the domain of admissible stress (plastically limited stress).
  subroutine eisdyn_sss (aa, ia, ja, meq, induvs, ppp, nn, induvs3,            &
                         uv3, uv4, uv5, uv6, uv7, uv8 )

    implicit none

    integer(4), intent(in)    :: nn, induvs3
    real(8),    intent(inout) :: aa(:)                       ! (9*nuvs)
    integer(4), intent(in)    :: ia(:), ja(:)                ! (nuv+1), (9*nuvs)
    integer(4), intent(in)    :: meq(:), induvs(:,:)         ! (nuvs), (3,nuvs)
    real(8),    intent(in)    :: ppp, uv3, uv4, uv5, uv6, uv7, uv8

    real(8), parameter     :: onemcor = one-corr
    real(8), parameter     :: qdelast = quart/elast
    real(8)                :: sig, sss, qqq
    integer(4)             :: nnn
    intrinsic              :: sqrt

    sig = sqrt( uv3**2 + uv4**2 + uv5**2 )
    sss = sqrt( uv6**2 + uv7**2 + uv8**2 )
  
    sss = sss*onemcor + eee*(elast*sig + ppp)

    if (sss >= ppp) then
      qqq = qdelast*(one - ppp/sss)
      do nnn=ia(nn),ia(nn+1)-1
        if (induvs(3,meq(ja(nnn))) == induvs3) aa(nnn) = aa(nnn) + qqq
      enddo
    endif

  end subroutine eisdyn_sss

  !-----------------------------------------------------------------------------

  subroutine press (nl, nu, casus, ice, eispr, ui, vi)
  
    implicit none
  
    integer(4), intent(in) :: nl, nu
    logical, intent(in)    :: casus(0:)
    real(8), intent(in)    :: ice(1:,0:)
    real(8), intent(out)   :: eispr(0:)
    real(8), intent(inout) :: ui(0:), vi(0:)
  
    integer(4)         :: mi
    real(8), parameter :: pstern = 2.e4_8
    real(8), parameter :: twenty = 20.0_8

!$OMP MASTER
    !  Ice dynamics model actually accesses ui/vi at land points, 
    !  so we better assign useful values. 
    !  For safety, we also initialize eispr at land.
    ui(0)    = zero
    vi(0)    = zero
    eispr(0) = zero
!$OMP END MASTER
  
    do mi=nl,nu
      if (casus(mi)) then
        eispr(mi) = zero
      else
        eispr(mi) = pstern*ice(1,mi)*exp(-twenty*(one-ice(2,mi)))
      endif
    enddo
  
  end subroutine press

  !-----------------------------------------------------------------------------

  subroutine init_uvse(nl, nu, idx, ss3, ss4, ss5, ss6, ss7, ss8)

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)  :: nl, nu
    integer(4), intent(in)  :: idx(3:,0:)
    real(8),    intent(out) :: ss3(0:), ss4(0:), ss5(0:)
    real(8),    intent(out) :: ss6(0:), ss7(0:), ss8(0:)

    !- local vars --------------------------------------------------------------
    integer(4) :: mi

!$OMP MASTER
    ss3(0) = zero
    ss4(0) = zero
    ss6(0) = zero
    ss7(0) = zero
    ss5(0) = zero
    ss8(0) = zero
!$OMP END MASTER

    do mi=nl,nu
      ss3(idx(3,mi)) = zero
      ss4(idx(4,mi)) = zero
      ss6(idx(6,mi)) = zero
      ss7(idx(7,mi)) = zero
      if (idx(5,mi) > 0) then
        ss5(idx(5,mi)) = zero
        ss8(idx(8,mi)) = zero
      endif
    enddo
      
  end subroutine init_uvse

  !-----------------------------------------------------------------------------

  subroutine limit_ice(nl, nu, casus, ice)

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: nl, nu
    logical,    intent(in)    :: casus(0:)
    real(8),    intent(inout) :: ice(1:,0:)

    !- local vars --------------------------------------------------------------
    integer(4)         :: mi
    real(8), parameter :: MinIce   = 1.e-6_8
    real(8), parameter :: icecomin = 1.e-3_8

    do mi=nl,nu
      if (casus(mi)) then
        ice(1,mi) = zero
        ice(2,mi) = zero
        ice(3,mi) = zero
      else
        ice(1,mi) = max(ice(1,mi),MinIce)
        ice(2,mi) = max(min(ice(2,mi),one),icecomin)
        ice(3,mi) = max(MinIce,min(ice(3,mi),ice(1,mi)))
      endif
    enddo
      
  end subroutine limit_ice

  !-----------------------------------------------------------------------------

  subroutine ice_stats( ss3, ss4, ss5, ss6, ss7, ss8, muvs, idx, ind, ieq, n2d,&
                        iuvp, ueis, veis, diffm, rms, it )
    use dmi_mpi_global, only : iu06

    implicit none 

    integer(4), intent(in) :: ind(:,:), muvs(:,0:,0:), ieq(:), n2d, iuvp(:)
    integer(4), intent(in) :: idx(3:,0:), it
    real(8),    intent(in) :: ueis(0:), veis(0:), diffm, rms
    real(8),    intent(in) :: ss3(0:), ss4(0:), ss5(0:)
    real(8),    intent(in) :: ss6(0:), ss7(0:), ss8(0:)

    integer(4) :: i, j, mi, ioff
    integer(4) :: n, nn, ind3
    real(8)    :: imx(8), imn(8), iav(8), irm(8), na(8), nap1, fac

    character(25), parameter :: f2 = '(a25,i1,a3,f16.11,f16.11)'
    character(23), parameter :: f4 = '(a25,i1,a3,f16.6,f16.6)'
    character(24), parameter :: f5 = '(a16,f16.8,f16.8,a18,i2)'

    imn(1) =  maxval(ueis)
    imx(1) =  minval(ueis)
    imn(2) =  maxval(veis)
    imx(2) =  minval(veis)
    imn(3) =  maxval(ss3)
    imx(3) =  minval(ss3)
    imn(4) =  maxval(ss4)
    imx(4) =  minval(ss4)
    imn(5) =  maxval(ss5)
    imx(5) =  minval(ss5)
    imn(6) =  maxval(ss6)
    imx(6) =  minval(ss6)
    imn(7) =  maxval(ss7)
    imx(7) =  minval(ss7)
    imn(8) =  maxval(ss8)
    imx(8) =  minval(ss8)
    iav(:) = zero
    irm(:) = zero
    na(:)  = zero

    do mi=1,n2d
      i = ind(1,mi)
      j = ind(2,mi)

      n = muvs(1,i,j)
      if (n > 0) then
        nn  = ieq(n)
        if (nn > 0) then
          imn(1) = min(ueis(mi),imn(1))
          imx(1) = max(ueis(mi),imx(1))
          nap1   = na(1) + one
          fac    = one/nap1
          iav(1) = (na(1)*iav(1) + ueis(mi))*fac
          irm(1) = (na(1)*irm(1) + ueis(mi)**2)*fac
          na(1)  = nap1
        endif
      endif

      n = muvs(2,i,j)
      if (n > 0) then
        nn  = ieq(n)
        if (nn > 0) then
          imn(2) = min(veis(mi),imn(2))
          imx(2) = max(veis(mi),imx(2))
          nap1   = na(2) + one
          fac    = one/nap1
          iav(2) = (na(2)*iav(2) + veis(mi))*fac
          irm(2) = (na(2)*irm(2) + veis(mi)**2)*fac
          na(2)  = nap1
        endif
      endif

      ioff = iuvp(1)
      n    = idx(3,mi)
      if (n > 0) then
        nn = ieq(n+ioff)
        if (nn > 0) then
          call stress_stats( ss3(n), na(3), iav(3), imn(3), imx(3), irm(3) )
        endif
      endif

      ioff = ioff + iuvp(2)
      n    = idx(4,mi)
      if (n > 0) then
        nn = ieq(n+ioff)
        if (nn > 0) then
          call stress_stats( ss4(n), na(4), iav(4), imn(4), imx(4), irm(4) )
        endif
      endif

      ioff = ioff + iuvp(2)
      n    = idx(5,mi)
      if (n > 0) then
        nn = ieq(n+ioff)
        if (nn > 0) then
          call stress_stats( ss5(n), na(5), iav(5), imn(5), imx(5), irm(5) )
        endif
      endif

      ioff = ioff + iuvp(3)
      n    = idx(6,mi)
      if (n > 0) then
        nn = ieq(n+ioff)
        if (nn > 0) then
          call stress_stats( ss6(n), na(6), iav(6), imn(6), imx(6), irm(6) )
        endif
      endif

      ioff = ioff + iuvp(2)
      n    = idx(7,mi)
      if (n > 0) then
        nn = ieq(n+ioff)
        if (nn > 0) then
          call stress_stats( ss7(n), na(7), iav(7), imn(7), imx(7), irm(7) )
        endif
      endif

      ioff = ioff + iuvp(2)
      n    = idx(8,mi)
      if (n > 0) then
        nn = ieq(n+ioff)
        if (nn > 0) then
          call stress_stats( ss8(n), na(8), iav(8), imn(8), imx(8), irm(8) )
        endif
      endif
    enddo

    do ind3=1,8
      irm(ind3) = sqrt(irm(ind3))
    enddo

    write(iu06,'(a72)')                                                        &
      '------------------------------------------------------------------------'
    write(iu06,'(a25)') 'Validation, ice dynamics:' 
    do ind3 = 1,2
      if (na(ind3) > zero) then
        write(iu06,f2)'Avr and rms for ice type ',ind3,':  ',iav(ind3),irm(ind3)
        write(iu06,f2)'Min and max for ice type ',ind3,':  ',imn(ind3),imx(ind3)
      endif
    enddo
    do ind3 = 3,8
      if (na(ind3) > zero) then
        write(iu06,f4)'Avr and rms for ice type ',ind3,':  ',iav(ind3),irm(ind3)
        write(iu06,f4)'Min and max for ice type ',ind3,':  ',imn(ind3),imx(ind3)
      endif
    enddo
    write(iu06,f5) 'RMSdiff and RMS:',diffm, rms,' at iteration No. ', it  

  end subroutine ice_stats

  !-----------------------------------------------------------------------------

  subroutine stress_stats( ss, na, iav, imn, imx, irm )
    implicit none

    real(8), intent(in)    :: ss
    real(8), intent(inout) :: na, iav, imn, imx, irm

    real(8) :: fac, nap1

    imn = min(ss,imn)
    imx = max(ss,imx)

    nap1 = na + one
    fac  = one/nap1
    iav  = (na*iav + ss)*fac
    irm  = (na*irm + ss*ss)*fac
    na   = nap1

  end subroutine stress_stats

end module eisdyn
