module exits

  !- directives ----------------------------------------------------------------
  implicit none
  private

  !- interfaces ----------------------------------------------------------------
  interface exitme
    module procedure exitme_txt
    module procedure exitme_val
    module procedure exitme_val_txt
    module procedure exitme_stdout
  end interface !f90 std exitme

  !- public vars & methods -----------------------------------------------------
  public  :: exitme

  !- private vars & methods ----------------------------------------------------
  private :: exitme_val, exitme_txt, exitme_val_txt, exitme_stdout, exit_hack
  private :: get_fmt
  integer(4),   private :: oval
  character(8), private :: cfmt

contains

  ! ----------------------------------------------------------------------------

  subroutine exitme_txt(txt)
    use dmi_mpi_global, only : iu06

    character (len=*), intent(in) :: txt

    write(iu06,'(a)') txt
    write(iu06,'(a14)') 'Program exited'
    call exit_hack(1_4)
  end subroutine exitme_txt

  ! ----------------------------------------------------------------------------

  subroutine exitme_val(val)
    use dmi_mpi_global, only : iu06

    integer(4), intent(in) :: val

    call get_fmt( val )

    write(iu06,cfmt) 'Program exited, return value=', oval
    call exit_hack(val)
  end subroutine exitme_val

  ! ----------------------------------------------------------------------------

  subroutine exitme_val_txt(val,txt)
    use dmi_mpi_global, only : iu06

    integer(4), intent(in)        :: val
    character (len=*), intent(in) :: txt

    call get_fmt( val )

    write(iu06,'(a)') txt
    write(iu06,cfmt) 'Program exited, return value=', oval
    call exit_hack(val)
  end subroutine exitme_val_txt

  ! ----------------------------------------------------------------------------

  subroutine exitme_stdout(val,txt,flag)
    use dmi_mpi_global, only : iu06

    integer(4), intent(in)        :: val
    character (len=*), intent(in) :: txt
    logical, intent(in)           :: flag

    call get_fmt( val )

    if (flag) then
      write(*,'(a)') txt
      write(*,cfmt) 'Program exited, return value=', oval
    else
      write(iu06,'(a)') txt
      write(iu06,cfmt) 'Program exited, return value=', oval
    endif
    call exit_hack(val,flag)
  end subroutine exitme_stdout

  ! ----------------------------------------------------------------------------

  subroutine exit_hack( returncode, flag )
    use port_exit, only : portable_exit
    use dmi_mpi_global, only : dmpi_close, dmpi_abort

    ! This is a hack to have only one single ansi f90/f95 warning about exit()
    integer(4), intent(in), optional :: returncode
    logical,    intent(in), optional :: flag
    integer(4) :: exit_status

    if (present(returncode)) then
      exit_status = returncode
    else
      exit_status = 0
    endif

    if (present(flag)) then
      if (exit_status == 0) then
        call dmpi_close( flag )
      else
        call dmpi_abort( flag )
      endif
    else
      if (exit_status == 0) then
        call dmpi_close( .false. )
      else
        call dmpi_abort( .false. )
      endif
    endif

    call portable_exit( exit_status ) 

  end subroutine exit_hack

  ! ----------------------------------------------------------------------------

  subroutine get_fmt( val )
    use dmi_mpi_global, only : iu06

    implicit none
    integer(4), intent(in) :: val

    oval = val
    if (val < 10) then
      cfmt = '(a29,'//'i1'//')'
    elseif (val < 100) then
      cfmt = '(a29,'//'i2'//')'
    elseif (val < 1000) then
      cfmt = '(a29,'//'i3'//')'
    elseif (val < 10000) then
      cfmt = '(a29,'//'i4'//')'
    else 
      cfmt = '(a29,'//'i4'//')'
      write(iu06,'(a39)') 'exit-value too large, set to 9999'
      oval = 9999
    endif

  end subroutine get_fmt

end module exits

