module geopotential
  implicit none
  private
  public  :: hydrostat
contains
subroutine hydrostat(mmk,ind,n2d,n3d,kh,h,rho,geopot,press)

  !- modules -------------------------------------------------------------------
  use constants, only : g, half
  use dmi_omp,   only : domp_get_domain
  
  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4),intent(in)           :: mmk(:,0:,0:)
  integer(4),intent(in)           :: ind(:,:) 
  integer(4),intent(in)           :: kh(0:)
  integer(4),intent(in)           :: n2d,n3d
  real(8),intent(in)              :: h(0:), rho(0:)
  real(8),intent(inout)           :: geopot(0:),press(0:)
  
  !- local vars ----------------------------------------------------------------
  integer(4)                      :: n2dl,n2du,n
  integer(4)                      :: i,j,k,kb
  
  call domp_get_domain (n3d, kh, 1, n2d, n2dl, n2du)

  do n=n2dl,n2du
    i=ind(1,n)
    j=ind(2,n)
    kb=kh(mmk(1,i,j))

    geopot(mmk(1,i,j)) = -g*half*h(mmk(1,i,j))
    press (mmk(1,i,j)) = g*half*h(mmk(1,i,j))*rho(mmk(1,i,j))
      
    if (kb < 2) cycle
    do k=2,kb
      geopot(mmk(k,i,j))=geopot(mmk(k-1,i,j))-g*half*                          &
                         (h(mmk(k-1,i,j))+h(mmk(k,i,j)) )
      ! assert(geopot (mmk(k,i,j)) <= 0.0)
      press (mmk(k,i,j))= press(mmk(k-1,i,j)) +                                &
                          g*half*( h(mmk(k-1,i,j))*rho(mmk(k-1,i,j))           &
                                +h(mmk(k,i,j ))*rho(mmk(k,i,j )) ) 
      ! assert(press (mmk(k,i,j)) >= 0.0))
    end do
  end do

end subroutine hydrostat
end module geopotential
