module intfeis_wrap 

  implicit none
  private
  public :: intfeis

contains

subroutine intfeis (mmaxp,nmaxp,me,muvse,induvse,iceite,uvp,nud,nvd,lud,lvd)

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in)    :: mmaxp,nmaxp,nud,nvd,lud(:),lvd(:)
  integer(4), intent(in)    :: me(0:,0:)
  integer(4), intent(out)   :: iceite, uvp(:)
  integer(4), intent(out)   :: muvse(:,0:,0:), induvse(:,:)

  !- local vars ----------------------------------------------------------------
  integer(4) :: i, j, nnn, lu, lv
  logical    :: uvdam(1:2,1:mmaxp-1,1:nmaxp-1)

  ! ----------------------------------------------------------------------------
  
  ! first, setup possible dams:
  uvdam(:,:,:) = .false.
  do lu=1,nud
    i = int(lud(lu)/1000)
    j = lud(lu)-i*1000
    uvdam(1,i,j) = .true.
  enddo
  do lv=1,nvd
    i = int(lvd(lv)/1000)
    j = lvd(lv)-i*1000
    uvdam(2,i,j) = .true.
  enddo

  ! let's be nice and initialize everything:
  muvse(:,:,:) = 0
  induvse(:,:) = 0
  iceite       = 0

  ! here we define muvse(1,:,:) and corresponding induvse(:,:)
  nnn=0
  do i=1,mmaxp-1
    do j=1,nmaxp-1
      if (me(i,j) > 0) then
        if (uvdam(1,i,j)) cycle
        if (me(i,j+1) > 0) then
          nnn = nnn+1
          muvse(1,i,j)   = nnn
          induvse(1,nnn) = i
          induvse(2,nnn) = j
          induvse(3,nnn) = 1
        endif
      endif
    enddo
  enddo

  ! muvse(2,:,:)
  do i=1,mmaxp-1
    do j=1,nmaxp-1
      if (me(i,j) > 0) then
        if (uvdam(2,i,j)) cycle
        if (me(i+1,j) > 0) then
          nnn = nnn+1
          muvse(2,i,j)   = nnn
          induvse(1,nnn) = i
          induvse(2,nnn) = j
          induvse(3,nnn) = 2
        endif
      endif
    enddo
  enddo
  uvp(1) = nnn

  ! here we define muvse(3,:,:) and corresponding induvse(:,:)
  do i=1,mmaxp-1
    do j=1,nmaxp-1
      if (me(i,j) > 0) then
        nnn = nnn+1
        muvse(3,i,j)   = nnn
        induvse(1,nnn) = i
        induvse(2,nnn) = j
        induvse(3,nnn) = 3
      endif
    enddo
  enddo
  uvp(2) = nnn - uvp(1)

  ! muvse(4,:,:)
  do i=1,mmaxp-1
    do j=1,nmaxp-1
      if (me(i,j) > 0) then
        nnn = nnn+1
        muvse(4,i,j)   = nnn
        induvse(1,nnn) = i
        induvse(2,nnn) = j
        induvse(3,nnn) = 4
      endif
    enddo
  enddo

  ! muvse(5,:,:)
  do i=1,mmaxp-1
    do j=1,nmaxp-1
      if (me(i,j) > 0) then
        if (uvdam(1,i,j) .or. uvdam(2,i,j)) cycle
        if (me(i,j+1) > 0 .and. me(i+1,j) > 0 .and. me(i+1,j+1) > 0) then
          if (uvdam(1,i+1,j) .or. uvdam(2,i,j+1)) cycle
          nnn = nnn+1
          muvse(5,i,j)   = nnn
          induvse(1,nnn) = i
          induvse(2,nnn) = j
          induvse(3,nnn) = 5
        endif
      endif
    enddo
  enddo
  uvp(3) = nnn - uvp(1) - 2*uvp(2)

  ! here we define muvse(6,:,:) and corresponding induvse(:,:)
  do i=1,mmaxp-1
    do j=1,nmaxp-1
      if (me(i,j) > 0) then
        nnn = nnn+1
        muvse(6,i,j)   = nnn
        induvse(1,nnn) = i
        induvse(2,nnn) = j
        induvse(3,nnn) = 6
      endif
    enddo
  enddo

  ! muvse(7,:,:)
  do i=1,mmaxp-1
    do j=1,nmaxp-1
      if (me(i,j) > 0) then
        nnn = nnn+1
        muvse(7,i,j)   = nnn
        induvse(1,nnn) = i
        induvse(2,nnn) = j
        induvse(3,nnn) = 7
      endif
    enddo
  enddo

  ! muvse(8,:,:)
  do i=1,mmaxp-1
    do j=1,nmaxp-1
      if (me(i,j) > 0) then
        if (uvdam(1,i,j) .or. uvdam(2,i,j)) cycle
        if (me(i,j+1) > 0 .and. me(i+1,j) > 0 .and. me(i+1,j+1) > 0) then
          if (uvdam(1,i+1,j) .or. uvdam(2,i,j+1)) cycle
          nnn = nnn+1
          muvse(8,i,j)   = nnn
          induvse(1,nnn) = i
          induvse(2,nnn) = j
          induvse(3,nnn) = 8
        endif
      endif
    enddo
  enddo

end subroutine intfeis 
end module intfeis_wrap 
