module old_ice_dyn
  !============================================================================+
  !
  !    **TEMPORARY** container for stuff related to the old ice model 
  !    It is not ment to be a permanent solution since this old model is in
  !    the process of being out phased.
  !
  !============================================================================+

  !- Modules -------------------------------------------------------------------
  use exits,          only : exitme
  use cmod_mem,       only : cmi1, cmi2, cmi3, cmr1
  use cmod_params,    only : me1, krz, mmxp, nmxp, iw2, nz, mmx, nmx
  use constants,      only : zero
  use dmi_mpi_global, only : iu06
  

  !- Implicit Directives -------------------------------------------------------
  implicit none
  private

  integer(4), parameter, public :: nicedynpar = 8
  integer(4), parameter, public :: niceidx    = 3

  integer(4), allocatable, save, public :: ne(:)
  integer(4), allocatable, save, public :: iwuvse(:)
  type(cmi3), allocatable, save, public :: muvse1(:)
  type(cmi2), allocatable, save, public :: induvse(:)
  type(cmi1), allocatable, save, public :: iuvp(:)
  integer(4), allocatable, save, public :: iceite(:)
  type(cmr1), allocatable, save, public :: taux(:), tauy(:) 


  !- Subroutines/functions -----------------------------------------------------
  public  :: OldIceModelParams
  private :: IceModelParams


contains

  !-----------------------------------------------------------------------------

  subroutine OldIceModelParams(narea)
    implicit none

    integer(4), intent(in) :: narea

    integer(4) :: ios, ia, nuvse, mnp, nnp, n2d

    !- Allocate the most basic stuff -------------------------------------------
    allocate( ne(narea), iwuvse(narea), muvse1(narea), induvse(narea),         &
              iceite(narea), iuvp(narea), taux(narea), tauy(narea), stat=ios)
    if (ios /= 0) then
      write(iu06,*) 'Error: Failed to allocate ice vars, status was: ', ios
      call exitme(1)
    endif
    iceite(:) = 0

    !- Find ice model parameters -----------------------------------------------
    do ia=1,narea
      call IceModelParams (ia)
      nuvse = iwuvse(ia)
      mnp   = mmxp(ia)
      nnp   = nmxp(ia)
      n2d   = iw2(ia)
      allocate( muvse1(ia)%p(1:nicedynpar,0:mnp,0:nnp), iuvp(ia)%p(3),         &
                taux(ia)%p(0:n2d), tauy(ia)%p(0:n2d),                          &
                induvse(ia)%p(1:niceidx,nuvse), stat=ios )
      taux(ia)%p(0) = zero
      tauy(ia)%p(0) = zero
    enddo

  end subroutine OldIceModelParams

  !-----------------------------------------------------------------------------


  !============================================================================+
  !
  !  Find parameters iwuvse and ne
  !
  !============================================================================+
  subroutine IceModelParams (ia)
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in) :: ia

    !- local vars --------------------------------------------------------------
    integer(4) :: h, i, j, k, l
    logical    :: flag


    !- run through grid --------------------------------------------------------
    l=0
    h=0
    flag = .false.
    do j=1,nmx(ia)
      do i=1,mmx(ia)

        if (me1(ia)%p(i,j) <= 0) cycle

        l=l+4

        if (me1(ia)%p(i,j+1) > 0) then
          l=l+1

          do k=1,nz(ia)
            if ( krz(ia)%p(1,k) == i .and.                                     &
               ( krz(ia)%p(2,k) == j .or. krz(ia)%p(2,k) == j+1) ) then
              flag = .true.
            endif
          enddo
          if (flag) h=h+1
          flag = .false.

        endif

        if (me1(ia)%p(i+1,j) > 0) then
          l=l+1

          do k=1,nz(ia)
            if ( krz(ia)%p(2,k) == j .and.                                     &
               ( krz(ia)%p(1,k) == i .or. krz(ia)%p(1,k) == i+1) ) then
              flag = .true.
            endif
          enddo
          if (flag) h=h+1
          flag = .false.

        endif

        if (me1(ia)%p(i,j+1) > 0 .and.                                         &
            me1(ia)%p(i+1,j) > 0 .and. me1(ia)%p(i+1,j+1) > 0) then
            l=l+2
        endif

      enddo
    enddo
    ne(ia)     = h
    iwuvse(ia) = l

  end subroutine IceModelParams

  !-----------------------------------------------------------------------------

end module old_ice_dyn

