module restart
  ! This is a hack caused by the sloppyness of the developer preventing
  ! him from maintaining shitloads of interface-blocks. 
  !     Hoping for forgiveness, Per

  use dmi_mpi_global, only : iu06
  use constants,      only : zero, ten

  implicit none
  private

  public  :: ReadRestart, WriteRestart
  private :: ReadInitST, InitTsoil, consistent, ReadDynHUVZ, InitST

  real(8), parameter, private :: s0 = 35.0_8, t0 = ten

contains

!===============================================================================
subroutine ReadRestart( narea, iw2, iw3, ind, kh,                              &
                        mmx, nmx, kmx, m1, hz,                                 &
                        starta,                                                &
                        h, u, v, z,                                            &
                        cmp, t_soil,                                           &
                        casus, ice, tsnei, ueis, veis, iota,                   &
                        tke, diss, avv,                                        &
                        a_lastasstime, a_errvar,                               &
                        aufdat, max_iw2 )

  !- modules -------------------------------------------------------------------
  use exits,       only : exitme
  use cmod_mem,    only : cml1, cmi1, cmi2, cmi3, cmr1, cmr2
  use io_subs,     only : io_new_unit
  use constants,   only : ldebug, itsfields, coldst, kalman, sstrs, cfreeze,   &
                          dyn, iotamodel, iotars, iota0
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4),    intent(in)    :: narea, max_iw2
  integer(4),    intent(in)    :: iw2(:), iw3(:), mmx(:), nmx(:), kmx(:)
  type (cmi1),   intent(in)    :: kh(:)    ! (narea)%p(0:)
  type (cmi2),   intent(in)    :: ind(:)   ! (narea)%p(:,:)
  type (cmi3),   intent(in)    :: m1(:)    ! (narea)%p(0:mmx,0:nmx,kmx)
  type (cmr1),   intent(in)    :: hz(:)    ! (narea)%p(0:)
  character(11), intent(in)    :: aufdat

  character(11), intent(inout) :: starta
  type (cmr2),   intent(inout) :: cmp(:)            ! (narea)%p(1:,0:)
  type (cmr1),   intent(inout) :: h(:), u(:), v(:), tke(:), diss(:), avv(:)
  type (cmr1),   intent(inout) :: a_errvar(:)
  type (cmr1),   intent(inout) :: z(:)                        ! (narea)%p(0:)
  type (cmr1),   intent(inout) :: tsnei(:), ueis(:), veis(:)
  type (cmr2),   intent(inout) :: ice(:), iota(:)
  type (cmr1),   intent(inout) :: a_lastasstime(:)
  type (cml1),   intent(inout) :: casus(:)
  type (cmr2),   intent(inout) :: t_soil(:)                 ! (narea)%p(3,iw2)

  !- locals --------------------------------------------------------------------
  integer(4) :: lun, ia, n2d, ios
  integer(4) :: idum(0:max_iw2)

  real(8), parameter :: errv0 = 0.4_8

  !-----------------------------------------------------------------------------
  ! Previous revisions also had atmospherical data and seegang data in the 
  ! restart file which, however, is not necessary here and therefore we do no 
  ! longer support that.
  !-----------------------------------------------------------------------------

  if (.not. coldst) then
    !- read variables for the restart:

    lun = io_new_unit()
    open(lun,file='aufpkt',form='unformatted',iostat=ios,                      &
         action='read',status='old')
    if (ios /= 0) call exitme (1, 'Could not open aufpkt' )
    read(lun,iostat=ios) starta
    if (ios /= 0) then
       write(iu06,*) 'I/O problem - iostat is:',ios
       call exitme (2, 'Could not read starta from aufpkt' )
    endif
    
    if (dyn) then
      call ReadDynHUVZ( lun, narea, h, u, v, z )
    else  ! in the static case  
      do ia=1,narea
        read(lun,iostat=ios) u(ia)%p(:), v(ia)%p(:), z(ia)%p(:)
        if (ios /= 0) then
          write(iu06,*) 'I/O problem - iostat is:',ios
          call exitme (3, 'Could not read u, v , z from aufpkt' )
        endif
      enddo
    endif
      
    do ia=1,narea
      read(lun,iostat=ios) cmp(ia)%p(2,:), cmp(ia)%p(1,:)
      if (ios /= 0) then
        write(iu06,*) 'I/O problem - iostat is:',ios
        call exitme (4, 'Could not read aufpkt' )
      endif
    enddo
    do ia=1,narea
      n2d = iw2(ia)
      read(lun,iostat=ios) idum(0:n2d),                                        &
                                  ice(ia)%p(1,:), ice(ia)%p(2,:),              &
                                  ice(ia)%p(3,:), tsnei(ia)%p(:),              &
                                  ueis(ia)%p(:),  veis(ia)%p(:)
      if (ios /= 0) then
        write(iu06,*) 'I/O problem - iostat is:',ios
        call exitme (5, 'Could not read aufpkt' )
      endif

      !- transfer to logical arrays:
      casus(ia)%p(0:n2d) = ( idum(0:n2d) /= 0)

      !- avoid ice on land 
      !  (some older restart files have that, beleave it or not)
      casus(ia)%p(0) = .true.
    enddo
    do ia=1,narea
      read(lun,iostat=ios) t_soil(ia)%p(:,:)
      if (ios /= 0) then 
        write(iu06,*) 'I/O problem - iostat is:',ios
        call exitme (6, 'Could not read aufpkt' )
      endif
    enddo

    do ia=1,narea
      read(lun,iostat=ios) tke(ia)%p(:), diss(ia)%p(:)
      if (ios /= 0) then
        write(iu06,*) 'I/O problem - iostat is:',ios
        call exitme (9, 'Could not read aufpkt' )
      endif
    enddo
    do ia=1,narea
      read(lun,iostat=ios) avv(ia)%p(:)
      if (ios /= 0) then
        write(iu06,*) 'I/O problem - iostat is:',ios
        call exitme (10, 'Could not read aufpkt' )
      endif
    enddo
    if (kalman) then
      if (sstrs) then
        ! read sst assim data from restart file:
        do ia=1,narea
          read(lun,iostat=ios) a_lastasstime(ia)%p(:), a_errvar(ia)%p(:)
          if (ios /= 0) then
            write(iu06,*) 'I/O problem - iostat is:',ios
            call exitme (12, 'Could not read aufpkt' )
          endif
        enddo
      else
        ! initialise sst assim data:
        do ia=1,narea
          a_lastasstime(ia)%p(:) = zero
          a_errvar(ia)%p(:)      = errv0
        enddo
      endif
    endif
    if (iotamodel) then
      ! initialise global iota array:
      do ia=1,narea
        iota(ia)%p(:,0)  = iota0
        iota(ia)%p(:,1:) = zero
      enddo
      ! read iota from restart file:
      if (iotars) then
        do ia=1,narea
          read(lun,iostat=ios) iota(ia)%p(:,:)
          if (ios /= 0) call exitme (15, 'Could not read aufpkt' )
        enddo
      endif
    else
      ! initialise global iota array:
      do ia=1,narea
        iota(ia)%p(:,0) = iota0
      enddo  
    endif

    close(lun)

  else

    !- For a cold start, use zero velocities, etc ...
    !
    do ia=1,narea
      call InitST( cmp(ia)%p )
    enddo

    if (itsfields > 0) then
      !- read initial T/S from file:

      if (itsfields == 1) then
        !- read from a restart file:
        lun = io_new_unit()
        open(lun,file='aufpkt',form='unformatted',iostat=ios,                  &
             action='read',status='old')
        if (ios /= 0) call exitme (1, 'Could not open aufpkt' )
        !- skip vars from a dummy restart file:
        read(lun,iostat=ios) starta
        write(iu06,*) "Dummy T/S restart file is valid at ",starta(2:11)
        if (ios /= 0) call exitme (1, 'Could not read aufpkt' )
        if (dyn) then
          call ReadDynHUVZ( lun, narea, h, u, v, z )
        else  ! in the static case
          do ia=1,narea
            read(lun,iostat=ios) u(ia)%p(:), v(ia)%p(:), z(ia)%p(:)
            if (ios /= 0) call exitme (1, 'Could not read aufpkt' )
          enddo
        endif
        do ia=1,narea
          read(lun,iostat=ios) cmp(ia)%p(2,:), cmp(ia)%p(1,:)
          if (ios /= 0) call exitme (1, 'Could not read aufpkt' )
        enddo
        close(lun)
      elseif (itsfields == 2) then
        !- read T and S from ascii data files:
        do ia=1,narea
          call ReadInitST ( ia, mmx(ia), nmx(ia), kmx(ia), m1(ia)%p, cmp(ia)%p )
        enddo
      endif
    endif

    starta = aufdat

    do ia=1,narea
      if (dyn) then
        h(ia)%p(:)   = hz(ia)%p(:)
      endif
      u(ia)%p(:)     = zero
      v(ia)%p(:)     = zero
      z(ia)%p(:)     = zero
      casus(ia)%p(:) = .true.
      ice(ia)%p(:,:) = zero
      tsnei(ia)%p(0) = zero
      tsnei(ia)%p(1:iw2(ia)) = -cfreeze*cmp(ia)%p(2,1:iw2(ia))
    enddo

    !- initialise t_soil as bottom water temperature:
    do ia=1,narea
      call InitTsoil( m1(ia)%p, cmp(ia)%p, iw2(ia),                            &
                      ind(ia)%p, kh(ia)%p, t_soil(ia)%p)
    enddo

    if (kalman) then
      do ia=1,narea
        a_lastasstime(ia)%p(:) = zero
        a_errvar(ia)%p(:)      = errv0
      enddo
    endif

    if (iotamodel) then
      ! initialise global iota array
      do ia=1,narea
        iota(ia)%p(:,0)  = iota0
        iota(ia)%p(:,1:) = zero
      enddo
    endif

  endif

  !- consistency check
  if (ldebug) then
    if ( consistent(narea, iw2, iw3, ind, u, v, z, cmp) ) then
      write(iu06,'(a21)') 'cmod state consistent'
    else
      write(iu06,*) 'error in cmod state; check file: consistency_errors '
    endif
  endif

  !- Check time stamp in restart file ------------------------------------------
  if (starta(2:11) /= aufdat(2:11)) then
    call exitme (5,'Time stamp in restart file does not match starting time')
  endif


end subroutine ReadRestart

!===============================================================================

subroutine ReadInitST ( ia, mno, nno, kno, m1, ts )

  !- modules -------------------------------------------------------------------
  use exits,     only : exitme
  use constants, only : ldebug
  use io_subs,   only : io_new_unit

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in) :: ia, mno, nno, kno
  integer(4), intent(in) :: m1(0:,0:,:)
  real(8), intent(inout) :: ts(1:,0:)

  !- local vars ----------------------------------------------------------------
  character(8) :: fnam
  character(3) :: c3
  integer(4)   :: i, j, k, idx, lun, ios
  logical      :: bmiss = .false.
  real(8)      :: s(1:mno-1, 1:nno-1, 1:kno), t(1:mno-1, 1:nno-1, 1:kno)

  real(8), parameter :: missing = -5.0_8

  !- read and check salt anbd temp from ascii file -----------------------------
  if (ia < 10) then
    write(c3,'(i1)') ia
  elseif (ia < 100) then
    write(c3,'(i2)') ia
  elseif (ia < 1000) then
    write(c3,'(i3)') ia
  else
    call exitme(1,'Cannot handle more than 999 areas.')
  endif
  fnam = 'sinit'//trim(c3)
  lun = io_new_unit()
  open(lun,file=trim(fnam),iostat=ios,action='read',status='old')
  if (ios /= 0) call exitme (1, 'Could not open '//trim(fnam) )
  read(lun,'(10f8.4)') (((s(i,j,k), i=1,mno-1), j=1,nno-1), k=1,kno)
  close(lun)
  do k=1,kno
    do j=1,nno-1
      do i=1,mno-1
        idx = m1(i,j,k)
        if (idx > 0) then
          if (s(i,j,k) <= missing) then
            if (ldebug .or. k ==1) then
              write(iu06,'(a,a,a,3i5)') 'Missing value in ',fnam,' at ',i,j,k
            endif
            if (k == 1) then
              bmiss = .true.
            else
              ts(2,idx) = ts(2,m1(i,j,k-1))
            endif
          else
            ts(2,idx) = s(i,j,k)
          endif
        endif
      enddo
    enddo
  enddo
  fnam = 'tinit'//trim(c3)
  lun = io_new_unit()
  open(lun,file=trim(fnam),iostat=ios,action='read',status='old')
  if (ios /= 0) call exitme (1, 'Could not open '//trim(fnam) )
  read(lun,'(10f8.4)') (((t(i,j,k), i=1,mno-1), j=1,nno-1), k=1,kno)
  close(lun)
  do k=1,kno
    do j=1,nno-1
      do i=1,mno-1
        idx = m1(i,j,k)
        if (idx > 0) then
          if (t(i,j,k) <= missing) then
            if (ldebug .or. k ==1) then
              write(iu06,'(a,a,a,3i5)') 'Missing value in ',fnam,' at ',i,j,k
            endif
            if (k == 1) then
              bmiss = .true.
            else
              ts(1,idx) = ts(1,m1(i,j,k-1))
            endif
          else
            ts(1,idx) = t(i,j,k)
          endif
        endif
      enddo
    enddo
  enddo

  !- That's it! ----------------------------------------------------------------
  if (bmiss) then
    call exitme ('Missing value(s) in initial S/T')
  endif

end subroutine ReadInitST

!===============================================================================

subroutine InitTsoil ( mx1, tw, n2d, ind, kh, tsoil )

  !- modules -------------------------------------------------------------------
  !  none

  !- directives ----------------------------------------------------------------
  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4), intent(in) :: n2d
  integer(4), intent(in) :: mx1(0:,0:,:), kh(0:), ind(:,:)
  real(8), intent(in)    :: tw(1:,0:)
  real(8), intent(out)   :: tsoil(:,1:)

  !- local vars ----------------------------------------------------------------
  integer(4) :: ii

  !- initialise tsoil as bottom water temperature ------------------------------
  do ii=1,n2d
    tsoil(1,ii) = tw(1,mx1(ind(1,ii),ind(2,ii),kh(ii)))
  enddo
  tsoil(2,:) = tsoil(1,:)
  tsoil(3,:) = tsoil(1,:)

end subroutine InitTsoil

!===============================================================================

  subroutine WriteRestart ()
 
    use exits,         only : exitme
    use io_subs,       only : io_new_unit
    use perm,          only : de_permute2, de_permute_omp, de_permute_2Dvars,  &
                              u_o, v_o, s_o, t_o, h_o, z_o, tsoil_o,           &
                              ice_o, ts_o, ui_o, vi_o, casus_o, turb1_o, turb2_o
    use constants,     only : aufdat, kalman, dyn, iotamodel
    use cmod_miscvars, only : ihour,ctim
    use cmod_params,   only : narea, mmx, nmx, kmx, iw2, iw3, mm1
    use cmod_arrays,   only : u, v, cmp, h, z, ice, tsnei, casus,              &
                              ueis, veis, tke, diss, t_soil, avv,              &
                              a_lastasstime, a_errvar,                         &
                              kh, ind, mm1k, iota
    use compress,      only : mm_op, idx_op


    integer(4) :: lun, ia, ios, lw, up
    logical    :: do_alloc

    aufdat(2:5)  = ctim(1:4)
    aufdat(6:10) = ctim(6:10)
    write(aufdat(11:11),'(I1.1)') ihour/3

    lun = io_new_unit()
    open (lun,file='restart',form='unformatted',status='replace', iostat=ios)
    if (ios /= 0) call exitme(1, 'Cannot open restart file')
    write(iu06,'(a33,a11)') 'writing restart file restart at: ',aufdat
    write(lun) aufdat

    do ia=1,narea
      ! de-permute u,v,s,t:
      do_alloc = .true.
      if (allocated(u_o)) then
        do_alloc = .false.
        if (iw3(ia) > size(u_o(:)) - 1) then
          deallocate(u_o,v_o,s_o,t_o,h_o)
          do_alloc = .true.
        endif
      endif
      if (do_alloc) then
        allocate(u_o(0:iw3(ia)),v_o(0:iw3(ia)),s_o(0:iw3(ia)),t_o(0:iw3(ia)))
        if (dyn) then
          allocate(h_o(0:iw3(ia)))
        else
          allocate(h_o(0:0))
        endif
      endif
!$OMP PARALLEL DEFAULT (shared)
      call de_permute_omp( iw2(ia), mm1k(ia)%p, mm_op(ia)%p, ind(ia)%p,        &
                           kh(ia)%p, idx_op(ia)%p, u(ia)%p, u_o, v(ia)%p, v_o)
      if (dyn) then
        call de_permute_omp( iw2(ia), mm1k(ia)%p, mm_op(ia)%p, ind(ia)%p,      &
                             kh(ia)%p, idx_op(ia)%p, h(ia)%p, h_o )
      endif
!$OMP END PARALLEL

      ! de-permute z:
      do_alloc = .true.
      if (allocated(z_o)) then
        do_alloc = .false.
        if (iw2(ia) > size(z_o(:)) - 1) then
          deallocate(z_o)
          do_alloc = .true.
        endif
      endif
      if (do_alloc) then
        allocate(z_o(0:iw2(ia)))
      endif
      lw = 0
      up = iw2(ia)
      call de_permute_2Dvars(mmx(ia),nmx(ia),mm1(ia)%p,z(ia)%p)

      if (dyn) then
        write(lun) h_o(0:iw3(ia)),u_o(0:iw3(ia)),v_o(0:iw3(ia)),z_o(lw:up)
      else
        write(lun) u_o(0:iw3(ia)),v_o(0:iw3(ia)),z_o(lw:up)
      endif
    enddo
    do ia=1,narea
      ! s_o, t_o were allocated in previous loop
!$OMP PARALLEL DEFAULT (shared)
      call de_permute_omp( iw2(ia), mm1k(ia)%p, mm_op(ia)%p, ind(ia)%p,        &
                           kh(ia)%p, idx_op(ia)%p, cmp(ia)%p, t_o, s_o )
!$OMP END PARALLEL
      write(lun) s_o(0:iw3(ia)),t_o(0:iw3(ia))
    enddo
    do ia=1,narea
      do_alloc = .true.
      if (allocated(ui_o)) then
        do_alloc = .false.
        if (iw2(ia) > size(ui_o(:)) - 1) then
          deallocate(ice_o,ts_o,ui_o,vi_o,casus_o)
          do_alloc = .true.
        endif
      endif
      if (do_alloc) then
        allocate(ice_o(1:3,0:iw2(ia)),ts_o(0:iw2(ia)),                         &
                 ui_o(0:iw2(ia)),vi_o(0:iw2(ia)),casus_o(0:iw2(ia)))
      endif
      lw = 0
      up = iw2(ia)
      call de_permute_2Dvars(mmx(ia),nmx(ia),mm1(ia)%p,ice(ia)%p,tsnei(ia)%p,  &
                             ueis(ia)%p,veis(ia)%p,casus(ia)%p)

      write(lun) casus_o(lw:up),ice_o(1,lw:up),ice_o(2,lw:up),ice_o(3,lw:up),  &
                 ts_o(lw:up),ui_o(lw:up),vi_o(lw:up)
    enddo
    do ia=1,narea
      do_alloc = .true.
      if (allocated(tsoil_o)) then
        do_alloc = .false.
        if (iw2(ia) > ubound(tsoil_o,2) ) then
          deallocate(tsoil_o)
          do_alloc = .true.
        endif
      endif
      if (do_alloc) then
        allocate(tsoil_o(1:3,1:iw2(ia)))
      endif
      lw = 1       
      up = iw2(ia)
      call de_permute_2Dvars(mmx(ia),nmx(ia),mm1(ia)%p,t_soil(ia)%p)
      write(lun) tsoil_o(1:3,lw:up)
    enddo

    do ia=1,narea
      do_alloc = .true.
      if (allocated(turb1_o)) then
        do_alloc = .false.
        if (iw3(ia) > size(turb1_o(:)) - 1) then
          deallocate(turb1_o,turb2_o)
          do_alloc = .true.
        endif
      endif
      if (do_alloc) allocate(turb1_o(0:iw3(ia)),turb2_o(0:iw3(ia)))
      call de_permute2(mmx(ia),nmx(ia),kmx(ia),mm1(ia)%p,                      &
                       tke(ia)%p,diss(ia)%p )
      write(lun) turb1_o(0:iw3(ia)),turb2_o(0:iw3(ia))
    enddo
    do ia=1,narea
      call de_permute2(mmx(ia),nmx(ia),kmx(ia),mm1(ia)%p,avv(ia)%p)
      write(lun) turb1_o(0:iw3(ia))
    enddo

    if (kalman) then
      do ia=1,narea
        !  re-use z_o and turb1_o which were allocated above
        call de_permute_2Dvars(mmx(ia),nmx(ia),mm1(ia)%p,a_lastasstime(ia)%p)
        call de_permute2(mmx(ia),nmx(ia),kmx(ia),mm1(ia)%p,a_errvar(ia)%p)
        write(lun) z_o(0:iw2(ia)), turb1_o(0:iw3(ia))
      enddo
    endif

    ! depermute and write iota into restart file
    ! use ice_o file for iota
    if (iotamodel) then
      do ia=1,narea
        do_alloc = .true.
        if (allocated(ice_o)) then
          do_alloc = .false.
          if (iw2(ia) > size(ice_o(1,:)) - 1) then
            deallocate(ice_o)
            do_alloc = .true.
          endif
        endif
        if (do_alloc) then
          allocate(ice_o(1:3,0:iw2(ia)))
        endif
        lw = 0
        up = iw2(ia)
        call de_permute_2Dvars(mmx(ia),nmx(ia),mm1(ia)%p,iotamodel,iota(ia)%p)
        write(lun) ice_o(1:3,lw:up)
      enddo
    endif

    close(lun)

  endsubroutine WriteRestart

!===============================================================================

logical function consistent( narea, iw2, iw3, ind, u, v, z, st )
!-------------------------------------------------------------------------------
!
! Checks whether cmod state is sane. Currently checks for
! temperature, salinity, speed, and sea level are implemented.
!
! Based on public domain code from: enkf.nersc.no
!
! Jesper Larsen, 2006/09/07
!-------------------------------------------------------------------------------

  !- modules -------------------------------------------------------------------
  use dmi_mpi_global, only : iu06
  use dmi_mpi,        only : mpi_io_serial
  use cmod_mem,       only : cmi2, cmr1, cmr2
  use io_subs,        only : io_new_unit

  implicit none

  !- arguments -----------------------------------------------------------------
  integer(4),    intent(in) :: narea
  integer(4),    intent(in) :: iw2(:), iw3(:)
  type (cmi2),   intent(in) :: ind(:)                !(narea)%p(:,:)
  type (cmr1),   intent(in) :: u(:), v(:)            !(narea)%p(0:)
  type (cmr1),   intent(in) :: z(:)                  !(narea)%p(0:)
  type (cmr2),   intent(in) :: st(:)                 !(narea)%p(1:,0:)


  !- local vars ----------------------------------------------------------------
  logical :: error_here                         ! error locally at certain i,j,k
  integer(4) :: i, count_weird                     ! counting weird
  integer(4) :: count_z_toosmall, count_z_toolarge
  integer(4) :: count_T_toosmall, count_T_toolarge
  integer(4) :: count_S_toosmall, count_S_toolarge
  integer(4) :: count_vel_toolarge
  integer(4) :: lun, ia

  ! hard coded file name:
  character(18), parameter :: filename = 'consistency_errors'

  ! formats:
  character(5), parameter :: e = '(a60)', f = '(a62)'
  character(4), parameter :: x = '(a1)' 

  ! give up reporting after 100 weird values
  integer(4), parameter :: maxweird = 100

  ! reasonable boundaries
  real(8), parameter :: z_max   = 10.0_8 ! maximum sea level
  real(8), parameter :: z_min   =-10.0_8 ! minimum sea level
  real(8), parameter :: t_min   = -3.0_8
  real(8), parameter :: t_max   = 35.0_8
  real(8), parameter :: s_min   = -0.00001_8
  real(8), parameter :: s_max   = 45.0_8
  real(8), parameter :: vel_max =  5.0_8 ! in m/s

  ! Initialisations
  consistent         = .true.
  count_z_toosmall   = 0
  count_z_toolarge   = 0
  count_T_toosmall   = 0
  count_T_toolarge   = 0
  count_S_toosmall   = 0
  count_S_toolarge   = 0
  count_vel_toolarge = 0

  ! Write out file header
  if (mpi_io_serial) then
    lun = io_new_unit()
    write(iu06,'(a14,a18)') 'opening file: ', trim(filename)
    open(lun, file=trim(filename), access='sequential',status='replace')
    write(lun,f)'**************************************************************'
    write(lun,f)'THIS FILE CONTAINS ERRORS DETECTED IN THE CMOD STATE VARIABLE '
    write(lun,f)'Go to the end of this file for a summary of all errors        '
    write(lun,f)'**************************************************************'
    write(lun,f)' '
    write(lun,f)'The following bounds are being tested:                        '
    write(lun,f)' 2D wet points variables                                      '
    write(lun,x)' '
    write(lun,'(a9,2a8,a5)') 'Variable','min','max','unit' 
    write(lun,x)' '
    write(lun,'(a9,2f8.1,a5)') 'z',z_min,z_max,'m'
    write(lun,x)' '
    write(lun,'(a,f8.2,a1)') ' 3D points in layers:'
    write(lun,'(a9,2a8,a5)') 'Variable','min','max','unit' 
    write(lun,x)' '
    write(lun,'(a9,2f8.1,a5)') 'T',t_min,t_max,'degC'
    write(lun,'(a9,2f8.1,a5)') 'S',s_min,s_max,'o/oo'
    write(lun,'(a9,8x,f8.1,a5)') '|u|',vel_max,'m/s'
    write(lun,x)' '
    write(lun,f)'**************************************************************'
    write(lun,x)' '
    write(lun,f)'  2D grid   i   j          z                                  '
    write(lun,x)' '
  endif

  ! Searching for possibly unphysical values their position. -------------------

  ! 2D variables 
  count_weird=0 !reinitializing
  do ia=1,narea
    do i=1,iw2(ia)
      if (count_weird > maxweird) then
        if (mpi_io_serial) then
          write(lun,'(a10,i5,a30)')                                            &
            'Exceeding ', maxweird,' weird 2D-values, exiting test'
        endif
        exit
      endif
      error_here = .false.           ! no error yet in present member at i,j,k 
      if (z(ia)%p(i) > z_max) then
        error_here       = .true.
        count_weird      = count_weird + 1
        count_z_toolarge = count_z_toolarge + 1
      end if
      if (z(ia)%p(i) < z_min) then
        error_here       = .true.
        count_weird      = count_weird + 1
        count_z_toosmall = count_z_toosmall + 1
      end if
      if (error_here .and. mpi_io_serial) then
        write(lun,'(a6,a5,i4,i4,1f10.2)') '2D', 'mm1',                         &
          ind(ia)%p(1,i), ind(ia)%p(2,i), z(ia)%p(i)
      endif
    enddo
  enddo
  consistent = (count_weird <= 0)

  ! 3D variables 
  count_weird=0 !reinitializing
  if (mpi_io_serial) then
    write(lun,x)' '
    write(lun,f)'iens 3D   i   j       u         v         T         S         '
  endif
  do ia=1,narea
    do i=1,iw3(ia)
      if (count_weird > maxweird) then
        if (mpi_io_serial) then
          write(lun,'(a10,i5,a30)')                                            &
            'Exceeding ', maxweird,' weird 3D-values, exiting test'
        endif
        exit
      endif
      error_here = .false.           ! no error yet in present member at i,j,k 
      if (st(ia)%p(1,i) < t_min) then
        error_here       = .true.
        count_weird      = count_weird + 1
        count_T_toosmall = count_T_toosmall + 1
      endif
      if (st(ia)%p(1,i) > t_max) then
        error_here       = .true.
        count_weird      = count_weird + 1
        count_T_toolarge = count_T_toolarge + 1
      endif
      if (st(ia)%p(2,i) < s_min) then
        error_here       = .true.
        count_weird      = count_weird + 1
        count_S_toosmall = count_S_toosmall + 1
      endif
      if (st(ia)%p(2,i) > s_max) then
        error_here       = .true.
        count_weird      = count_weird + 1
        count_S_toosmall = count_S_toolarge + 1
      endif
      if (sqrt(u(ia)%p(i)**2+v(ia)%p(i)**2) > vel_max) then
        error_here         = .true.
        count_weird        = count_weird + 1
        count_vel_toolarge = count_vel_toolarge + 1
      endif
      if (error_here .and. mpi_io_serial) then
        write(lun,'(a6,a5,i4,i4,4f10.2)') '3D','mm1',                          &
          ind(ia)%p(1,i), ind(ia)%p(2,i),                                      &
          u(ia)%p(i), v(ia)%p(i), st(ia)%p(1,i), st(ia)%p(2,i)
      endif
    enddo 
  enddo
  consistent = (count_weird <= 0)

  if (.not.consistent .and. mpi_io_serial) then
    write(iu06,e) '************************************************************'
    write(iu06,e) '***  ERROR   ERROR   ERROR   ERROR   ERROR   ERROR   ERROR  '
    write(iu06,e) '************************************************************'
    write(iu06,'(a16,a18,a22)')                                                &
                  'please see file ',trim(filename),' for more information '
    write(iu06,e) '************************************************************'
    if (count_T_toosmall   > 0) write(lun,'(i8,a,f9.1,a)') count_T_toosmall,   &
                                ' occurences          T <',t_min,' deg'
    if (count_T_toolarge   > 0) write(lun,'(i8,a,f9.1,a)') count_T_toolarge,   &
                                ' occurences          T >',t_max,' deg'
    if (count_S_toosmall   > 0) write(lun,'(i8,a,f9.1,a)') count_S_toosmall,   &
                                ' occurences          S <',s_min,' o/oo'
    if (count_S_toolarge   > 0) write(lun,'(i8,a,f9.1,a)') count_S_toolarge,   &
                                ' occurences          S >',s_max,' o/oo'
    if (count_vel_toolarge > 0) write(lun,'(i8,a,f9.1,a)') count_vel_toolarge, &
                                ' occurences   velocity >',vel_max,' m/s'
    if (count_z_toosmall   > 0) write(lun,'(i8,a,f9.1,a)') count_z_toosmall,   &
                                ' occurences sea level <',z_min,' m'
    if (count_z_toolarge   > 0) write(lun,'(i8,a,f9.1,a)') count_z_toolarge,   &
                                ' occurences sea level >',z_max,' m'
    write(lun,x) ' '
    close(lun)
  endif
 
end function consistent

!===============================================================================

  subroutine ReadDynHUVZ( lun, narea, h, u, v, z )

    use cmod_mem,    only : cmr1
    use exits,       only : exitme

    implicit none

    integer(4),  intent(in)    :: lun, narea
    type (cmr1), intent(inout) :: h(:), u(:), v(:), z(:)

    integer(4) :: ia, ios

    do ia=1,narea
      read(lun,iostat=ios) h(ia)%p(:), u(ia)%p(:), v(ia)%p(:), z(ia)%p(:)
      if (ios /= 0) then
        write(iu06,*) 'I/O problem - iostat is:',ios
        call exitme (3, 'Could not read h, u, v , z from aufpkt' )
      endif
    enddo

  end subroutine ReadDynHUVZ

!===============================================================================

  subroutine InitST ( st )

    implicit none

    real(8), intent(out) :: st(1:,0:)

    st(2,0:) = s0
    st(1,0:) = t0
    
  end subroutine InitST

!===============================================================================

end module restart
