module transports2zero

  implicit none
  private
  
  public  :: modtrans2zero, transecttrans2zero

 !- parameter ------------------------------------------------------------------
 real(8), parameter, private :: zero = 0.0_8

  contains
  
    subroutine modtrans2zero(n3d, utr, vtr, temputr, tempvtr, saltutr, saltvtr )

      !- modules ---------------------------------------------------------------
      use dmi_omp, only : domp_get_domain

      !- dircetives ------------------------------------------------------------
      implicit none

      !- arguments -------------------------------------------------------------
      integer(4), intent(in) :: n3d
      real(8), intent(inout) :: utr(0:), vtr(0:)
      real(8), intent(inout) :: temputr(0:), tempvtr(0:)
      real(8), intent(inout) :: saltutr(0:), saltvtr(0:)
      
      !- local vars ------------------------------------------------------------
      integer(4)  :: nn, nl, nu
      
      call domp_get_domain(1,n3d, nl, nu)

!$OMP BARRIER
!$OMP MASTER
        !- initialize on-land values:
        utr(0)     = zero
        vtr(0)     = zero
        temputr(0) = zero
        tempvtr(0) = zero
        saltutr(0) = zero
        saltvtr(0) = zero
!$OMP END MASTER
      
      do nn=nl,nu
        utr(nn)     = zero
        vtr(nn)     = zero
        temputr(nn) = zero
        tempvtr(nn) = zero
        saltutr(nn) = zero
        saltvtr(nn) = zero
      enddo
      
    end subroutine modtrans2zero


!===============================================================================
    

    subroutine transecttrans2zero ( n2d, kmax, utr, vtr, temputr, tempvtr,     &
                                    saltutr, saltvtr )

      !- modules ---------------------------------------------------------------
      use dmi_omp, only : domp_get_domain

      !- dircetives ------------------------------------------------------------
      implicit none

      !- arguments -------------------------------------------------------------
      integer(4), intent(in) :: n2d
      integer(4), intent(in) :: kmax
      real(8), intent(inout) :: utr(:,0:), vtr(:,0:)
      real(8), intent(inout) :: temputr(:,0:), tempvtr(:,0:)
      real(8), intent(inout) :: saltutr(:,0:), saltvtr(:,0:)
      
      !- local vars ------------------------------------------------------------
      integer(4)  :: nn, nl, nu, kt
      
      call domp_get_domain(1,n2d, nl, nu)

      do nn=nl,nu
        do kt = 1, kmax  
          utr(kt,nn)     = zero
          vtr(kt,nn)     = zero
          temputr(kt,nn) = zero
          tempvtr(kt,nn) = zero
          saltutr(kt,nn) = zero
          saltvtr(kt,nn) = zero
        enddo
      enddo
      
    end subroutine transecttrans2zero
    
end module transports2zero
