module updatearrays

  !- modules -------------------------------------------------------------------
  use constants, only : zero

  !- directives ----------------------------------------------------------------
  implicit none
  private

  !- public routines -----------------------------------------------------------
  public :: zerofy, new2old, h_srfplusconst, hplusconst, update_h, update_z,   &
            set_h_srf, set_h, update_h_ref

contains

  !-----------------------------------------------------------------------------

  subroutine zerofy ( a1, a2, a3, a4, a5, a6, kh, n2, mk, ind, idx )
    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: n2
    integer(4), intent(in)    :: kh(0:), mk(:,0:,0:), ind(:,:)
    real(8),    intent(inout) :: a1(0:), a2(0:), a3(0:), a4(0:), a5(0:), a6(0:)
    integer(4), intent(inout) :: idx(:,:)

    !- local variables ---------------------------------------------------------
    integer(4) :: np, nl, nu
    integer(4) :: i, j, mi, mi0, kb

!$OMP BARRIER
!$OMP MASTER
    a1(0) = zero
    a2(0) = zero
    a3(0) = zero
    a4(0) = zero
    a5(0) = zero
    a6(0) = zero
!$OMP END MASTER

    call domp_get_domain(kh, 1, n2, nl, nu, idx)

    do np=nl,nu
      a1(np) = zero
      a2(np) = zero
      a3(np) = zero
      a4(np) = zero
      a5(np) = zero
      a6(np) = zero
    enddo
     
    do np=nl,nu
      kb = kh(np)
      if (kb < 2) cycle
      i = ind(1,np)
      j = ind(2,np)
      mi0 = mk(2,i,j)
      do mi=mi0,mi0+kb-2
        a1(mi) = zero
        a2(mi) = zero
        a3(mi) = zero
        a4(mi) = zero
        a5(mi) = zero
        a6(mi) = zero
      enddo
    enddo

  end subroutine zerofy

  !-----------------------------------------------------------------------------

  subroutine new2old (un, vn, n2d, mk, ind, kh, uo, vo, idx)

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: n2d
    integer(4), intent(in)    :: kh(0:), mk(:,0:,0:), ind(:,:)
    real(8),    intent(in)    :: un(0:), vn(0:)
    real(8),    intent(inout) :: uo(0:), vo(0:)
    integer(4), intent(inout) :: idx(:,:)
      
    !- local variables ---------------------------------------------------------
    integer(4) :: np, nl, nu, i, j, mi

    call domp_get_domain(kh, 1, n2d, nl, nu, idx)
    do np=nl,nu
      uo(np) = un(np)
      vo(np) = vn(np)
    enddo
    do np=nl,nu
      if (kh(np)<=1) cycle
      i = ind(1,np)
      j = ind(2,np)
      do mi= mk(2,i,j),mk(kh(np),i,j)
        uo(mi) = un(mi)
        vo(mi) = vn(mi)
      enddo
    enddo

  end subroutine new2old

  !-----------------------------------------------------------------------------

  subroutine h_srfplusconst (iwet2, kh, CONST, hold, hnew, idx)

    !- modules -----------------------------------------------------------------
    use dmi_omp,   only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: iwet2, kh(0:)
    real(8),    intent(in)    :: CONST
    real(8),    intent(inout) :: hold(0:), hnew(0:)
    integer(4), intent(inout) :: idx(:,:)
      
    !- local variables ---------------------------------------------------------
    integer(4) :: nsrf, nl, nu

    call domp_get_domain(kh, 1, iwet2, nl, nu, idx)

    do nsrf=nl,nu
      hold(nsrf) = hold(nsrf) + CONST
      hnew(nsrf) = hnew(nsrf) + CONST
    enddo

  end subroutine h_srfplusconst

  !-----------------------------------------------------------------------------

  subroutine hplusconst (mmk, ind, kh, iwet2, CONST, hold, hnew, idx)

    !- modules -----------------------------------------------------------------
    use dmi_omp,   only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: mmk(:,0:,0:)
    integer(4), intent(in)    :: ind(:,:)
    integer(4), intent(in)    :: kh(0:)
    integer(4), intent(in)    :: iwet2
    real(8),    intent(in)    :: CONST
    real(8),    intent(inout) :: hold(0:), hnew(0:)
    integer(4), intent(inout) :: idx(:,:)
      
    !- local variables ---------------------------------------------------------
    integer(4) :: nsrf, nl, nu
    integer(4) :: i, j, mi, mi0

    !FIXME: Using dynamical vert. coord. another domp_get_domain probably more
    !       sensible - this has to be tested!!
    call domp_get_domain(kh, 1, iwet2, nl, nu, idx)

    do nsrf=nl,nu
      hold(nsrf) = hold(nsrf) + CONST
      hnew(nsrf) = hnew(nsrf) + CONST
    enddo
    do nsrf=nl,nu
      if (kh(nsrf) < 2) cycle
      i = ind(1,nsrf)
      j = ind(2,nsrf)
      mi0 = mmk(2,i,j)
      do mi=mi0,mi0+kh(nsrf)-2
        hold(mi) = hold(mi) + CONST
        hnew(mi) = hnew(mi) + CONST
      enddo
    enddo

  end subroutine hplusconst
  
  !-----------------------------------------------------------------------------

  subroutine update_h (hz, iwet2, z, h, kh, idx)

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: iwet2, kh(0:)
    real(8),    intent(in)    :: hz(0:), z(0:)
    real(8),    intent(inout) :: h(0:)
    integer(4), intent(inout) :: idx(:,:)
      
    !- local variables ---------------------------------------------------------
    integer(4) :: nl, nu

    call domp_get_domain(kh, 1, iwet2, nl, nu, idx)

    h(nl:nu)  = hz(nl:nu) + z(nl:nu)

  end subroutine update_h
  !-----------------------------------------------------------------------------

  subroutine update_h_ref (hz, iwet2, z, elev_ref, h, kh, idx)

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: iwet2, kh(0:)
    real(8),    intent(in)    :: hz(0:), z(0:), elev_ref(0:)
    real(8),    intent(inout) :: h(0:)
    integer(4), intent(inout) :: idx(:,:)
      
    !- local variables ---------------------------------------------------------
    integer(4) :: nsrf, nl, nu

    call domp_get_domain(kh, 1, iwet2, nl, nu, idx)

    do nsrf=nl,nu
      h(nsrf)  = hz(nsrf) + z(nsrf) - elev_ref(nsrf)
    enddo

  end subroutine update_h_ref

  !-----------------------------------------------------------------------------

  subroutine update_z (zn, iwet2, z, kh, idx)

    !- modules -----------------------------------------------------------------
    use dmi_omp, only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: iwet2, kh(0:)
    real(8),    intent(in)    :: zn(0:)
    real(8),    intent(inout) :: z(0:)
    integer(4), intent(inout) :: idx(:,:)
      
    !- local variables ---------------------------------------------------------
    integer(4) :: nsrf, nl, nu

    call domp_get_domain(kh, 1, iwet2, nl, nu, idx)

    do nsrf=nl,nu
      z(nsrf)  = zn(nsrf)
    enddo

  end subroutine update_z
  
  !-----------------------------------------------------------------------------

  subroutine set_h_srf (h_act, iwet2, h_set, kh, idx)

    !- modules -----------------------------------------------------------------
    use dmi_omp,   only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: iwet2, kh(0:)
    real(8),    intent(in)    :: h_act(0:)
    real(8),    intent(inout) :: h_set(0:)
    integer(4), intent(inout) :: idx(:,:)
      
    !- local variables ---------------------------------------------------------
    integer(4) :: nsrf, nl, nu

    call domp_get_domain(kh, 1, iwet2, nl, nu, idx)

    do nsrf=nl,nu
      h_set(nsrf) = h_act(nsrf)
    enddo

  end subroutine set_h_srf
  
  !-----------------------------------------------------------------------------

  subroutine set_h (mmk, ind, kh, h_act, iwet2, h_set, idx)

    !- modules -----------------------------------------------------------------
    use dmi_omp,   only : domp_get_domain

    !- directives --------------------------------------------------------------
    implicit none

    !- arguments ---------------------------------------------------------------
    integer(4), intent(in)    :: iwet2, mmk(:,0:,0:), ind(:,:), kh(0:)
    real(8),    intent(in)    :: h_act(0:)
    real(8),    intent(inout) :: h_set(0:)
    integer(4), intent(inout) :: idx(:,:)
      
    !- local variables ---------------------------------------------------------
    integer(4) :: nsrf, nl, nu
    integer(4) :: i, j, mi, mi0

    !FIXME: Using dynamical vert. coord. another domp_get_domain probably more
    !       sensible - this has to be tested!!
    call domp_get_domain(kh, 1, iwet2, nl, nu, idx)

    do nsrf=nl,nu
      h_set(nsrf) = h_act(nsrf)
    enddo
    do nsrf=nl,nu
      if (kh(nsrf) < 2) cycle
      i = ind(1,nsrf)
      j = ind(2,nsrf)
      mi0 = mmk(2,i,j)
      do mi=mi0,mi0+kh(nsrf)-2
        h_set(mi) = h_act(mi)
      enddo
    enddo

  end subroutine set_h


end module updatearrays
