module weireqs

  implicit none
  private

  public :: weir, weir_init, weir_diff

contains

subroutine weir (nweir, widx, mmk, ty, tx, z, zweir, wflowr, un, vn)

  use constants,   only : g, zero, milli

  implicit none
  
  integer(4), intent(in) :: nweir
  integer(4), intent(in) :: widx(:,0:)
  integer(4), intent(in) :: mmk(:,0:,0:)
  real(8), intent(in)    :: ty
  real(8), intent(in)    :: tx(:)
  real(8), intent(in)    :: z(0:)
  real(8), intent(in)    :: zweir(0:)
  real(8), intent(in)    :: wflowr(0:)
  real(8), intent(inout) :: un(0:)
  real(8), intent(inout) :: vn(0:)

  integer(4)    :: nw, i, j, mi
  real(8)       :: dpress, dzweir, zaff, zw, umi, vmi, ept  

  do nw = 1,nweir

    i = widx(1,nw)
    j = widx(2,nw)

    zaff     = zweir(nw)

    dpress   = zero

    !- Weir intercepts northward flow
    if(widx(3,nw) == 1) then
      mi     = mmk(1,i,j)
      zw     = min(zaff, z(mmk(1,i+1,j)))
      dzweir = max(zero, (zw-z(mmk(1,i,j))))
      dpress = -(g*ty*dzweir)
      ept    = max(zaff - z(mmk(1,i+1,j)), zero)

      !---- v-equation  (u-eqation not needed) -------------------------------
      vmi = min(vn(mi),vn(mi)+dpress)
      if( z(mmk(1,i,j)) < zaff ) then
        vmi = max(zero,vmi)
        if( z(mmk(1,i+1,j)) < zaff ) vmi = max(zero, vmi*exp(-ept)-milli)
      endif
      vn(mi) = wflowr(nw)*vmi

    !- Weir intercepts eastward flow
    elseif(widx(3,nw) == 2) then
      mi     = mmk(1,i,j)
      zw     = min(zaff, z(mmk(1,i,j)))
      dzweir = max(zero,(zw-z(mmk(1,i,j+1))))
      dpress = -(g*tx(i)*dzweir)
      ept    = max(zaff - z(mmk(1,i,j)), zero)

      !---- u-equation (v-eqation not needed) --------------------------------
      umi = min(un(mi),un(mi)+dpress)
      if( z(mmk(1,i,j)) < zaff ) then
        umi = max(zero,umi)
        if( z(mmk(1,i,j+1)) < zaff ) umi = max(zero, umi*exp(-ept)-milli)
      endif
      un(mi) = wflowr(nw)*umi

    !- Weir intercepts southward flow
    elseif(widx(3,nw) == 3) then
      mi     = mmk(1,i,j)
      zw     = min(zaff, z(mmk(1,i,j)))
      dzweir = max(zero, (zw-z(mmk(1,i+1,j))))
      dpress = g*ty*dzweir
      ept    = max(zaff - z(mmk(1,i,j)), zero)

      !---- v-equation  (u-eqation not needed) --------------------------------
      vmi = max(vn(mi),vn(mi)+dpress)
      if( z(mmk(1,i+1,j)) < zaff ) then
        vmi = min(zero,vmi)
        if( z(mmk(1,i,j)) < zaff ) vmi = min(zero, vmi*exp(-ept)+milli)
      endif
      vn(mi) = wflowr(nw)*vmi

    !- Weir intercepts westward flow
    elseif(widx(3,nw) == 4) then
      mi     = mmk(1,i,j)
      zw     = min(zaff, z(mmk(1,i,j+1)))
      dzweir = max(zero,(zw-z(mmk(1,i,j))))
      dpress = g*tx(i)*dzweir
      ept    = max(zaff - z(mmk(1,i,j+1)), zero)

      !---- u-equation (v-eqation not needed) --------------------------------
      umi = max(un(mi),un(mi)+dpress)
      if( z(mmk(1,i,j)) < zaff ) then
        umi = min(zero,umi)
        if( z(mmk(1,i,j+1)) < zaff ) umi = min(zero, umi*exp(-ept)+milli)
      endif
      un(mi) = wflowr(nw)*umi

    endif

  enddo

end subroutine weir

!=============================================================================

subroutine weir_init (ia, nweir, widx, me, khu, khv, zweir, wflowr)

    use cmod_date,     only : jultimea
    use constants,     only : TIM, zero, one, twenty
    use dmi_mpi_global,only : iu06
    use exits,         only : exitme
    use io_subs,       only : io_new_unit

    implicit none

    integer(4), intent(in)    :: ia, nweir
    integer(4), intent(in)    :: widx(:,0:)
    integer(4), intent(in)    :: me(0:,0:)
    integer(4), intent(in)    :: khu(0:)
    integer(4), intent(in)    :: khv(0:)
    real(8),    intent(in)    :: wflowr(0:)
    real(8),    intent(inout) :: zweir(0:)

    character(1)  :: c1
    character(2)  :: c2
    character(3)  :: c3
    character(4)  :: iformat
    character(5)  :: lin, col, lay
    character(13) :: cformat
    character(14) :: fname
    character(19) :: cdat
    integer(4)    :: lun, ios, i, j, n, kb
    real(8)       :: timfl


    !- Check the weir set up-in weir points only one (flow) layer is acceptet
    do n = 1, nweir
      i = widx(1,n)
      j = widx(2,n)

      !- Weir intercepts north- or southward flow
      if ( widx(3,n) == 1 .or. widx(3,n) == 3) then
        kb = khv(me(i,j))
      !- Weir intercepts east- or westward flow
      elseif(widx(3,n) == 2 .or. widx(3,n) == 4) then
        kb = khu(me(i,j))
      endif

      !- Check if weir points have exactly one velocity-layer
      if (kb /= 1) then
        write(lin,*) i
        write(col,*) j
        write(lay,*) kb
        call exitme(1,'Layer number in weir point: line:'//lin//' column:'// &
                       col//' is '//lay//' and not 1...stop')
      endif

      !- Check if flow rate of weirs is filled with sane values (0<wflowr<=1)
      !- Because of weir buildings the cross-section area is restricted, so
      !- that also the flow rate should be restricted.
      !- In this code the flow rate will be multiplied with the current 
      !- through the weir - physical correct would be a multiplication with 
      !- the cross-section area
      if ( wflowr(n) <= zero .or. wflowr(n) > one ) call exitme(1,           &
                             'weir flowrate must be > zero and <= one...stop')
    enddo

    !- Read in the actual weir heights
    !- set up file name, format, etc ... -------------------------------------
    if (nweir < 1) then
      return
    elseif (nweir < 10) then
      iformat = '(i1)'
      write(c1,iformat) nweir
      cformat = '(a19, '//c1//'f8.1)'
    elseif (nweir < 100) then
      iformat = '(i2)'
      write(c2,iformat) nweir
      cformat = '(a19,'//c2//'f8.1)'
    elseif (nweir < 1000) then
      iformat = '(i3)'
      write(c3,iformat) nweir
      cformat = '(a19,'//c3//'f8.1)'
    else
      ! trap
      call exitme(1,'ERROR: Number of weir points too large.')
    endif
    if (ia < 10) then
      iformat = '(i1)'
      write(c1,iformat) ia
      fname   = 'weirheights_0'//c1
    elseif (ia < 100) then
      iformat = '(i2)'
      write(c2,iformat) ia
      fname   = 'weirheights_'//c2
    else
      ! trap
      call exitme(1,'ERROR: Number of areas too large.')
    endif

    !- open the file and read it ---------------------------------------------
    lun = io_new_unit()
    open (lun,file=fname,action='read',status='old',iostat=ios)
    if (ios /= 0) then
      write(iu06,'(a49)') 'Unable to open '//fname//', continuing without'
    else
      read (lun,*)
      read (lun,'(a19)') cdat
      call jultimea(cdat,timfl)
      if ( abs(TIM-timfl) <= twenty ) then
        backspace (lun)
        read (lun,cformat) cdat, (zweir(i),i=1,nweir)
        close (lun)
      endif
    endif

end subroutine weir_init

!=============================================================================

subroutine weir_diff (nweir, widx, mmk, u, v, uvdam)

  !--------------------------------------------------------------------------!
  !- This subroutine checks if the weir is closed and acts like a dam or if -!
  !- the weir is open so that also diffusion across the weir is possible    -!
  !--------------------------------------------------------------------------!

  use constants,   only : zero, one

  implicit none

  integer(4), intent(in) :: nweir
  integer(4), intent(in) :: widx(:,0:)
  integer(4), intent(in) :: mmk(:,0:,0:)
  real(8), intent(in)    :: u(0:)
  real(8), intent(in)    :: v(0:)
  real(8), intent(inout) :: uvdam(:,0:)

  integer(4) :: nw, i, j

  do nw = 1,nweir

    i = widx(1,nw)
    j = widx(2,nw)

    !- Weir intercepts north- or southward flow
    if(widx(3,nw) == 1 .or. widx(3,nw) == 3) then
      if (v(mmk(1,i,j)) == zero) then
        uvdam(2,mmk(1,i,j)) = zero
      else
        uvdam(2,mmk(1,i,j)) = one
      endif

    !- Weir intercepts east- or westward flow
    elseif(widx(3,nw) == 2 .or. widx(3,nw) == 4) then
      if (u(mmk(1,i,j)) == zero) then
        uvdam(1,mmk(1,i,j)) = zero
      else
        uvdam(1,mmk(1,i,j)) = one
      endif

    endif

  enddo

end subroutine weir_diff

end module weireqs
