module writegrib

  use cmod_mem,  only : cmr2, cmr3
  implicit none
  private
  public  :: grb_init,grb_close,grb_write
  integer, parameter, private       :: NOT_USED=-1  ! are they initialized
  logical, save, private            :: initialized
  integer(4), private, parameter :: zmiss  = -99999
  integer, allocatable, save, private :: idclone(:), outfile(:)
  character(LEN=256), allocatable, save, private :: ofilename(:)
  type(cmr2), allocatable, private, save ::  tmp2d(:)  
  type(cmr3), allocatable, private, save ::  tmp3d(:)  
  interface vec2mat
    module procedure vec2mat_2d
    module procedure vec2mat_3d
  end interface  

contains

!===============================================================================

  subroutine grb_init ()
     ! for each area do, open one relating to this
     ! clone all files from generic grib header
     ! generate a specific header for each area that we can use throughout the
     ! this module
    use grib_api, only : grib_open_file, grib_new_from_file, grib_release,     &
                         grib_set_int, grib_clone, grib_close_file
    use exits,          only : exitme
    use cmod_params,    only : narea, cfgfile, mmx, nmx, kmx
    use dmi_mpi_global, only : iu06

    implicit none
    integer(4)    :: ios, ia, infile, igrib
    integer(4)    :: npparallel,npmeridian

    allocate( tmp2d(narea), tmp3d(narea), idclone(narea),outfile(narea),       &
              ofilename(narea) )
    do ia=1,narea
      allocate( tmp3d(ia)%p(1:nmx(ia),1:mmx(ia),1:kmx(ia)),                    &
                tmp2d(ia)%p(1:nmx(ia),1:mmx(ia)) )
    enddo
    do ia=1,narea
      ofilename(ia) = 'area_'//trim(cfgfile(ia)(6:))//'.grb'
    enddo
    call grib_open_file(infile,'grb_clone.dat','r',ios)
    if (ios.ne. 0) then
      stop 'ERROR: Can not open grb_clone.dat - bailing out'
    endif
    call grib_new_from_file(infile,igrib,ios)
    do ia=1,narea
      call grib_clone(igrib, idclone(ia))
    enddo
    call grib_release(igrib)
    call grib_close_file(infile)

    write(iu06,'(a21)') 'Opening grib output files : '
    do ia=1,narea
      call grib_open_file(outfile(ia),ofilename(ia),'w',ios)
      if (ios.ne. 0) then
        stop 'ERROR: Can not open grib output file - bailing out'
      endif
    enddo

    ! proper initialization of grib_headers
!    do ia=1,narea
      ! FIXME set the area local stuff
!      call grib_set_int(idclone(ia),'numberOfPointsAlongAParallel',npparallel)
!      call grib_set_int(idclone(ia),'numberOfPointsAlongAMeridian',npmeridian)
!   enddo

    initialized  = .true.
  end subroutine grb_init

!===============================================================================

  subroutine grb_close()
    use cmod_params,   only : narea
    use grib_api,      only : grib_close_file
    use exits,         only : exitme
    implicit none
    integer(4) :: ia, ios
    do ia=1,narea
      call grib_close_file(outfile(ia),ios)
      if (ios.ne. 0) then
        stop 'ERROR: Can not close grib file - bailing out'
      endif
    enddo
  end subroutine grb_close 

!===============================================================================

  subroutine grb_write (ctim,tim,izeits)
    use grib_api,       only : grib_set, grib_write
    use exits,          only : exitme
    use cmod_params,    only : narea, mmx, nmx, kmx
    use dmi_mpi,        only : mpi_io_serial
    use dmi_mpi_global, only : iu06
    use cmod_arrays,    only : u, mm1k

    implicit none
    character(len=2)    :: step
    character(19), intent(in) :: ctim
    real(8),       intent(in) :: tim
    integer(4),    intent(in) :: izeits
    integer(4) :: param
    integer(4) :: ia
    integer(4)    :: year,month,day,hour,yoc

    if (mpi_io_serial) then
      write(iu06,'(a,a19)') 'write on foobar at time: ',ctim
      write(*,*) ctim,tim

      ! set date stuff
      do ia=1,narea
        call grib_set(idclone(ia),'yearOfCentury',yoc)
        call grib_set(idclone(ia),'month',month)
        call grib_set(idclone(ia),'day',day)
        call grib_set(idclone(ia),'hour',hour)
      enddo
      ! prepare field for writing
      ! wu, wv, u, v, w, h, z, u, v, ww, z, s, t
      ! write field
      do ia=1,narea
        call vec2mat(tmp3d(ia)%p, u(ia)%p, mm1k(ia)%p, mmx(ia),nmx(ia),kmx(ia))
        call grib_set(idclone(ia),'indicatorOfParameter',param)
        call grib_set(idclone(ia),'stepRange',adjustl(step))
        call grib_set(idclone(ia),'values',pack(tmp3d(ia)%p,.true.))
        call grib_write(idclone(ia),outfile(ia))
      enddo
    endif
  end subroutine grb_write 

!===============================================================================
  
  subroutine vec2mat_2d(outvar, var, mm, mmx, nmx)
    ! convert compressed 1D vector into 2D matrix + transpose matrix
    implicit none
    real(8), intent(out)   :: outvar(:,:)
    real(8), intent(in)    :: var(0:)
    integer(4), intent(in) :: mm(:,0:,0:)
    integer(4), intent(in) :: mmx, nmx
    integer(4)             :: i,j
    do j = 1,nmx
      do i = 1, mmx
        if(mm(1, i,j) == 0) then
          outvar(j,i) = zmiss
        else
          outvar(j,i) = var(mm(1, i,j))
        endif
      enddo
    enddo
  end subroutine vec2mat_2d
  
!===============================================================================

  subroutine vec2mat_3d (outvar, var, mm, mmx, nmx, kmx)
    ! convert compressed 1D vector into 3D matrix + transpose matrix
    implicit none
    real(8), intent(out)   :: outvar(:,:,:)
    real(8), intent(in)    :: var(0:)
    integer(4), intent(in) :: mm(:,0:,0:)
    integer(4), intent(in) :: mmx, nmx, kmx
    integer(4)             :: i,j,k
    do k = 1,kmx
      do j = 1, nmx
        do i = 1, mmx
          if (mm(k,i,j) == 0) then
            outvar(j,i,k) = zmiss
          else
            outvar(j,i,k) = var(mm(k,i,j))
          endif  
        enddo
      enddo
    enddo
  end subroutine vec2mat_3d    

!===============================================================================

end module writegrib
