module writenetcdf_BSH
  ! methods to write netcdf-output-files
  
  use constants, only : zero, one, ten
  use cmod_mem,  only : cmi2, cmi3, cmr1
  use compress,  only : wuvscale, uvscale, stscale, wscale, hscale, zscale,    &
                       edscale, ecscale, sdscale, uvescale, tsscale, iotascale,&
                       wu_i4, wv_i4, u_i4, v_i4, w_i4, h_i4, z_i4, s_i4, t_i4, &
                       eisdi_i4, eisco_i4, snowd_i4, tsnei_i4, ueis_i4,veis_i4,&
                       iota1_i4, iota2_i4, iota3_i4, mm_op,                    &
                       allocate_compressed_arrays,allocate_compressed_ice_arrays

  implicit none
  private
  
  !- interfaces ----------------------------------------------------------------
  interface Fill_dummy_3D
    module procedure Fill_dummy_3D_int
    module procedure Fill_dummy_3D_logical
  end interface  

  public  :: netcdf_new, netcdf_next, netcdf_close, netcdf_write 
  public  :: netcdf_icefile_new, netcdf_icefile_next, write_netcdf_icefile,    &
             netcdf_icefile_close,arc2deg, calculate_depth
  public  :: round
  private :: Fill_dummy_3D, Fill_dummy_3D_int, Fill_dummy_3D_logical,          &
             Fill_dummy_4D

  logical,       public,  save :: write_c_file, write_m_file, write_t_file,    &
                                  write_z_file, write_i_file
  integer(4),    private, save :: incdf, icencdf
  character(40), private, save :: ncdfnam
  integer(4),    allocatable, private, save :: lunmncdf(:), luncncdf(:),       &
                                               luntncdf(:), lunzncdf(:),       &
                                               lunincdf(:), luniotancdf(:)
  integer(4),    allocatable, private, save :: lonmid(:),latmid(:),dpthmid(:), &
                                               tmmid(:), loncid(:), latcid(:), &
                                               dpthcid(:), tmcid(:), lontid(:),&
                                               lattid(:), dpthtid(:), tmtid(:),&
                                               lonzid(:),latzid(:),dpthzid(:), &
                                               tmzid(:), loniid(:), latiid(:), &
                                               dpthiid(:), tmiid(:),           &
                                               loniotaid(:), latiotaid(:),     &
                                               dpthiotaid(:), tmiotaid(:)
  integer(4),    allocatable, private, save :: lonmvi(:),latmvi(:),dpthmvi(:), &
                                               tmmvi(:), loncvi(:), latcvi(:), &
                                               dpthcvi(:), tmcvi(:), lontvi(:),&
                                               lattvi(:), dpthtvi(:), tmtvi(:),&
                                               lonzvi(:),latzvi(:),dpthzvi(:), &
                                               tmzvi(:), lonivi(:), lativi(:), &
                                               dpthivi(:), tmivi(:),           &
                                               loniotavi(:), latiotavi(:),     &
                                               dpthiotavi(:), tmiotavi(:)
  integer(4),    allocatable, private, save :: wuid(:),wvid(:),uid(:),vid(:),  &
                                               wid(:),hid(:),zid(:),sid(:),    &
                                               tid(:),casusid(:),icethkid(:),  &
                                               icecoid(:),snowthkid(:),        &
                                               tsneiid(:),uiceid(:),viceid(:), &
                                               iota1id(:),iota2id(:),iota3id(:)
  character(40), allocatable, private, save :: mncdfnam(:), cncdfnam(:),       &
                                               tncdfnam(:), zncdfnam(:),       &
                                               incdfnam(:), iotancdfnam(:)

  !- Parameter -----------------------------------------------------------------
  integer(4),            private, parameter :: delete_value  = -2000000000
  real(8), dimension(1), private, parameter :: surface_depth = (/ 0.0_8 /) 
   
  !- Scale factors for extract -------------------------------------------------
  real(8), parameter :: ext_wuvscale = one / wuvscale   ! wind scalefactor
  real(8), parameter :: ext_uvscale  = one / uvscale    ! current scalefactor
  real(8), parameter :: ext_stscale  = one / stscale    ! salt/temp scalefactor
  real(8), parameter :: ext_wscale   = one / wscale     ! vertical vel. scalef.
  real(8), parameter :: ext_hscale   = one / hscale     ! layer thkn. scalef.
  real(8), parameter :: ext_zscale   = one / zscale     ! elevation scalefactor
  real(8), parameter :: ext_edscale  = one / edscale    ! ice thickness scalef.
  real(8), parameter :: ext_ecscale  = one / ecscale    ! ice cover scalefactor
  real(8), parameter :: ext_sdscale  = one / sdscale    ! snow thickn. scalef.
  real(8), parameter :: ext_uvescale = one / uvescale   ! ice velocity scalef.
  real(8), parameter :: ext_tsscale  = one / tsscale    ! tsnei scalefactor
  real(8), parameter :: ext_iotascale= one / iotascale  ! iota scalefactor
  
  !- Real arrays for output format ---------------------------------------------
  type(cmr1), allocatable, private, save :: lon(:), lat(:), depth(:)

  !- Integer dummy arrays used for output --------------------------------------
  type(cmi2), allocatable, private, save ::  dummy_3D(:)  
  type(cmi3), allocatable, private, save ::  dummy_4D(:)  

contains

!===============================================================================

  subroutine netcdf_new ()

    use constants,      only : archive, toutput, filesizemax
    use cmod_arrays,    only : dz, xkoor, ykoor
    use cmod_params,    only : narea, cfgfile, iw2, iw3, mmx, nmx, kmx, mm1,   &
                               nthreads
    use dmi_mpi_global, only : iu06
    use cmod_miscvars,  only : schedule

    implicit none

    logical, save :: FirstEntry = .true.
    character(2)  :: cncdf
    integer(4)    :: ia, ifsh, nt, idiv
    real(8)       :: fsize, ofsize, osch, ifsv


    if (FirstEntry) then
      FirstEntry = .false.
      
      call allocate_compressed_arrays( )

      ! allocate on first entry:
      ia = narea
      allocate( lon(ia), lat(ia), depth(ia) )
      allocate( dummy_3D(ia), dummy_4D(ia) )
      allocate( lunmncdf(ia), luncncdf(ia), luntncdf(ia), lunzncdf(ia),        &
                mncdfnam(ia), cncdfnam(ia), tncdfnam(ia), zncdfnam(ia) )
      allocate( lonmid(ia), latmid(ia), dpthmid(ia), tmmid(ia), loncid(ia),    &
                latcid(ia), dpthcid(ia), tmcid(ia), lontid(ia), lattid(ia),    &
                dpthtid(ia), tmtid(ia), lonzid(ia), latzid(ia), dpthzid(ia),   &
                tmzid(ia) )
      allocate( lonmvi(ia), latmvi(ia), dpthmvi(ia), tmmvi(ia), loncvi(ia),    &
                latcvi(ia), dpthcvi(ia), tmcvi(ia), lontvi(ia), lattvi(ia),    &
                dpthtvi(ia), tmtvi(ia), lonzvi(ia), latzvi(ia), dpthzvi(ia),   &
                tmzvi(ia) )
      allocate( wuid(ia), wvid(ia), uid(ia), vid(ia), wid(ia), hid(ia),zid(ia),&
                sid(ia), tid(ia) )
      do ia=1,narea
        allocate( lon(ia)%p(1:nmx(ia)), lat(ia)%p(1:mmx(ia)),                  &
                  depth(ia)%p(1:kmx(ia)) )

        allocate( dummy_4D(ia)%p(1:nmx(ia),1:mmx(ia),1:kmx(ia)),               &
                  dummy_3D(ia)%p(1:nmx(ia),1:mmx(ia)) )

        call arc2deg( mmx(ia), nmx(ia), xkoor(ia)%p, ykoor(ia)%p, lon(ia)%p,   &
                      lat(ia)%p )
        call calculate_depth( kmx(ia), dz(ia)%p, depth(ia)%p )
      enddo

      incdf = 0
      if (archive) then
        ncdfnam = 'ncdf_archive.nc'
        do ia=1,narea
           mncdfnam(ia) = 'm_file_'//trim(cfgfile(ia)(6:))//'_archive.nc'
           cncdfnam(ia) = 'c_file_'//trim(cfgfile(ia)(6:))//'_archive.nc'
           tncdfnam(ia) = 't_file_'//trim(cfgfile(ia)(6:))//'_archive.nc'
           zncdfnam(ia) = 'z_file_'//trim(cfgfile(ia)(6:))//'_archive.nc'
        enddo
      else
        write(cncdf,'(i2.2)') incdf
        ncdfnam = 'ncdf_'//cncdf//'.nc'
        do ia=1,narea
          mncdfnam(ia) = 'm_file_'//trim(cfgfile(ia)(6:))//'_'//cncdf//'.nc'
          cncdfnam(ia) = 'c_file_'//trim(cfgfile(ia)(6:))//'_'//cncdf//'.nc'
          tncdfnam(ia) = 't_file_'//trim(cfgfile(ia)(6:))//'_'//cncdf//'.nc'
          zncdfnam(ia) = 'z_file_'//trim(cfgfile(ia)(6:))//'_'//cncdf//'.nc'
        enddo
      endif

!FIXME: The calculation of the file-size has to be reworked
!      ! Calculate file size in Bytes of 8bit:
!      !    ifsh is size of the static data (the header),
!      !    ifsv is size of the dynamic data (i.e. the vars) for each time step.
!      ! Total size of file is ifsh + nt*ifsv where nt is No. of time steps
!      ! in the file.
!      ifsh = 4 + 11 + 4                                 ! aufdat header
!      ifsv = 4. + 19. + 8. + 4.                         ! ctim, tim
!      ifsv = ifsv + 4.
!      do ia=1,narea
!        ifsv = ifsv + 4.*2.*real(iw2(ia)+1,8)           ! Wu,Wv
!      enddo
!      ifsv = ifsv + 4.
!      do ia=1,narea
!        ifsv = ifsv + 4.
!        ifsv = ifsv + 4.*real(3*(iw3(ia)+1) + (iw2(ia)+1),8)   ! u,v,w,z
!        ifsv = ifsv + 4.
!      enddo
!      do ia=1,narea
!        ifsv = ifsv + 4.
!        ifsv = ifsv + 4.*2.*real(iw3(ia)+1,8)           ! s,t
!        ifsv = ifsv + 4.
!      enddo
!      nt    = schedule*nint(3600._8/toutput)
!      osch  = 3600.*real(schedule,8)
!      fsize = real(ifsh,8) + real(nt,8)*ifsv
!      if (fsize > (1024.*1024.*1024.*1024.)) then
!        ofsize = fsize/(1024.*1024.*1024.*1024.)
!        ctmp   = 'TB'
!      elseif (fsize > (1024.*1024.*1024.)) then
!        ofsize = fsize/(1024.*1024.*1024.)
!        ctmp   = 'GB'
!      elseif (fsize > (1024.*1024.)) then
!        ofsize = fsize/(1024.*1024.)
!        ctmp   = 'MB'
!      elseif (fsize > 1024.) then
!        ofsize = fsize/1024.
!        ctmp   = 'KB'
!      else
!        ofsize = fsize
!        ctmp   = 'B '
!      endif
!      write(iu06,'(a17,1x,f5.1,1x,a2)') 'Output file size:', ofsize, ctmp 
!      if (fsize/(1024.*1024.*1024.) > filesizemax) then
!        write(iu06,'(a7,1x,f8.3,1x,a6)') 'exceeds', filesizemax, 'GB, so'
!!FIXME: re-scheduling is actually NOT implemented yet, therefore the next
!!       two lines are just suggestions:
!        write(iu06,'(a33)') 'filesizemax should be higher or  '
!!        write(iu06,'(a33)') 'output should be re-scheduled to:'
!!        write(iu06,'(a31)') 'output will be re-scheduled to:'
!        idiv   = 1
!        osch   = 0.5*osch
!        ofsize = 0.5*fsize
!        do while (ofsize/(1024.*1024.*1024.) > filesizemax)
!          idiv   = idiv + 1
!          osch   = 0.5*osch
!          ofsize = 0.5*ofsize
!        enddo
!        write(iu06,'(f8.3,1x,a3)') real(schedule,8)/(2.0**idiv), 'hrs'
!!FIXME: pseudo-code drafting the treatment of too large files:
!!
!!       each output chunk of 'schedule' hrs should be subdivided into idiv+1
!!         sub-intervals each of length schedule/(2**idiv) 
!!         
!!       this should somehow be communicated to where cmod_main calls
!!         tempdat_next
!!       
!!       calc ntempdat, No. of tempdat files needed
!!       if (ntempdat >= 100) call exitme(1,'Too many outputfiles needed')
!!
!      endif

    else
      incdf = incdf + 1
      write(cncdf,'(i2.2)') incdf
      ncdfnam = 'ncdf_'//cncdf//'.nc'
      do ia=1,narea
        mncdfnam(ia) = 'm_file_'//trim(cfgfile(ia)(6:))//'_'//cncdf//'.nc'
        cncdfnam(ia) = 'c_file_'//trim(cfgfile(ia)(6:))//'_'//cncdf//'.nc'
        tncdfnam(ia) = 't_file_'//trim(cfgfile(ia)(6:))//'_'//cncdf//'.nc'
        zncdfnam(ia) = 'z_file_'//trim(cfgfile(ia)(6:))//'_'//cncdf//'.nc'
      enddo
    endif


  end subroutine netcdf_new

!===============================================================================

  subroutine netcdf_close()

    use cmod_params,      only : narea
    use netcdf_interface, only : ncdf_close_file 
    
    implicit none
    
    integer(4)    :: ia
    
    do ia=1,narea
      if (write_m_file) call ncdf_close_file (lunmncdf(ia))
      if (write_c_file) call ncdf_close_file (luncncdf(ia))
      if (write_t_file) call ncdf_close_file (luntncdf(ia))
      if (write_z_file) call ncdf_close_file (lunzncdf(ia))
    enddo
  end subroutine netcdf_close

!===============================================================================

  subroutine netcdf_icefile_close()

    use cmod_params,      only : narea
    use constants,        only : iotamodel
    use netcdf_interface, only : ncdf_close_file 

    implicit none

    integer(4)    :: ia
    
    do ia=1,narea
      call ncdf_close_file (lunincdf(ia))
      if ( iotamodel ) call ncdf_close_file (luniotancdf(ia))
    enddo

  end subroutine netcdf_icefile_close  

!===============================================================================

  subroutine netcdf_next ()

    use dmi_mpi_global, only : iu06
    implicit none

    call netcdf_close ()

    write(iu06,'(a33,a)') 'output stored on: m_,c_,t_ and z_',trim(ncdfnam)

    call netcdf_new ()

  end subroutine netcdf_next
  
!===============================================================================

  subroutine netcdf_icefile_next ()

    use dmi_mpi_global, only : iu06
    implicit none

    call netcdf_icefile_close ()

    write(iu06,'(a27,a)') 'icedata-output stored on: i_',trim(ncdfnam)

    call netcdf_icefile_new ()

  end subroutine netcdf_icefile_next
  
!===============================================================================

  subroutine netcdf_write (ctim,tim,aufdat,izeits,lmn2)

    use exits,            only : exitme
    use io_subs,          only : io_new_unit, flush_unit
    use cmod_arrays,      only : ti, xkoor, ykoor
    use cmod_params,      only : narea, nmx, mmx, kmx
    use constants,        only : dyn
    use dmi_mpi,          only : mpi_io_serial
    use dmi_mpi_global,   only : iu06
    use netcdf_interface, only : ncdf_create_new, ncdf_set_global_attr,        &
                                 ncdf_set_dimensions, ncdf_set_dim_vars,       &
                                 ncdf_save_dim_vars, ncdf_set_dim_attr,        &
                                 ncdf_set_3D_variables, ncdf_set_4D_variables, &
                                 ncdf_set_variable_attr, ncdf_end_definitions, &
                                 ncdf_save_time, ncdf_save_3D_vars,            &
                                 ncdf_save_4D_vars

    implicit none
    
    character(19), intent(in) :: ctim
    real(8),       intent(in) :: tim
    character(11), intent(in) :: aufdat
    integer(4),    intent(in) :: izeits,lmn2

    integer(4) :: ia
    real(8)    :: roundtim

    integer(4), save   :: counter
 
    if (mpi_io_serial) then
      write(iu06,'(a9,a,a10,a19)') 'write on ', trim(ncdfnam),                 &
                                   '-files at time: ',ctim
      !- create the netcdf-files and write the header when needed --------------
      if (izeits == lmn2) then
        counter = 1      

        do ia=1,narea
          if (write_m_file) then
            !- Create a new file
            call ncdf_create_new( mncdfnam(ia), lunmncdf(ia) )
            !- Set global attributes
            call ncdf_set_global_attr( lunmncdf(ia), minval(lon(ia)%p),        &
                                       maxval(lon(ia)%p), minval(lat(ia)%p),   &
                                       maxval(lat(ia)%p), zero, zero )  
            !- Set dimensions
            call ncdf_set_dimensions( lunmncdf(ia), nmx(ia), lonmid(ia),       &
                                      mmx(ia), latmid(ia), 1, dpthmid(ia),     &
                                      tmmid(ia))
            !- define spatial and temporal coordinates:
            call ncdf_set_dim_vars( lunmncdf(ia), lonmid(ia), lonmvi(ia),      &
                                    latmid(ia), latmvi(ia),dpthmid(ia),        &
                                    dpthmvi(ia),tmmid(ia), tmmvi(ia) )
            !- set dimension attributes
            call ncdf_set_dim_attr ( lunmncdf(ia), lonmvi(ia), latmvi(ia),     &
                                     dpthmvi(ia), tmmvi(ia) )
            !- define variables of the meteorological netcdf-files
            call ncdf_set_3D_variables(lunmncdf(ia), "wuvel", lonmvi(ia),      &
                                       latmvi(ia), tmmvi(ia), wuid(ia) )
            call ncdf_set_3D_variables(lunmncdf(ia), "wvvel", lonmvi(ia),      &
                                       latmvi(ia), tmmvi(ia), wvid(ia) )
            !- set variable attributes of the meteorological netcdf-files      &
            call ncdf_set_variable_attr( lunmncdf(ia), wuid(ia),               &
                                         'eastward_wind_velocity', 'm/s',      &
                                         'meters per second','eastward_wind',  &
                                         delete_value, ext_wuvscale )
            call ncdf_set_variable_attr( lunmncdf(ia), wvid(ia),               &
                                         'northward_wind_velocity', 'm/s',     &
                                         'meters per second','northward_wind', &
                                         delete_value, ext_wuvscale )
            !- end of definitions
            call ncdf_end_definitions(lunmncdf(ia))

            !- write coordinates:
            call ncdf_save_dim_vars(lunmncdf(ia), lonmvi(ia), lon(ia)%p,       &
                                    latmvi(ia), lat(ia)%p, dpthmvi(ia),        &
                                    surface_depth  )
          endif
 
          if (write_c_file) then 
            !- Create a new file
            call ncdf_create_new( cncdfnam(ia), luncncdf(ia) )
            !- Set global attributes     
            call ncdf_set_global_attr( luncncdf(ia), minval(lon(ia)%p),        &
                                       maxval(lon(ia)%p), minval(lat(ia)%p),   &
                                       maxval(lat(ia)%p),zero,maxval(ti(ia)%p) )
            !- Set dimensions
            call ncdf_set_dimensions(luncncdf(ia), nmx(ia), loncid(ia),mmx(ia),&
                                     latcid(ia), kmx(ia),dpthcid(ia), tmcid(ia))
            !- define spatial and temporal coordinates:
            call ncdf_set_dim_vars( luncncdf(ia),loncid(ia),loncvi(ia),        &
                                    latcid(ia), latcvi(ia),dpthcid(ia),        &
                                    dpthcvi(ia),tmcid(ia), tmcvi(ia) )
            !- set dimension attributes
            call ncdf_set_dim_attr ( luncncdf(ia), loncvi(ia), latcvi(ia),     &
                                     dpthcvi(ia), tmcvi(ia) )
            !- define variables
            call ncdf_set_4D_variables(luncncdf(ia), "uvel", loncvi(ia),       &
                                       latcvi(ia),dpthcvi(ia),tmcvi(ia),uid(ia))
            call ncdf_set_4D_variables(luncncdf(ia), "vvel", loncvi(ia),       &
                                       latcvi(ia),dpthcvi(ia),tmcvi(ia),vid(ia))
            !- set variable attributes of the current netcdf-files
            call ncdf_set_variable_attr(luncncdf(ia), uid(ia),                 &
                                        'eastward_sea_water_velocity', 'm/s',  &
                                        'meters per second','eastward_current',&
                                        delete_value, ext_uvscale )
            call ncdf_set_variable_attr(luncncdf(ia), vid(ia),                 &
                                        'northward_sea_water_velocity', 'm/s', &
                                        'meters per second',                   &
                                        'northward_current', delete_value,     &
                                        ext_uvscale )
            !- end of definitions
            call ncdf_end_definitions(luncncdf(ia))
            
            !- write coordinates:
            call ncdf_save_dim_vars(luncncdf(ia), loncvi(ia), lon(ia)%p,       &
                                  latcvi(ia),lat(ia)%p,dpthcvi(ia),depth(ia)%p )
          endif

          if (write_t_file) then
            !- Create a new file
            call ncdf_create_new( tncdfnam(ia), luntncdf(ia) )
            !- Set global attributes     
            call ncdf_set_global_attr( luntncdf(ia), minval(lon(ia)%p),        &
                                       maxval(lon(ia)%p), minval(lat(ia)%p),   &
                                       maxval(lat(ia)%p),zero,maxval(ti(ia)%p) )
            !- Set dimensions
            call ncdf_set_dimensions(luntncdf(ia), nmx(ia),lontid(ia), mmx(ia),&
                                     lattid(ia), kmx(ia),dpthtid(ia), tmtid(ia))
            !- define spatial and temporal coordinates:
            call ncdf_set_dim_vars( luntncdf(ia),lontid(ia),lontvi(ia),        &
                                    lattid(ia), lattvi(ia),dpthtid(ia),        &
                                    dpthtvi(ia),tmtid(ia), tmtvi(ia) )
            !- set dimension attributes
            call ncdf_set_dim_attr ( luntncdf(ia), lontvi(ia), lattvi(ia),     &
                                     dpthtvi(ia), tmtvi(ia) )
            !- define variables of the teperature/salinity netcdf-files
            call ncdf_set_4D_variables(luntncdf(ia), "temp", lontvi(ia),       &
                                      lattvi(ia),dpthtvi(ia),tmtvi(ia),tid(ia) )
            call ncdf_set_4D_variables(luntncdf(ia), "salt", lontvi(ia),       &
                                      lattvi(ia),dpthtvi(ia),tmtvi(ia),sid(ia) )
            !- set variable attributes of the temperature / salinity ncdf-files 
            call ncdf_set_variable_attr(luntncdf(ia), tid(ia),                 &
                                        'sea_water_potential_temperature',     &
                                        'degree_Celsius', 'degree Celsius',    &
                                        'potential temperature', delete_value, &
                                        ext_stscale )
            call ncdf_set_variable_attr(luntncdf(ia), sid(ia),                 &
                                        'sea_water_salinity', 'PSU',           &
                                        'practical salinity unit','salinity',  &
                                        delete_value, ext_stscale )
            !- end of definitions
            call ncdf_end_definitions(luntncdf(ia))
           
            !- write coordinates:
            call ncdf_save_dim_vars(luntncdf(ia), lontvi(ia), lon(ia)%p,       &
                                  lattvi(ia),lat(ia)%p,dpthtvi(ia),depth(ia)%p )
          endif

          if (write_z_file) then
            !- Create a new file
            call ncdf_create_new( zncdfnam(ia), lunzncdf(ia) )
            !- Set global attributes     
            call ncdf_set_global_attr( lunzncdf(ia), minval(lon(ia)%p),        &
                                     maxval(lon(ia)%p), minval(lat(ia)%p),     &
                                     maxval(lat(ia)%p), zero, maxval(ti(ia)%p) )
            !- Set dimensions
            call ncdf_set_dimensions(lunzncdf(ia), nmx(ia), lonzid(ia),mmx(ia),&
                                     latzid(ia), kmx(ia),dpthzid(ia), tmzid(ia))
            !- define spatial and temporal coordinates:
            call ncdf_set_dim_vars(lunzncdf(ia),lonzid(ia),lonzvi(ia),         &
                                   latzid(ia), latzvi(ia),dpthzid(ia),         &
                                   dpthzvi(ia),tmzid(ia), tmzvi(ia) )
            !- set dimension attributes
            call ncdf_set_dim_attr ( lunzncdf(ia), lonzvi(ia), latzvi(ia),     &
                                     dpthzvi(ia), tmzvi(ia) )
            !- define variables of the water level netcdf-files
            call ncdf_set_3D_variables(lunzncdf(ia), "elev", lonzvi(ia),       &
                                       latzvi(ia), tmzvi(ia), zid(ia) )
            if (dyn) then
              call ncdf_set_4D_variables( lunzncdf(ia), "h", lonzvi(ia),       &
                                          latzvi(ia), dpthzvi(ia),tmzvi(ia),   &
                                          hid(ia) )
            endif
            call ncdf_set_4D_variables(lunzncdf(ia), "wvel", lonzvi(ia),       &
                                     latzvi(ia),dpthzvi(ia),tmzvi(ia),wid(ia) )
            !- set variable attributes of the water level netcdf-files
            call ncdf_set_variable_attr(lunzncdf(ia), wid(ia),                 &
                                      'upward_sea_water_velocity', 'm/s',      &
                                      'meters per second', 'vertical velocity',&
                                      delete_value, ext_wscale )
            if (dyn) then
              call ncdf_set_variable_attr(lunzncdf(ia), hid(ia),               &
                                          'layer_thicknesses', 'm', 'meters',  &
                                          'layer thicknesses', delete_value,   &
                                          ext_hscale )
            endif
            call ncdf_set_variable_attr(lunzncdf(ia), zid(ia),                 &
                                      'sea_surface_height_above_sea_level','m',&
                                      'meters', 'sea surface elevation',       &
                                      delete_value, ext_zscale )
            !- end of definitions
            call ncdf_end_definitions(lunzncdf(ia))

            !- write coordinates:
            call ncdf_save_dim_vars(lunzncdf(ia), lonzvi(ia), lon(ia)%p,       &
                                  latzvi(ia),lat(ia)%p,dpthzvi(ia),depth(ia)%p )
          endif
        enddo

      endif


      do ia=1,narea
        !- save time step information
        roundtim = tim
        call round(5,roundtim)
        if (write_m_file) then
          call ncdf_save_time (lunmncdf(ia), tmmvi(ia), roundtim, counter )

          !- wind-u-velocity
          call Fill_dummy_3D (dummy_3D(ia)%p, wv_i4(ia)%p,mm_op(ia)%p,mmx(ia), &
                              nmx(ia))
          call ncdf_save_3D_vars (lunmncdf(ia),wvid(ia),counter,dummy_3D(ia)%p )
          !- wind-v-velocity      
          call Fill_dummy_3D (dummy_3D(ia)%p, wu_i4(ia)%p, mm_op(ia)%p,mmx(ia),&
                              nmx(ia))
          call ncdf_save_3D_vars (lunmncdf(ia),wuid(ia),counter,dummy_3D(ia)%p )
        endif

        if (write_c_file) then
          call ncdf_save_time (luncncdf(ia), tmcvi(ia), roundtim, counter )

          !- eastward velocity
          call Fill_dummy_4D (dummy_4D(ia)%p, u_i4(ia)%p, mm_op(ia)%p,mmx(ia), &
                               nmx(ia), kmx(ia) )
          call ncdf_save_4D_vars (luncncdf(ia),uid(ia),counter,dummy_4D(ia)%p )
          !- northward velocity
          call Fill_dummy_4D (dummy_4D(ia)%p, v_i4(ia)%p,mm_op(ia)%p,mmx(ia),  &
                              nmx(ia), kmx(ia) )
          call ncdf_save_4D_vars (luncncdf(ia),vid(ia),counter,dummy_4D(ia)%p )
        endif


        if (write_t_file) then
          call ncdf_save_time (luntncdf(ia), tmtvi(ia), roundtim, counter )

          !- water temperature
          call Fill_dummy_4D (dummy_4D(ia)%p, t_i4(ia)%p, mm_op(ia)%p,mmx(ia), &
                              nmx(ia), kmx(ia) )
          call ncdf_save_4D_vars (luntncdf(ia),tid(ia),counter,dummy_4D(ia)%p )
          !- salinity
          call Fill_dummy_4D (dummy_4D(ia)%p, s_i4(ia)%p, mm_op(ia)%p,mmx(ia), &
                              nmx(ia), kmx(ia) )
          call ncdf_save_4D_vars (luntncdf(ia),sid(ia),counter,dummy_4D(ia)%p )
        endif


        if (write_z_file) then
          call ncdf_save_time (lunzncdf(ia), tmzvi(ia), roundtim, counter )

          !- water level      
          call Fill_dummy_3D (dummy_3D(ia)%p, z_i4(ia)%p, mm_op(ia)%p,mmx(ia), &
                              nmx(ia))
          call ncdf_save_3D_vars (lunzncdf(ia),zid(ia),counter,dummy_3D(ia)%p )
          if (dyn) then
            !- layer thickness
            call Fill_dummy_4D (dummy_4D(ia)%p, h_i4(ia)%p, mm_op(ia)%p,       &
                                mmx(ia), nmx(ia), kmx(ia) )
            call ncdf_save_4D_vars (lunzncdf(ia),hid(ia),counter,              &
                                    dummy_4D(ia)%p )
          endif
          !- vertical velocity
          call Fill_dummy_4D (dummy_4D(ia)%p, w_i4(ia)%p, mm_op(ia)%p,mmx(ia), &
                              nmx(ia), kmx(ia) )
          call ncdf_save_4D_vars (lunzncdf(ia),wid(ia),counter,dummy_4D(ia)%p )
        endif
      enddo
      counter = counter + 1
     
    endif

  end subroutine netcdf_write
  
!===============================================================================

  subroutine write_netcdf_icefile (ctim,tim,izeits,lmn1)

    use cmod_params,      only : narea, nmx, mmx
    use cmod_arrays,      only : casus, dz, xkoor, ykoor
    use constants,        only : iotamodel
    use dmi_mpi,          only : mpi_io_serial
    use dmi_mpi_global,   only : iu06
    use netcdf_interface, only : ncdf_create_new, ncdf_set_global_attr,        &
                                 ncdf_set_dimensions, ncdf_set_dim_vars,       &
                                 ncdf_save_dim_vars, ncdf_set_dim_attr,        &
                                 ncdf_set_3D_variables, ncdf_set_variable_attr,&
                                 ncdf_end_definitions, ncdf_save_time,         &
                                 ncdf_save_3D_vars

    implicit none

    character(19), intent(in) :: ctim
    real(8),       intent(in) :: tim
    integer(4),    intent(in) :: izeits,lmn1

    integer(4) :: ia
    real(8)    :: roundtim
    
    integer(4), save   :: counter

    write(iu06,'(a26,a19)') 'write on netcdf_icefile at time: ',ctim
    if (iotamodel) then
      write(iu06,'(a34,a19)') 'write on netcdf_iotafile at time: ',ctim
    endif
    
    !- create the netcdf-icefiles and write the header when needed -------------
    if (izeits == lmn1) then
      counter = 1

      do ia=1,narea
        !- Create the ice-file
        call ncdf_create_new( incdfnam(ia), lunincdf(ia) )

        !- Set global attributes     
        call ncdf_set_global_attr( lunincdf(ia), minval(lon(ia)%p),            &
                                   maxval(lon(ia)%p), minval(lat(ia)%p),       &
                                   maxval(lat(ia)%p), zero, zero )  

        !- Set dimensions
        call ncdf_set_dimensions(lunincdf(ia), nmx(ia), loniid(ia), mmx(ia),   &
                                 latiid(ia), 1, dpthiid(ia), tmiid(ia) )
        
        !- define spatial and temporal coordinates:
        call ncdf_set_dim_vars(lunincdf(ia),loniid(ia),lonivi(ia),latiid(ia),  &
                               lativi(ia),dpthiid(ia),dpthivi(ia),tmiid(ia),   &
                               tmivi(ia) )
 

        !- set dimension attributes
        call ncdf_set_dim_attr(lunincdf(ia),lonivi(ia),lativi(ia),dpthivi(ia), &
                               tmivi(ia) )
 
        !- define variables of the ice netcdf-files (the i-file)
        call ncdf_set_3D_variables(lunincdf(ia),"casus",lonivi(ia),lativi(ia), &
                                   tmivi(ia), casusid(ia), (.true.) )
        call ncdf_set_3D_variables(lunincdf(ia), "ice_thk", lonivi(ia),        &
                                   lativi(ia), tmivi(ia), icethkid(ia) )
        call ncdf_set_3D_variables(lunincdf(ia), "ice_cov", lonivi(ia),        &
                                   lativi(ia), tmivi(ia), icecoid(ia) )
        call ncdf_set_3D_variables(lunincdf(ia), "snow_thk", lonivi(ia),       &
                                   lativi(ia), tmivi(ia), snowthkid(ia) )
        call ncdf_set_3D_variables(lunincdf(ia), "tsnei", lonivi(ia),          &
                                   lativi(ia), tmivi(ia), tsneiid(ia) )
        call ncdf_set_3D_variables(lunincdf(ia), "ice_uvel", lonivi(ia),       &
                                   lativi(ia), tmivi(ia), uiceid(ia) )
        call ncdf_set_3D_variables(lunincdf(ia), "ice_vvel", lonivi(ia),       &
                                   lativi(ia), tmivi(ia), viceid(ia) )
   
        !- set variable attributes of the ice netcdf-files (the i-file)
        call ncdf_set_variable_attr(lunincdf(ia), casusid(ia),                 &
                                    'casus', '1-ice/0-no ice','Does ice exist',&
                                    -1 )
        call ncdf_set_variable_attr(lunincdf(ia), icethkid(ia),                &
                                    'sea_ice_thickness', 'm', 'meters',        &
                                    'sea ice thickness', delete_value,         &
                                    ext_edscale )
        call ncdf_set_variable_attr(lunincdf(ia), icecoid(ia),                 &
                                    'sea_ice_area_fraction', 'percent',        &
                                    'percent', 'sea ice cover', delete_value,  &
                                    ext_ecscale )
        call ncdf_set_variable_attr(lunincdf(ia), snowthkid(ia),               &
                                    'snow_thickness', 'm', 'meters',           &
                                    'snow thickness',delete_value,ext_sdscale )
        call ncdf_set_variable_attr(lunincdf(ia), tsneiid(ia),                 &
                                    'surface_temperature_snow_ice',            &
                                    'degree_Celsius', 'degree Celsius',        &
                                    'srf temp snow-ice', delete_value,         &
                                    ext_tsscale )
        call ncdf_set_variable_attr(lunincdf(ia), uiceid(ia),                  &
                                    'eastward_ice_velocity', 'm/s',            &
                                    'meters per second',                       &
                                    'eastward ice velocity', delete_value,     &
                                    ext_uvescale )
        call ncdf_set_variable_attr(lunincdf(ia), viceid(ia),                  &
                                    'northward_ice_velocity', 'm/s',           &
                                    'meters per second',                       &
                                    'northward ice velocity', delete_value,    &
                                    ext_uvescale )


        !- end of definitions
        call ncdf_end_definitions(lunincdf(ia))

        !- write coordinates:
        call ncdf_save_dim_vars(lunincdf(ia),lonmvi(ia), lon(ia)%p, latmvi(ia),&
                                lat(ia)%p, dpthmvi(ia), surface_depth  )


        if (iotamodel) then
          !- Create the iota-file
          call ncdf_create_new( iotancdfnam(ia), luniotancdf(ia) )

          !- Set global attributes     
          call ncdf_set_global_attr( luniotancdf(ia), minval(lon(ia)%p),       &
                                     maxval(lon(ia)%p), minval(lat(ia)%p),     &
                                     maxval(lat(ia)%p), zero, zero )  

          !- Set dimensions
          call ncdf_set_dimensions(luniotancdf(ia), nmx(ia), loniotaid(ia),    &
                                   mmx(ia), latiotaid(ia), 1, dpthiotaid(ia),  &
                                   tmiotaid(ia) )
        
          !- define spatial and temporal coordinates:
          call ncdf_set_dim_vars(luniotancdf(ia), loniotaid(ia), loniotavi(ia),&
                                 latiotaid(ia), latiotavi(ia), dpthiotaid(ia), &
                                 dpthiotavi(ia), tmiotaid(ia), tmiotavi(ia) )
 

          !- set dimension attributes
          call ncdf_set_dim_attr(luniotancdf(ia), loniotavi(ia), latiotavi(ia),&
                                 dpthiotavi(ia), tmiotavi(ia) )

          !- define variables of the iota netcdf-files (the iota-file)
          call ncdf_set_3D_variables(luniotancdf(ia), "intactness",            &
                                     loniotavi(ia), latiotavi(ia),tmiotavi(ia),&
                                     iota1id(ia) )
          call ncdf_set_3D_variables(luniotancdf(ia), "len_sca", loniotavi(ia),&
                                     latiotavi(ia), tmiotavi(ia), iota2id(ia) )
          call ncdf_set_3D_variables(luniotancdf(ia), "fragility",             &
                                     loniotavi(ia), latiotavi(ia),tmiotavi(ia),&
                                     iota3id(ia) )
   
          !- set variable attributes of the iota netcdf-files (the iota-file)
          call ncdf_set_variable_attr(luniotancdf(ia), iota1id(ia),            &
                                      'ice_intactness', 'no unit', 'no unit',  &
                                      'ice intactness', delete_value,          &
                                      ext_iotascale )
          call ncdf_set_variable_attr(luniotancdf(ia), iota2id(ia),            &
                                      'length_scale', 'm', 'Meter',            &
                                      'length scale', delete_value,            &
                                      ext_iotascale )
          call ncdf_set_variable_attr(luniotancdf(ia), iota3id(ia),            &
                                      'fragility', 'no unit', 'no unit',       &
                                      'fragility',delete_value, ext_iotascale )

          !- end of definitions
          call ncdf_end_definitions(luniotancdf(ia))

          !- write coordinates:
          call ncdf_save_dim_vars(luniotancdf(ia), lonmvi(ia), lon(ia)%p,      &
                                  latmvi(ia), lat(ia)%p, dpthmvi(ia),          &
                                  surface_depth  )

        endif
      enddo
    endif


    if (mpi_io_serial) then
      do ia=1,narea
        !- save time step information
        roundtim = tim
        call round(5,roundtim)
        call ncdf_save_time (lunincdf(ia), tmivi(ia), roundtim, counter )

        !- write the ice netcdf-files (the i-file)
        !- casus
        call Fill_dummy_3D(dummy_3D(ia)%p,casus(ia)%p,mm_op(ia)%p,mmx(ia),     &
                           nmx(ia))
        call ncdf_save_3D_vars (lunincdf(ia), casusid(ia), counter,            &
                                dummy_3D(ia)%p)
        !- ice thickness
        call Fill_dummy_3D (dummy_3D(ia)%p,eisdi_i4(ia)%p, mm_op(ia)%p,mmx(ia),&
                            nmx(ia))
        call ncdf_save_3D_vars(lunincdf(ia),icethkid(ia),counter,dummy_3D(ia)%p)
        !- ice cover      
        call Fill_dummy_3D (dummy_3D(ia)%p,eisco_i4(ia)%p ,mm_op(ia)%p,mmx(ia),&
                            nmx(ia))
        call ncdf_save_3D_vars(lunincdf(ia),icecoid(ia),counter,dummy_3D(ia)%p)
        !- snow thickness      
        call Fill_dummy_3D (dummy_3D(ia)%p,snowd_i4(ia)%p ,mm_op(ia)%p,mmx(ia),&
                            nmx(ia))
        call ncdf_save_3D_vars(lunincdf(ia), snowthkid(ia), counter,           &
                               dummy_3D(ia)%p)
        !- surface temperature snow /ice      
        call Fill_dummy_3D (dummy_3D(ia)%p,tsnei_i4(ia)%p, mm_op(ia)%p,mmx(ia),&
                            nmx(ia))
        call ncdf_save_3D_vars(lunincdf(ia),tsneiid(ia),counter,dummy_3D(ia)%p)
        !- eastward ice velocity      
        call Fill_dummy_3D (dummy_3D(ia)%p,ueis_i4(ia)%p, mm_op(ia)%p, mmx(ia),&
                            nmx(ia))
        call ncdf_save_3D_vars(lunincdf(ia),uiceid(ia),counter,dummy_3D(ia)%p)
        !- northward ice velocity      
        call Fill_dummy_3D (dummy_3D(ia)%p,veis_i4(ia)%p, mm_op(ia)%p, mmx(ia),&
                            nmx(ia))
        call ncdf_save_3D_vars(lunincdf(ia),viceid(ia),counter,dummy_3D(ia)%p)

        if (iotamodel) then
          !- write the iota netcdf-files (the iota-file)
          !- store intactness
          call Fill_dummy_3D (dummy_3D(ia)%p,iota1_i4(ia)%p, mm_op(ia)%p,      &
                              mmx(ia), nmx(ia))
          call ncdf_save_3D_vars(luniotancdf(ia), iota1id(ia), counter,        &
                                 dummy_3D(ia)%p)
          !- length scale
          call Fill_dummy_3D (dummy_3D(ia)%p,iota2_i4(ia)%p, mm_op(ia)%p,      &
                              mmx(ia), nmx(ia))
          call ncdf_save_3D_vars(luniotancdf(ia), iota2id(ia), counter,        &
                                 dummy_3D(ia)%p)
          !- fragility
          call Fill_dummy_3D (dummy_3D(ia)%p,iota3_i4(ia)%p, mm_op(ia)%p,      &
                              mmx(ia), nmx(ia))
          call ncdf_save_3D_vars(luniotancdf(ia), iota3id(ia), counter,        &
                                 dummy_3D(ia)%p)
        endif
      enddo
      counter = counter + 1
    endif

  end subroutine write_netcdf_icefile

!===============================================================================

  subroutine netcdf_icefile_new ()

    use dmi_mpi,        only : mpi_io_serial
    use cmod_params,    only : narea, cfgfile
    use constants,      only : archive, iotamodel

    implicit none

    logical, save :: FirstEntry = .true.
    character(2)  :: cicencdf
    integer(4)    :: ia


    ! allocate on first entry:
    if (FirstEntry) then

      FirstEntry = .false.
      call allocate_compressed_ice_arrays( )

      ia = narea
      allocate( lunincdf(ia), incdfnam(ia) )
      allocate( loniid(ia), latiid(ia), dpthiid(ia), tmiid(ia) )
      allocate( lonivi(ia), lativi(ia), dpthivi(ia), tmivi(ia) )
      allocate( casusid(ia),icethkid(ia),icecoid(ia),snowthkid(ia),tsneiid(ia),&
                uiceid(ia), viceid(ia) )
      if ( iotamodel ) then
        allocate( luniotancdf(ia), iotancdfnam(ia) )
        allocate( loniotaid(ia), latiotaid(ia), dpthiotaid(ia), tmiotaid(ia) )
        allocate( loniotavi(ia), latiotavi(ia), dpthiotavi(ia), tmiotavi(ia) )
        allocate( iota1id(ia), iota2id(ia), iota3id(ia) )
      endif

      do ia=1,narea
        icencdf = 0
        if (archive) then
          incdfnam(ia) = 'i_file_'//trim(cfgfile(ia)(6:))//'_archive.nc'
          if ( iotamodel ) then
            iotancdfnam(ia)='iota_file_'//trim(cfgfile(ia)(6:))//'_archive.nc'
          endif
        else
          write(cicencdf,'(i2.2)') icencdf
          incdfnam(ia) = 'i_file_'//trim(cfgfile(ia)(6:))//'_'//cicencdf//'.nc'
          if ( iotamodel ) iotancdfnam(ia) =                                   &
                      'iota_file_'//trim(cfgfile(ia)(6:))//'_'//cicencdf//'.nc'
        endif
      enddo

    
    else
    
      icencdf = icencdf + 1
      write(cicencdf,'(i2.2)') icencdf
      do ia=1,narea
        incdfnam(ia) = 'i_file_'//trim(cfgfile(ia)(6:))//'_'//cicencdf//'.nc'
        if ( iotamodel ) iotancdfnam(ia) =                                     &
                      'iota_file_'//trim(cfgfile(ia)(6:))//'_'//cicencdf//'.nc'
      enddo    
    
    endif

  end subroutine netcdf_icefile_new

!===============================================================================

  subroutine arc2deg(mmx, nmx, xkoor, ykoor, lon, lat)

    use constants,        only : grarad
    

    implicit none
    
    integer(4), intent(in) :: mmx, nmx
    real(8), intent(in)    :: xkoor(:), ykoor(:)
    real(8), intent(out)   :: lon(:), lat(:)
    
    integer(4)  :: i
    
    do i = 1, mmx
      lat(i) = ykoor(i)/grarad
      call round(4, lat(i))
    enddo

    do i = 1,nmx
      lon(i) = xkoor(i)/grarad
      call round(4, lon(i))      
    enddo
    
  end subroutine arc2deg
    
!===============================================================================

  subroutine calculate_depth(kmx, dz, depth)
    
    implicit none
    
    integer(4), intent(in) :: kmx
    real(8), intent(in)    :: dz(:)
    real(8), intent(out)   :: depth(:)
    
    integer(4)  :: i
    real(8)     :: dpt
    
    dpt = zero
    do i = 1, kmx
      dpt = dpt + dz(i)
      depth(i) = dpt
    enddo
    
  end subroutine  calculate_depth 

!===============================================================================
  
  subroutine Fill_dummy_3D_int (dummy_3D, var, mm, mmx, nmx)
  
    implicit none
  
    integer(4), intent(out) :: dummy_3D(:,:)
    integer(4), intent(in)  :: var(0:)
    integer(4), intent(in)  :: mm(0:,0:,:)
    integer(4), intent(in)  :: mmx, nmx
  
    integer(4)             :: i,j
  
    do j = 1,nmx
      do i = 1, mmx
        if(mm(i,j,1) == 0) then
          dummy_3D(j,i) = delete_value
        else
          dummy_3D(j,i) = var(mm(i,j,1))
        endif
      enddo
    enddo
  
  end subroutine Fill_dummy_3D_int
  
!===============================================================================
  
  subroutine Fill_dummy_3D_logical (dummy_3D, var, mm, mmx, nmx)
  
    implicit none
  
    integer(4), intent(out) :: dummy_3D(:,:)
    logical, intent(in)     :: var(0:)
    integer(4), intent(in)  :: mm(0:,0:,:)
    integer(4), intent(in)  :: mmx, nmx
  
    integer(4)              :: i,j
  
    do j = 1,nmx
      do i = 1, mmx
        if(mm(i,j,1) == 0) then
          dummy_3D(j,i) = -1
        else
          if (var(mm(i,j,1))) then
            dummy_3D(j,i) = 1
          elseif (.not.var(mm(i,j,1))) then
            dummy_3D(j,i) = 0
          endif
        endif
      enddo
    enddo
  
  end subroutine Fill_dummy_3D_logical 

!===============================================================================
 
  subroutine Fill_dummy_4D (dummy_4D, var, mm, mmx, nmx, kmx)
  
    implicit none
  
    integer(4), intent(out) :: dummy_4D(:,:,:)
    integer(4), intent(in)  :: var(0:)
    integer(4), intent(in)  :: mm(0:,0:,:)
    integer(4), intent(in)  :: mmx, nmx, kmx
  
    integer(4)              :: i,j,k
  
    do k = 1,kmx
      do j = 1, nmx
        do i = 1, mmx
          if (mm(i,j,k) == 0) then
            dummy_4D(j,i,k) = delete_value
          else
            dummy_4D(j,i,k) = var(mm(i,j,k))
          endif  
        enddo
      enddo
    enddo
  
  end subroutine Fill_dummy_4D    

!===============================================================================

  subroutine round (dig, num)
  !-----------------------------------------------------------------------------
  !-
  !- This subroutine rounds real values, so that they have just the stated
  !- number of digits. Otherwise the output files look partly very strange
  !- Hopefully on a new machine this routine isn't needed anymore!!
  !-
  !-----------------------------------------------------------------------------
    
    integer(4), intent(in)     :: dig
    real(8), intent(inout)     :: num
    
    integer(4)    :: numb_int, numb_int_float, numb_int_test, internal_digits
      
    if (dig > 10) then
      internal_digits = 10
    else
      internal_digits = dig
    endif
      
    numb_int = int(num, kind = 4)
    numb_int_test = int((num - real(numb_int, kind = 8)) *                     &
                    (ten**internal_digits))

    if(numb_int_test /= 0 .or. dig <= 10) then
      if(numb_int /= 0) num = num - real(numb_int, kind = 8)

      num = num * (ten**internal_digits)

      numb_int_float = nint(num, kind = 4)
      num = real(numb_int_float, kind = 8) / ten**internal_digits
      num = num + real(numb_int, kind = 8)
    endif
    
  end subroutine round

!===============================================================================
    
end module writenetcdf_BSH
