module writetempdat
  ! methods to write tempdat files and ice data files

  use constants, only : half, two, four, eight, nineteen, sph, kbyte
  use compress,  only : wu_i4, wv_i4, u_i4, v_i4, w_i4, h_i4, z_i4, s_i4, t_i4,&
                       eisdi_i4, eisco_i4, snowd_i4, tsnei_i4, ueis_i4,veis_i4,&
                       casus_out, iota1_i4, iota2_i4, iota3_i4,                &
                       allocate_compressed_arrays,                             &
                       allocate_compressed_ice_arrays

  implicit none
  private

  public  :: tempdat_new,tempdat_next,tempdat_close,tempdat_write
  public  :: write_icefile,icefile_close,icefile_new

  integer(4),    private, save :: itempd, luntempdat, ios, luneispar,luniotapar
  character(40), private, save :: tdnam

contains

!===============================================================================

  subroutine tempdat_new ()

    use constants,      only : archive, toutput, filesizemax
    use cmod_params,    only : narea, iw2, iw3
    use dmi_mpi_global, only : iu06
    use cmod_miscvars,  only : schedule

    implicit none

    logical, save :: FirstEntry = .true.
    character(2)  :: ctmp
    integer(4)    :: ia, ifsh, nt, idiv
    real(8)       :: fsize, ofsize, osch, ifsv


    if (FirstEntry) then
      FirstEntry = .false.

      call allocate_compressed_arrays( )

      itempd = 0
      if (archive) then
        tdnam = 'archive'
      else
        write(ctmp,'(i2.2)') itempd
        tdnam = 'tempdat.'//ctmp
      endif

      ! Calculate file size in Bytes of 8bit:
      !    ifsh is size of the static data (the header),
      !    ifsv is size of the dynamic data (i.e. the vars) for each time step.
      ! Total size of file is ifsh + nt*ifsv where nt is No. of time steps
      ! in the file.
      ifsh = 4 + 11 + 4                                 ! aufdat header
      ifsv = four + nineteen + eight + four             ! ctim, tim
      ifsv = ifsv + four
      do ia=1,narea
        ifsv = ifsv + four*two*real(iw2(ia)+1,8)        ! Wu,Wv
      enddo
      ifsv = ifsv + four
      do ia=1,narea
        ifsv = ifsv + four
        ifsv = ifsv + four*real(3*(iw3(ia)+1) + (iw2(ia)+1),8)   ! u,v,w,z
        ifsv = ifsv + four
      enddo
      do ia=1,narea
        ifsv = ifsv + four
        ifsv = ifsv + four*two*real(iw3(ia)+1,8)        ! s,t
        ifsv = ifsv + four
      enddo
      nt    = schedule*nint(sph/toutput)
      osch  = sph*real(schedule,8)
      fsize = real(ifsh,8) + real(nt,8)*ifsv
      if (fsize > (kbyte*kbyte*kbyte*kbyte)) then
        ofsize = fsize/(kbyte*kbyte*kbyte*kbyte)
        ctmp   = 'TB'
      elseif (fsize > (kbyte*kbyte*kbyte)) then
        ofsize = fsize/(kbyte*kbyte*kbyte)
        ctmp   = 'GB'
      elseif (fsize > (kbyte*kbyte)) then
        ofsize = fsize/(kbyte*kbyte)
        ctmp   = 'MB'
      elseif (fsize > kbyte) then
        ofsize = fsize/kbyte
        ctmp   = 'KB'
      else
        ofsize = fsize
        ctmp   = 'B '
      endif
      write(iu06,'(a17,1x,f5.1,1x,a2)') 'Output file size:', ofsize, ctmp 
      if (fsize/(kbyte*kbyte*kbyte) > filesizemax) then
        write(iu06,'(a7,1x,f8.3,1x,a6)') 'exceeds', filesizemax, 'GB, so'
!FIXME: re-scheduling is actually NOT implemented yet, therefore the next
!       two lines are just suggestions:
        write(iu06,'(a33)') 'filesizemax should be higher or  '
        write(iu06,'(a33)') 'output should be re-scheduled to:'
!        write(iu06,'(a31)') 'output will be re-scheduled to:'
        idiv   = 1
        osch   = half*osch
        ofsize = half*fsize
        do while (ofsize/(kbyte*kbyte*kbyte) > filesizemax)
          idiv   = idiv + 1
          osch   = half*osch
          ofsize = half*ofsize
        enddo
        write(iu06,'(f8.3,1x,a3)') real(schedule,8)/(two**idiv), 'hrs'
!FIXME: pseudo-code drafting the treatment of too large files:
!
!       each output chunk of 'schedule' hrs should be subdivided into idiv+1
!         sub-intervals each of length schedule/(2**idiv) 
!         
!       this should somehow be communicated to where cmod_main calls
!         tempdat_next
!       
!       calc ntempdat, No. of tempdat files needed
!       if (ntempdat >= 100) call exitme(1,'Too many outputfiles needed')
!
      endif

    else
      itempd = itempd + 1
      write(ctmp,'(i2.2)') itempd
      tdnam = 'tempdat.'//ctmp
    endif


  end subroutine tempdat_new

!===============================================================================

  subroutine tempdat_close()
    implicit none
    close (luntempdat)
  end subroutine tempdat_close

!===============================================================================

  subroutine icefile_close()
    use constants, only : iotamodel
    implicit none
    close (luneispar)
    if (iotamodel) then
      close (luniotapar)
    endif
  end subroutine icefile_close

!===============================================================================

  subroutine tempdat_next ()

    use dmi_mpi_global, only : iu06
    implicit none

    call tempdat_close ()

    write(iu06,'(a21,a)') 'hd output stored on: ',trim(tdnam)

    call tempdat_new ()

  end subroutine tempdat_next

!===============================================================================

  subroutine tempdat_write (ctim,tim,aufdat,izeits,lmn2)

    use exits,          only : exitme
    use io_subs,        only : io_new_unit, flush_unit
    use cmod_params,    only : narea
    use constants,      only : dyn
    use dmi_mpi,        only : mpi_io_serial
    use dmi_mpi_global, only : iu06

    implicit none

    character(19), intent(in) :: ctim
    real(8),       intent(in) :: tim
    character(11), intent(in) :: aufdat
    integer(4),    intent(in) :: izeits,lmn2

    integer(4) :: ia

    if (mpi_io_serial) then
      write(iu06,'(a9,a,a10,a19)') 'write on ',trim(tdnam),' at time: ',ctim
      !- open file and write analysis header first time only (in each file) ----
      if (izeits == lmn2) then
        luntempdat = io_new_unit()
        open(luntempdat,file=tdnam,form='unformatted',status='replace',        &
             iostat=ios)
        if (ios /= 0) call exitme(1, 'Cannot open '//trim(tdnam))
        write(luntempdat,iostat=ios) aufdat
        if (ios /= 0) call exitme(1,'Could not write header in '//trim(tdnam))
      endif

      write(luntempdat,iostat=ios) ctim,tim
      if (ios /= 0) call exitme(1,'Could not write time stamp in '//trim(tdnam))

      write(luntempdat,iostat=ios) (wu_i4(ia)%p, wv_i4(ia)%p, ia=1,narea)
      if (ios /= 0) call exitme(1,'Could not write wind data in '//trim(tdnam))
      do ia=1,narea
        if (dyn) then
          write(luntempdat,iostat=ios) u_i4(ia)%p, v_i4(ia)%p, w_i4(ia)%p,     &
                                       h_i4(ia)%p, z_i4(ia)%p
          if(ios /= 0) call exitme(1,'Could not write HD data in '//trim(tdnam))
        else
          write(luntempdat,iostat=ios) u_i4(ia)%p, v_i4(ia)%p, w_i4(ia)%p,     &
                                       z_i4(ia)%p
          if(ios /= 0) call exitme(1,'Could not write HD data in '//trim(tdnam))
        endif
      enddo
      do ia=1,narea 
        write(luntempdat,iostat=ios) s_i4(ia)%p, t_i4(ia)%p
        if (ios /= 0) call exitme(1,'Could not write S/T data in '//trim(tdnam))
      enddo

      call flush_unit(luntempdat)
    endif

  end subroutine tempdat_write

!===============================================================================

  subroutine write_icefile (ctim,tim)

    use cmod_params,    only : narea
    use dmi_mpi,        only : mpi_io_serial
    use dmi_mpi_global, only : iu06
    use constants,      only : iotamodel

    implicit none

    character(19), intent(in) :: ctim
    real(8),       intent(in) :: tim

    integer(4) :: ia

    write(iu06,'(a26,a19)') 'write on icefile at time: ',ctim

    if (iotamodel) then
      write(iu06,'(a27,a19)') 'write on iotafile at time: ',ctim
    endif

    if (mpi_io_serial) then
      write(luneispar,'(a19,f16.8)') ctim,tim

      do ia=1,narea
        write(luneispar,'(250l1)') casus_out(ia)%p
        write(luneispar,'(62i4)') eisdi_i4(ia)%p
        write(luneispar,'(83i3)') eisco_i4(ia)%p
        write(luneispar,'(62i4)') snowd_i4(ia)%p, tsnei_i4(ia)%p
        write(luneispar,'(50i5)') ueis_i4(ia)%p, veis_i4(ia)%p
      enddo

      if (iotamodel) then
        write(luniotapar,'(a19,f16.8)') ctim,tim
        do ia=1,narea
          write(luniotapar,'(41i6)') iota1_i4(ia)%p
          write(luniotapar,'(41i6)') iota2_i4(ia)%p
          write(luniotapar,'(41i6)') iota3_i4(ia)%p
        enddo
      endif
    endif

  end subroutine write_icefile

!===============================================================================

  subroutine icefile_new (aufdat)

    use dmi_mpi,        only : mpi_io_serial
    use exits,          only : exitme
    use io_subs,        only : io_new_unit
    use constants,      only : iotamodel

    implicit none

    character(11), intent(in) :: aufdat

    logical, save :: FirstEntry = .true.


    ! allocate on first entry:
    if (FirstEntry) then
      FirstEntry = .false.
      call allocate_compressed_ice_arrays( )
    endif
    
    if (mpi_io_serial) then
      luneispar = io_new_unit()
      open(luneispar,file='i'//aufdat(2:11),status='replace',iostat=ios )
      if (ios /= 0) call exitme(1, 'Cannot open '//'i'//aufdat(2:11) )
      write(luneispar,'(a)') aufdat

      if (iotamodel) then
        luniotapar = io_new_unit()
        open(luniotapar,file='iota'//aufdat(2:11),status='replace',iostat=ios )
        if (ios /= 0) call exitme(1, 'Cannot open '//'iota'//aufdat(2:11) )
        write(luniotapar,'(a)') aufdat
      endif

    endif

  end subroutine icefile_new

!===============================================================================

end module writetempdat
